
cpm(Network, Graph, Latest) :-
	build(Network, Graph),
	early_late(Graph, Graph, End, Latest),
	Latest >= End,
	analyse(Graph, Graph).
	
cpm(Network, Graph) :-
	build(Network, Graph),
	early_late(Graph, Graph, End),
	analyse(Graph, Graph).

% build an adjacency graph out of the network
build([], Graph) :-
	buildv([], _, Graph).
build([[I, J, C]|T], Graph) :-
	buildv(ed(I, J, C), to, Graph),
	buildv(ed(I, J, C), from, Graph),
	build(T, Graph).

buildv([], _, []).
buildv([], _, [ad(_, _, _, To, From)|T]) :-
	 addedg([], _, To),
	addedg([], _, From),
	buildv([], _, T).
buildv(ed(I, J, C), to, [ad(I, Es, Lc, To, From)|T]) :-
	 addedg(J, C, To).
buildv(Edge, to, [H|T]) :-
	 buildv(Edge, to, T).
buildv(ed(I, J, C), from, [ad(J, Es, Lc, To, From)|T]) :-
	 addedg(I, C, From).
buildv(Edge, from, [H|T]) :-
	 buildv(Edge, from, T).

addedg([], _, []) .
addedg([], _, [H|T]) :-  
	addedg([], _, T).
addedg(V, C, [ed(V, C, _, _, _, _)|T]) .
addedg(V, C, [H|T]) :-
	addedg(V, C, T).

% Get early start times and latest completion times
% early/4 is used when a ending time is given
% otherwise early/3 assumes that the early start time
% for the end node is equal to the latest completion time
early_late([], _, _, _).
early_late([ad(I, Es, Lc, To, From)|T], G, End, Latest) :-
	setearly(From, To, G, End, Es),
	setlate(To, G, Latest, Lc),
	early_late(T, G, End, Latest).

early_late([], _, _).
early_late([ad(I, Es, Lc, To, From)|T], G, End) :-
	setearly(From, To, G, End, Es),
	setlate(To, G, End, Lc),
	early_late(T, G, End).

setearly([], _, _, _, X) :- X=:=0.
setearly([ed(V, C, _, _, _, _)|T], [], G, Es, Es) :-
	
	getnode(V, G, Es1, _),
	Es2=:=Es1+C,	
	setmax(T, G, Es2, Es).
setearly([ed(V, C, _, _, _, _)|T], _, G, End, Es) :-
	getnode(V, G, Es1, _),
	Es2=:=Es1+C,
	setmax(T, G, Es2, Es).

setmax([], _, Max, Max).
setmax([ed(V, C, _, _, _, _)|T], G, Max0, Max) :-
	getnode(V, G, Es1, _),
	Es2=:=Es1+C, max(Max0, Es2,Max1),
	setmax(T, G, Max1, Max).

setlate([], _, Last, Last).
setlate([ed(V, C, _, _, _, _)|T], G, Last, Lc) :-
	getnode(V, G, _, Lc1),
	Lc2=:=Lc1-C,	
	setmin(T, G, Lc2, Lc).

setmin([], _, Min, Min).
setmin([ed(V, C, _, _, _, _)|T], G, Min0, Min) :-
	getnode(V, G, _, Lc1),
	Lc2=:=Lc1-C, min(Min0, Lc2,Min1),
	setmin(T, G, Min1 , Min).

% Search graph for the early & late times for a node
getnode(I, [ad(I, Es, Lc, _, _)|T], Es, Lc).
getnode(I, [H|T], Es, Lc) :-
	getnode(I, T, Es, Lc).

% Compute the other times :
%		Ls - latest start time
%		Ec - earliest completion time
%		Tf - total float time
%		Ff - free float time
analyse([], G).
analyse([ad(I, Es, Lc, To, _)|T], G) :-
	analyse_times(To, Es, Lc, G), 
	analyse(T, G).

analyse_times([], _, _, _).
analyse_times([ed(V, C, Ls, Ec, Tf, Ff)|T], Esi, Lci, G) :-
	getnode(V, G, Esj, Lcj), 
	compute(Ls, Ec, Tf, Ff, Esj, Lcj, Esi, Lci, C), 
	analyse_times(T, Esi, Lci, G).

% Indirect way of doing the calculation just to speed things up
% can be removed and inserted directly in analyse_times
compute(Ls, Ec, Tf, Ff, Esj, Lcj, Esi, Lci, C) :-
	X =:= Esi+C, 
	Ls =:= Lcj-C, 
	Ec =:= Esi+C, 
	Tf =:= Lcj-X, 
	Ff =:= Esj-X.

min(X,Y,X) :- X=<Y.
min(X,Y,Y) :- Y<X.

max(X,Y,X) :- X>=Y.
max(X,Y,Y) :- Y>X.