%%% fft4(P1,P2,P3,P4,FFT) :- fft(4,[P1,P2,P3,P4],FFT).
%%% fft(4,[c(1.65,3.123),c(34.22,12.1),c(22.6,7.5),c(23.5,55.32)],FFT).



/* fft(N, F, Ft) is true if the list Ft contains the Fourier transform of  */
/*   the N -- a power of two -- samples in the list F.  Each sample is a   */
/*   complex number represented by c(RealPart, ImaginaryPart).             */
fft(N, F, F):- N=:=1.

fft(N, F, Ft):-
  N > 1,  N1 =:= N/2,
  evens_and_odds(F, E, O),
  fft(N1, E, Et),
  fft(N1, O, Ot),
  K1=:=1, K2=:=2,
  w(K1, W1),
  w(K2, W2),
  w(N, Wn),
  fft_1(Et, Ot, W2, Wn, Gt, []),
  fft_1(Et, Ot, W1, Wn, Ft, Gt).

fft_1([], [], _, _, Ft, Ft).
fft_1([E|Et], [O|Ot], Wk, Wn, [F|Ft], Fu):-
  product(O, Wk, Temp),
  sum(E, Temp, F),
  product(Wk, Wn, Wk1),
  fft_1(Et, Ot, Wk1, Wn, Ft, Fu).

/* evens_and_odds(Xs, Evens, Odds) is true if Evens is the list of the     */
/*   even-positioned elements of the list Xs, and Odds is the list of the  */
/*   odd-positioned elements of the list Xs, where Xs has an even number   */
/*   of elements, and the first element of Xs is considered to be at an    */
/*   even position.                                                        */
evens_and_odds([], [], []).
evens_and_odds([Even,Odd|Xs], [Even|Evens], [Odd|Odds]):-
  evens_and_odds(Xs, Evens, Odds).

/* sum(A, B, C) is true if C is the sum of the complex numbers A and B.    */
sum(c(Ra,Ia), c(Rb,Ib), c(Rc,Ic)):-
  Rc =:= Ra + Rb, Ic =:= Ia + Ib.

/* product(A, B, C) is true if C is the product of the complex numbers A   */
/*   and B.                                                                */
product(c(Ra,Ia), c(Rb,Ib), c(Rc,Ic)):-
  Rc =:= Ra*Rb - Ia*Ib, Ic =:= Ra*Ib + Ia*Rb.

/* w(N, c(Cosine, Sine)) is true if Cosine = cos(2 * pi / N) and           */
/*   Sine = sin(2 * pi / N), where N is a power of two between 1 and 16384 */
/*   inclusive.                                                            */

w(N,c(Cosine, Sine)) :-
	Pi=:=3.14 , Cosine =:= cos(2 * Pi / N), Sine =:= sin(2 * Pi / N).

%%%w(    1, c( 1.0, 0.0)).
%%%w(    2, c(-1.0, 0.0)).
%%%w(    4, c( 0.0, 1.0)).
%%%w(    8, c( 0.70710678119, 0.70710678119)).
%%%w(   16, c( 0.92387953251, 0.38268343237)).
%%%w(   32, c( 0.9807852804,  0.19509032202)).
%%%w(   64, c( 0.99518472667, 0.09801714033)).
%%%w(  128, c( 0.99879545621, 0.049067674327)).
%%%w(  256, c( 0.9996988187,  0.024541228523)).
%%%w(  512, c( 0.99992470184, 0.012271538286)).
%%%w( 1024, c( 0.99998117528, 0.0061358846492)).
%%%w( 2048, c( 0.99999529381, 0.003067956763)).
%%%w( 4096, c( 0.99999882345, 0.0015339801863)).
%%%w( 8192, c( 0.99999970586, 0.00076699031874)).
%%%w(16384, c( 0.99999992647, 0.00038349516882)).