%% ADDED BY ROBERTO BAGNARA

:- use_module(library(clpr)).

scale_color(rgb(A,B,C),D,rgb(E,F,G)) :-
  {H = A*D,
  I = B*D,
  J = C*D,
  max(H,255,E),
  max(I,255,F),
  max(J,255,G)}.

max(A,B,A) :- {A>=B}.
max(A,B,B) :- {A=<B}.


combine_ref_shad(A,rgb(B,C,D),rgb(E,F,G),rgb(H,I,J)) :-
  {H = (A*B+(1-A)*E),
  I = (A*C+(1-A)*F),
  J = (A*D+(1-A)*G)}.

color_add(rgb(A,B,C),rgb(D,E,F),rgb(G,H,I)) :-
  {J = A+D,
  K = B+E,
  L = C+F,
	Top = 255},
  max(J,Top,G),
  max(K,Top,H),
  max(L,Top,I).

%% END ADDED BY ROBERTO BAGNARA


acc_calc_intersections([], _, _, _, A, A).
acc_calc_intersections([v(A,B)|C], D, E, F, G, H) :-
	calc_sh(B, I),
	acc_calc_intersections_1([v(A,B)|C], D, E, F, G, H,I).

acc_calc_intersections_1([v(A,B)|C], D, E, F, G, H,I) :-	
	{I=F},
	acc_calc_intersections(C, A, B, F, G, H).

acc_calc_intersections_1([v(A,B)|C], D, E, F, G, H,I) :-
	{I<F},
	acc_calc_intersections_2([v(A,B)|C], D, E, F, G, H,I).
acc_calc_intersections_1([v(A,B)|C], D, E, F, G, H,I) :-
	{I>F},
	acc_calc_intersections_2([v(A,B)|C], D, E, F, G, H,I).

acc_calc_intersections_2([v(A,B)|C], D, E, F, G, H,I) :-
	{A>0,D>0},
	{J=G+1},
	acc_calc_intersections(C, A, B, I, J, H).
acc_calc_intersections_2([v(A,B)|C], D, E, F, G, H,I) :-
	{A=<0},
	acc_calc_intersections_3([v(A,B)|C], D, E, F, G, H,I).
acc_calc_intersections_2([v(A,B)|C], D, E, F, G, H,I) :-
	{D=<0},
	acc_calc_intersections_3([v(A,B)|C], D, E, F, G, H,I).

acc_calc_intersections_3([v(A,B)|C], D, E, F, G, H,I) :-
	{A*D<0, D-E*(A-D)/(B-E)>0},
	{J=G+1},
	acc_calc_intersections(C, A, B, I, J, H).
acc_calc_intersections_3([v(A,B)|C], D, E, F, G, H,I) :-
	{A*D>=0},
	acc_calc_intersections(C, A, B, I, G, H).
acc_calc_intersections_3([v(A,B)|C], D, E, F, G, H,I) :-
	{D-E*(A-D)/(B-E)=<0},
	acc_calc_intersections(C, A, B, I, G, H).


append([], A, A).
append([A|B], C, [A|D]) :-
        append(B, C, D).

calc_intersections([v(A,B)|C], D) :-
	{G = 0},
	calc_sh(B, E),
	acc_calc_intersections(C, A, B, E, G, D).
calc_polygon_normal([A,B,C|_], D) :-
	vector_sub(B, A, E),
	vector_sub(C, B, F),
	cross_product(E, F, G),
	normalize_vector(G, D).

calc_sh(A, B) :- {A<0,B= -1}.
calc_sh(A, B) :- {A>=0,B=1}.


combine(A, B, C, D) :-
	{Col=0},
	combine_hits(A, B, C, rgb(Col,Col,Col), D).
combine_hits([], _, _, A, A).
combine_hits([hit|B], C, D, E, F) :-
	combine_hits(B, C, D, E, F).
combine_hits([ray(_,G)|B], C, D, E, F) :-
	vector_mult(C, G, H),
	vector_length(C, I),
	vector_length(G, J),
	{K = abs(H/(I*J))},
	scale_color(D, K, L),
	color_add(L, E, M),
	combine_hits(B, C, D, M, F).


cross_product(v(A,B,C), v(D,E,F), v(G,H,I)) :-
        {G = B*F-C*E,
        H = C*D-A*F,
        I = A*E-B*D}.

default(A, A) :- !.
default(_, _).


find_coord(A, B, C, D, E) :-
        vector_sub(B, A, F),
        vector_add(B, C, G),
        vector_sub(G, A, H),
        cross_product(F, H, I),
        vector_add(B, I, J),
        vector_sub(J, A, K),
        cross_product(K, F, L),
        normalize_vector(I, D),
        normalize_vector(L, E).
gen_light_rays([], _, []).
gen_light_rays([light(A)|B], C, [light(ray(C,F),G)|E]) :-
        vector_sub(A, C, H),
        vector_length(H, G),
        normalize_vector(H, F),
        gen_light_rays(B, C, E).

gen_ray_row(A, _, _, _, []) :- {A=0}.
gen_ray_row(A, B, C, D, [ray(B,E)|F]) :-
        vector_sub(C, B, G),
        normalize_vector(G, E),
        vector_sub(C, D, H),
        {I = A-1},
        gen_ray_row(I, B, H, D, F).
gen_ray_rows(A, _, _, _, _, _, []) :- {A=0}.
gen_ray_rows(A, B, C, D, E, F, [G|H]) :-
        {A>0},
        gen_ray_row(B, C, D, E, G),
        vector_sub(D, F, I),
        {J = A-1},
        gen_ray_rows(J, B, C, I, E, F, H).
gen_rays(A, B, C, D, E, F, G) :-
        gen_ray_rows(F, E, A, B, C, D, G).

get_map_color(A, B, map(size(C,D),E), surface(0,F)) :-
        {G = (C*A),
        H = (D*B)},
        my_nth(H, E, I),
        my_nth(G, I, F).

scale_color(rgb(A,B,C),D,rgb(E,F,G)) :-
	{H = (A*D),I = (B*D),J = (C*D), Top=255},
	max(H,Top,E),max(I,Top,F),max(J,Top,G).

combine_ref_shad(A,rgb(B,C,D),rgb(E,F,G),rgb(H,I,J)):-
	{H = (A*B+(1-A)*E),I = (A*C+(1-A)*F),J = (A*D+(1-A)*G)}.

color_add(rgb(A,B,C),rgb(D,E,F),rgb(G,H,I)) :-
	{J = A+D,K = B+E, L = C+F, Top=255},
	max(J,Top,G),max(K,Top,H),max(L,Top,I).

init_rays(camera(A,B,C,D,E,F,G), ray(A,X,X,Q,R,I,J,I)) :-
	{I = (E/G),J = (F/G)},
        vector_sub(B, A, K),
        normalize_vector(K, L),
        vector_scalar_mult(L, C, M),
        vector_add(A, M, N),
        find_coord(A, N, D, O, P),
        scale_vector_length(O, G, Q),
        scale_vector_length(P, G, R),
        {S = E/2,
        T = F/2},
        scale_vector_length(O, S, U),
        scale_vector_length(P, T, V),
        vector_add(N, U, W),
        vector_add(W, V, X).

init_rays(camera(A,B,C,D,E,F,G), H) :-
        vector_sub(B, A, K),
        normalize_vector(K, L),
        vector_scalar_mult(L, C, M),
        vector_add(A, M, N),
        find_coord(A, N, D, O, P),
        scale_vector_length(O, G, Q),
        scale_vector_length(P, G, R),
        S is E/2,
        T is F/2,
        scale_vector_length(O, S, U),
        scale_vector_length(P, T, V),
        vector_add(N, U, W),
        vector_add(W, V, X),
        H=ray(A,X,X,Q,R,I,J,I).


% intersect(Obj,Ray1,_,Surface,Ray2,Success)
intersect(sphere(A,B,C), ray(D,E), F, C, ray(G,H),y) :-
        vector_sub(B, D, I),
        vector_mult(I, I, J),
        vector_mult(I, E, K),
		  intersect_1(A,J,K,F),
        {abs(F)>1.0e-05},
        vector_scalar_mult(E, F, M),
        vector_add(D, M, G),
        vector_sub(G, B, N),
        {O = 1/A},
        vector_scalar_mult(N, O, H).

intersect(sphere_map(A,B,C,D,E,F), ray(G,H), I, J, ray(K,L),y) :-
        vector_sub(B, G, M),
        vector_mult(M, M, N),
        vector_mult(M, H, O),
		  intersect_1(A,N,O,I),
        {abs(I)>1.0e-05},
        vector_scalar_mult(H, I, Q),
        vector_add(G, Q, K),
        vector_sub(K, B, R),
        {S = 1/A},
        vector_scalar_mult(R, S, L),
        vector_mult(L, C, T),
        {U = acos(-1*T),
        V = U/3.141592654},
        vector_mult(L, D, W),
        {X = acos(W/sin(U))/(2*3.141592654)},
        vector_mult(E, L, Y),
		  intersect_2(Y,X,Z),
        get_map_color(Z, V, F, J).
intersect(plane(A,B,C), ray(D,E), F, C, ray(G,H),y) :-
        intersect_plane(A, B, E, D, F, G, H).
intersect(triangle(A,B,C,D,E,F), ray(G,H), I, F, ray(J,K),y) :-
        intersect_plane(D, E, H, G, I, J, K),
        vector_sub(J, A, L),
        vector_mult(C, C, M),
        vector_mult(B, B, N),
        vector_mult(L, L, O),
        {2*O<M+N},
        vector_mult(L, B, P),
        vector_mult(B, C, Q),
        vector_mult(L, C, R),
        {S =(M*P-Q*R)/(N*M-Q*Q),
        S>=0,
        S=<1,
        T =(R-S*Q)/M,
        T>=0,
        T=<1,
        S+T=<1}.
intersect(polygon(A,B,C,D,E), ray(F,G), H, E, ray(I,J),y) :-
        intersect_plane(A, B, G, F, H, I, J),
        map_point(D, I, K),
        translate_points(C, K, L),
        calc_intersections(L, M),
        odd(M).
intersect(triangle_map(A,B,C,D,E,F), ray(G,H), I, J, ray(K,L),y) :-
        intersect_plane(D, E, H, G, I, K, L),
        vector_sub(K, A, M),
        vector_mult(C, C, N),
        vector_mult(B, B, O),
        vector_mult(M, M, P),
        {2*P<N+O},
        vector_mult(M, B, Q),
        vector_mult(B, C, R),
        vector_mult(M, C, S),
        {T =(N*Q-R*S)/(O*N-R*R),
        T>=0,
        T=<1,
        U =(S-T*R)/N,
        U>=0,
        U=<1,
        T+U=<1},
        get_map_color(T, U, F, J).

intersect(_,_,_,_,_,n).


intersect_1(A,J,K,F) :-
	{J>=A*A,
	K>0,
	L = A*A-J+K*K,
	L>0,
	 SL*SL=L,
	F = K-SL}.
intersect_1(A,J,K,F) :-
	{J<A*A,
	 L = A*A-J+K*K,
	 	 SL*SL=L,
	 F = K+SL}.

intersect_2(Y,X,Z) :-
	{Y>0,Z=X}.
intersect_2(Y,X,Z) :-
	{Y=<0,Z=1-X}.

intersect_plane(A, B, C, D, E, F, G) :-
        vector_mult(A, C, H),
        {Zero=0.0},
		  not_eq(H,Zero),
        vector_mult(A, D, I),
        {J = B-I,
        E = J/H,
        E>0},
        vector_scalar_mult(C, E, K),
        vector_add(K, D, F),
		  intersect_plane_1(H,A,G).

intersect_plane_1(H,A,G) :-
	{H<0,G=A}.
intersect_plane_1(H,A,G) :-
	{H>=0,M_one= -1},  vector_scalar_mult(A, M_one, G).


not_eq(X,Y) :- {X>Y}.
not_eq(X,Y) :- {X<Y}.

is_digit(A) :-
        A=<[57],
        A>=[48].

load_obj(sphere_map(A,B,C,D,E), sphere_map(A,B,C,D,G,H)) :- 
        read_map(E, H),
        cross_product(C, D, G).
load_obj(plane(A,B,C), plane(D,B,C)) :-
        normalize_vector(A, D).
load_obj(triangle(v(K,L,M),B,C,D), triangle(v(K,L,M),B,C,v(H,I,J),F,D)) :-
        cross_product(B, C, G),
        normalize_vector(G, v(H,I,J)),
        {F = H*K+I*L+J*M}.
load_obj(triangle_map(v(L,M,N),B,C,D), triangle_map(v(L,M,N),B,C,v(I,J,K),F,G)) :-
        cross_product(B, C, H),
        normalize_vector(H, v(I,J,K)),
        {F = I*L+J*M+K*N},
        read_map(D, G).
load_obj(polygon([v(J,K,L)|A1],B), polygon(v(G,H,I),D,E,F,B)) :-
        calc_polygon_normal([v(J,K,L)|A1], v(G,H,I)),
        {D = G*J+H*K+I*L},
        polygon_dominator(v(G,H,I), F),
        append([v(J,K,L)|A1], [v(J,K,L)], M),
        project_points(M, F, E).
load_obj(A, A).

map_point(One, v(_,A,B), v(A,B)) :- {One=1}.
map_point(Two, v(A,_,B), v(A,B)) :- {Two=2}.
map_point(Three, v(A,B,_), v(A,B)) :- {Three=3}.

map_size([D|Ds], B, C) :-
        length([D|Ds], C),
        length(D, B).


next_ray(ray(A,B,C,D,E,F,G,H), ray(A,B,C,D,E,F,G,H), []) :-
	{F=0,K=G-1,K=0}.
next_ray(ray(A,B,C,D,E,F,G,H), ray(A,L,L,D,E,H,K,H), []) :-
	{F=0,K=G-1,Zero=0},not_eq(K,Zero),
	vector_sub(C, E, L).
next_ray(ray(A,B,C,D,E,F,G,H), ray(A,O,C,D,E,P,G,H), ray(A,N)) :-
	{F>0},
	vector_sub(B, A, M),
	normalize_vector(M, N),
	vector_sub(B, D, O),
	{P = F-1}.	
normalize_vector(A, B) :-
	{One=1},scale_vector_length(A, One, B).


my_nth(Zero, [B2|_], B1) :- {Zero=0,B1=B2}.
my_nth(A, [B2|B], B1) :- {D = A-1}, not_eq(B2,B1),
        my_nth(D, B, B1).

%odd(A) :- {Zero=0,Two=2}, mod(A,Two,B), 0=\=A mod 2, not_eq(B,Zero).
odd(Y) :- {X=(Y+1)/2,Z=0}, bb_inf([X],X,X,_,Z).


polygon_dominator(v(A,B,C), D) :-
	{AA=abs(A),AB=abs(B),AC=abs(C)},
	polygon_dominator_1(v(AA,AB,AC), D).

polygon_dominator_1(v(A,B,C), D) :-
	{A>B,A>C,D=1}.
polygon_dominator_1(v(A,B,C), D) :-
	{A=<B}, polygon_dominator_2(v(A,B,C), D).
polygon_dominator_1(v(A,B,C), D) :-
	{A=<C}, polygon_dominator_2(v(A,B,C), D).
polygon_dominator_2(v(A,B,C), D) :-
	{B>A,B>C,D=2}.
polygon_dominator_2(v(A,B,C), D) :-
	{B=<A,D=3}.
polygon_dominator_2(v(A,B,C), D) :-
	{B=<C,D=3}.


process_world([], []).
process_world([A|B], [C|D]) :-
        load_obj(A, C),
        process_world(B, D).
project_points([], _, []).
project_points([A|B], C, [D|E]) :-
        map_point(C, A, D),
        project_points(B, C, E).

rays(v(Zero,Zero,M_eight), v(Zero,Zero,Zero), v(Zero,One,Zero), D, E, F, G) :-
		{Zero=0,One=1,M_eight= -8,Ten=10,
		 H = (D/F),
		 I = (E/F)},
        find_coord(v(Zero,Zero,M_eight), v(Zero,Zero,Zero), v(Zero,One,Zero), J, K),
        scale_vector_length(J, F, L),
        scale_vector_length(K, F, M),
        {N = D/2,
        O = E/2},
        scale_vector_length(J, N, P),
        scale_vector_length(K, O, Q),
        vector_add(v(Zero,Zero,Zero), P, R),
        vector_add(R, Q, S),
        gen_rays(v(Zero,Zero,M_eight), S, L, M, H, I, G).


%%%read_chars(0, _, A) :- !,
%%%        A=[].
%%%read_chars(A, B, [C|D]) :-
%%%        A>0,
%%%        get0(B, C),
%%%        E = A-1,
%%%        read_chars(E, B, D).
%%%read_digit(A, B) :-
%%%        skip_nondigits(A, C),
%%%        read_digits(A, D),
%%%        number_chars(B, [C|D]).
%%%read_digits(A, B) :-
%%%        get0(A, C),
%%%        (   is_digit(C) ->
%%%            B=[C|D],
%%%            read_digits(A, D)
%%%        ;   B=[]
%%%        ).
%%%read_map(A, map(size(B,C),D)) :-
%%%        read_ppm(A, D),
%%%        map_size(D, B, C).
%%%read_ppm(A, B) :-
%%%        open(A, read, C),
%%%        read_ppm_stream(C, B),
%%%        close(C).
%%%read_ppm_ascii(0, _, A, _) :- !,
%%%        A=[].
%%%read_ppm_ascii(A, B, [C|D], E) :-
%%%        read_ppm_ascii_row(B, C, E),
%%%        F = A-1,
%%%        read_ppm_ascii(F, B, D, E).
%%%read_ppm_ascii_row(0, A, _) :- !,
%%%        A=[].
%%%read_ppm_ascii_row(A, [rgb(B,C,D)|E], F) :-
%%%        read_digit(F, B),
%%%        read_digit(F, C),
%%%        read_digit(F, D),
%%%        G = A-1,
%%%        read_ppm_ascii_row(G, E, F).
%%%read_ppm_raw(0, _, A, _) :- !,
%%%        A=[].
%%%read_ppm_raw(A, B, [C|D], E) :-
%%%        read_ppm_raw_row(B, C, E),
%%%        F = A-1,
%%%        read_ppm_raw(F, B, D, E).
%%%read_ppm_raw_row(0, A, _) :- !,
%%%        A=[].
%%%read_ppm_raw_row(A, [rgb(B,C,D)|E], F) :-
%%%        get0(F, B),
%%%        get0(F, C),
%%%        get0(F, D),
%%%        G = A-1,
%%%        read_ppm_raw_row(G, E, F).
%%%read_ppm_spec(A, B, C, D) :-
%%%        read_digit(D, A),
%%%        skip_to_nl(D),
%%%        read_digit(D, B),
%%%        read_digit(D, C),
%%%        read_digit(D, _).
%%%read_ppm_stream(A, B) :-
%%%        read_ppm_spec(C, D, E, A),
%%%        (   C=6 ->
%%%            read_ppm_raw(E, D, B, A)
%%%        ;   read_ppm_ascii(E, D, B, A)
%%%        ).


%%% world(Objs,Lights)
read_world(world(H,G), C) :-
	world(world(F,G)),
	camera(C),
	process_world(F, H).

reflection_ray(ray(_,A), ray(B,C), ray(B,D)) :-
        vector_mult(C, A, E),
        {F = 2*E},
        vector_scalar_mult(C, F, G),
        vector_sub(A, G, D).
scale_vector_length(v(A,B,C), D, v(E,F,G)) :-
        {T*T = A*A+B*B+C*C,
			H = D/T,
        E = A*H,
        F = B*H,
        G = C*H}.

shadow_hit_any([], B, _, B).
shadow_hit_any([E|F], B, C, hit) :-
   {G<C},intersect(E, B, G, _, _,y).
shadow_hit_any([E|F], B, C, D) :-
	intersect(E, B, G, _, _,n),
	shadow_hit_any(F, B, C, D).

size([H|T], B, C) :-
        my_length([H|T], C),
        my_length(H, B).

my_length([],Z) :- {Z=0}.
my_length([H|T],L) :- {L=L1+1}, my_length(T,L1).

skip_nondigits(A, B) :-
        get0(A, C),
        (   is_digit(C) ->
            B=C
        ;   skip_nondigits(A, B)
        ).
skip_to_char(A, B) :-
        get0(A, C),
        (   C=B ->
            true
        ;   skip_to_char(A, B)
        ).
skip_to_nl(A) :-
        skip_to_char(A, 10).

main :-
        read_world(D, E),
        init_rays(E, F),
        trace_rays(F, G, D),
		  writeq(D).

trace_light_rays([], _, []).
trace_light_rays([light(A,B)|C], D, [E|F]) :-
        shadow_hit_any(D, A, B, E),
        trace_light_rays(C, D, F).

trace_ray(A, B, C) :-
	{D=5},
	trace_ray(A, B, C, D).

trace_ray(A, world(B,C), D, E) :-
	{Z=0,Million=999999},
        F=rgb(0,0,0),
        trace_ray(A, B, Million, surface(Z,rgb(Z,Z,Z)), surface(G,H), nohit, ray(J,K),y),
		  trace_ray_2(A,E,G,J,K,I,N),
            (   {G=\=1} ->
                trace_shadow(J, K, B, C, H, O)
            ;
					 O=rgb(0,0,0)
            ),

				
            (   {G=0} ->
                D=O
            ;
					 {G=1} ->
                D=N
            ;
					 
					 combine_ref_shad(G, N, O, D)
            ).


trace_ray(A, world(B,C), rgb(Z,Z,Z), E) :-
	{Z=0,Million=999999},
	trace_ray(A, B, Million, surface(Z,rgb(Z,Z,Z)), surface(G,H), nohit, nohit,y).


trace_ray_2(A,E,G,J,K,I,N) :-
	{Z=0,E>0},
	not_eq(G,Z),
	{L = E-1},
	reflection_ray(A, ray(J,K), M),
	trace_ray(M, world(B,C), N, L).
trace_ray_2(A,E,G,J,K,I,rgb(Z,Z,Z)) :-
	{Z=0,E=<0}.
trace_ray_2(A,E,G,J,K,I,rgb(Z,Z,Z)) :-
	{Z=0,G=0}.

trace_ray_3(G,J, K, B, C, H, O) :-
	{One=1,Z=0},
	not_eq(G,One),
	trace_shadow(J, K, B, C, H, O).
trace_ray_3(G,J, K, B, C, H, rgb(Z,Z,Z)) :-
	{G=1,Z=0}.


trace_ray_4(G,N,O,D) :-
	{G=0,D=O}.
trace_ray_4(G,N,O,D) :-
	{G=1,D=N}.
trace_ray_4(G,N,O,D) :-
	{G<0},
	combine_ref_shad(G, N, O, D).
trace_ray_4(G,N,O,D) :-
	{G>1},
	combine_ref_shad(G, N, O, D).
trace_ray_4(G,N,O,D) :-
	{G>0,G<1},
	combine_ref_shad(G, N, O, D).



trace_ray(_, [], _, B, B, D, D,y).
trace_ray(A, [H|I], C, D, E, F, G,S) :-
	{J<C,
	 M=J},
	intersect(H, A, J, K, L,y),
	trace_ray(A, I, M, N, E, L, G,S).
trace_ray(A, [H|I], C, D, E, F, G,S) :-
	{M=C,N=D, O=F},
	intersect(H, A, J, K, L,n),
	trace_ray(A, I, M, N, E, O, G,S).
trace_ray(A, [H|I], C, D, E, F, G,S) :-
	{M=C,N=D, O=F,J>=C},
	intersect(H, A, J, K, L,y),
	trace_ray(A, I, M, N, E, O, G,S).
trace_ray(_,_,_,_,_,_,_,n).

	
trace_ray_rows([], _, [], _).
trace_ray_rows(A, B, [C|D], E) :-
        trace_row(A, B, F, C, E),
        next_ray(F, G, H),
        trace_ray_rows(H, G, D, E).
trace_rays(A, B, C) :-
        next_ray(A, D, E),
        trace_ray_rows(E, D, B, C).
trace_row([], A, A, [], _).
trace_row(A, B, C, [D|E], F) :-
        trace_ray(A, F, D),
        next_ray(B, G, H),
        trace_row(H, G, C, E, F).
trace_shadow(A, B, C, D, E, F) :-
	{Eps=0.0001},
        vector_scalar_mult(B, Eps, G),
        vector_add(A, G, H),
        gen_light_rays(D, H, I),
        trace_light_rays(I, C, J),
        combine(J, B, E, F).
translate_points([], _, []).
translate_points([A|B], C, [D|E]) :-
        vector_sub(A, C, D),
        translate_points(B, C, E).

vector_add(v(A,B,C), v(D,E,F), v(G,H,I)) :-
        {G = A+D,
        H = B+E,
        I = C+F}.
vector_length(v(A,B,C), D) :-
        {D*D = A*A+B*B+C*C}.
vector_length_sq(A, B) :-
        vector_mult(A, A, B).
vector_mult(v(A,B,C), v(D,E,F), G) :-
        {G = A*D+B*E+C*F}.
vector_scalar_mult(v(A,B,C), D, v(E,F,G)) :-
        {E = A*D,
        F = B*D,
        G = C*D}.
vector_sub(v(A,B,C), v(D,E,F), v(G,H,I)) :-
        {G = A-D,
        H = B-E,
        I = C-F}.
vector_sub(v(A,B), v(C,D), v(E,F)) :-
        {E = A-C,
        F = B-D}.


%%%write_ppm_ascii(A, B) :-
%%%        open(A, write, C),
%%%        write_ppm_stream(B, C),
%%%        close(C).
%%%write_ppm_bitmap_rows([], _).
%%%write_ppm_bitmap_rows([A|B], C) :-
%%%        write_ppm_row(A, C),
%%%        write_ppm_bitmap_rows(B, C).
%%%write_ppm_bitmap_rows_raw([], _).
%%%write_ppm_bitmap_rows_raw([A|B], C) :-
%%%        write_ppm_row_raw(A, C),
%%%        write_ppm_bitmap_rows_raw(B, C).
%%%write_ppm_raw(A, B) :-
%%%        open(A, write, C),
%%%        write_ppm_stream_raw(B, C),
%%%        close(C).
%%%write_ppm_row([], A) :-
%%%        nl(A).
%%%write_ppm_row([rgb(A,B,C)|D], E) :-
%%%        write(E, A),
%%%        write(E, ' '),
%%%        write(E, B),
%%%        write(E, ' '),
%%%        write(E, C),
%%%        write(E, ' '),
%%%        write_ppm_row(D, E).
%%%write_ppm_row_raw([], _).
%%%write_ppm_row_raw([rgb(A,B,C)|D], E) :-
%%%        put(E, A),
%%%        put(E, B),
%%%        put(E, C),
%%%        write_ppm_row_raw(D, E).
%%%write_ppm_stream(A, B) :-
%%%        size(A, C, D),
%%%        write(B, 'P3'),
%%%        nl(B),
%%%        write(B, '# CREATOR: Prolog ray tracer'),
%%%        nl(B),
%%%        format(B, '~w ~w~n255~n', [C,D]),
%%%        write_ppm_bitmap_rows(A, B).
%%%write_ppm_stream_raw(A, B) :-
%%%        size(A, C, D),
%%%        write(B, 'P6'),
%%%        nl(B),
%%%        write(B, '# CREATOR: Prolog ray tracer'),
%%%        nl(B),
%%%        format(B, '~w ~w ~n255~n', [C,D]),
%%%        write_ppm_bitmap_rows_raw(A, B).


world(world(A,B)) :-
	world(A,B).
world([  
	sphere(12,v(0,-5,116),surface(0,rgb(155,60,60))),
        sphere(8,v(-10,7,118),surface(0.8,rgb(10,160,10))),
        sphere(8,v(10,7,118),surface(0.8,rgb(10,160,10))),
       sphere(1.5,v(4,-2,105),surface(0,rgb(40,40,40))),
       sphere(1.5,v(-4,-2,105),surface(0,rgb(40,40,40))),
       sphere(2.5,v(0,-5,104),surface(0,rgb(90,40,40))),
       sphere(0.5,v(-6,-9,105),surface(0,rgb(0,0,0))),
       sphere(0.5,v(6,-9,105),surface(0,rgb(0,0,0))),
       sphere(0.5,v(0,-11,105),surface(0,rgb(0,0,0))),
       plane(v(-1,0,2),150,surface(0,rgb(60,60,250)))
	],[
       light(v(10,0,50)),
       light(v(-20,70,50))
      ]).
	
camera(camera(X1,X2,X3,X4,X5,X6,X7)) :-
	camera(X1,X2,X3,X4,X5,X6,X7).

camera(v(-30,50,100),v(-10,10,114),7,v(0,1,0),6,6,0.05).
