%%%mode(in_language,[+,+]).
%%%mode(accepts,[+,+]).
%%%mode(app,[-,-,+]).
%%%mode(ne_app,[-,-,+]).
%%%mode(symbol,[+]).
%%%mode(string,[+]).



in_language(E,S) :- string(S), accepts(E,S).

%accepts( eq(S1) ,[S2]) :- {S2 =:= S1}.
accepts( lt(S1) ,[S2]) :- {S2 < S1}.
%accepts( leq(S1) ,[S2]) :- {S2 =< S1}.
accepts( gt(S1) ,[S2]) :- {S2 > S1}.
%accepts( geq(S1) ,[S2]) :- {S2 >= S1}.
accepts( int(A,B) ,[S2]) :- {S2 > A, S2 < B}.

accepts( (E1,E2), S) :-
        app(S1,S2,S),
        accepts(E1,S1),
        accepts(E2,S2).

accepts( E1+E2, S) :- accepts(E1,S).
accepts( E1+E2, S) :- accepts(E2,S).

accepts( star(E), []).

%accepts( star(E), S) :-     This tautology can be discarded
%        accepts(E,[]),
%        accepts(star(E),S).

accepts( star(E), S) :-
        ne_app(S1,S2,S),
        accepts(E,S1),
        accepts(star(E),S2).

app([],L,L).
app([A|L1],L2,[A|L3]):-
        app(L1,L2,L3).

ne_app([A],L,[A|L]).
ne_app([A|L1],L2,[A|L3]) :-
        ne_app(L1,L2,L3).

string([]).
string([_|S]) :- string(S).




%%% solver:clpq
%%% uf: leftmost_consumer
%%% wqounf: none

%%% INITIAL DEFINITION




in_lang1(W) :- {U=:=1, D=:=2}, in_language(((gt(U),star(gt(U)))+(gt(U),(star(gt(U)),lt(D)))),W).

sp_in_lang1(A) :- newinlang1_2(A).
newinlang1_2([A|B]) :- {A>1}, newinlang1_3(B).
newinlang1_3([]).
newinlang1_3([A]) :- {A<2}.
newinlang1_3([A|B]) :- {A>1}, newinlang1_4(B).
newinlang1_4([]).
newinlang1_4([A]) :- {A<2}.
newinlang1_4([A|B]) :- {A>1}, newinlang1_4(B).

%%% queries
/*
{D=:=2, T=:=3}, in_lang1([T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,D]).

% above query takes 2560 ms (repeated 1000 times)
% specialized version takes 1920ms
% Speedup 1.33


{U=:=1, T=:=3}, in_lang1([T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,U]).



% above query takes 6660 ms (repeated 10 times)
% specialized version takes 20ms
% Speedup 333
*/			

in_lang2(W) :- {U=:=1, D=:=2, T=:=3}, in_language(( star(int(U,D)) + star(int(U,T) )),W).


sp_in_lang2(A) :- newinlang2_2(A).
newinlang2_2([]).
newinlang2_2([A|B]) :- {A>1, A<2}, newinlang2_3(B).
newinlang2_2([A|B]) :- {A>=2, A<3}, newinlang2_4(B).
newinlang2_3([]).
newinlang2_3([A|B]) :- {A>1, A<2}, newinlang2_3(B).
newinlang2_3([A|B]) :- {A>=2, A<3}, newinlang2_4(B).
newinlang2_4([]).
newinlang2_4([A|B]) :- {A>1, A<3}, newinlang2_4(B).


/*

{D=:=1.9}, in_lang2([D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,D]).



% above query takes 3390ms  (repeated 1000  times)
% specialized version takes 2670ms
% Speedup 1.27

{T=:=2.3, D=:=1.9}, in_lang2([D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,T]).



% above query takes 2520ms  (repeated 100  times)
% specialized version takes 280ms
% Speedup 9
*/


in_lang3(W) :- {U=:=1, D=:=2}, in_language(((gt(U),star(gt(U))+(gt(U),(star(gt(U)),lt(D))))),W).

sp_in_lang3(A) :- newinlang3_2(A).
newinlang3_2([A|B]) :- {A>1}, newinlang3_3(B).
newinlang3_3([]).
newinlang3_3([A|B]) :- {A>1}, newinlang3_4(B).
newinlang3_4([]).
newinlang3_4([A]) :- {A<2}.
newinlang3_4([A|B]) :- {A>1}, newinlang3_5(B).
newinlang3_5([]).
newinlang3_5([A]) :- {A<2}.
newinlang3_5([A|B]) :- {A>1}, newinlang3_5(B).



in_lang4(W) :- {Z=:=0, U=:=1, D=:=2, T=:=3}, in_language(( star(int(Z,D)) + star(int(U,T) )),W).

sp_in_lang4(A) :- newinlang4_2(A).
newinlang4_2([]).
newinlang4_2([A|B]) :- {A>1, A<2}, newinlang4_3(B).
newinlang4_2([A|B]) :- {A>=2, A<3}, newinlang4_4(B).
newinlang4_2([A|B]) :- {A>0, A=<1}, newinlang4_5(B).
newinlang4_5([]).
newinlang4_5([A|B]) :- {A>0, A<2}, newinlang4_5(B).
newinlang4_4([]).
newinlang4_4([A|B]) :- {A>1, A<3}, newinlang4_4(B).
newinlang4_3([]).
newinlang4_3([A|B]) :- {A>1, A<2}, newinlang4_3(B).
newinlang4_3([A|B]) :- {A>0, A=<1}, newinlang4_5(B).
newinlang4_3([A|B]) :- {A>=2, A<3}, newinlang4_4(B).



/*

{D=:=1.9}, in_lang4([D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,D]).



% above query takes 3820 ms  (repeated  1000 times)
% specialized version takes 3310  ms
% Speedup 1.15

{T=:=2.3, D=:=1.9}, in_lang4([D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,
		 D,D,D,D,D,D,D,D,D,D,T]).



% above query takes 2620 ms  (repeated 100  times)
% specialized version takes 350 ms
% Speedup 7.49
*/




in_lang5(W) :- {Z=:=0, U=:=1}, in_language(((gt(U),star(gt(U))+(gt(U),(star(gt(U)),lt(Z))))),W).

sp_in_lang5(A) :- newinlang5_2(A).
newinlang5_2([A|B]) :- {A>1}, newinlang5_3(B).
newinlang5_3([]).
newinlang5_3([A|B]) :- {A>1}, newinlang5_4(B).
newinlang5_4([]).
newinlang5_4([A]) :- {A<0}.
newinlang5_4([A|B]) :- {A>1}, newinlang5_5(B).
newinlang5_5([]).
newinlang5_5([A]) :- {A<0}.
newinlang5_5([A|B]) :- {A>1}, newinlang5_5(B).


%%% queries
/*
{D=:=2, T=:=3}, in_lang5([T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,D]).

% above query takes  2580 ms (repeated 1000 times)
% specialized version takes 1910 ms
% Speedup 1.35


{MU=:= -1, T=:=3}, in_lang5([T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,
		 T,T,T,T,T,T,T,T,T,T,MU]).



% above query takes 6490 ms (repeated 10 times)
% specialized version takes 20 ms
% Speedup 324
*/			