:- use_module(library(lists), [ append/3,
				memberchk/2,
				member/2,
				select/3
			      ]).
:- use_module(library(terms), [ term_variables/2,
				term_subsumer/3]).


time(G,T) :-
	time(G,T,_,_,_).

time(Goal,T,M,P,C):-
        statistics(runtime,[T1,_]),
        Goal,
        statistics(runtime,[T2,_]),
        statistics(memory,[M,_]),
        statistics(program,[P,_]),
        statistics(choice,[C,_]),
        %T is ((T2-T1)//100)/10.
        T is T2-T1.

% ptime(G) G is a goal
%
ptime(G) :- G.
%%%ptime(G):-
%%%  time(G,T,_,_,_),
%%%  nl,
%%%  functor(G,F,_),write(F),write(' execution time: '),write(T),
%%%  nl.


min(X,Y,Z) :- (X=<Y ->
		  Z=X
	      ;
		  Z=Y).





% conj_to_list(S,L): S a conjunction of the form (A1,...,An), and L is the list [A1,...,An].
conj_to_list(B,L) :-
   ( (nonvar(B), functor(B,',',_)) ->
       (   B = (B1,B2),
	   L=[B1|H], conj_to_list(B2,H))
   ;   L=[B]).

%constr_conj_to_list({B},L) :- conj_to_list((B),L).


count_from_to(X,X,[X]).
count_from_to(X,Y,[X|Xs]) :-
	X2 is X + 1,
	count_from_to(X2,Y,Xs).

pos(X,[Y|_],1) :- X==Y, !.
pos(X,[_|L],M) :- pos(X,L,N), M is N+1.

% split(L,N,L1,E,L2):
% N is a positive integer. E is the element having position N
% in L, and L is the concatenation of L1, [E], and L2.

split(L,N,L1,E,L2) :-
	M is N-1, length1(L1,M), append(L1,[E|L2],L), nonvar(E),!.
split(L,N,L1,E,L2) :- M is N+1, split(L,M,L1,E,L2).

% length1(L,N):
% N is a not negative integer. L is a list made out of N variables.
length1([],0).
length1([_|L],N) :- N>0, N1 is N-1, length1(L,N1).



% replace_in_list(L1,L2,L3,L4):
% L4 is the result of replacing in L2 the sublist L1 by L3.
replace_in_list([],P,Ns,P1) :-
	append(P,Ns,P1).
replace_in_list( [E|L], P, Ns, P1):-
  delete_sublist( L, P, P2), replace_elt_in_list_(E, P2, Ns, P1).

replace_elt_in_list_(E,[E|P2], Ns, P):-	!, append(Ns,P2,P).
replace_elt_in_list_(E,[F|P2], Ns, [F|P]):- replace_elt_in_list_(E, P2, Ns, P).


% delete_sublist(L1,L2,L):
% L is the result of deleting from L2 the first occurrence of the elements
% which occurr also in L1. 
% Ex: delete_sublist([a,b],[c,a,b,a,c,b],[c,a,c,b]) is true.
delete_sublist([],L,L).
delete_sublist([X|R],L,L1) :- del_occ(X,L,L2), delete_sublist(R,L2,L1).


%del_occ(X,[X|L],L) :- !.
%del_occ(X,[Y|L],[Y|L1]) :- del_occ(X,L,L1).


del_occ(E,L,L2) :-
	select(E,L,L2).

%%%del_occ(_,[],[]).
%%%del_occ(X,[Y|L],L1) :-
%%%	(X==Y ->
%%%		 L2==L
%%%	;
%%%		 del_occ(X,L,L1)
%%%	).






% delete_elts_sublist(L1,L2,L):
% L is the result of deleting from L2 all the occurrences of the elements
% which occurr also in L1. 
% Ex: delete_elts_sublist([a,b],[c,a,b,a,c,b],[c,c]) is true.
delete_elts_sublist([],L,L).
delete_elts_sublist([X|R],L,L1) :- del_occs(X,L,L2), delete_elts_sublist(R,L2,L1).

del_occs(_,[],[]).
del_occs(X,[Y|L],L2) :- X==Y, !, del_occs(X,L,L2).
del_occs(X,[Y|L],[Y|L1]) :- del_occs(X,L,L1).



% This sort does not remove duplicates.
sort_list([],[]).
sort_list([A],[A]) :- !.
sort_list([A|L],L2) :- sort_list(L,L3), ord_add_elt_list(L3,A,L2).

ord_add_elt_list([], Element, [Element]).
ord_add_elt_list([Head|Tail], Element, Set) :-
   Head=.. [FHead|_],    Element=.. [FElement|_], 
   compare(Order, FHead, FElement), ord_add_elt_list(Order, Head, Tail, Element,
 Set).
 
ord_add_elt_list(<, Head, Tail, Element, [Head|Set]) :- !,
        ord_add_elt_list(Tail, Element, Set).
ord_add_elt_list(>, Head, Tail, Element, [Element,Head|Tail]) :- !.
ord_add_elt_list(_, Head, Tail, Element, [Head|Set]) :-
        ord_add_elt_list(Tail, Element, Set).
ord_add_elt_list(_, Head, Tail, Element, [Element,Head|Tail]).


append(L1,L2,L3,R) :- append(L2,L3,Z), append(L1,Z,R).


% F: append/2 flattens one level
append([], []).
append([L|Ls], List0) :- append(L, List1, List0), append(Ls, List1).


lists_to_pairs([],[],[]).
lists_to_pairs([X|Xs],[Y|Ys],[(X,Y)|Zs]) :-
	lists_to_pairs(Xs,Ys,Zs).

lists_to_assoc_pairs([],[],[]).
lists_to_assoc_pairs([X|Xs],[Y|Ys],[X-Y|Zs]) :-
	lists_to_assoc_pairs(Xs,Ys,Zs).


append_pairs([],[]).
append_pairs([(X,Y)|XYs],[Z|Zs]) :-
	append(X,Y,Z),
	append_pairs(XYs,Zs).



change_arg(T1,Old,New,NewTree) :-
	change_arg(T1,Old,New,NewTree,_).
change_arg(T1,Old,New,NewTree,Max) :-
	change_arg(T1,Old,New,NewTree,0,_,Max).


change_arg(Tree,_,_,NewTree,Occ,_,Max) :-
	Occ==Max, !,
	Tree=NewTree.
	
change_arg(Tree,Old,New,NewTree,Occ,NewOcc,_) :-
	Tree==Old, !,
	NewTree=New,
	NewOcc is Occ+1.
	
change_arg(Tree,Old,New,NewTree,Occ,NewOcc,Max) :-
	compound(Tree), !,
	same_functor(Tree,NewTree,N),
	change_arg(Tree,Old,New,NewTree,Occ,NewOcc,Max,N).
change_arg(Tree,_,_,Tree,Occ,Occ,_).

change_arg(_,_,_,_,Occ,Occ,_,0) :- !.
change_arg(Tree,Old,New,NewTree,Occ,NewOcc,Max,N) :-
	arg(N,Tree,ArgTree),
	arg(N,NewTree,ArgNewTree),
	change_arg(ArgTree,Old,New,ArgNewTree,Occ,Occ2,Max),
	N2 is N-1,
	change_arg(Tree,Old,New,NewTree,Occ2,NewOcc,Max,N2).

	
same_functor(T1,T2,N) :-
	functor(T1,F,N),
	functor(T2,F,N).




subterm(X,T,N) :- N > 0,
   arg(N,T,Arg),
   subterm(X,Arg).
subterm(X,T,N) :- N > 0, N2 is N-1, subterm(X,T,N2).

subterm(X,Y) :- var(Y), !, X==Y.
subterm(X,T) :-
   T=..[_|L],
   subterm_l(X,L).
subterm_l(X,[Y|_]) :- subterm(X,Y).
subterm_l(X,[_|L]) :- subterm_l(X,L).

%% Precondition: Pos is an increasing list of integers
%%
%% ord_positions_to_elts(L,Pos,L2,Rest) holds iff
%% L2 is the list obtained by selecting the elements of L
%% occurring at the positions "contained" in the list Pos and Rest is the rest of L.
ord_positions_to_elts(L,Pos,L2,Rest) :-
	ord_positions_to_elts(L,1,Pos,L2,Rest).

ord_positions_to_elts(L,_,[],[],L) :-!.
ord_positions_to_elts([EX|L],X,[P|Pos],[EX|L2],Rest) :-
	X==P, !,
	X2 is X+1,
	ord_positions_to_elts(L,X2,Pos,L2,Rest).
ord_positions_to_elts([EX|L],X,Pos,L2,[EX|Rest]) :-
	X2 is X+1,
	ord_positions_to_elts(L,X2,Pos,L2,Rest).

duplicate_elts_at_positions(B,L,B1) :-
	duplicate_elts_at_positions(B,1,L,B1).

duplicate_elts_at_positions(B,_,[],B) :- !.
duplicate_elts_at_positions([EX|L],X,[P|Pos],[EX,EX|L2]) :-
	X==P, !,
	X2 is X+1,
	duplicate_elts_at_positions(L,X2,Pos,L2).
duplicate_elts_at_positions([EX|L],X,Pos,[EX|L2]) :-
	X2 is X+1,
	duplicate_elts_at_positions(L,X2,Pos,L2).

%%% int_var(T,TL,V): V is the list of variables occurring in TL and not occurring in T.
int_var(T,TL,V) :- free_variables(T,[],[],V1), free_variables(TL,V1,[],V).

%%% vars(LV) = vars(T1) \cap (intersection) vars(T2)
linking_variables(T1,T2,LV) :- int_var(T2,T1,V1), int_var(V1,T1,LV).




%%% common_vars/3 similar to linking_variables/3
%%% but more efficient?
common_vars(L1,L2,L3) :-
	list_intersection(L1,L2,L3).


list_intersection([],_,[]).
list_intersection([X|Xs],Ys,[X|Zs]):-   
    belongs(X,Ys), !,
    list_intersection(Xs,Ys,Zs).   
list_intersection([_|Xs],Ys,Zs):-   
    list_intersection(Xs,Ys,Zs).   





intersection_vars_empty(L1,L2) :- member(E,L1), member(F,L2), E==F, !, fail.
intersection_vars_empty(_,_).

member_rest(X,[X|L],L).
member_rest(X,[Y|L],[Y|L2]) :- member_rest(X,L,L2).


%%% list_difference(A,B,D): D = A \setminus B
%%% 
list_difference([A|As],[B|Bs],D) :-
	!,
	( belongs(A,[B|Bs]) ->
		 list_difference(As,[B|Bs],D)
	;
		 D=[A|D1],
		 list_difference(As,[B|Bs],D1)
	).
list_difference(A,[],A) :- !.
list_difference([],_,[]).

vars_seteq(L1,L2) :-
	vars_subseteq(L1,L2), vars_subseteq(L2,L1).

vars_subseteq(X,Y) :- belong(X,Y).


% belongs(E,L): E is an element of list L.
belongs(X,[Y|_]) :- X==Y, !.
belongs(X,[_|T]) :- belongs(X,T).

belong([],_).
belong([X|Xs],L) :- belongs(X,L), belong(Xs,L).




%Borrowed from Quintus libraries

%From library(occurs): 




free_variables(Term, Bound, Vars0, Vars) :-
	term_variables(Term,T),
	term_variables(Bound,B),	
	list_difference(T,B,Vars1),
	(Vars0 == [] -> 
	    Vars = Vars1
	;
	    %%% Vars0 is a term, not necessarily a list
	    term_variables((Vars0,Vars1),Vars)
	).


%%%%   free_variables(+Goal, +Bound, +Vars0, -Vars) binds Vars to the union of Vars0 with
%%%%   the set of *FREE* variables in Goal, that is the set of variables which are captured
%%%%   neither by Bound nor by any internal quantifiers or templates in Goal.


%%%free_variables(Term, Bound, Vars0, Vars) :-
%%%   ( nonvar(Term) ->
%%%		 free_variables_1(Term, Bound, Vars0, Vars) 
%%%   ;
%%%		 term_is_free_of(Bound, Term),
%%%		 (list_is_free_of(Vars0, Term) ->
%%%			  Vars = [Term|Vars0]
%%%		 ;
%%%			  Vars = Vars0)
%%%	).

%%%free_variables_1((Conjunct,Conjuncts), Bound) --> !,
%%%   free_variables(Conjunct,  Bound),	free_variables(Conjuncts, Bound).
%%%free_variables_1((Disjunct ; Disjuncts), Bound) --> !,
%%%   free_variables(Disjunct,  Bound), free_variables(Disjuncts, Bound).
%%%free_variables_1(NormalGoal, Bound) --> data_variables(NormalGoal, Bound).
%%%%   data_variables(+Term, +Bound, +Vars0, -Vars) binds Vars to the union of Vars0 with the
%%%%   set of variables in Term which do not occur in (are not "captured by") Bound.
%%%%   When the Bound contains no variables, it delivers the same results as
%%%% term_variables(Term, Vars0, Vars).
%%%data_variables(Term, Bound, Vars0, Vars) :-
%%%   ( nonvar(Term) -> functor(Term, _, N), data_variables(N, Term, Bound, Vars0, Vars)
%%%   ; term_is_free_of(Bound, Term), list_is_free_of(Vars0, Term) ->  Vars = [Term|Vars0]
%%%   ;   Vars = Vars0 ).

%%%data_variables(N, Term, Bound) -->
%%%   (  { N =:= 0 } -> [] ; { arg(N, Term, Arg), M is N-1 },
%%%   data_variables(Arg, Bound), data_variables(M, Term, Bound)).

%%%%   term_is_free_of(+Term, +Var) is a meta-logical predicate which is true when the variable
%%%%   Var does not occur anywhere in the term Term.  It is used when the Term is a tree
%%%% built from all the existential quantifiers and Templates dominating
%%%%   (the goal containing) this variable.
%%%term_is_free_of(Term, Var) :-
%%%   ( var(Term) -> Term \== Var ; functor(Term, _, N), term_is_free_of(N, Term, Var) ).
%%%term_is_free_of(N, Term, Var) :-
%%%   ( N =:= 0 -> true ; arg(N, Term, Arg), term_is_free_of(Arg, Var), M is N-1,
%%%    term_is_free_of(M, Term, Var) ).

%%%%   list_is_free_of(+Vars0, +Var) is a meta-logical predicate which is true when the
%%%%   variable Var is not an element of Vars0, which is known to be a list of variables.
%%%%   It is used when Vars0 is the set of free variables which have been built up so far,
%%%%   and we are considering whether Var should be added to this set.
%%%list_is_free_of([], _).
%%%list_is_free_of([Head|Tail], Var) :- Head \== Var, list_is_free_of(Tail, Var).


%%% restrict_to_vars(L1,T,L2)
%%% L2 contains the elements of L1 whose variables are contained
%%% in those of T
restrict_to_vars([],_T,[]).
restrict_to_vars([X|Xs],T,L2) :-
	free_variables(X,T,[],ExtraVars),
	(ExtraVars == [] ->
	    L2 = [X|L3]
	;
	    L2 = L3
	),
	restrict_to_vars(Xs,T,L3).
	





flatten([],[]).
flatten([[]|L], L2) :- !, flatten(L,L2).
flatten([[E|L]|L2], L3) :-
  !, flatten([E|L],EL), flatten(L2,L4), append(EL,L4,L3).
flatten([E|L], [E|L2]) :- flatten(L,L2).

insert_once(E,L,L) :-
	member(E,L), !.
insert_once(E,L,[E|L]).

insert_after_elt(E1,E2,[E2|L],[E2,E1|L]) :- !.
insert_after_elt(E1,E2,[E|L],[E|L2]) :-
	insert_after_elt(E1,E2,L,L2).



insert_at_position([X],_,[],[X]) :- !.
insert_at_position(L1,1,L2,L3) :- !,
	append(L1,L2,L3).
insert_at_position(L1,N,[E|L2],[E|L3]) :-
	N2 is N-1,
	insert_at_position(L1,N2,L2,L3).



%From library(sets): 
seteq(Set1, Set2) :- subset(Set1, Set2), subset(Set2, Set1).
subset([], _).
subset([Element|Elements], Set) :- memberchk(Element, Set),  subset(Elements, Set).

subtract([], _, [], []).
subtract([Element|Elements], Set, [Element|Del], Dif) :-
     memberchk(Element, Set), !, subtract(Elements, Set, Del, Dif).
subtract([Element|Elements], Set, Del, [Element|Dif]) :- subtract(Elements, Set, Del, Dif).

move_first_pos(E1,L,[E1|NewL]) :- del_occ(E1,L,NewL).

move_after_elt(E1,E2,[E1|L],NewL) :- !, insert_after_elt(E1,E2,L,NewL).
move_after_elt(E1,E2,[E2|L],[E2,E1|NewL]) :- !, del_occ(E1,L,NewL).
move_after_elt(E1,E2,[E|L],[E|NewL]) :- move_after_elt(E1,E2,L,NewL).

%% make_sublist_count(N,Length,Pos) holds iff
%% Pos is a list of Length distinct integers in the interval 1..N
%% N must be ground, and if Length is not ground then it returns only the empty string
%% Ex.
%% make_sublist_count(3,L,P). L = 0,P = []
%% bagof(P,make_sublist_count(3,2,P),Ps). Ps = [[2,3],[1,3],[1,2]] 
%% make_sublist_count(N,L,[1,2]). {INSTANTIATION ERROR: _62 is _59-1 - arg 2}

make_sublist_count(N,Length,Pos) :-
	make_list_count(N,Pos2), % Pos2 is the list [1,..,N]
	sublist(Pos2,Length,Pos). % Pos is  a sublist of Pos2 of length Length


%% sublist(+L,+N,-Pos) holds iff Pos is a sublist (not necessarily contiguous) of L of length N
%% Ex.
%% bagof(L,sublist([1,a,4],2,L),LL) returns LL = [[a,4],[1,4],[1,a]]

sublist(_,0,[]) :-!.
sublist([_|L],N,Pos) :-
	sublist(L,N,Pos).
sublist([E|L],N,[E|Pos]) :-
	N2 is N-1,
	sublist(L,N2,Pos).
	

%% make_list_count(N,Pos) holds iff Pos is the list [1,...,N]
%% Ex.
%%  make_list_count(N,P) returns N = 1, P = [1] 
%%  make_list_count(5,P) returns P = [1,2,3,4,5]
make_list_count(N,Pos) :-
	make_list_count(1,N,Pos).

make_list_count(N,N,[N]):-!.
make_list_count(N,Max,[N|Pos]) :-
	N2 is N+1,
	make_list_count(N2,Max,Pos).

% from Quintus library
print_length(Command, Length) :-
   open_null_stream(Stream), 'print length'(Command, user, Stream),
   character_count(Stream, CharsWritten), line_count(Stream, LineNumber), close(Stream),
%	LineNumber = 1,
	LineNumber = 0,      %%%%% Quintus -> Sicstus
	Length = CharsWritten.

'print length'(Module:Command, _, Stream) :- !, 'print length'(Command, Module, Stream).
'print length'(write(X), _, Stream) :- !, write(Stream, X).
'print length'(writeq(X), _, Stream) :- !, writeq(Stream, X).
'print length'(display(X), _, Stream) :- !, write_canonical(Stream, X).
'print length'(write_canonical(X), _, Stream) :- !, write_canonical(Stream, X).
'print length'(print(X), _, Stream) :- !, print(Stream, X).
'print length'(format(F,L), _, Stream) :- !, format(Stream, F, L).
'print length'(Command, Module, Stream) :-
   current_output(OldOutput), set_output(Stream),
   (   Module:Command -> true
   ;   set_output(OldOutput), close(Stream), fail ), set_output(OldOutput).






anti_unify_list([T|Ts],G) :-
	anti_unify_list(Ts,T,G).

anti_unify_list([],T,T).
anti_unify_list([T|Ts],InT,OutT) :-
	anti_unify(T,InT,OutT1),
	anti_unify_list(Ts,OutT1,OutT).



%From library(anti_unify): 
/* Anti-unification is the mathematical dual of unification:  given two terms T1 and T2
it returns the most specific term which generalises them, T.  T is the most specific term
which unifies with both T1 and T2.  A common use for this is in learning; the idea of using
it that way comes from Gordon Plotkin. The code here is based on a routine called
'generalize/5' written by Fernando Pereira. The name was changed because there are other
ways of generalising things, but there is nly one dual of unification.
*/
%   anti_unify(+Term1, +Term2, -Term)
%   binds Term to a most specific generalisation of Term1 and Term2.
%   When you call it, Term should be a variable.
%%% anti_unify(Term1, Term2, Term) :- anti_unify(Term1, Term2, _, _, Term).
anti_unify(Term1, Term2, Term) :- term_subsumer(Term1, Term2, Term).

%   anti_unify(+Term1, +Term2, -Subst1, -Subst2, -Term)
%   binds Term to a most specific generalisation of Term1 and Term2,
%   and Subst1 and Subst2 to substitutions such that
%       Subst1(Term) = Term1
%       Subst2(Term) = Term2
%   Substitutions are represented as lists of Var=Term pairs, where
%   Var is a Prolog variable, and Term is the term to substitute for Var.
%   When you call it, Subst1, Subst2, and Term should be variables.
anti_unify(Term1, Term2, Phi1, Phi2, Term) :-
	anti_unify(Term1, Term2, [], [], Phi1, Phi2, Term).
anti_unify(Term1, Term2, Phi1, Phi2, Theta1, Theta2, Term) :-
   (nonvar(Term1), nonvar(Term2), functor(Term1, F, N), functor(Term2, F, N) ->
    functor(Term, F, N), anti_unify(N, Term1, Term2, Phi1, Phi2, Theta1, Theta2, Term)
   ; Term1 == Term2 -> Theta1 = Phi1, Theta2 = Phi2, Term = Term1
   ; already_covered(Phi1, Phi2, Term1, Term2, V) -> Theta1 = Phi1, Theta2 = Phi2, Term = V
   ;   /* Not same functor, not identical, not covered by an existing V */
    Theta1 = [V = Term1|Phi1], Theta2 = [V = Term2|Phi2], Term = V ).

anti_unify(0, _, _, SS, ST, SS, ST, _) :- !.
anti_unify(N, S, T, SS0, ST0, SS, ST, U) :-
   arg(N, S, Sn), arg(N, T, Tn), arg(N, U, Un),
   anti_unify(Sn, Tn, SS0, ST0, SS1, ST1, Un),
   M is N-1, anti_unify(M, S, T, SS1, ST1, SS, ST, U).

already_covered([V=S0|_], [V=T0|_], S, T, V) :- S0 == S, T0 == T, !.
already_covered([_|Phi1], [_|Phi2], S, T, V) :-	already_covered(Phi1, Phi2, S, T, V).


%From library(occurs): 
%   free_of_var(+Variable, +Term)
%   is true when the given Term contains no sub-term identical to the
%   given Variable (which may actually be any term, not just a var).
%   For variables, this is precisely the "occurs check" which is
%   needed for sound unification.
free_of_var(Variable, Term) :-	Term == Variable, !, fail.
free_of_var(Variable, Term) :- 
   compound(Term), !, functor(Term, _, Arity), free_of_var(Arity, Term, Variable).
free_of_var(_, _).
free_of_var(1, Term, Var) :- !, arg(1, Term, Arg), free_of_var(Var, Arg).
free_of_var(N, Term, Var) :-
   arg(N, Term, Arg), free_of_var(Var, Arg), M is N-1, !, free_of_var(M, Term, Var).


%   occurrences_of_var(+Term, +Variable, ?Tally)
%   is true when the given Variable occurs exactly Tally times in
%   Term.  It can only be used to calculate or test Tally;
%   to enumerate Variables you'll have to use sub_term/2 and then
%   test them with this routine.  If you just want to find out
%   whether Variable occurs in Term or not, use contains_var/2
%   or free_of_var/2.  This used to be called occ/3.
occurrences_of_var(Variable, Term, Occurrences) :-
   occurrences_of_var(Term, Variable, 0, Tally), Occurrences = Tally.
occurrences_of_var(Term, Var, SoFar, Tally) :- Term == Var, !, Tally is SoFar+1.
occurrences_of_var(Term, Variable, SoFar, Tally) :-
   compound(Term), !, functor(Term, _, Arity),
   occurrences_of_var(Arity, Term, Variable, SoFar, Tally).
occurrences_of_var(_, _, Tally, Tally).
occurrences_of_var(1, Term, Variable, SoFar, Tally) :- !,
   arg(1, Term, Argument), occurrences_of_var(Argument, Variable, SoFar, Tally).
occurrences_of_var(N, Term, Variable, SoFar, Tally) :-
   arg(N, Term, Argument), occurrences_of_var(Argument, Variable, SoFar, Accum),
   M is N-1, occurrences_of_var(M, Term, Variable, Accum, Tally).



/*
msg(A,B,C) : C is the least common anti-instance 
   (generalisation) of A and B.  
   Algorithm from Reynolds, Machine Intelligence 5
*/ 

msg(T1,T2,T3) :-
	msg1(T1,T2,T3,_).

msg1(T1,T2,T3,S) :-
	msgDiffer(T1,T2),
	msgOccur(subst(T1,T2,T3),S),
	!.
msg1(T1,T2,T3,S) :-
	nonvar(T1),
	nonvar(T2),
	!,
	T1 =.. [F|Args1],
	T2 =.. [F|Args2],
	msgArgs(Args1,Args2,Args3,S),
	T3 =.. [F|Args3].
msg1(T1,_,T1,_).

msgArgs([X|Xs],[Y|Ys],[Z|Zs],S) :-
	msg1(X,Y,Z,S),
	msgArgs(Xs,Ys,Zs,S).
msgArgs([],[],[],_).

msgDiffer(X,Y) :-
	var(X),
	!,
	X \== Y.
msgDiffer(X,Y) :-
	var(Y),
	!,
	X \== Y.
msgDiffer(X,Y) :-
	functor(X,F,N),
	functor(Y,G,M),
	msgDiffTerm(F,G,N,M).

msgDiffTerm(F,G,_,_) :-
	F \== G,
	!.
msgDiffTerm(_,_,N,M) :-
	N =\= M.

msgOccur(subst(X,Y,Z),[U|_]) :-
	var(U),
	U = subst(X,Y,Z),
	!.
msgOccur(subst(X,Y,Z),[subst(U,V,Z)|_]) :-
	X == U,
	Y == V,
	!.
msgOccur(subst(X,Y,Z),[_|S]) :-
	msgOccur(subst(X,Y,Z),S).

/* setMsg(Xs,A) : A is the msg of a set of terms Xs */

setMsg([X],X) :-
	!.
setMsg([X,Y|Xs],W) :-
	msg(X,Y,V),
	setMsg([V|Xs],W).



% eq_variables(A,B,V) proceeeds by induction on the
% structure of the terms A and B
% If it finds two variables it unifies them
% provided that it is not the case that A and B are both in the list V
%
% i.e. eq_variables([A1,A2,A3],[X,Y],V) will bind A1=X and A2=Y
% eq_variables([A1,A2|L],[X,Y],V) will bind A1=X and A2=Y but not L=[]


eq_variables(A,B,V) :-
	( var(A) ->
	    ( var(B) ->
		( \+ belong([A,B],V) ->
		    A=B
		;
		    true
		)
	    ;
		true
	    )
	;
	    ( nonvar(B) ->
		A =.. [F|Args1],
		B =.. [F|Args2],
		eq_variables_args(Args1,Args2,V)
	    ;
		true
	    )
	).
	    
eq_variables_args([],[],_).
eq_variables_args([X|Xs],[Y|Ys],V) :-
	eq_variables(X,Y,V),
	eq_variables_args(Xs,Ys,V).


% AG is not more specific than AU
% V is the set of variables occurring in AG
% which generalize terms in AU sharing variables with VHU
gen_head_vars(AG,AU,VHU,V) :-
	gen_head_vars(AG,AU,VHU,[],V).

gen_head_vars(AG,AU,VHU,InV,OutV) :-
	( var(AG) ->
	    free_variables(AU,[],[],VAU),
	    linking_variables(VAU,VHU,VAHU),
	    ( VAHU==[] ->
		OutV=InV
	    ;
		OutV=[AG|InV]
	    )
	;
	    AG =.. [F|Args1],
	    AU =.. [F|Args2],
	    gen_head_vars_args(Args1,Args2,VHU,InV,OutV)
	).
	    
gen_head_vars_args([],[],_,V,V).
gen_head_vars_args([X|Xs],[Y|Ys],VHU,InV,OutV) :-
	gen_head_vars(X,Y,VHU,InV,OutV1),
	gen_head_vars_args(Xs,Ys,VHU,OutV1,OutV).



