:- use_module(library(clpfd)).

:- dynamic stratum_constraint/1.
:- dynamic num_of_clauses/2.
:- dynamic fails/1.
:- dynamic valid/1.
:- dynamic is_useful/1.


%%% compute_stratification(Cls,Strat)
%%% Cls is a list of clauses
%%% Strat is an ordered list of stratum(Stratum,ListofPredSymbinStrat)
compute_stratification([],[]).
compute_stratification(Cls,Strata) :-
	Cls \== [],
	retractall(stratum_constraint(_)),
	retractall(num_of_clauses(_,_)),
%	retractall(strat_pred_defined(_)),
%	retractall(strat_pred_used(_)),
	assert_stratum_constraints(Cls),
	setof(C1,stratum_constraint(C1),L),
	ground_to_var_list(L,Ass,C),
	term_variables(Ass,VC),
	length(VC,MaxStrat1), MaxStrat is MaxStrat1-1,
	domain(VC,0,MaxStrat),
	post(C),
	(labeling([],VC) ->
	    true
	;
	    throw(stratification_exception),
	    fail
	),
	remove_duplicates(VC,VC1),
	sort(VC1,VC2),
	get_strata(VC2,Ass,Strata).


assert_stratum_constraints([]).
assert_stratum_constraints([(HD,BD)|Cls]) :-
	functor(HD,HF,_),
	incr_num_of_clauses(HF),
	assert(stratum_constraint(HF#>=0)),
%	assert_fact_once(strat_pred_defined(HF)),
	BD = bd(_,As),
	assert_stratum_constraints_body(As,HF),
	assert_stratum_constraints(Cls).

assert_stratum_constraints_body([],_).
assert_stratum_constraints_body([L|Ls],HF) :-
	( is_a_negative_literal(L,A1) ->
	    functor(A1,F,_),
	    assert(stratum_constraint(HF#>F))
	;
	    functor(L,F,_),
	    assert(stratum_constraint(HF#>=F))	    
	),
%	assert_fact_once(strat_pred_used(F)),
	assert(stratum_constraint(F#>=0)),
	assert_stratum_constraints_body(Ls,HF).

get_strata([],_,[]).
get_strata([S|S1],Ass,[stratum(S,L)|Strata]) :-
	setof(F,member(gtv(F,S),Ass),L),
	get_strata(S1,Ass,Strata).



%%% ground_to_var_list(L1,Ass,L2)
%%% Ass is a funct. represented as a list of pairs gtv(ground,Var) s.t.
%%% L2 contains the terms in L1 where each ground leaf atom
%%% has been replaced by a variable according to Ass
%%% the el. of L1 are ground (because of the use of memberchk)

ground_to_var_list(L1,Ass,L2) :-
	ground_to_var_list(L1,[],L2,Ass).

%%% ground_to_var_list(L1,InAss,L2,OutAss)
ground_to_var_list([],Ass,[],Ass).
ground_to_var_list([T|Ts],InAss,[TV|TVs],OutAss) :-
	ground_to_var(T,InAss,TV,InAss1),
	ground_to_var_list(Ts,InAss1,TVs,OutAss).


ground_to_var(T,InAss,T1,OutAss) :-
	ground(T),
	(atom(T) ->
	    (memberchk(gtv(T,V),InAss) ->
		T1 = V, OutAss=InAss
	    ;
		T1 = V1, OutAss=[gtv(T,V1)|InAss]
	    )
	;
	    T =.. [F|Args],
	    ground_to_var_list(Args,InAss,Args1,OutAss),
	    T1 =.. [F|Args1]
	).



strata_bup(S,PCls) :-
	retractall(fails(_)),
	retractall(valid(_)),
	retractall(is_useful(_)),
	strata_bup_(S,PCls).

strata_bup_([],_PCls).
strata_bup_([stratum(S,L)|Strata],PCls) :-
	debug((
	       print('Processing stratum '),print(S),nl
	      )),
	predicate_list_clauses(L,PCls,Clss,RestPCls),
	flatten(Clss,Cls),
	clauses_bup(Cls,Cls1),
	
	find_useless_predicates(Cls1),
	debug((
	       portray_clauses(Cls1),
	       nl,
	       print('Valid predicates '),nl,
	       (setof(VF,(member(VF,L),valid(VF),print(VF),nl),_) -> true ; print('none'),nl),
	       print('Failed predicates '),nl,
	       (setof(FF,(member(FF,L),fails(FF),print(FF),nl),_) -> true ; print('none'),nl)
	      )),
	strata_bup_(Strata,RestPCls).



predicate_list_clauses([],_PCls,[],_PCls).
predicate_list_clauses([F|Fs],PCls,[Cls|Cls1],RestPCls) :-
	predicate_clauses(F,PCls,Cls,PCls1),
	length(Cls,N),
	(N==0 ->
	    assert(fails(F))
	;
	    true
	),
	assert(num_of_clauses(F,N)),
	predicate_list_clauses(Fs,PCls1,Cls1,RestPCls).

predicate_clauses(_F,[],[],[]).
predicate_clauses(F,[(H,B)|PCls],Cls,PCls1) :-
	(functor(H,F,_) ->
	    Cls = [(H,B)|NCls],
	    PCls1 = NPCls1
	;
	    Cls = NCls,
	    PCls1 = [(H,B)|NPCls1]
	),
	predicate_clauses(F,PCls,NCls,NPCls1).
	



%%%predicate_list_clauses(Fs,Cls) :-
%%%	current_program(P,_),
%%%	predicate_list_clause_n_(Fs,P,Ls1),
%%%	flatten(Ls1,Ls),
%%%	clause_hd_bd_list(Ls,Cls).

%%%predicate_list_clause_n_([],_,[]).
%%%predicate_list_clause_n_([F|Fs],P,[L|Ls]) :-
%%%	predicate_clause_n(F,P,L),
%%%	length(L,N),assert(num_of_clauses(F,N)),
%%%	predicate_list_clause_n_(Fs,P,Ls).

%%%predicate_clause_n(F,P,L) :-
%%%	(setof(CN,
%%%	       Hd^Bd^X^(clause_definition(CN, Hd, Bd, X),
%%%		   member(prog(P),X),functor(Hd,F,_)),
%%%	       L) ->
%%%	    true
%%%	;
%%%	    L=[],assert(fails(F))
%%%	).






clauses_bup(Cls,Cls1) :-
	clauses_bup_(Cls,Cls2),
	(Cls2 == Cls ->
		 Cls1 = Cls
	;
		 clauses_bup(Cls2,Cls1)
	).

clauses_bup_([],[]).
clauses_bup_([Cl|Cls],Cls1) :-
	is_a_fact(Cl),
	!,
	Cl = (H,_B),functor(H,F,_),
	assert(valid(F)),
	debug((
	       print('asserting valid '),print(F),nl
	      )),
	clauses_bup_(Cls,Cls1).
clauses_bup_([Cl|Cls],Cls1) :-
	current(solver,S),
	is_a_constrained_fact(Cl,S),
	Cl = (H,bd(C,[])),
	free_variables(C,H,[],[_|_]), % there are local variables in C
	!,
	solve_clause(Cl,head_vars,S,NewCl),
	Cls1 = [NewCl|Cls2],
	clauses_bup_(Cls,Cls2).
clauses_bup_([Cl|Cls],Cls1) :-  % naive subsumption
	Cl = (H,_B),functor(H,F,_),valid(F),
	!,
	clauses_bup_(Cls,Cls1).
clauses_bup_([Cl|Cls],Cls1) :-
	Cl = (H,bd(_C,Lits)), % remove failed clauses (provided C is satisfiable)
	failed_body(Lits),
	!,
	functor(H,HF,_),decr_num_of_clauses(HF),
	clauses_bup_(Cls,Cls1).
clauses_bup_([Cl|Cls],Cls1) :-
	Cl = (H,bd(C,Lits)),
	simplify_body(Lits,Lits1),
	Cls1 = [(H,bd(C,Lits1))|Cls2],
	clauses_bup_(Cls,Cls2).


failed_body(Lits) :- member(L,Lits), failed_lit(L).

failed_lit(L) :-
	(is_a_negative_literal(L,A) ->
		 functor(A,F,_),
		 valid(F)
	;
		 functor(L,F,_),
		 fails(F)
	).

simplify_body([],[]).
simplify_body([L|Ls],L1) :-
	(valid_lit(L) ->
		 simplify_body(Ls,L1)
	;
		 L1 = [L|L2],
		 simplify_body(Ls,L2)
	).

valid_lit(L) :-
	(is_a_negative_literal(L,A) ->
		 functor(A,F,_),
		 fails(F)
	;
		 functor(L,F,_),
		 valid(F)
	).

decr_num_of_clauses(F) :-
	(num_of_clauses(F,N) ->
		 (N == 1 ->
		     retract(num_of_clauses(F,N)),
		     assert(num_of_clauses(F,0)),
		     assert(fails(F)),
		     debug((
			    print('asserting fails '),print(F),nl
			   ))
		 ;
			  retract(num_of_clauses(F,N)),
			  M is N-1,
			  assert(num_of_clauses(F,M))
		 )
	;
		 assert(fails(F))
	).

incr_num_of_clauses(F) :-
	(num_of_clauses(F,N) ->
		 retract(num_of_clauses(F,N)),
		 M is N+1,
		 assert(num_of_clauses(F,M))
	;
		 assert(num_of_clauses(F,1))
	).



find_useless_predicates(Cls) :-
	find_useless_predicates(Cls,_).


find_useless_predicates(Cls,UselessCls) :-
	find_useless_predicates_(Cls,Cls1),
	(Cls == Cls1 ->
		 UselessCls = Cls1,
		 % assert fails(F) for each useless predicate
	    %%% IMPROVE
		 (setof(F,H^(member((H,_),UselessCls),functor(H,F,_),assert(fails(F)),
			     debug((
				    print(F),print(' is useless'),nl
				   ))
			    ),_);true)
	;
		 find_useless_predicates(Cls1,UselessCls)
	).
	
find_useless_predicates_([],[]).
find_useless_predicates_([Cl|Cls],Useless) :-
	Cl = (H,_),
	functor(H,F,_),
	is_useful(F),
	!,
	find_useless_predicates_(Cls,Useless).
find_useless_predicates_([Cl|Cls],Useless) :-
	Cl = (H,bd(_C,Lits)),
	pos_neg_literals(Lits,PLits,_NLits),
	( (PLits == [] ; all_useful(PLits)) ->
		 %%% only for clarity
       %%% Plits == [] implies all_useful(PLits)
	    functor(H,F,_),
	    assert(is_useful(F)),
	    find_useless_predicates_(Cls,Useless)
	;
	    
	    Useless = [Cl|Useless1],
	    find_useless_predicates_(Cls,Useless1)
	).

all_useful([]).
all_useful([A|As]) :-	% A is a positive literal
	functor(A,F,_),
	is_useful(F),
	all_useful(As).
