% Constrained Atomic Folding

:- use_module(library(lists), [ append/3,
				memberchk/2,
				nth/4
			      ]).
:- use_module(library(terms), [ subsumes_chk/2 ]).


% CafV encodes the type of c.a.folding (soft or hard)
caf(Solver,CafV,ToFoldN,ToUseN,Success) :-  
   check_input_clauses_for_caf(Solver,ToFoldN,ToUseN,Success),
   (var(Success) ->
       try_caf(Solver,CafV,ToFoldN,ToUseN,Success)
   ;
       true).



try_caf(Solver,CafV,ToFoldN,ToUseN,Out) :-
   (apply_caf(Solver,CafV,ToFoldN,ToUseN,NewClause) ->
       Out=out(label(caf(Solver,CafV,ToFoldN,ToUseN)),ToFoldN,[NewClause]) ;
       Out=error('Impossible constrained atomic folding.')).

check_input_clauses_for_caf(_Solver,[ToFold],[ToUse],Success) :-
	current_program(_PN,Cls),
	(memberchk(ToFold,Cls) ->
	    (is_a_definition_clause(ToUse) ->
			  true
	    ;	  Success=error('The clause used for folding must belong to a definition.'))
	;	 Success=error('The clause to fold must belong to the current program.')).


apply_caf(Solver,CafV,[ToFoldN],[ToUseN],NewClause) :-
	clause_hd_bd(ToFoldN,H,bd(C,BdwoC)),
	clause_hd_bd(ToUseN,HD,bd(CD,BdDwoCD)),
	apply_caf__(Solver,CafV,(H,bd(C,BdwoC)),(HD,bd(CD,BdDwoCD)),NewClause).
	
apply_caf_(Solver,CafV,ToFold,[ToUseN],NewClause) :-
	clause_hd_bd(ToUseN,HD,bd(CD,BdDwoCD)),
	ToUse = (HD,bd(CD,BdDwoCD)),
	apply_caf__(Solver,CafV,ToFold,ToUse,NewClause).

apply_caf__(Solver,CafV,ToFold,ToUse,NewClause) :-
	ToFold = (_,bd(_,BdwoC)),
	ToUse = (_,bd(_,[AD])),
	% find the first (potentially) foldable position in BdwoC
	same_pred_pos(AD,BdwoC,AN),
	apply_caf___(Solver,CafV,ToFold,AN,ToUse,NewClause).

apply_caf___(Solver,CafV,ToFold,LN,ToUse,NewClause) :-
	ToFold = (H,bd(C,BdwoC)),
	ToUse = (HD,bd(CD,[AD])),
	nth(LN,BdwoC,Lit,Rest),
%	ord_positions_to_elts(BdwoC,[LN],[Lit],Rest),
	(is_a_negative_literal(Lit,PAtom1) ->
	    Atom=PAtom1
	;
	    Atom=Lit
	),

	(caf_conditions(Solver,touse(HD,CD,AD),tofold(H,C,Lit,Rest)) ->
	    true
	;
	    % for debugging purposes, indeed we don't want an exception
	    % to be thrown, rather we want enumeration on same_pred_pos
	    U=(HD :- CD,AD),
	    F=(H :- C,Lit,Rest),
	    portray_clause(F), portray_clause(U),
	    throw(caf_conditions(ToFold, ToUse,F,U))
	),


	(is_a_negative_literal(Lit,PAtom1) ->
	    functor(HD1,not,1),
	    arg(1,HD1,HD),
	    nth(LN,Lits,HD1,Rest)
	;
	    nth(LN,Lits,HD,Rest)
	),	
	%%% manage constraints
	( CafV == soft ->
	    NewClause=(H,bd(C,Lits))
	;
	    throw(only_soft_caf_implemented)
	).


caf_conditions(Solver,touse(HD,CD,AD),tofold(H,C,Lit,Rest)) :-
	(is_a_negative_literal(Lit,Atom1) ->
	    Atom=Atom1
	;
	    Atom=Lit
	),
	subsumes_chk(AD,Atom),
	unify_with_occurs_check(AD,Atom),
	entails(Solver,C,CD),

%%%	int_var(HD,(CD,AD),V),    
%%%	free_variables(H,[],[],V1),
%%%	intersection_vars_empty(V,V1),
%%%	free_variables([H|Rest],[],C,V2),
%%%	intersection_vars_empty(V,V2).	

	( Atom == Lit -> %%%  Lit is an atom
	    free_variables((CD,AD),HD,[],DLocalVars),
	    free_variables((H,C,Rest),[],[],ContextVars),
	    (intersection_vars_empty(DLocalVars,ContextVars) ->
		true
	    ;
		linking_variables(DLocalVars,ContextVars,LV),
		determined_vars(LV,Solver,C)
	    )
	;
	    %%% Lit is a negated atom
	    term_variables(Lit,VLit),
	    term_variables(HD,VHD),
	    free_variables(VLit, VHD, [], VLitnotHD),
	    ( VLitnotHD = [] ->
		true
	    ;
		determined_vars(VLitnotHD,Solver,C)
	    )
	).


determined_vars(Vars,S,C) :-
	list_to_constraint(S,C,C1),
	copy_term(term(Vars,C1),term(Varscopy,C1copy)),
	setof(Varscopy,C1copy,[ListofValues]),
	ground(ListofValues).
