do_ccr([],_,[]).
do_ccr([CN|CNs],S,OutCNs) :-
	clause_hd_bd_list([CN|CNs],Cls),
	do_ccr_(Cls,S,OutCls),
	update_transf_tree(ccr([CN|CNs],OutCNs),[CN|CNs],OutCls,OutCNs).


do_ccr_([],_,[]).
do_ccr_([Cl|Cls],S,OutCls) :-
	init_ccr([Cl|Cls],S),
	ccr_clauses([Cl|Cls],S,OutCls),
	retractall(call_pattern(_,_,_)).


init_ccr([],_S).
init_ccr([Cl|Cls],S) :-
	retractall(call_pattern(_,_,_)),
	assert_call_patterns([Cl|Cls],S). %%% ADD CALL PATTERNS FOR INIT DEFS

%%% S is a solver
assert_call_patterns([],_S).
assert_call_patterns([(_,bd(C,As))|Cls],S) :-
	assert_call_patterns_body(As,C,S),
	assert_call_patterns(Cls,S).

assert_call_patterns_body([],_,_).
assert_call_patterns_body([L|Ls],C,S) :-
	( is_a_negative_literal(L,A1) ->
		 A=A1
	;
		 A=L
	),
	functor(A,F,Ar),
	free_variables(A,[],[],VA),
	solve(S,C,VA,CA),
	assert(call_pattern(F,Ar,cp(CA,A))),
	assert_call_patterns_body(Ls,C,S).

%%% setof_call_patterns(H,C,S,CPs)
%%% retrieves the set of asserted call patterns cp(D,A)
%%% for H,C such that the constraint (D and A=H and C) is satisfiable
%%%
%%% NOTE:
%%% keep CPs as small as possible, because for removing a constraint C
%%% from a clause, u need all the elements of CPs to entail C
setof_call_patterns(H,_C,_S,CPs) :-
	functor(H,F,Ar),
	
	( setof(CP,call_pattern(F,Ar,CP),CPs1) ->
	    CPs = CPs1
	;
	    CPs = []
	).



%%% ccr_clauses(Cls,Solver,NewCls)
%%% Cls is a list of clauses
%%% NewCls is a list of clauses
ccr_clauses([],_S,[]).
ccr_clauses([(H,bd(C,As))|Cls],S,NewCls) :-
	( C \== [] ->
		 setof_call_patterns(H,C,S,CPs),
		 (CPs == [] ->
			  NewCl = (H,bd(C,As)),
			  NewCls = [NewCl|NewCls1],
			  ccr_clauses(Cls,S,NewCls1)
		 ;
			  ccr(C,H,CPs,S,C1),
			  NewCl = (H,bd(C1,As)),
			  NewCls = [NewCl|NewCls1],
			  ccr_clauses(Cls,S,NewCls1)
		 )
	;
	%%% if there's nothing to remove, go on
		 NewCl = (H,bd(C,As)),
		 NewCls = [NewCl|NewCls1],
		 ccr_clauses(Cls,S,NewCls1)
	).




ccr([],_,_,_,[]).
ccr([C|Cs],H,CPs,S,C1) :-
	( all_call_patterns_entail(CPs,S,H,C) ->
	    ccr(Cs,H,CPs,S,C1)
	;
	    C1=[C|C2],
	    ccr(Cs,H,CPs,S,C2)
	).

all_call_patterns_entail([],_S,_H,_C).
all_call_patterns_entail([cp(D,A)|CPs],S,H,C) :-
	copy_term((H,C),(H1,C1)), %%% may be copy_term is not needed?
	copy_term((A,D),(A1,D1)),
	(unify_with_occurs_check(A1,H1) ->
		 ( entails(S,D1,[C1]) ->
			  all_call_patterns_entail(CPs,S,H,C)
		 ;
			  fail
		 )
	;
		 all_call_patterns_entail(CPs,S,H,C)
	).

