:- use_module(library(lists), [ nth/3,nth/4 ]).
:- use_module(library(terms), [
			       subsumes/2,
			       subsumes_chk/2,
			       term_subsumer/3
			      ]).



contextual_specialization([],[]).
contextual_specialization(_Defs,_FinalCls) :-
	no_solver_selected, !, fail.
contextual_specialization(_Defs,_FinalCls) :-
	no_uf_selected, !, fail.
contextual_specialization(_Defs,_FinalCls) :-
	no_wqounf_selected, !, fail.
contextual_specialization(_Defs,_FinalCls) :-
	no_wqogen_selected, !, fail.
contextual_specialization(_Defs,_FinalCls) :-
	no_widening_selected, !, fail.

contextual_specialization(Defs,FinalCls) :-
	cs_env(Solver,Uf,Wqounf,Wqogen,Widening),
	contextual_specialization(Defs,Solver,Uf,Wqounf,Wqogen,Widening,FinalCls).


cs_env(Solver,Uf,Wqounf,Wqogen,Widening):-
	current(solver,Solver),
	current(uf,Uf),
	current(wqounf,Wqounf),
	current(wqogen,Wqogen),
	current(widening,Widening).

%
% contextual_specialization/7
%
contextual_specialization([],_Solver,_Uf,_Wqounf,_Wqogen,_Widening,[]).
contextual_specialization(Defs,Solver,Uf,Wqounf,Wqogen,Widening,FinalCls) :-
	(cs_conditions(Solver,Uf,Wqounf,Wqogen,Widening) ->
	    create_forest(Defs,DefsForest),
	    contextual_specialization(Defs,[],DefsForest,Solver,Uf,Wqounf,Wqogen,Widening,CFactss,FldClsss),
	    flatten(FldClsss,FoldedCls), flatten(CFactss,CFacts),
	    append(CFacts,FoldedCls,FinalCls1),

	    debug((  write('** Contextual Specialization Cls **'),nl, portray_clauses(FinalCls1)  )),
	    
	    (current(ccr,yes) ->
		(ptime(do_ccr_(FinalCls1,Solver,CCRCls)) ->
		    debug(( write('** CCRCls **'), nl, portray_clauses(CCRCls)  )),
		    true
		;
		    trace,do_ccr_(FinalCls1,Solver,CCRCls)
		)
	    ;
		(current(ccr,no) ->
		    CCRCls=FinalCls1
		;
		    throw(current_ccr_undefined)
		)
	    ),
	    
	    (current(bup,yes) -> 
		(ptime(tcl_bottom_up_phase(CCRCls)) ->
		    debug(( write('** BUP Cls **'), nl, portray_clauses(CCRCls)  )),
		    true
		;
		    trace,tcl_bottom_up_phase(CCRCls)
		)
	    ;
		(current(bup,no) ->	
		    true
		;
		    throw(current_bup_undefined)
		)
	    ),
	    FinalCls=CCRCls
	;
	    throw(cs_conditions(Solver,Uf,Wqounf,Wqogen,Widening))
	).

cs_conditions(Solver,Uf,Wqounf,Wqogen,Widening) :-
	is_a_solver(Solver),
	is_a_uf(Uf,Solver),
	is_a_wqounf(Wqounf,Solver),
	is_a_wqogen(Wqogen,Solver),
	is_a_widening(Widening,Solver).

%
% contextual_specialization/10
% contextual_specialization(Defs,InDefs,DefsForest,Solver,Uf,Wqounf,Wqogen,Widening,CFactss,FldClsss)
% FldClsss is a list of lists of lists of folded clauses
contextual_specialization([],Defs,_DefsForest,_Solver,_Uf,_Wqounf,_Wqogen,_Widening,[],[]) :-
%	debug(( write('** DefsForest **'),nl,writeq(DefsForest),nl )),
	 debug((
		write('** Defs **'),nl,portray_clauses(Defs),nl
	)).
	
contextual_specialization([Def|Defs],InDefs,InDefsForest,Solver,Uf,Wqounf,Wqogen,Widening,[CFacts|NewCFs],[FldClss|NewFldCls]) :-

	ptime(unfold_replace([Def|Defs],Solver,Uf,Wqounf,UForest)),
	
	forest_leaves(UForest,Solver,RootLeaves,CFacts),

	debug((
	       nl,
	       write('*** Unfolding ***'), nl,
	       write('CFacts:'),nl,
	       flatten(CFacts,CF),
	       portray_clauses(CF), nl,
	       write('RootLeaves:'),nl,
	       show_rootleaves(RootLeaves),
	       write('*** End Unfolding ***'), nl	
	      )),
	
	ptime(define_fold_forest(RootLeaves,InDefs,InDefsForest,Solver,Wqogen,Widening,OutDefs,OutDefsForest,FldClss)),
	delete_elts_sublist(InDefs,OutDefs,NewDefs),

%%%	newdefs_forest(NewDefsForest,DefsForest,NewDefs),

	debug((
	       write('*** Define Fold ***'), nl,
	       write('FldClss:'),nl,
	       flatten(FldClss,FC),
	       portray_clauses(FC), nl,
	       write('NewDefs:'),nl,
	       portray_clauses(NewDefs),
	       write('*** End Define Fold ***'), nl
	      )),
	contextual_specialization(NewDefs,OutDefs,OutDefsForest,Solver,Uf,Wqounf,Wqogen,Widening,NewCFs,NewFldCls).

unfold_replace([],_Solver,_Uf,_Wqounf,[]).
unfold_replace([Def|Defs],Solver,Uf,Wqounf,[Utree1|Utrees]) :-
	create_forest([Def],[Utree]),

	debug((
	       write('*UTree '), writeq(Def),  nl
	      )),
	
	build_unfolding_tree(Utree,Solver,Uf,Wqounf,Utree1),

	debug((
	       writeq(Utree1), nl,
	       write('*End UTree '), writeq(Def),  nl
	      )),
	unfold_replace(Defs,Solver,Uf,Wqounf,Utrees).


% a tree si represented as the list of the paths
% from the leaves to the root
build_unfolding_tree([],_Solver,_Uf,_Wqounf, []).
build_unfolding_tree([[Leaf|Ancs]|Paths],Solver,Uf,Wqounf,[[Leaf|Ancs]|NewPaths]) :-
	do_not_unfold([[Leaf|Ancs]|Paths],Solver,Wqounf), !,
	build_unfolding_tree(Paths,Solver,Uf,Wqounf,NewPaths).
build_unfolding_tree([[Leaf|Ancs]|Paths],Solver,Uf,Wqounf,OutUtree) :-
%%% to fix: there are clauses with failed sons only,
%%% which belong to the final program
%%% hint: check uf
	(uf(Uf,Leaf,[[Leaf|Ancs]|Paths],Solver,AN) ->
	% a potentially unfoldable position has been found
	    unfold_init(Leaf,AN,UCls),
	    ( UCls==[] ->
%%%		% stop unfolding this leaf
%%%		OutUtree=[[Leaf|Ancs]|NewPaths],
		%%% remove this leaf from the program
		%%%	  update_transf_tree(Label,[Leaf],UCls,[]),
		%%% remove this leaf and its path to the root from the tree
			  OutUtree=NewPaths,
			  build_unfolding_tree(Paths,Solver,Uf,Wqounf,NewPaths)
	    ;
		%%% unfolding was successful
		(current(quick_solve,yes) ->
                %%% check satisfiability only
		    quick_solve_clauses(UCls,Solver,NFCls)
		;
		    solve_clauses(UCls,linking_vars,Solver,SCls),
		    nonfailed_clauses(SCls,Solver,NFCls)
		),
%%%		update_transf_tree(Label,[Leaf],NFCls,NewCNs),
		(NFCls==[] ->
		    build_unfolding_tree(Paths,Solver,Uf,Wqounf,NewPaths),
		    OutUtree=NewPaths
		;
		    tree_addleaves(NFCls,[Leaf|Ancs],NewPaths),
		    append(NewPaths,Paths,Utree),
		    build_unfolding_tree(Utree,Solver,Uf,Wqounf,OutUtree)
		)
	    )
	;
%%%	    (Ancs==[] ->
%%%		%%% Leaf is the root of the unfolding tree
%%%		%%% if uf/5 fails should we delete Leaf
%%%	        %%% from the current program?
%%%		true
%%%		;
%%%		true
%%%		),
	    build_unfolding_tree(Paths,Solver,Uf,Wqounf,NewPaths),
	    OutUtree=[[Leaf|Ancs]|NewPaths]
	).

%%%build_unfolding_tree([[L|P]|Paths],Solver,Uf,Wqounf,NewPaths) :-
%%%%	throw(build_unfolding_tree(L)).
%%%	trace.

do_not_unfold([[Leaf|_Ancs]|_Paths],Solver,_Wqounf) :-
	is_a_constrained_fact(Leaf,Solver), !.
do_not_unfold([[Leaf|_Ancs]|_Paths],Solver,_Wqounf) :-
	has_failing_body(Leaf,Solver), !.
do_not_unfold([[Leaf|Ancs]|_Paths],Solver,Wqounf) :-
	embeds_ancestor_clause(Leaf,Ancs,Solver,Wqounf,_EmbCl)
%%%	,write('*** Embedding found while unfolding *** '), nl,
%%%	show_numbered_clauses([EmbCN,Leaf]), nl
	.



% embeds_ancestor_clause(Leaf,Ancs,Solver,Wqo,EmbddCl)
% Leaf,EmbddCl are clauses of the form (Hd,Bd)
embeds_ancestor_clause(_Leaf,[],_Solver,_Wqo,_EmbCl) :- fail, !.
embeds_ancestor_clause((LHd,LBd),[Parent|_Ancs],Solver,Wqo,EmbCl) :-
%%%	clause_hd_bd(Parent,PHd,PBd),
	Parent = (PHd,PBd),
	embeds_clause(Wqo,Solver,(LHd,LBd),(PHd,PBd)), ! ,
	EmbCl=(PHd,PBd).
embeds_ancestor_clause((LHd,LBd),[_Parent|Ancs],Solver,Wqo,EmbCl) :-
	embeds_ancestor_clause((LHd,LBd),Ancs,Solver,Wqo,EmbCl).



embeds_clause(Wqo,Solver,(_LeafH,LeafB),(_ParentH,ParentB)) :-
	embeds(Wqo,Solver,LeafB,ParentB).


% forest_leaves(UForest,Solver,RootFoldableLeaves,CFacts)
% UForest is a list of unfolding trees 
% RootFoldableLeaves is a list of pairs (Root,[L1,...,LN]), one for each unfolding tree Utree, where:
% - Root is the root of Utree
% - [L1,...,LN] is the list of the leaves of Utree which we have to fold
% CFacts is a list of constrained facts
%

forest_leaves([],_Solver,[],[]).
forest_leaves([ [] |Utrees],Solver,RLss,CFss) :-
	forest_leaves(Utrees,Solver,RLss,CFss).
forest_leaves([Utree|Utrees],Solver,[RootLeaves| RLss],[CFs|CFss]) :-
	tree_leaves(Utree,Solver,RootLeaves,CFs),
	forest_leaves(Utrees,Solver,RLss,CFss).


tree_leaves(Utree,Solver,(Root,FLeaves),CFacts) :-
	(Utree==[] ->
	    throw(tree_leaves(empty_tree))
	;
	    true
	),
	tree_root(Utree,Root),
	tree_leaves(Utree,Leaves),
	foldable_clauses(Leaves,Solver,FLeaves,CFacts).




foldable_clauses_n([],_Solver,[],[]).
foldable_clauses_n([Leaf|Leaves],Solver,Fleaves,[Leaf|CFacts]) :-
	is_a_constrained_fact_n(Leaf,Solver), !,
	foldable_clauses_n(Leaves,Solver,Fleaves,CFacts).
foldable_clauses_n([Leaf|Leaves],Solver,[Leaf|Fleaves],CFacts) :-
	foldable_clauses_n(Leaves,Solver,Fleaves,CFacts).



nonfailed_clauses([],_Solver,[]).
nonfailed_clauses([Leaf|Leaves],Solver,NFLeaves) :-
	has_failing_body(Leaf,Solver), !,	
	nonfailed_clauses(Leaves,Solver,NFLeaves).
nonfailed_clauses([Leaf|Leaves],Solver,[Leaf|NFleaves]) :-
	nonfailed_clauses(Leaves,Solver,NFleaves).


foldable_clauses([],_Solver,[],[]).
foldable_clauses([Leaf|Leaves],Solver,Fleaves,[Leaf|CFacts]) :-
	is_a_constrained_fact(Leaf,Solver), !,
	foldable_clauses(Leaves,Solver,Fleaves,CFacts).
foldable_clauses([Leaf|Leaves],Solver,[Leaf|Fleaves],CFacts) :-
	foldable_clauses(Leaves,Solver,Fleaves,CFacts).



% define_fold_forest/9
% 	define_fold_forest(RootLeaves,InDefs,DefsForest,Solver,Wqogen,Widening,NewDefs,NewDefsForest,FldClss)
% RootLeaves is a list of pairs (Root,[L1,...,LN]) where
% [L1,...,LN] is a list of clauses
% representing the leaves to be folded in a Utree with root Root
% FldClss is a list of lists of folded clauses
define_fold_forest([],Defs,DefsForest, _Solver,_Wqogen,_Widening,Defs,DefsForest,[]).
define_fold_forest([(Root,Leaves)|RLs],InDefs,InDefsForest,Solver,Wqogen,Widening,OutDefs,OutDefsForest,[FldCls|FldLeaves]) :-
%%%	forest_nodes(InDefsForest,InDefs),
	define_fold_tree((Root,Leaves),InDefs,InDefsForest,Solver,Wqogen,Widening,Defs,DefsForest,FldCls),
%%%	update_transf_tree(def_fold(Leaves,FldCNs),Leaves,FldCls,FldCNs),
	define_fold_forest(RLs,Defs,DefsForest,Solver,Wqogen,Widening,OutDefs,OutDefsForest,FldLeaves).


%
% define_fold_tree/9
%
define_fold_tree((_Root,[]),Defs,DefsForest,_Solver,_Wqogen,_Widening,Defs,DefsForest,[]).
define_fold_tree((Root,[L|Leaves]),InDefs,InDefsForest,Solver,Wqogen,Widening,OutDefs,OutDefsForest,[FldL|FldLeaves]) :-
%%%	forest_nodes(InDefsForest,InDefs),
	L = (_,bd(_,Lits)),
	length(Lits,MaxAN),
	(current(mc,yes) ->
	    ptime(define_fold_leaf_mc(Root,L,InDefs,InDefsForest,Solver,Wqogen,Widening,Defs,DefsForest,1,MaxAN,FldL))
	;
	    ptime(define_fold_leaf(Root,L,InDefs,InDefsForest,Solver,Wqogen,Widening,Defs,DefsForest,1,MaxAN,FldL))
	),
	define_fold_tree((Root,Leaves),Defs,DefsForest,Solver,Wqogen,Widening,OutDefs,OutDefsForest,FldLeaves).


%
% define_fold_leaf/12
%
define_fold_leaf(_Root,Leaf,Defs,DefsForest,_Solver,_Wqogen,_Widening,Defs,DefsForest,AN,MaxAN,Leaf) :- AN > MaxAN,!.
define_fold_leaf(Root,Leaf,InDefs,InDefsForest,Solver,Wqogen,Widening,OutDefs,OutDefsForest,AN,MaxAN,FldL) :-
	(ptime(select_definition(AN,Leaf,Root,InDefs,InDefsForest,Solver,Wqogen,Widening,Def,OldNew)) ->
	    true
	;
	    throw(select_definition(AN,Leaf,Root,InDefs,InDefsForest,Solver,Wqogen,Widening,Def,OldNew))
	),
	(OldNew==old ->
	    DefsForest=InDefsForest,
	    Defs=InDefs
	;
	    update_forest(Def,Root,InDefsForest,DefsForest),
	    Defs=[Def|InDefs]
	),
	copy_term(Def,Def1),
	ptime(apply_caf___(Solver,soft,Leaf,AN,Def1,NewL)),
	AN1 is AN+1,
	define_fold_leaf(Root,NewL,Defs,DefsForest,Solver,Wqogen,Widening,OutDefs,OutDefsForest,AN1,MaxAN,FldL).


%
% define_fold_leaf_mc/12
%
define_fold_leaf_mc(_Root,Leaf,Defs,DefsForest,_Solver,_Wqogen,_Widening,Defs,DefsForest,AN,MaxAN,Leaf) :- AN > MaxAN,!.
define_fold_leaf_mc(Root,Leaf,InDefs,InDefsForest,Solver,Wqogen,Widening,OutDefs,OutDefsForest,AN,MaxAN,FldL) :-
	(ptime(select_definition_mc(AN,Leaf,Root,InDefs,Solver,Wqogen,Widening,Def,OldNew)) ->
	    true
	;
	    throw(select_definition_mc(AN,Leaf,Root,InDefs,Solver,Wqogen,Widening,Def,OldNew))
	),
	(OldNew==old ->
	    DefsForest=InDefsForest,
	    Defs=InDefs
	;
	    InDefsForest = [InDefsTree],
	    DefsForest = [[[Def]|InDefsTree]],
	    Defs=[Def|InDefs]
	),
	copy_term(Def,Def1),
	ptime(apply_caf___(Solver,soft,Leaf,AN,Def1,NewL)),
	AN1 is AN+1,
	define_fold_leaf_mc(Root,NewL,Defs,DefsForest,Solver,Wqogen,Widening,OutDefs,OutDefsForest,AN1,MaxAN,FldL).

%%%first_foldable_pos_clause((_H,bd(_C,Lits)),Solver,AN) :-
%%%	first_foldable_pos_goal(Lits,Solver,AN).

%%%first_foldable_pos_goal([],_Solver,_AN) :- fail, !.
%%%first_foldable_pos_goal([AC|ACs],_Solver,AN) :-
%%%	nth(AN,[AC|ACs],A),
%%%	(is_a_negative_literal(A,PA) ->
%%%	    A1=PA
%%%	;
%%%	    A1=A
%%%	),
%%%	functor(A1,F,Ar),
%%%	is_defined_in_prog(F/Ar,1).


select_definition(AN,Leaf,Root,Defs,DefsForest,Solver,Wqogen,Widening,Def,OldNew) :-
	(candidate_def(Leaf,Solver,AN,CandD) ->
		 true
	;
		 throw(candidate_def(Leaf,Solver,AN,CandD))
	),
	(folding_equivalent_list(CandD,Defs,Solver,FEDef) ->
	    Def=FEDef, OldNew=old
	;
	    node_ancestors_forest(Root,DefsForest,RootAncs),
	    generalize_or_define(CandD,[Root|RootAncs],Defs,Solver,Wqogen,Widening,Def,OldNew)
	).


%	Notice that the candidate definition is of the form (Head,Body).
%  It is not a clause number!
candidate_def(Leaf,Solver,AN,Def) :-
	Leaf = (Hd,bd(C,Bd)),
	nth(AN,Bd,Atom1,Rest),
	(is_a_negative_literal(Atom1,PAtom1) ->
	    Atom=PAtom1
	;
	    Atom=Atom1
	),
	% constraint of the candidate def
	free_variables(Atom,[],[],AVars),
	solve(Solver,C,AVars,C1),
	% head of the candidate definition
	linking_variables((C1,Atom),(Hd,C,Rest),LV),
	functor(Atom,AF,_AA),
	gen_new_name_functor(HeadF,new,AF),
	Head =.. [HeadF|LV],
	Body = bd(C1,[Atom]),
	copy_term((Head,Body),Def).


% generalize_or_define(CandD,RootAncs,Defs,Solver,Wqogen,Widening,Def,OldNew)
% CandD is a clause: (H,B)
% Def is a clause number
% OldNew in {old,new}

%generalize_or_define(CandD,_RootAncs,Defs,Solver,_Wqogen,_Widening,FEDefN,old) :-
%	folding_equivalent_list(CandD,Defs,Solver,FEDefN), !
%	,write('old definition '), write(FEDefN), write(CandD)
%	.
generalize_or_define(CandD,RootAncs,Defs,Solver,Wqogen,Widening,Def,OldNew) :-
	(embeds_ancestor_clause(CandD,RootAncs,Solver,Wqogen,ToGen) ->
	    (generalize_def(ToGen,CandD,Solver,Widening,GenDef) ->
		true
	    ;
		throw(generalize_def(ToGen,CandD,Solver,Widening,GenDef))
	    ),
	    (folding_equivalent_list(GenDef,Defs,Solver,FEDef) ->
		copy_term(FEDef,Def),
		OldNew=old
	    ;
%%%		GenDef=(Hd,_Bd),
%%%		functor(Hd,F,A), Label=label(definition(F/A)),
%%%		update_transf_tree(Label,[],[GenDef],[DefN]),
		copy_term(GenDef,Def),
		OldNew=new
	    )
	;
%%%	    CandD=(Hd,_Bd), functor(Hd,F,A), Label=label(definition(F/A)),
%%%	    update_transf_tree(Label,[],[CandD],[DefN]),
	    copy_term(CandD,Def),
	    OldNew=new
        %%% ,write('new definition '), write(DefN), write(CandD)
	).
	
%generalize_or_define(Def,_RootAncs,_Defs,_Solver,_Wqogen,_Widening,DefN,new) :-
%	Def=(Hd,_Bd), functor(Hd,F,A), Label=label(definition(F/A)),
%	update_transf_tree(Label,[],[Def],[DefN]), OldNew=new
%	write('new definition '), write(DefN), write(Def)
%	.



% generalize_def(Source,Use,Solver,Widening,Gen)
% Gen is the generalization of clause Source w.r.t. clause Use
% 
generalize_def((_HS,BS),(_HU,BU),Solver,Widening,(HG,BG)):-
	BS = bd(CS,BdSwoCS),
	BU = bd(CU,BdUwoCU),
	% check whether Source and Use are atomic defs...
	BdSwoCS=[AS], BdUwoCU=[AU],
	% ...and they contain the same predicate in the body
	functor(AS,Pred,Ar), functor(AU,Pred,Ar),
%%%	msg(AS,AU,AG),
	term_subsumer(AS,AU,AG),
	free_variables(AG,[],[],VAG), 
	eq_variables(AG,AS,VAG),
	eq_variables(AG,AU,VAG),

	% body of the generalized def
	widening(Widening,Solver,CS,CU,CG),
	BG = bd(CG,[AG]),
	% head of the generalized definition
	gen_new_name_functor(HGF,gen,Pred),

	free_variables(AG,[],[],VHG), % troppe!


%%%	free_variables(HU,[],[],VHU),	
%%%	( AG==AU ->
%%%	    free_variables(BG,[],[],VBG),
%%%	    linking_variables(VHU,VBG,VHG)
%%%	;
%%%	% AG is more general than AU
%%%	    gen_head_vars(AG,AU,VHU,VHG1),
%%%	    free_variables(BG,[],[],VBG),
%%%	    linking_variables(VHU,VBG,VHG2),
%%%	    free_variables(VHG1,[],VHG2,VHG) % VHG = (VHG1 union VHG2)
%%%	),
	    
	HG =.. [HGF|VHG].




%%% Tailored to verification
select_definition_mc(AN,Leaf,_Root,Defs,Solver,_Wqogen,Widening,Def,OldNew) :-
	Leaf = (_Hd,bd(C,Bd)),
	nth(AN,Bd,Atom1,_Rest),
	(is_a_negative_literal(Atom1,PAtom1) ->
	    Atom=PAtom1
	;
	    Atom=Atom1
	),
	% constraint of the candidate def
	free_variables(Atom,[],[],AVars),

	solve(Solver,C,AVars,Ctemp),
	Widening = pred_abs,
	arg(1,Atom,State), %%% we assume that the 1st arg of Atom encodes the state of the system
	widening(Widening,Solver,State,Ctemp,C1),
	
	%%% head of the candidate definition
%	linking_variables((C1,Atom),(Hd,C,Rest),LV),

	LV=AVars, %%% DEFAULT
%%%	free_variables(Atom2,[],[],LV),%%% REMOVE (only for trace2)
	
	functor(Atom,AF,_AA),
	gen_new_name_functor(HeadF,new,AF),
	Head =.. [HeadF|LV],
	Body = bd(C1,[Atom]), %%% DEFAULT
%%%	Atom =.. [Sat,State,Formula,_],%%% REMOVE (only for trace2)
%%%	Atom2 =.. [Sat,State,Formula,_FreshVar],%%% REMOVE (only for trace2)
%%%	Body = bd(C1,[Atom2]), %%% REMOVE (only for trace2)

	CandD = (Head,Body),

	(ptime(folding_equivalent_list(CandD,Defs,Solver,FEDef)) ->
        %%% ALTERNATIVE: (folding_subsumer_list(CandD,Defs,Solver,FEDef) ->
	    copy_term(FEDef,Def),
	    OldNew=old
	;
%	    length(LV,A), Label=label(definition(HeadF/A)),
%	    update_transf_tree(Label,[],[CandD],[Def]),
	    copy_term(CandD,Def),
	    OldNew=new
	).



folding_equivalent_list(CandD,[Def|_Defs],Solver,Def) :-
	folding_equivalent(CandD,Def,Solver), ! ,
	debug((
	       write('Folding equivalent'),nl, portray_clauses([CandD,Def])
	      ))
	.
folding_equivalent_list(CandD,[_Def|Defs],Solver,Def) :-
	folding_equivalent_list(CandD,Defs,Solver,Def).


folding_equivalent((OHd1,OBd1),(OHd2,OBd2),Solver) :-
	call_residue((copy_term((OHd1,OBd1),(Hd1,Bd1))),_R1),
	call_residue((copy_term((OHd2,OBd2),(Hd2,Bd2))),_R2),
	Bd1 = bd(C1,Bd1woC1), Bd2 = bd(C2,Bd2woC2), 

	variant(Bd1woC1,Bd2woC2),
	% is it necessary to solve again?
	linking_variables(C1,(Hd1,Bd1woC1),V1),
	solve(Solver,C1,V1,D1),
	linking_variables(C2,(Hd2,Bd2woC2),V2),
	solve(Solver,C2,V2,D2),

	Bd1woC1 = Bd2woC2, % unify bodies
	% same variables in the head
	free_variables(Hd1,[],[],VHd1), 
	free_variables(Hd2,[],[],VHd2), 
	% VHd1 and VHd2 are identical _sets_ of variables
	vars_seteq(VHd1,VHd2), 
	
	equivalent_constraints(Solver,D1,D2).




folding_subsumer_list(CandD,[Def|_Defs],Solver,Def) :-
	folding_subsumed(CandD,Def,Solver), ! ,
	debug((
	       write('Folding subsumed'),nl, portray_clauses([CandD,Def])
	      ))
	.
folding_subsumer_list(CandD,[_Def|Defs],Solver,Def) :-
	folding_subsumer_list(CandD,Defs,Solver,Def).


folding_subsumed((OHd1,OBd1),(OHd2,OBd2),Solver) :-
	call_residue((copy_term((OHd1,OBd1),(Hd1,Bd1))),_R1),
	call_residue((copy_term((OHd2,OBd2),(Hd2,Bd2))),_R2),
	Bd1 = bd(C1,Bd1woC1), Bd2 = bd(C2,Bd2woC2), 

	subsumes(Bd1woC1,Bd2woC2),
	% is it necessary to solve again?
%	linking_variables(C1,(Hd1,Bd1woC1),V1),
%	solve(Solver,C1,V1,D1),
%	linking_variables(C2,(Hd2,Bd2woC2),V2),
%	solve(Solver,C2,V2,D2),

	
%%%	free_variables(Hd1,[],[],VHd1), 
%%%	free_variables(Hd2,[],[],VHd2), 
%%%	vars_subseteq(VHd1,VHd2), 
	free_variables(Hd1,Hd2,[],[]),
	
%	entails(Solver,D1,D2).
	entails(Solver,C1,C2).
