:- use_module(library(conversion), [atoms_terms/2]).

definition(X,Y) :- 
	apply_definition(X,Y).

apply_definition(Output,File) :-
	parseNewDefinition(File,NewPred,Defs,Success), !,
	(Success == error -> Output=error('Syntax error in definition clauses.') ;
	    (Success == wait -> fail ;
		Output=out(label(definition(NewPred)),[],Defs),
		tcl_eval('destroy .wdefinition'))).

% How to write directly in the text widget ?
parseNewDefinition(File,NewPred,Defs,Success) :-
%	concat_atoms(['../',File],F),
	see(File),
	parseNewDef(Defs,NewPred,SingVars,UndefLits,Success),
	seen,
	flatten(SingVars,LSingVars),
	flatten(UndefLits,LUndefLits2),
	list_to_ord_set(LUndefLits2,LUndefLits),
	tcl_warning_new_definition(LSingVars,LUndefLits,Answer),
	(Answer=="no" -> Success=wait ; true).

parseNewDef([Def|Defs],P/A,[S|Ss],[U|UndefLits],Success) :-
	read_term(Cl,[singletons(S)]),
	Cl\==end_of_file, !,
	(Cl= (Head :- Body_Conj) ->
	    conj_to_list(Body_Conj,Body),
	    undefined_lits_in_body(Body,U),
 		 %%% may be get_current_solver?	    
	    find_constraints(_Solver,Body,Constrs,Atoms),
	    Def=(Head,bd(Constrs,Atoms))
	;
	    Cl=Head, Def=(Cl,bd([],[]))
	),
	atom_pred_arity(Head,P,A),
	(defined_predicate(P/A,WhichP) ->  
	    Success=error, tcl_error_pred_exists(P/A,WhichP)
	; 
	    parseNewDef2(P/A,Defs,Ss,UndefLits,Success)
	).
parseNewDef(_,_,_,_,error).
	
parseNewDef2(P/A,[Def|Defs],[S|Ss],[U|UndefLits],Success) :-
	read_term(Cl,[singletons(S)]),
	Cl\==end_of_file, !,
	(Cl= (Head :- Body_Conj) ->
	    conj_to_list(Body_Conj,Body),
	    undefined_lits_in_body(Body,U),
 		 %%% may be get_current_solver?
	    find_constraints(_Solver,Body,Constrs,Atoms),
	    Def=(Head,bd(Constrs,Atoms))
	;
	    Cl=Head, Def=(Cl,bd([],[]))
	),
	(atom_pred_arity(Head,P,A) -> 
	    parseNewDef2(P/A,Defs,Ss,UndefLits,Success)
	;
	    Success=error, tcl_error_def_not_same_pred
	).
parseNewDef2(_,[],[],[],_):-!.
parseNewDef2(_,_,_,_,error):-!.
    
    
restrict(Out,Preds) :- apply_restrict(Out,Preds).

apply_restrict(Out,Preds) :-
	current_program(_,Cls),
	list_to_ord_set(Preds,Preds2),
	apply_restriction(Preds2,Cls,[],CNs2),
	subtract(Cls,CNs2,_,L),
	Out=out(label(restriction(Preds2,L)),L,[]).


apply_restriction(_, [], Cls, Cls) :- !.
apply_restriction(L, Cls, Inter, Cls2) :-
   member_rest(C,Cls,Cls3), clause_hd_bd(C,Head,_), atom_pred_arity(Head,P,A),
   member(P/A,L),!, get_predicates_body(L,C,L2), append(Inter, [C], Inter2),
   apply_restriction(L2, Cls3, Inter2, Cls2).
apply_restriction(_, _, Cls, Cls).




tcl_prepare_new_definition(L) :-
	get_clause_numbers(L,CNs),
	gen_new_name_functor(X,new,pred),
	anti_unify_heads(CNs,Head,U),
	nonvar(Head),
	Head =.. [_|L2],
        new_defs_general_case(U,X,L2,NewDefs2),
        rearrange_definitions(NewDefs2,NewDefs3),
        try_to_drop_arguments(NewDefs3, NewDefs),
	tcl_add_clauses_definition(NewDefs).
tcl_prepare_new_definition(_).

tcl_add_clauses_definition([]).
tcl_add_clauses_definition([(H,B)|L]) :-
	make_clause_readable_(H,B,Head,Body),
        concat_atoms(['addClause ','.wdefinition.text ',' { ',Head,'} ',Body],Cmd),
	tcl_eval(Cmd),
	tcl_add_clauses_definition(L).



gen_new_name_functor(NewDef,Prefix,BaseName) :-

%%%	gen_new_name_functor(NewDef,Prefix,BaseName,1).
%%%gen_new_name_functor(NewDef,Prefix,BaseName,N) :-

%%%	name(N,X),
	name(Prefix,PrefixCh),
	(var(BaseName) ->
%%%		 append(PrefixCh,X,NewDef1Ch)
	    NewDef1Ch = PrefixCh
	;
		 name(BaseName,BNCh),
		 append(PrefixCh,BNCh,NewDef1Ch)
	),
	incr_max_number_defs,
	max_number_defs(M),
	name(M,M1),
	append(NewDef1Ch,M1,NewDef2Ch),
	name(NewDef,NewDef2Ch).
%%%	(clause(defined_predicate(NewDef1/_,_),_) ->
%%%	    (N2 is N + 1, gen_new_name_functor(NewDef,Prefix,BaseName,N2)) ;
%%%	    NewDef1=NewDef).

anti_unify_heads([N], CH, [(H,CG)]):-
   (clause_hd_bd(N,H,bd(Co,B)) ; N= (H,bd(Co,B))),
	append(Co,B,CG),
   copy_term((H,CG),(CH,CCG)), bind_freevars_in_head(CH,CCG), !.

anti_unify_heads([N1,N2], H, [(H1,CG1), (H2,CG2)] ) :-
   (clause_hd_bd(N1,H1,bd(Co1,B1)) ; N1=(H1,bd(Co1,B1))),
   (clause_hd_bd(N2,H2,bd(Co2,B2)) ; N2=(H2,bd(Co2,B2))),
	append(Co1,B1,CG1), append(Co2,B2,CG2),
	copy_term((H1,CG1),(CH1,CCG1)), copy_term((H2,CG2),(CH2,CCG2)),
	bind_freevars_in_head(CH1,CCG1), bind_freevars_in_head(CH2,CCG2),
	anti_unify(H1,H2,H), !.

anti_unify_heads([N1,N2|L], H, [(H1,CG1), (H2,CG2)|L2] ) :-
   (clause_hd_bd(N1,H1,bd(Co1,B1)) ; N1=(H1,bd(Co1,B1))),
   (clause_hd_bd(N2,H2,bd(Co2,B2)) ; N2=(H2,bd(Co2,B2))),
	append(Co1,B1,CG1), append(Co2,B2,CG2),
	copy_term((H1,CG1),(CH1,CCG1)), copy_term((H2,CG2),(CH2,CCG2)),
	bind_freevars_in_head(CH1,CCG1), bind_freevars_in_head(CH2,CCG2),
	anti_unify(CH1,CH2,H3), anti_unify_heads(L, H4, L2), anti_unify(H3,H4,H).

bind_freevars_in_head(H,IneqsB) :-
    body_without_inequalities(IneqsB,_,B), free_variables(B,[],[],VarsB),
    free_variables(H,VarsB,[],FreeVarsH), bind_vars(H,FreeVarsH).

bind_vars(_,[]).
bind_vars(H,[X|L]):- (occurrences_of_var(X,H,1) ->
			 X=fv ; true), bind_vars(H,L).

bind_vars([]).
bind_vars([fv|L]):- bind_vars(L).


new_defs_general_case([],_,_,[]).
new_defs_general_case([(H,B1)],X,L2,[(NewH,B2)]) :-!,
	new_defs_general_case_body(B1,B2), H =..[_|L1],
	new_defs_general_case_args(L1,L2,L3), NewH =..[X|L3].
new_defs_general_case([(H,B1)|LL1],X,L2,[(NewH,B2)|LL2]) :-
	new_defs_general_case_body(B1,B2), H =..[_|L1],
	new_defs_general_case_args(L1,L2,L3),
	NewH =..[X|L3], new_defs_general_case(LL1,X,L2,LL2).
new_defs_general_case_args([],[],[]).
new_defs_general_case_args([A1|L1],[A2|L2], [A1|L3]) :-
	var(A2), !, new_defs_general_case_args(L1, L2, L3).
new_defs_general_case_args([_|L1],[fv|L2], L3) :-
 	new_defs_general_case_args(L1, L2, L3).
new_defs_general_case_args([A1|L1],[A2|L2], L3) :-
	 A1=..[F|LL1],  A2=..[F|LL2],
	append(LL1,L1,NewL1), append(LL2,L2,NewL2),
	new_defs_general_case_args(NewL1, NewL2, L3).
new_defs_general_case_body([],[]).
%new_defs_general_case_body([E|B],B2):-
%	(functor(E, \+,_);functor(E,=,_)), !,
%	new_defs_general_case_body(B,B2).
new_defs_general_case_body([Lit|B],B2):-
	(is_a_negative_literal(Lit,Atom) -> true ; Lit=Atom),
	functor(Atom,=,_), !,
	new_defs_general_case_body(B,B2).
new_defs_general_case_body([E|B],[E|B2]):-
	new_defs_general_case_body(B,B2).


rearrange_definitions([],[]).
rearrange_definitions([D|L],[(Head_def,Body_list)|L2]) :-
   (D=(Head_def :- Body_def) ->
	   conj_to_list((Body_def),Body_list);
           D=(Head_def,Body_list)), rearrange_definitions(L,L2).
rearrange_definition(D2, (Head,Body_list)) :-
    (D2=(Head:- Body) -> conj_to_list((Body),Body_list)
    ;	Head=D2, Body_list=[]).



try_to_drop_arguments(Defs, NewDefs) :-
   get_arity_new_def(Defs,Arity),
   try_to_drop_arguments(Defs, Arity,NewDefs).
try_to_drop_arguments(Defs, Arity, NewDefs) :-
   useless_arguments(Defs,Arity,Args),
   drop_arguments(Defs, Args, NewDefs).

get_arity_new_def([(H,_)|_], A) :- atom_pred_arity(H,_,A).

useless_arguments(_,0,[]) :- !.
useless_arguments(Defs,Arity,[Arity|Args]) :-
   is_a_useless_argument(Defs,Arity), !, X is Arity - 1,
   useless_arguments(Defs,X,Args).
useless_arguments(Defs,Arity,Args) :-
   X is Arity - 1, useless_arguments(Defs,X,Args).

% corrected on the 19/2/98
is_a_useless_argument([D],Arity) :- !,
   D = (H,B), H =.. [_|Args],
   del_4(Var,Arity,Args,NewArgs), var(Var),
%   free_of_var(Var,B). % changed 6/3/98
   (free_of_var(Var,B) ; (pos(Var,NewArgs,OtherOcc), OtherOcc<Arity)).

is_a_useless_argument(Defs,A) :- is_a_useless_argument2(Defs,A).

is_a_useless_argument2([],_).
is_a_useless_argument2([D|Defs],Arity) :-
   D = (H,B), H =.. [_|Args],
   del_4(Var,Arity,Args,NewArgs), var(Var),
   free_of_var(Var,(NewArgs,B)), is_a_useless_argument2(Defs,Arity).

drop_arguments(Defs, [], Defs).
drop_arguments(Defs, [A|L], NewDefs) :-
   drop_argument(Defs, A, NewDefs1),  % must be in decreasing order
   drop_arguments(NewDefs1, L, NewDefs).
drop_argument([],_,[]).
drop_argument([D|Defs], A, [NewD|NewDefs]) :-
   D = (H,B), H =.. [F|Args], del_4(A,Args,NewArgs),
   NewH =.. [F|NewArgs], NewD = (NewH,B),
   drop_argument(Defs, A, NewDefs).


del_4(X,1,[X|L],L).
del_4(X,N,[Y|L],[Y|L1]) :- N > 1, M is N-1, del_4(X,M,L,L1).
del_4(1,[_|L],L).
del_4(N,[Y|L],[Y|L1]) :- N > 1, M is N-1, del_4(M,L,L1).
