:- use_module(library(lists), [ nth/4, sum_list/2 ]).
:- use_module(library(terms), [ term_hash/2 ]).



determ_strategy([],[]).
determ_strategy(_Defs,_FinalCls) :-
	no_solver_selected, !, fail.
determ_strategy(_Defs,_FinalCls) :-
	no_uf_selected, !, fail.
determ_strategy(_Defs,_FinalCls) :-
	no_wqounf_selected, !, fail.
determ_strategy(Defs,FinalCls) :-
	ds_env(Solver,Uf,Wqounf),
	determ_strategy(Defs,Solver,Uf,Wqounf,FinalCls).


ds_env(Solver,Uf,Wqounf):-
	current(solver,Solver),
	current(uf,Uf),
	current(wqounf,Wqounf).

%
% determ_strategy/5
%
determ_strategy([],_Solver,_Uf,_Wqounf,[]).
determ_strategy(Defs,Solver,Uf,Wqounf,FinalCls) :-
	(ds_conditions(Solver,Uf,Wqounf) ->
	    determ_strategy([Defs],[Defs],Solver,Uf,Wqounf,CFactss,FldClsss),
	    flatten(FldClsss,FoldedCls), flatten(CFactss,CFacts),
	    append(CFacts,FoldedCls,FinalCls)
	;
	    throw(ds_conditions(Solver,Uf,Wqounf))
	).

ds_conditions(Solver,Uf,Wqounf) :-
	is_a_solver(Solver),
	is_a_uf(Uf,Solver),
	is_a_wqounf(Wqounf,Solver).

%
% determ_strategy/7
% determ_strategy(Defs,InDefs,Solver,Uf,Wqounf,CFactss,FldClsss)
% FldClsss is a list of lists of lists of folded clauses
determ_strategy([],_Defs,_Solver,_Uf,_Wqounf,[],[]).
determ_strategy([Defs|Defss],InDefss,Solver,Uf,Wqounf,[CFactss1|NewCFs],[FldCls|NewFldCls]) :-
	ptime(unfold_replace_general([Defs|Defss],Solver,Uf,Wqounf,UForests)),
	forests_leaves(UForests,Solver,Leavess,CFactss),
	write('By unfolding:'),nl,portray_clauses_lists(CFactss),nl,portray_clauses_lists(Leavess),nl,

        %%% safety check 
	%length(Leavess,L),length(CFactss,L),
	
	lists_to_pairs(CFactss,Leavess,CFs_NonCFs),
	subsumption_strategy(CFs_NonCFs,Solver,ToSplitPairss),
	(CFs_NonCFs \== ToSplitPairss ->
	    lists_to_pairs(CFactss2,Leavess2,ToSplitPairss)
	    ,write('By subsumption:'),nl,portray_clauses_lists(CFactss2),nl,portray_clauses_lists(Leavess2),nl
	;
	    CFactss2 = CFactss,
	    Leavess2 = Leavess
	),

	(current(split_cfacts,yes) ->
	    %%% split also constrained facts
	    append_pairs(ToSplitPairss,ToSplitss),
	    case_split_strategy_(ToSplitss,Solver,Packets2),
	    foldable_clauses_packets(Packets2,Solver,CFactss3,Packets3),
	    lists_to_pairs(CFactss3,Packets3,Pairs3),
	    subsumption_strategy(Pairs3,Solver,Pairs4),
	    lists_to_pairs(CFactss1,Packets4,Pairs4),
	    remove_empty_lists(Packets4,Packets)
	;
	    %%% split and fold non constrained facts only
	    ToSplitss = Leavess2,
	    CFactss1 = CFactss2,
	    case_split_strategy_(ToSplitss,Solver,Packets)
	),

	
	write('*** '),length(Packets,NumPack),write(NumPack),write(' packets found ***'),nl,
	
%	free_variables(Leavess,[],[],VL),free_variables(Packets,[],[],VP),vars_seteq(VL,VP),  % for debugging purposes
%	write('By case split:'),nl,portray_clauses_lists(Packets),nl,
	
	ptime(define_fold_packets(Packets,InDefss,Solver,OutDefss,FldCls)),
	write('By folding:'),nl, portray_clauses(FldCls),nl,
	delete_elts_sublist(InDefss,OutDefss,NewDefss),
	write('New defs:'),nl,portray_clauses_lists(NewDefss),nl,
	determ_strategy(NewDefss,OutDefss,Solver,Uf,Wqounf,NewCFs,NewFldCls).



remove_empty_lists([],[]).
remove_empty_lists([[]|Ls],L2s) :- !,
	remove_empty_lists(Ls,L2s).
remove_empty_lists([X|Ls],[X|L2s]) :-
	remove_empty_lists(Ls,L2s).


foldable_clauses_packets([],_,[],[]).
foldable_clauses_packets([P|Ps],S,[CF|CFs],[P1|P1s]) :-
	foldable_clauses(P,S,P1,CF),
	foldable_clauses_packets(Ps,S,CFs,P1s).



subsumption_strategy([],_,[]).
subsumption_strategy([Pair|Ps],S,[NewPair|NewPs]) :-
	subsumption_strategy_pair(Pair,S,NewPair),
	subsumption_strategy(Ps,S,NewPs).

subsumption_strategy_pair((CFs,NonCFs),S,(NewCFs,NewNonCFs)) :-
	apply_subsumption_clauses(CFs,S,NewCFs),
	apply_subsumption_clauses(NonCFs,NewCFs,S,NewNonCFs1),
	apply_subsumption_clauses(NewNonCFs1,S,NewNonCFs).


/*-----------
apply_subsumption_clauses(L1,S,L2): L2 is the list L1 in which the clauses which are subsumed
by another clause in L1 are deleted. S is a solver.
------------*/


apply_subsumption_clauses(L1,S,L2):- apply_subsumption_clauses_(L1,S,[],L2).

apply_subsumption_clauses_([],_,L,L).
apply_subsumption_clauses_([E|L],S,I,F) :-
   is_subsumed(E,S,L),!,
   apply_subsumption_clauses_(L,S,I,F).
apply_subsumption_clauses_([E|L],S,I,F) :-
   is_subsumed(E,S,I),!,
   apply_subsumption_clauses_(L,S,I,F).
apply_subsumption_clauses_([E|L],S,Inter,F) :-
    append(Inter,[E],Inter2), apply_subsumption_clauses_(L,S,Inter2,F).


/*-----------
apply_subsumption_clauses(L1,L,S,L2): L2 is the list L1 in which the clauses which are subsumed
by a clause in L are deleted. S is a solver.
------------*/

apply_subsumption_clauses(L1,L,S,L2):- apply_subsumption_clauses(L1,L,S,[],L2).

apply_subsumption_clauses([],_L,_S,L,L).
apply_subsumption_clauses([E|Es],L,S,I,F) :-
	is_subsumed(E,S,L),!,
	apply_subsumption_clauses(Es,L,S,I,F).
apply_subsumption_clauses([E|Es],L,S,Inter,F) :-
	apply_subsumption_clauses(Es,L,S,[E|Inter],F).


is_subsumed(E,S,[F|_]):- is_subsumed_clause(E,S,F),!.
is_subsumed(E,S,[_|L]):- is_subsumed(E,S,L).
is_subsumed_clause((H1,bd(C1,L1)),S,(H2,bd(C2,L2))):-
	length(L2,N),
	length(F,N),
	append(F,_,L1),
	subsumes_chk((H2,L2), (H1,F)),
	call_residue(copy_term((H1,bd(C1,F)),(H3,bd(C3,F3))),_),
	call_residue(copy_term((H2,bd(C2,L2)),(H4,bd(C4,L4))),_),
	unify_with_occurs_check((H4,L4), (H3,F3)),
	solve_clause((H3,bd(C3,F3)),linking_vars,S,(_,bd(D3,_))),
	solve_clause((H4,bd(C4,L4)),linking_vars,S,(_,bd(D4,_))),
	entails(S,D3,D4).



unfold_replace_general([],_Solver,_Uf,_Wqounf,[]).
unfold_replace_general([Defs|Defss],Solver,Uf,Wqounf,[UForest|UForests]) :-
	unfold_replace(Defs,Solver,Uf,Wqounf,UForest),
	unfold_replace_general(Defss,Solver,Uf,Wqounf,UForests).


forests_leaves([],_Solver,[],[]).
forests_leaves([UForest|UForests],Solver,[Leaves|Leavess],[CFacts|CFactss]) :-
	forest_leaves(UForest,Solver,RootLeaves,CFacts1),
	get_leaves(RootLeaves,Leaves),
	append(CFacts1,CFacts),
	forests_leaves(UForests,Solver,Leavess,CFactss).

get_leaves([],[]).
get_leaves([(_R,L)|RLs],L2) :-
	append(L,L1,L2),
	get_leaves(RLs,L1).


% define_fold_packets/5
% 	define_fold_packets(Packets,InDefs,Solver,OutDefs,FldCls)
% Packets is a list of packets
% FldClss is a list of folded clauses
define_fold_packets([],Defs,_Solver,Defs,[]).
define_fold_packets([Pack|Packs],InDefs,Solver,OutDefs,[FldCl|FldCls]) :-
	define_fold_packet(Pack,InDefs,Solver,subsumed,Defs,FldCl),% portray_clause(FldCl),nl,
	define_fold_packets(Packs,Defs,Solver,OutDefs,FldCls).


%
% define_fold_packet/6
%

define_fold_packet([],Defs,_Solver,_Type,Defs,[]).
define_fold_packet([Cl|Cls],InDefs,Solver,Type,OutDefs,FldPack) :-
	unify_heads_input_args([Cl|Cls]), %%% heads of clauses in the same packet are variant
	candidate_defs_packet([Cl|Cls],Solver,CandDefs),
	(all_folding_list(CandDefs,Solver,Type,InDefs,FECandDefs) ->
	    ToUse = FECandDefs,
	    OutDefs = InDefs
	;
	    ToUse = CandDefs,
	    OutDefs = [ToUse|InDefs]
	),
	disjunctive_clause([Cl|Cls],DToFold),
	disjunctive_clause(ToUse,DToUse),
	general_recover_folded_clause(DToFold,DToUse,FldPack).


all_folding_list(Defs,S,T,Defss,FEDefs) :-
	length(Defs,L),
	get_defs_hash(Defs,H),
	member(Defs2,Defss),
	length(Defs2,L),
	get_defs_hash(Defs2,H),
%	permutation(Defs2,Defs3),

	compute_defs_hash_(Defs,Hs),
	lists_to_assoc_pairs(Hs,Defs,HDefs),
	keysort(HDefs,HDefsS),
	equivalence_classes_key(HDefsS,ECDefs),
	lists_to_assoc_pairs(_,DefsPerm,HDefsS),
	find_positions(Defs,DefsPerm,Perm),

	compute_defs_hash_(Defs2,Hs2),
	lists_to_assoc_pairs(Hs2,Defs2,HDefs2),
	keysort(HDefs2,HDefsS2),
	equivalence_classes_key(HDefsS2,ECDefs2),
	
%	permutation_hash(Defs2,Hs,Defs3),
%	write('*** all_folding_list/5: trying a different permutation ***'), nl,
%	all_folding(Defs,Defs3,S,T,Defs4),

	all_folding_classes(ECDefs,ECDefs2,S,T,Defss3),
	flatten(Defss3,Defs3),
	recover_positions(Defs4,Defs3,Perm),
	!,
	write('*** all_folding_list/5: FOUND ***'),nl,
	FEDefs = Defs4.

%%%all_folding_list(Defs,S,T,Defss,FEDefs) :-
%%%	length(Defs,L),
%%%	gen_defs_keys_pos(Defs,DefsK),
%%%	keysort(DefsK,DefsKS),
%%%	equivalence_classes_key(DefsKS,EqClasses),
%%%	length(EqClasses,NEC),

	
%%%	member(Defs2,Defss),
%%%	length(Defs2,L),
%%%	gen_defs_keys_pos(Defs2,Defs2K),
%%%	keysort(Defs2K,Defs2KS),
%%%	equivalence_classes_key(Defs2KS,EqClasses2),
%%%	length(EqClasses2,NEC),
	
%%%	%write('*** all_folding_list/5: trying a different permutation ***'), nl,
%%%	all_folding_classes(EqClasses,S,T,EqClasses2,Ass),
%%%	!,
%%%	write('*** all_folding_list/5: FOUND ***'),nl,
%%%	reconstruct_defs(Ass,Defs2,FEDefs).


get_defs_hash([(A,B)|Defs],H) :-
	functor(A,F,_),
	(defs_hash(F,H1) ->
	    H=H1
	;
	    compute_defs_hash([(A,B)|Defs],H),
	    assert(defs_hash(F,H))
	).


compute_defs_hash(Defs,H) :-
	compute_defs_hash_(Defs,Hs),
	sum_list(Hs,H).

compute_defs_hash_([],[]).
compute_defs_hash_([D|Ds],[H|Hs]) :-
	compute_def_hash(D,H),
	compute_defs_hash_(Ds,Hs).

compute_def_hash((_,bd(_,G)),H) :-
	body_without_equalities(G,_,G1),
	body_without_inequalities(G1,_,G2),
	preds_body(G2,P),
	term_hash(P,H).

permutation_hash([],[],[]).
permutation_hash(L1,[H|Hs],[D1|L2]) :-
	nth(_,L1,D1,R1),
	compute_def_hash(D1,H),
	permutation_hash(R1,Hs,L2).




reconstruct_defs(Ass,Defs2,FEDefs) :-
	keysort(Ass,Ass1),
	reconstruct_defs_(Ass1,Defs2,FEDefs).
reconstruct_defs([],_,[]).
reconstruct_defs([_-P2|Ass],Defs,[D|Ds]) :-
	nth(P2,Defs,D),
	reconstruct_defs(Ass,Defs,Ds).
	

equivalence_classes_key([],[]).
equivalence_classes_key([K-V|Ps],[[K-V|Equiv]|Rest]) :-
	same_key(Ps,K,Equiv,NotEquiv),
	equivalence_classes_key(NotEquiv,Rest).

same_key([],_,[],[]).
same_key([K1-V|Ps],K,[K1-V|Equiv],NotEquiv) :- K1 == K, !,
	same_key(Ps,K,Equiv,NotEquiv).
same_key([K1-V|Ps],K,Equiv,[K1-V|NotEquiv]) :- 
	same_key(Ps,K,Equiv,NotEquiv).


gen_defs_keys_pos(Defs,DefsKP) :-
	gen_defs_keys_pos(Defs,DefsKP,1).

	
gen_defs_keys_pos([],[],_).
gen_defs_keys_pos([(H,bd(C,G))|Cls],[P-(N,(H,bd(C,G)))|Ps],N) :-
	preds_body(G,P),
	N1 is N+1,
	gen_defs_keys_pos(Cls,Ps,N1).


preds_body([],[]).
preds_body([L|Ls],[F/A|Ps]) :-
	functor(L,F,A),
	preds_body(Ls,Ps).


all_folding_classes([],[],_,_,[]).
all_folding_classes([C1|C1s],[C2|C2s],S,T,[Defs3|Ds]) :-
	lists_to_assoc_pairs(_,Defs1,C1),
	lists_to_assoc_pairs(_,Defs2,C2),
	all_folding(Defs1,Defs2,S,T,Defs3),
	all_folding_classes(C1s,C2s,S,T,Ds).
	



all_folding_([],_S,_T,[],[]).
all_folding_([K-(P1,D1)|D1s],S,subsumed,D2s,[P1-P2|Ass]) :-
	nth(_,D2s,(K-(P2,D2)),R),
	folding_subsumed(D1,D2,S),
	all_folding_(D1s,S,subsumed,R,Ass).


all_folding([],[],_S,_T,[]).
all_folding([D1|D1s],Defs2,S,T,[D2|D2s]) :-
	nth(_,Defs2,D2,R),
	(T == subsumed ->
	    folding_subsumed(D1,D2,S)
	;
	    T == equivalent ->
	    folding_equivalent(D1,D2,S)
	),
	all_folding(D1s,R,S,T,D2s).

%%%all_folding([],_S,_T,[]).
%%%all_folding([D1|D1s],S,T,[D2|D2s]) :-
%%%	(T == subsumed ->
%%%	    folding_subsumed(D1,D2,S)
%%%	;
%%%	    T == equivalent ->
%%%	    folding_equivalent(D1,D2,S)
%%%	),
%%%	all_folding(D1s,S,T,D2s).

			       
candidate_defs_packet([Cl|Cls],S,CandDefs) :-
	common_constr_eqsineqs_([Cl|Cls],_,_,S,C,EqsIneqs),
	Cl = (H,_),
	free_variables(H,[],[],VH),
	defs_bodies_unnecessary_vars([Cl|Cls],C,EqsIneqs,VH,DefsBodies,UnHV),
	free_variables(VH,UnHV,[],NewHV),
	gen_new_name_functor(CDP,new,p),
	NewHd =.. [CDP|NewHV],
	
	assert_newpred_modes(CDP,NewHV,DefsBodies),
	
	disjunctive_clause(CandDefs,(NewHd,DefsBodies)).


assert_newpred_modes(CDP,NewHV,DefsBodies) :-
	leftmost_atom_input_args(DefsBodies,InArgs),
	free_variables(InArgs,[],[],InVars),
	newpred_modes(NewHV,InVars,Modes),
	assert(mode_definition(CDP,Modes)).


newpred_modes([],_,[]).
newpred_modes([X|Xs],Vars,[M|Modes]) :-
	(belongs(X,Vars) ->
	    M = '+'
	;
	    M = '-'
	),
	newpred_modes(Xs,Vars,Modes).
	

leftmost_atom_input_args([],[]).
leftmost_atom_input_args([bd(_,Lits)|Bs],InArgs) :-
	(leftmost_atom(Lits,_,1,_,A)->
	    input_args(A,InArgsA),
	    InArgs = [InArgsA|InArgs2]
	;
	    InArgs = InArgs2
	),
	leftmost_atom_input_args(Bs,InArgs2). 



defs_bodies_unnecessary_vars([],_,_,Vars,[],Vars).
defs_bodies_unnecessary_vars([(H,bd(C,Lits))|Cls],D,EqsIneqs,InVars,[bd(DefC,DefLits)|DefBodies],OutVars) :-
	list_difference(C,D,DefC),
	list_difference(Lits,EqsIneqs,DefLits),
	free_variables(H,bd(DefC,DefLits),[],UV),
	linking_variables(UV,InVars,OutVars1),
	defs_bodies_unnecessary_vars(Cls,D,EqsIneqs,OutVars1,DefBodies,OutVars).

%%% REMOVE
%%%build_candidate_defs([],_,[]).		     
%%%build_candidate_defs([Cl|Cls],H,[D|Ds]) :-
%%%	build_candidate_def(Cl,H,D),
%%%	build_candidate_defs(Cls,H,Ds).

%%%build_candidate_def(Cl,H,D) :- build_candidate_def(Cl,H,D,_,_).

%%%build_candidate_def(Cl,H,D,Eq,Ineq) :-
%%%	Cl = (_,bd(C,Lits)),
%%%	body_without_inequalities(Lits,Ineq,Lits1),
%%%	body_without_equalities(Lits1,Eq,Lits2),
%%%	D = (H,bd(C,Lits2)).


unify_heads_input_args([]).
unify_heads_input_args([_Cl]).
unify_heads_input_args([Cl1,Cl2|Cls]) :-
	Cl1 = (H1,_),
	Cl2 = (H2,_),
	input_args(H1,InArgs1),	input_args(H2,InArgs2),
	unify_with_occurs_check(InArgs1,InArgs2),
	unify_heads_input_args([Cl2|Cls]).
