
%%%:- dynamic initial_prog/1.
%%%:- dynamic clause_definition/4.
%%%:- dynamic mode_definition/2.
%%%:- dynamic type_definition/2.
%%%:- dynamic max_number_clauses/1.
%%%:- dynamic max_number_laws/1.
%%%:- dynamic defined_predicate/2.
%%%:- dynamic max_number_program/1.
%%%:- dynamic current_program/2.
%%%:- dynamic node/5.
%%%:- dynamic law_definition/4.


:- use_module(library(lists), [ non_member/2 ]).
:- use_module(library(charsio), [atom_to_chars/2,
				 read_from_chars/2
				]).


:- multifile current/2.
:- multifile portray/1.


:- dynamic current/2.



:- discontiguous button_name_rule/2.


sh :-listing_transf_tree.

%%%listing_transf_tree :- 
%%%	listing(clause_definition),
%%%	listing(mode_definition),
%%%	listing(type_definition),
%%%	listing(law_definition),
%%%	listing(max_number_clauses),
%%%	listing(max_number_program),
%%%	listing(node),
%%%	listing(defined_predicate),
%%%	listing(max_number_laws),
%%%	listing(current_program).


listing_transf_tree :- 
	predicate_property(Hd,dynamic),
	functor(Hd,P,A),
	listing(P/A),
	fail.
listing_transf_tree.

erase_environment :-
	dynamic_relation_prog_state(R),
	retractall_definitions_for_rel(R),
	fail.
erase_environment :- tcl_eval(destroy4).

erase_environment_not_theory :-
	dynamic_relation_prog_state(R),
	((R=law_definition/4 ; R=law_applied/1 ; R=max_number_laws/1) ->
    	    erase_environment_for_laws ;
	    retractall_definitions_for_rel(R)),
	fail.
erase_environment_not_theory  :- tcl_eval(destroy4).

erase_environment_for_laws :-
	clause(law_definition(_,N,_,_),_),
	(clause(law_applied(N),_) -> true ;
	    retract(law_definition(_,N,_,_))),
	fail.
erase_environment_for_laws :-
	retractall(law_applied(_)).

erase_environment_ask_theory :-
	clause(law_applied(_),_), !,
	Msg='{Do you want to keep the current theory ?}',
	tcl_warning_yesno(Msg,Answer),
	(Answer=="yes" -> erase_environment_not_theory ;
	    erase_environment).
erase_environment_ask_theory :-
	erase_environment.

dynamic_relation_prog_state(initial_prog/1).
dynamic_relation_prog_state(clause_definition/4).
dynamic_relation_prog_state(mode_definition/2).
dynamic_relation_prog_state(type_definition/2).
dynamic_relation_prog_state(max_number_clauses/1).
dynamic_relation_prog_state(max_number_laws/1).
dynamic_relation_prog_state(defined_predicate/2).
dynamic_relation_prog_state(max_number_program/1).
dynamic_relation_prog_state(current_program/2).
dynamic_relation_prog_state(law_applied/1).
dynamic_relation_prog_state(node/5).
dynamic_relation_prog_state(law_definition/4).

dynamic_relation_prog_state(predicate_program_clause/3).
dynamic_relation_prog_state(initial_clauses/1).
dynamic_relation_prog_state(max_number_defs/1).
dynamic_relation_prog_state(defs_hash/2).



retractall_definitions_for_rel(P/A) :-
	functor(T,P,A),
	retractall(T).



load_initial_program :-
	load_initial_program(1).
load_initial_program(N) :-
	read(C),
	C\==end_of_file, !,
%%% case of a type declaration
	(C=..[type,Pred,Type] -> 
		 assert_relation_prog_state(type_definition,[Pred,Type]), M = N
	;
%%% case of a mode declaration
%%% mode(P,L) represents the modes associated to predicate P. L
%%% L is a list with elements in {+,-} s.t. length(L)=arity(P).
%%% The i-th element of L corresponds to the i-th argument of P.
%%% `+' specifies an input argument, that
%%% is, an argument which should contain a ground term at runtime.

		 ( C=..[mode,Pred,Mode] -> 
			  assert_relation_prog_state(mode_definition,[Pred,Mode]), M = N
		 )
	;
%%% case of predicate abstraction
		 ( C=..[pred_abs,State,Constr] -> 
			  assert_relation_prog_state(pred_abs,[State,Constr]), M = N
		 )
	;		 
%%% case of a clause
		 M is N + 1, parse_init_prog_clause(N, C)
	),
	load_initial_program(M).

load_initial_program(N) :-
	M is N-1,
        assert_relation_prog_state(max_number_clauses,[M]),
        assert_relation_prog_state(max_number_program,[1]),
	count_from_to(1,M,L),
	assert_relation_prog_state(current_program,[1,L]),
%	assert_relation_prog_state(program_clauses,[1,L]),
%	assert_relation_prog_state(transformation_tree,[1,init,[],tt(1,init,current)]).
	assert_relation_prog_state(node,[1,0,[],label(init),L]).





incr_max_number_defs :-
	(max_number_defs(N) ->
		 retract(max_number_defs(N)),
		 M is N+1,
		 assert(max_number_defs(M))
	;
		 assert(max_number_defs(1))
	).



assert_fact_once(R) :-
	ground(R),
	(call(R) ->
	    true
	;
	    assert(R)
	).

program_clauses(P,CNs) :-
	node(P,_,_,_,CNs).

parse_init_prog_clause(N, C) :-
        (C= (Head :- Body_Conj) ->
            conj_to_list(Body_Conj,Body)
        ;
				Head=C, Body=[]
	),
        clause_skeleton([Head|Body],HdPred,Skeleton,LengthBd),
	find_constraints(Solver,Body,Constrs,Atoms),
	(var(Solver) -> 
	    true
	;
	    ((\+ current(solver,Solver)) ->
		set_current(solver,Solver)
	    ;
		true
	    )
	),
        assert_once_relation_prog_state(defined_predicate,[HdPred,1]),
        assert_relation_prog_state(clause_definition,N,Head,bd(Constrs,Atoms),
                   [prog(1),skeleton(Skeleton),length_bd(LengthBd)]).

assert_relation_prog_state(clause_definition,N,Head,Body, Params_C) :- !,
        assert(clause_definition(N,Head,Body, Params_C)).
assert_relation_prog_state(Rel,Args) :-
        Cl=..[Rel|Args],
        assert(Cl).
assert_once_relation_prog_state(Rel,Args) :-
        Cl=..[Rel|Args],
        (call(clause(Cl,_)) -> true ; assert(Cl)).

retract_defined_predicate(Rel) :-
	retract(defined_predicate(Rel,_)).
retract_current_program(N) :-
	retract(current_program(N,_)).
retract_node(N) :-
	retract(node(N,_,_,_,_)).
delete_son(N,Son) :-
	node(N,Father,Sons,LabelN,CNs),
	del_occ(Son,Sons,NewSons),
	retract(node(N,Father,Sons,LabelN,CNs)),
	assert(node(N,Father,NewSons,LabelN,CNs)).
	
retract_clauses([]).
retract_clauses([C|L]) :-
	retract(clause_definition(C,_,_,_)),
	retract_clauses(L).


cache_predicate_program_clause(_,[]).
cache_predicate_program_clause(P,[CN|CNs]) :-
	clause_hd_bd(CN,H,_),
	functor(H,Fun,Ar),
	assert_relation_prog_state(predicate_program_clause,[Fun/Ar,P,CN]),
	cache_predicate_program_clause(P,CNs).

clauses_for_predicate(_/_,[],[]).
clauses_for_predicate(Fun/Ar,[CN|CNs],PredDefs) :-
	clause_hd_bd(CN,H,_),
	(functor(H,Fun,Ar) ->
		 PredDefs = [CN|RDefs],
		 clauses_for_predicate(Fun/Ar,CNs,RDefs)
	;
		 clauses_for_predicate(Fun/Ar,CNs,PredDefs)
	).

get_defined_predicates([],[]).
get_defined_predicates([CN|CNs],DefinedPreds) :-
	(setof(F/A,(CN1^H^B^Sk^(member(CN1,[CN|CNs]),clause_definition(CN1,H,B,Sk),functor(H,F,A))),DefinedPreds) ->
		 true
	;
		 DefinedPreds = []
	).

get_failed_predicates(DefinedPreds,FailedPreds) :-
	( setof(F/A,(CP^(call_pattern(F,A,CP),non_member(F/A,DefinedPreds))),FailedPreds) ->
		 true
	;
		 FailedPreds = []
	).

show_initial_program :- show_program(1).
show_program(PN) :-
	program_clauses(PN,ClNs),
	concat_atoms(['addProofText2 ',
					  '{Program ',PN,':\n}'],CmdProgramLine),
	tcl_eval(CmdProgramLine),
	tcl_add_clauses_proofText(ClNs),
	tcl_eval('clearProofSelText').

set_current_program(N) :-
	node_clauses(N,CNs),
	retract(current_program(_,_)), assert(current_program(N,CNs)),
	tcl_infonote('Current program changed.').

undo :-   current_program(N,_),
	  node_label(N,Label),
	  node_father(N,Father),
%	  node_clauses(N,Cs1),
	  node_clauses(Father,Cs2),
          new_clauses_at_prog(N,L),
          ( max_number_program(N) ->
	      (new_max_number_clauses(Cs2,L,NewMax),
	      update_max_number_clauses(NewMax),
	      update_max_number_program(Father)) ; true),
	  ( Label=definition(Rel) ->
               retract_defined_predicate(Rel) ; true),
	  retract_current_program(N),
	  delete_son(Father,N),
	  assert_relation_prog_state(current_program,[Father,Cs2]),
	  retract_node(N),
	  compoundtoatom(L,Old1),
	  compoundtoatom(Cs2,RecCs2),
	  concat_atoms(['removeTagsCls ',Old1, ' ',RecCs2],RemTags),
   	  tcl_eval(RemTags),
	  retract_clauses(L),
	  tcl_infonote('Last transformation undone.').

new_max_number_clauses(L,NewMax) :-
	max_number_clauses(Max),
	length(L,N),
	NewMax is Max-N.
    
new_max_number_clauses(LFather,L,NewMax) :-
	max_number_clauses(Max),
	findall(X, (member(X,L), member(X,LFather)), L2),
	length(L2,N),
	NewMax is Max-N.
    	  
tcl_alertnote(Msg) :-
	flag(map_tcl_interp,I),
	concat_all(['set msg ','"',Msg,'"'], Cmd),
	tcl_eval(I, Cmd,_),
        tcl_eval(I,'alertnote $msg',_).

tcl_infonote(Msg) :-
	concat_all(['set msg ','"',Msg,'\n"'], Cmd),
	tcl_eval(Cmd),
        tcl_eval('addProofText2 $msg').


is_a_transformation_rule(default,define).
is_a_transformation_rule(default,restrict).
is_a_transformation_rule(default,unfold).
is_a_transformation_rule(default,fold).
is_a_transformation_rule(default,replace_goal).
is_a_transformation_rule(default,move_goal).
is_a_transformation_rule(default,add_goal).
is_a_transformation_rule(default,delete_goal).
is_a_transformation_rule(default,replace_clauses).
is_a_transformation_rule(default,move_clause).
is_a_transformation_rule(default,add_clauses).
is_a_transformation_rule(default,delete_clauses).
is_a_transformation_rule(default,generalize).
is_a_transformation_rule(default,case_split).

is_a_transformation_rule(default,solve_constraints(linking_vars)).
is_a_transformation_rule(default,ccr).
is_a_transformation_rule(default,caf(soft)).
is_a_transformation_rule(default,split_determ).

button_name_rule('define',define).
button_name_rule('restrict',restrict).
button_name_rule('unfold',unfold).
button_name_rule('fold',fold).
button_name_rule('replace goal',replace_goal).
button_name_rule('move goal',move_goal).
button_name_rule('add goal',add_goal).
button_name_rule('delete goal',delete_goal).
button_name_rule('replace clauses',replace_clauses).
button_name_rule('move clause',move_clause).
button_name_rule('add clauses',add_clauses).
button_name_rule('delete',delete_clauses).
button_name_rule('generalize',generalize).
button_name_rule('case split', case_split).

button_name_rule('solve constr.', solve_constraints(linking_vars)).
button_name_rule('constr. folding', caf(soft)).
button_name_rule('ccr', ccr).
button_name_rule('clause split', split_determ).





debug(G) :-
	(current(debug,yes) ->
	    call(G)
	;
	    true
	).


set_current(Type,What) :-
	retractall(current(Type,_)),
	assert(current(Type,What)).

is_a(Type,A,L) :-
	ground(Type),
	atom_to_chars(is_a_,ISA),
	atom_to_chars(Type,T),
	atom_to_chars(.,Dot),	
	flatten([ISA,T,Dot],ISAT),
	read_from_chars(ISAT,Pred),
	functor(G,Pred,A),
	arg(1,G,What),
	( A = 2 ->
	    arg(2,G,What2), G1=What2^G
	;
	    G1=G
	), % currently supported <= 2 args... Generalize if needed.
	( setof(What,G1,L) ->
	    true
	;
	    L = []
	).



%%% START STRATEGIES

is_a_strategy(ds).
is_a_strategy(cs).

button_name_rule('determiniz.',ds).
button_name_rule('specialization',cs).



%%% END STRATEGIES





/*
showRules :-
	bagof(R, is_a_transformation_rule(R), L),
	compoundtoatom(L,L_tcl),
	flag(map_tcl_interp,I),
	concat_all(['foreach i ',L_tcl, ' {.bot.leftbot.config2.f.list insert end $i} '],Cmd),
	tcl_eval(I, Cmd,_).
*/

assert_new_clauses(P,LC,LN,Param) :-
	max_number_clauses(Max),
	assert_new_clauses(P,Max,LC,LN,Param).

assert_new_clauses(_,Max,[],[],_) :-
	update_max_number_clauses(_,Max).

assert_new_clauses(P,Max,[(Hd,bd(C,G))|Cs],[M|Ms],[Param]) :-
     M is Max+1,
	  append(C,G,Bd),
     clause_skeleton([Hd|Bd],_,Skeleton,LengthBd),
	  find_constraints(_Solver,Bd,Constrs,Atoms),	  
     assert_relation_prog_state(clause_definition,M,Hd,bd(Constrs,Atoms),
		   [Param,prog(P),skeleton(Skeleton),length_bd(LengthBd)]),
     assert_new_clauses(P,M,Cs,Ms,[Param]).


update_transf_tree(Label,OldCNs,NewCls,NewCNs) :-
%	retract(current_program(P,L)),
	current_program(P,L),
	update_max_number_program(_,NewMaxP),
	assert_new_clauses(NewMaxP,NewCls,NewCNs,[Label]),
	(Label=label(move_clause(CN,CN2)) ->
	    (CN2==0 ->
			  move_first_pos(CN,L,NewL)
		 ;
			  move_after_elt(CN,CN2,L,NewL)
		 )
	;

	    replace_in_list(OldCNs,L,NewCNs,NewL)
	
%%%	    % the following test should always succeed - FIX!
%%%	     (memberchk(L,OldCNs) ->
%%%		 replace_in_list(OldCNs,L,NewCNs,NewL)
%%%	     ;
%%%		 append(OldCNs,[L],NewL)
%%%	     )
	),
 %	assert(current_program(NewMaxP,NewL)),
	retract(node(P,Father,Sons,LabelP,CNsP)),
	assert(node(P,Father,[NewMaxP|Sons],LabelP,CNsP)),
	assert(node(NewMaxP,P,[],Label,NewL)),
	( Label=label(definition(Rel)) ->
		 assert_once_relation_prog_state(defined_predicate,[Rel,NewMaxP])
	;
		 ( Label=label(add_clauses(Rels)) ->
			  assert_new_relations(Rels,NewMaxP)
		 ;
			  true)),
	retract(current_program(P,L)),
	assert(current_program(NewMaxP,NewL)).

assert_new_relations([],_).
assert_new_relations([Rel|Rels],NewMaxP) :-
	assert_once_relation_prog_state(defined_predicate,[Rel,NewMaxP]),
	assert_new_relations(Rels,NewMaxP).

update_max_number_clauses(New) :-
	update_max_number_clauses(_,New).
update_max_number_clauses(Old,New) :-
	retract(max_number_clauses(Old)),
	assert(max_number_clauses(New)).
update_max_number_laws(Old,New) :-
	(clause(max_number_laws(_),_) ->
		retract(max_number_laws(Old)) ; true),
	assert(max_number_laws(New)).

update_max_number_program(New) :-
	update_max_number_program(_,New).
update_max_number_program(Old,New) :-
	retract(max_number_program(Old)),
	(var(New) -> New is Old + 1 ; true),
	assert(max_number_program(New)).



portray(rat(N,D)) :- portray(N),write('/'),portray(D).


show_numbered_clauses([]).
show_numbered_clauses([CN|CNs]) :-
	clause_hd_bd(CN,Hd,Bd),
	write(CN), write('. '),
	print_length(write(CN),Length),
	Pos is Length + 2,
	show_clause((Hd,Bd),Pos,map),nl,
	show_numbered_clauses(CNs).

show_clauses_without_numbers([]).
show_clauses_without_numbers([CN|CNs]) :-
	clause_hd_bd(CN,Hd,Bd),
	show_clause((Hd,Bd)),nl,
	show_clauses_without_numbers(CNs).


show_program_as_term([]).
show_program_as_term([CN|CNs]) :-
	clause_hd_bd(CN,H,bd(C,As)),
	append(C,As,B),
	numbervars([H|B],0,_),
	write(':- assert(myclause(['),
	writeq_list([H|B]),
	write('])).'), nl,
	show_program_as_term(CNs).

show_program_for_sicstus([]).
show_program_for_sicstus([CN|CNs]) :-
	clause_hd_bd(CN,H,bd(C,As)),
	append(C,As,B),
	numbervars([H|B],0,_),
	show_clause_for_sicstus((H,bd(C,As))),nl,
	show_program_for_sicstus(CNs).

show_table_directives([]).
show_table_directives([P/A|Preds]) :-
	write(':- table '),write(P),write('/'),write(A),write('.'),nl,
	show_table_directives(Preds).

show_clauses_for_xsb([]).
show_clauses_for_xsb([CN|CNs]) :-
	clause_hd_bd(CN,H,B),
	B = bd(C,As),
	propagate_and_remove_equalities(C,C1),
	show_clause((H,bd(C1,As))),nl,
	show_clauses_for_xsb(CNs).


show_failed_clauses([]).
show_failed_clauses([F/A|FailPreds]) :-
	length(Args,A),
	numbervars(Args,0,_),
	H =.. [F|Args],
	write(H),write(' :- fail.'),nl,
	show_failed_clauses(FailPreds).


writeq_list([]).
writeq_list([A|As]) :-
	(portray(A) ->
		 true
	;
		 writeq(A)
	),
	( As\==[] ->
		 write(',')
	;
		 true
	),
	writeq_list(As).


show_clause(C) :-
	show_clause(C,0,map).
show_clause_for_sicstus(C) :-
	show_clause(C,0,sicstus).


show_clause((H,B),Pos,T):- 
	numbervars((H,B),0,_), 
	writeq(H),
	(B == bd([],[]) ->
	    write('.')
	;
	    print_length(write(H),LH),
	    LH2 is LH + 4 + Pos,
	    write(' :- '),
	    B = bd(Cs,As), 
	    (T == map -> 
		append(Cs,As,B1),
		show_body(B1,LH2)
	    ;
		T == sicstus ->
		( Cs \== [] -> 
		    write('{'),show_body_nodot(Cs,LH2),write('}')
		;
		    true
		),
		( As \== [] -> 
		    write(','),show_body(Cs,LH2)
		;
		    write('.')
		)	
	    )
	).


show_body(L,LH) :- show_body_lines(L,9,LH,80,dot).
show_body_nodot(L,LH) :- show_body_lines(L,9,LH,80,nodot).

show_body_lines([H], PosInitBody,PosLine, Max, Dot) :- !, 
	print_length(write(H),Length), NewPos is PosLine + Length + 1,
	(NewPos >= Max ->
		 nl, tab(PosInitBody)
	; true),
	print(H),
	( Dot == dot ->
	    write('.')
	;
	    true
	).
show_body_lines([H|T],PosInitBody,Pos,Max, Dot) :- 
	print_length(write(H),Length), NewPos is Pos + Length + 2,
	(NewPos >= Max ->
		 (nl, tab(PosInitBody), Pos2 is PosInitBody +  Length + 2)
	; Pos2=NewPos),
	print(H), write(', '),
	show_body_lines(T, PosInitBody,Pos2,Max, Dot).


read_session :-
	read(C), C\==end_of_file, !,
	assert(C), read_session.
read_session.


add_theory_old(Out,T,ToRetract,ToAssert) :-
	( clause(max_number_laws(N),_) ->
		 (N2 is N+1,
			  (is_user_defined_theory(T) ->
					add_theory(Out,T,N2,ToAssert)
			  ;
					check_and_add_theory(Out,T,N2,ToRetract,ToAssert)))
	;   (add_theory(Out,T,1,ToAssert), ToRetract=[])).


is_user_defined_theory(T) :- name(map_tmp_user_laws,F), name(T,NT), append(_,F,NT).

check_and_add_theory(Out,T,N,ToRetract,ToAssert) :-
	(clause(law_definition(T,_,_,_),_) ->
		 (tcl_ask_theory_overwrite(T,Answer),
			  (Answer=="yes" ->
					(bagof(X, Y^Z^law_definition(T,X,Y,Z), ToRetract),
						 retractall(law_definition(T,_,_,_)),
						 add_theory(Out,T,N,ToAssert))
			  ; true)) ;
	    (ToRetract=[], add_theory(Out,T,N,ToAssert))).


assert_laws([]).
assert_laws([Law|Laws]) :-
	assert(Law),
	assert_laws(Laws).

add_theory(Out,T,N,[law_definition(T,N,LLHS,LRHS)|ToAssert]) :-
	read(C),
	C\==end_of_file, !,
	(C=..['->',LHS,RHS] ->
		 (conj_to_list(LHS,LLHS),
			  conj_to_list(RHS,LRHS),
%	    assert_relation_prog_state(law_definition,[T,N,LLHS,LRHS]),
			  M is N + 1,
			  add_theory(Out,T,M,ToAssert))
	;
		 Out=error
).

add_theory(ok,_,N,[]) :-
	M is N-1,
	update_max_number_laws(_,M).

