:- use_module(library(system), [ file_exists/1,
                                 file_property/2, shell/2 ]).

:- use_module(library(terms), [ subsumes_chk/2 ]).

:- use_module(library(ordsets), [
				 list_to_ord_set/2,
				 ord_subtract/3
				]).


% apply_transformation_rule(Rule,,Args2,Args3) :- 

apply_transformation_rule(Rule,Args1,Args2,Args3) :- 
	apply_transformation_rule(Rule,Args1,Args2,Args3,Output),
%	(Rule==definition -> true ; tcl_eval(clearProofSelText)),
	tcl_eval(clearDefSelText),
	% added a test to reuse existing tcl code for lawSel
	% (Rule \= solve_constraints(_IntVarsCode) -> tcl_eval(clearLawSelText) ; true),
	tcl_eval(clearLawSelText),
	(Rule==definition ->
		 true ;
		 tcl_eval(clearProofSelText)
	),
	(Output=error(Alert) ->
		 tcl_alertnote(Alert), fail
	;
		 Output = out(Label,OldCNs,NewCls),
		 update_transf_tree(Label,OldCNs,NewCls,NewCNs),
		 update_proofText(Label,OldCNs,NewCNs)
	).


apply_transformation_rule(definition,_,_,_,Success) :-
%%% map_tmp_newdef is the name of the temporary file in which defs are stored.
	definition(Success,map_tmp_newdef). 
apply_transformation_rule(restrict,Args,_,_,Success) :-
	make_args_restrict(Args,Preds),
	restrict(Success,Preds).
	
apply_transformation_rule(unfold,[atom(CN,AN)],_,_,Success) :- !,
	clause_hd_bd(CN,_,bd(C,_)), length(C,LC), AN1 is AN-LC,
	unfold(Success,CN,AN1).
apply_transformation_rule(unfold,_,_,_,Success) :-
	Success = error('Bad selection for unfolding.').
	
apply_transformation_rule(fold, ToFold,ToUse,_,Success) :-
	get_clause_numbers(ToFold,L1),
	get_clause_numbers(ToUse,L2),
	general_fold(Success,L1,L2).

% Tamaki - Sato folding
%apply_transformation_rule(fold, ToFold,ToUse,_,Success) :-
%	get_clause_numbers(ToFold,[L1]),
%	get_clause_numbers(ToUse,[L2]),
%	fold(Success,L1,L2,R).


get_clause_numbers([],[]).
get_clause_numbers([cl(N)|L],[N|L2]) :-
	get_clause_numbers(L,L2).

apply_transformation_rule(replace_goal, L,_,[law(Law)],Success) :-	
	L=[atom(ToReplace,AN)|L2], !,
	make_args_replace_goal(ToReplace,ANs,L2,Success),
	(var(Success) -> replace_goal(Success,ToReplace,Law,[AN|ANs]) ; true).

apply_transformation_rule(replace_goal, L,_,[law(Law)],Success) :-	
	L=[cl(ToReplace)], !,
	(var(Success) -> replace_goal(Success,ToReplace,Law,[all]) ; true).

apply_transformation_rule(replace_goal, _,_,_,Success) :-
	Success= error('Bad selection.').


apply_transformation_rule(move_goal, L,_,_,Success) :-	
	L=[atom(ToMove,AN)|L2],
	make_args_move_goal(ToMove,ANs,L2,Pos,Success),
	(var(Success) -> move_goal(Success,ToMove,[AN|ANs],Pos) ; true).

apply_transformation_rule(add_goal, L,_,_,Success) :-	
	L=[atom(ToDupplicate,AN)|L2],
	make_args_replace_goal(ToDupplicate,ANs,L2,Success),
	(var(Success) -> add_goal(Success,ToDupplicate,[AN|ANs]) ; true).

apply_transformation_rule(delete_goal, L,_,_,Success) :-	
	L=[atom(ToDelete,AN)|L2],
	make_args_replace_goal(ToDelete,ANs,L2,Success),
	(var(Success) -> delete_goal(Success,ToDelete,[AN|ANs]) ; true).

apply_transformation_rule(move_clause, L,_,_,Success) :-	
	make_args_move_clause(L,ToMove,NewPos,Success),
	(var(Success) -> move_clause(Success,ToMove,NewPos) ; true).

apply_transformation_rule(add_clauses, _,_,_,Success) :-	
	(var(Success) -> add_clauses(Success) ; true).

apply_transformation_rule(delete_clauses, L,_,_,Success) :-	
	get_clause_numbers(L,ToDel),
	(var(Success) -> delete_clauses(Success,ToDel) ; true).

apply_transformation_rule(replace_clauses, L,_,_,Success) :-	
	get_clause_numbers(L,ToDel),
	(var(Success) -> replace_clauses(Success,ToDel) ; true).

apply_transformation_rule(ccr,_,_,_,Success) :-
	(current(solver,S) ->
	    current_program(_,CNs),
	    clause_hd_bd_list(CNs,Cls),
	    do_ccr_(Cls,S,OutCls),
	    Success = out(label(ccr),CNs,OutCls)
	;
	    Success = error('No solver selected')
	).


apply_transformation_rule(case_split,[cl(N)],_,_,Success) :-
	get_new_head(N,I,case_split),
	tk_next_event(I,E),
	(E==error ->
	    Success = error('abort case split.')
	;
	    (check_entry_case_split(N,E) ->
		(tcl_event(I, 'destroy .',_), tcl_delete(I),
		    split_clause(Success,N,E))
	    ;
		(tcl_alertnote('Bad selection. \
			      Enter an instance of the head of clause to split.'),
		tcl_event(I, 'destroy .',_),
		    tcl_delete(I),
		    apply_transformation_rule(case_split,[cl(N)],_,_,Success)
		)
	    )
	).

apply_transformation_rule(split_determ,[cl(N)],_,_,Success) :-
	get_new_head(N,I,split_determ),
	tk_next_event(I,E),
	(E==error ->
	    Success = error('abort clause split.')
	;
	    (check_entry_split_determ(N,E) ->
		(tcl_event(I, 'destroy .',_), tcl_delete(I),
		    apply_split_determ(Success,N,E))
	    ;
		(tcl_alertnote('Error in clause split'),
		tcl_event(I, 'destroy .',_),
		    tcl_delete(I),
		    apply_transformation_rule(split_determ,[cl(N)],_,_,Success)
		)
	    )
	).



apply_transformation_rule(generalize,[cl(N)],_,_,Success) :-
	get_generalized_clause(N,I),
	tk_next_event(I,E),
	(E==error -> Success = error('abort generalize.') ;
	    (check_entry_generalized_clause(N,E,Gen) ->
		(tcl_event(I, 'destroy .',_), 	tcl_delete(I),
		generalize_clause(Success,N,Gen)) ;
		(tcl_alertnote('Bad selection. \
Not a generalization of the selected clause.'),
                 tcl_event(I, 'destroy .',_), tcl_delete(I),
                 apply_transformation_rule(generalize,[cl(N)],_,_,Success)))).


% solve all constraints in a clause
apply_transformation_rule(solve_constraints(Solver,IntVarsCode), L,_,_,Success) :-
	L=[cl(ToSolve)], !,
	tcl_solve_clause(Success,ToSolve,IntVarsCode,Solver,_ANs).

apply_transformation_rule(solve_constraints(_,_), _,_,_,Success) :-
	Success= error('Bad selection.').

%% TODO clause solving

%apply_transformation_rule(solve_constraints, L,_,[solver(Solver)],Success) :-	
%	L=[cl(ToReplace)], !,
%	(var(Success) -> replace_goal(Success,ToReplace,Law,[all])
%	; true).


apply_transformation_rule(caf(Solver,CafV), ToFold,ToUse,_,Success) :-	
	get_clause_numbers(ToFold,L1),
	get_clause_numbers(ToUse,L2),
	caf(Solver,CafV,L1,L2,Success).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  STRATEGIES
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% TODO:
% - manage Output
apply_strategy(Strategy,Args1,Args2,Args3) :- 
	apply_strategy(Strategy,Args1,Args2,Args3,Output),
	tcl_eval(clearDefSelText),
	(Output=error(Alert) ->
	    tcl_alertnote(Alert), fail
	;
%%%	    ( functor(Strategy,cs,_) ->
%%%			  Output = out(Label,OldCNs,NewCNs),
%%%			  update_proofText(Label,OldCNs,NewCNs)
%%%		 ;
			  Output = out(Label,OldCNs,NewCls),
			  update_transf_tree(Label,OldCNs,NewCls,NewCNs),
			  update_proofText(Label,OldCNs,NewCNs)			  
%%%		 )

	).

apply_strategy(solve_program(Solver,IntVars),_Args1,_Args2,_Args3,Output) :-
	current_program(_P,CNs),
	(time(solve_clauses_n(CNs,IntVars,Solver,NewCls),T) ->
	    Output=out(label(solve_program(Solver,IntVars)),CNs,NewCls),
	    concat_all(['{Elapsed time ',T,' ms}'],TimeMsg),
	    tcl_warning_ok(TimeMsg)
	;
	    Output=error('Error in solve program'),
	    tcl_alertnote(Output)
	).


apply_strategy(make_prop,_Args1,_Args2,_Args3,Output) :-
	current_program(_P,CNs),
	(time(make_prop(CNs,NewCls),T) ->
	    Output=out(label(make_prop),CNs,NewCls),
	    concat_all(['{Elapsed time ',T,' ms}'],TimeMsg),
	    tcl_warning_ok(TimeMsg)
	;
	    Output=error('Error in making program propositional'),
	    tcl_alertnote(Output)
	).


apply_strategy(cs,_Args1,TclDefs,_Args3,Output) :-
	get_clause_numbers(TclDefs,DefsN),
	clause_hd_bd_list(DefsN,Defs),
	(contextual_specialization(Defs,FoldedCls) ->
	    Output=out(label(cs),DefsN,FoldedCls)
	;
	    Output=error('Error in contextual specialization'),
	    tcl_alertnote(Output)
	).

apply_strategy(cs(Solver,Uf,Wqounf,Wqogen,Widening),_Args1,TclDefs,_Args3,Output) :-
	get_clause_numbers(TclDefs,DefsN),
	clause_hd_bd_list(DefsN,Defs),
	(time(contextual_specialization(Defs,Solver,Uf,Wqounf,Wqogen,Widening,FoldedCls),T) ->
	    Output=out(label(cs,Solver,Uf,Wqounf,Wqogen,Widening),DefsN,FoldedCls),
	    concat_all(['{Elapsed time ',T,' ms}'],TimeMsg),
	    tcl_warning_ok(TimeMsg)
	;
	    Output=error('Error in contextual specialization')
	).

%%%apply_strategy(ds,_Args1,TclDefs,_Args3,Output) :-
%%%	get_clause_numbers(TclDefs,DefsN),
%%%	clause_hd_bd_list(DefsN,Defs),
%%%	(determ_strategy(Defs,FoldedCls) ->
%%%	    Output=out(label(ds),DefsN,FoldedCls)
%%%	;
%%%	    Output=error('Error in the determinization strategy'),
%%%	    tcl_alertnote(Output)
%%%	).

apply_strategy(ds(Solver,Uf,Wqounf),_Args1,TclDefs,_Args3,Output) :-
	get_clause_numbers(TclDefs,DefsN),
	clause_hd_bd_list(DefsN,Defs),
	(time(determ_strategy(Defs,Solver,Uf,Wqounf,FoldedCls),T) ->
	    Output=out(label(ds,Solver,Uf,Wqounf),DefsN,FoldedCls),
	    concat_all(['{Elapsed time ',T,' ms}'],TimeMsg),
	    tcl_warning_ok(TimeMsg)
	;
	    Output=error('Error in the determinization strategy')
	).


tcl_bottom_up_phase :-
	current_program(_,CurrCNs),
	program_clauses(1,InitCNs),
	list_to_ord_set(CurrCNs,CCNs),
	list_to_ord_set(InitCNs,ICNs),
	ord_subtract(CCNs,ICNs,NCNs),
	(NCNs \== [] ->
	    CNs = NCNs
	;
	    CNs = InitCNs
	),
	clause_hd_bd_list(CNs,Cls),
	tcl_bottom_up_phase(Cls).

tcl_bottom_up_phase(Cls) :-
	(compute_stratification(Cls,Strata) ->
	    print(Strata)
	;
	    tcl_warning_ok('{Program is not stratified.}'),
	    fail
	),
	strata_bup(Strata,Cls).
	


make_args_replace_goal(_,[],[],_) :- !.
make_args_replace_goal(CN,[AN|ANs],[atom(CN,AN)|L],Success) :- !,
	make_args_replace_goal(CN,ANs,L,Success).
make_args_replace_goal(_,_,_,Success) :-
	Success= error('Bad selection. Select one or more atoms in the body of one clause.').

make_args_move_goal(CN,[],[atom(CN,AN)],AN,_) :- !.
make_args_move_goal(CN,[AN|ANs],[atom(CN,AN)|L],Pos,Success) :- !,
	make_args_move_goal(CN,ANs,L,Pos,Success).
make_args_move_goal(_,_,_,_,Success) :-
	Success= error('Bad selection for move goal. Select one or more atoms in the body of one clause.').

make_args_move_clause([cl(CN)],CN,0,_) :-!.
make_args_move_clause([cl(CN),cl(CN2)],CN,CN2,_) :-!.
make_args_move_clause([cl(CN),atom(CN2,_)],CN,CN2,_) :-!.
make_args_move_clause(_,_,_,_,Success) :-
	Success= error('Bad selection for move clause.').

make_args_add_or_del_clauses([cl(CN)],CN,_) :-!.
make_args_add_or_del_clauses(_,_,Success) :-
	Success= error('Bad selection. Select one clause.').




		 
get_new_head(N,I,Rule) :-
	ground(Rule),
	(Rule == case_split ->
	    tk_new([name('case split')],I)
	;
	    Rule == split_determ ->
	    tk_new([name('split determ')],I)
	),
	tcl_eval(I, 'set w .entrycasesplit',_),
	concat_atoms(['set clause_num ',N], CmdCN),
	tcl_eval(I, CmdCN, _),
	tcl_eval(I, 'frame $w',_),
	tcl_eval(I, 'pack $w',_),
	(Rule == case_split ->
	    tcl_eval(I, 'set cslabel "Enter an instance of the head:"',_)
	;
	    Rule == split_determ ->
	    tcl_eval(I, 'set cslabel "Modify the constraint:"',_)
	),	
	tcl_eval(I, 'label  $w.msg -wraplength 4i  -justify left \
		-foreground black -text $cslabel', _),
	tcl_eval(I, 'pack  $w.msg -side top -pady 5 -padx 3 \
		-anchor w', _),
	tcl_eval(I, 'frame  $w.buttons',_),
	tcl_eval(I, 'pack $w.buttons -side bottom -fill x -pady 2m',_),
	tcl_eval(I, 'entry $w.e1 -relief sunken -background azure2\
		-foreground black\
	    -width 40  -textvariable newhd',_),
	tcl_eval(I, 'pack $w.e1 -side top -pady 5 -padx 5 -expand yes',_),
	tcl_eval(I, '$w.e1 delete 0 end',_),
	(Rule == case_split ->
	    tcl_eval(I, 'prolog make_entry_case_split($clause_num,NewHd)',_)
	;
	    Rule == split_determ ->
	    tcl_eval(I, 'prolog make_entry_split_determ($clause_num,NewHd)',_)
	),	
	tcl_eval(I, '$w.e1 insert end $prolog_variables(NewHd)', _),
	tcl_event(I, 'button $w.buttons.cancel -text Close -foreground black \
	    -command {destroy . ; prolog_event error}', _),
	tcl_event(I, 'button $w.buttons.ok -foreground black \
	    -text Split -command { prolog_event $newhd }',_),
	tcl_eval(I, 'pack $w.buttons.cancel $w.buttons.ok \
		-side left -expand 1',_).


make_args_case_split([cl(CN)],CN,NewHead,_) :-!,
	concat_atoms(['get_head_case_split ',CN], Cmd),
	tcl_eval(Cmd,NewHead).
make_args_case_split(_,_,_,Success) :-
	Success= error('Bad selection. Select one clause.').

make_entry_case_split(CN,Head) :-
	clause_hd_bd(CN,Hd,Bd),
	numbervars((Hd,Bd),0,_),
	term_atom(Hd,Head).
	


make_entry_split_determ(CN,TclTerm) :-
	clause_hd_bd(CN,Hd,bd(C,_Lits)),
	TclTerm1 = (Hd,C),
	numbervars(TclTerm1,0,_),
	term_atom(TclTerm1,TclTerm).


all_clauses_selection([],0).
all_clauses_selection([cl(_)|Cs],N2) :-
	all_clauses_selection(Cs,N),
	N2 is N+1.

split_in_two(L,N,ToFold,ToUse,_) :-
	N2 is N//2,
	make_gen_list(L1,N2),
	make_gen_list(L2,N2),
	append(L1,L2,L),
	get_cl_num(L1,ToUse),
	get_cl_num(L2,ToFold).
split_in_two(_,_,_,_,Success) :-
	Success= error('Bad selection for folding. Select an even number of clauses').
get_cl_num([],[]).
get_cl_num([cl(N)|L],[N|L2]) :-
	get_cl_num(L,L2).

make_gen_list([],0).
make_gen_list([_|L],N) :-
	N2 is N-1,
	make_gen_list(L,N2).

make_args_restrict([cl(CN)|Args],[P/A|Preds]) :-
	pred_hd_of_clause(CN,P,A),
	make_args_restrict(Args,Preds).
make_args_restrict([atom(CN,AN)|Args],[P/A|Preds]) :-
	pred_hd_of_atom(CN,AN,P,A),
	make_args_restrict(Args,Preds).
make_args_restrict([],[]).
		  

% ToMove somewhere else ?

apply_print(Cmd,What) :- 
	tell(map_tmp_print),
	(What==program ->
	    (current_program(P,CNs), show_clauses_without_numbers(CNs)) ;
        (What==init_defs ->
	    (program_clauses_defs(1,All),
	    show_clauses_without_numbers(All)) ;
	(What==theory ->
	    (current_theory_empty ->
		(concat_atoms(['alertnote {',
		       'Theory is empty. Printing aborted.}'],Cmd3),
		tcl_eval(Cmd3), fail) ; show_law_definitions_print );
	(What==all ->
	    (write('Theory:'), nl, show_law_definitions_print, nl,
	     program_clauses_defs(1,All),
	     current_program(P,CNs),
	     write('Init+Defs:'),nl, 		
     	     show_clauses_without_numbers(All),	nl,
	     write('Program:'), nl,		
     	     show_clauses_without_numbers(CNs)) ;
	(What==history ->
	    (current_program(P,_),
	     ancestors(P,Ancestors),
	     show_short_history([P|Ancestors]))))))),
	told,
	file_property(map_tmp_print,size(Size)),
	(Size==0 -> tcl_alertnote('Empty file. No printing.') ; 
	    (name(Cmd,CmdS),
	     name(' map_tmp_print',FS),
	     append(CmdS,FS,PrintS),
     	     name(Unix1,PrintS),
 	     shell(Unix1,X),
	     (X==0 ->
		 (What==history -> tcl_infonote('Printing history...') ;
     		 (What==theory -> tcl_infonote('Printing theory...') ;
     		 (What==program -> tcl_infonote('Printing program...') ;
     		 (What==program -> tcl_infonote('Printing init + defs...') ;
     		 (What==all ->
		     tcl_infonote('Printing theory + init + defs + program...'))))))
	     ;
	    tcl_alertnote('Problem with printing.\
	    Check the printing command.')))).

apply_save_program(AbsFilename) :-
	current_program(_,CNs),
	tell(AbsFilename),
	show_clauses_without_numbers(CNs),
	told,
	tcl_infonote('Program saved.').


apply_save_program_as_term(AbsFilename) :-
	current_program(_,CNs),
	tell(AbsFilename),
	show_program_as_term(CNs),
	told,
	tcl_infonote('Program saved as assertions.').

apply_save_program_for_sicstus(AbsFilename) :-
	current_program(_,CNs),
	tell(AbsFilename),
	show_program_for_sicstus(CNs),
	told,
	tcl_infonote('Program saved for sicstus.').


apply_save_program_for_xsb(AbsFilename) :-
	current_program(_,CNs),
	current(solver,S),
	retractall(call_pattern(_,_,_)),
	assert_call_patterns(CNs,S),
	get_defined_predicates(CNs,DefinedPreds),
	get_failed_predicates(DefinedPreds,FailedPreds),
	append(DefinedPreds,FailedPreds,AllPreds),
	tell(AbsFilename),
	show_table_directives(AllPreds), nl,nl,
	show_clauses_for_xsb(CNs), nl,nl,
	show_failed_clauses(FailedPreds),
	told,
	retractall(call_pattern(_,_,_)),
	tcl_infonote('Program saved for XSB.').


apply_save_history(AbsFilename) :-
	current_program(P,_),
	ancestors(P,Ancestors),
	tell(AbsFilename),
	show_short_history([P|Ancestors]),
	told,
	tcl_infonote('History saved.').

apply_save_session(AbsFilename) :-
	tell(AbsFilename),
	listing_transf_tree,
	told,
	tcl_infonote('Session saved.').

apply_save_theory(AbsFilename) :-
	tell(AbsFilename),
	show_law_definitions_save,
	told,
	tcl_infonote('Theory saved.').


apply_restart(P) :-
        erase_environment_ask_theory,
	see(P), load_initial_program, seen,
	assert_relation_prog_state(initial_prog,[P]),
	show_initial_program.

apply_start(P) :-
	erase_environment,
	see(P), load_initial_program, seen,
	assert_relation_prog_state(initial_prog,[P]),
	show_initial_program.

apply_load_session(AbsFilename) :-
	erase_environment,
	see(AbsFilename),
	read_session,
	seen,
	current_program(PN,_),
	show_program(PN).

apply_add_theory(AbsFilename) :-
	see(AbsFilename),
	add_theory_old(Output,AbsFilename,_,ToAssert),
	seen,
	(Output== ok ->
	    (assert_laws(ToAssert),
	    tcl_infonote('Theory updated.'),
	    tcl_eval('destroy .wlaws'),
	    tcl_eval(showCurrentTheoryIf)) ;
	    tcl_syntax_error_theory).
	    
apply_delete_laws([]) :-
     tcl_infonote('Theory updated.').
%    tcl_eval(showCurrentTheory).
     
apply_delete_laws([LN|Ls]) :-
	(check_possible_law_deletion(LN) ->
	    retract(law_definition(_,LN,_,_)) ;
	    (concat_atoms(['{Law ',LN,' has been used and cannot be deleted.}'],Msg),
	    tcl_warning_ok(Msg) )),
	apply_delete_laws(Ls).




check_possible_law_deletion(LN) :-
	clause(law_applied(LN),_), !, fail.
check_possible_law_deletion(_).	


	
 

