generalize_clause(X,Y,Z) :- apply_generalize_clause(X,Y,Z).

apply_generalize_clause(error(Alert),CN,_) :-
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.
%%% This can only be applied to non-clp programs
apply_generalize_clause(Output,CN,Gen) :-
	clause_hd_bd(CN,H,bd(C,B)),
	anti_unify((H,B),Gen,Eqs,_,Z),
	Z=(NewH,NewB),
	add_equalities(Eqs,[NewH|NewB],New),
	New=[GenH|GenB],
	Output=out(label(generalize(CN)), [CN], [(GenH,bd(C,GenB))]).

add_equalities([],Z,Z).
add_equalities([X=Y|Eqs],Z,New) :-
	var(Y), !, X=Y,
	add_equalities(Eqs,Z,New).

add_equalities([X=Y|Eqs],Z,New) :-
	add_equality(X=Y,Z,Z2),
	add_equalities(Eqs,Z2,New).

add_equality(X=Y,[L|Z],[L|Z2]) :-
	 free_of_var(X,L), !,
	 add_equality_bd(X=Y,Z,Z2).

add_equality(X=Y,[L|Z],[L,X=Y|Z]).

add_equality_bd(X=Y,[L|Z],[L|Z2]) :-
	 free_of_var(X,L), !,
	 add_equality_bd(X=Y,Z,Z2).

add_equality_bd(X=Y,[L|Z],[X=Y,L|Z]).

get_generalized_clause(N,I) :-
	tk_new([name('generalize')],I),
	tcl_eval(I, 'set w .entrygeneralization',_),
	concat_atoms(['set clause_num ',N], CmdCN),
	tcl_eval(I, CmdCN, _),
	tcl_eval(I, 'frame $w',_),
	tcl_eval(I, 'pack $w',_),
	tcl_eval(I, ' prolog make_clause_readable_entry($clause_num,Entry)',_),
	tcl_eval(I, 'label  $w.msg -wraplength 7i  -justify left -foreground black \
		-text "enter a generalization of the selected clause:"', _),
	tcl_eval(I, 'pack  $w.msg -side top -pady 5 -padx 3 \
		-anchor w', _),
	tcl_eval(I, 'frame  $w.buttons',_),
	tcl_eval(I, 'pack $w.buttons -side bottom -fill x -pady 2m',_),
	tcl_eval(I, 'entry $w.e1 -relief sunken -xscrollcommand "$w.hscroll set"\
		-background azure2 -foreground black \
		-textvariable gencl -width 100',_),
	tcl_eval(I, 'pack $w.e1 -side top -padx 10 -pady 10 ',_),
	tcl_eval(I, 'scrollbar $w.hscroll -orient horiz -command "$w.e1 xview"',_),
	tcl_eval(I, 'pack $w.hscroll  -padx 10 -pady 10 -side bottom -fill x',_),
	tcl_eval(I, '$w.e1 delete 0 end',_),
        tcl_eval(I, '$w.e1 insert end $prolog_variables(Entry)', _),
	tcl_event(I, 'button $w.buttons.cancel -text Close -foreground black \
	    -command {destroy . ; prolog_event error}', _),
	tcl_event(I, 'button $w.buttons.ok -foreground black \
	    -text Generalize -command { prolog_event $gencl }',_),
	tcl_eval(I, 'pack $w.buttons.cancel $w.buttons.ok \
		-side left -expand 1',_).

check_entry_generalized_clause(N,E,(Head,Body)) :-
	clause_hd_bd(N,H,bd(_C,B)),
	(E= (Head :- Body_Conj) ->
            conj_to_list(Body_Conj,BodyList),
	    ( current(solver,Solver) ->
		find_constraints(Solver,BodyList,_GC,Body)
	    ;
		Body = BodyList
	    )
        ;
	    (Head=E, Body=[])
	),
	Gen=(Head,bd([],Body)), Inst=(H,bd([],B)),
	subsumes_chk(Gen,Inst),
	\+ variant(Gen,Inst).