replace_goal(X,Y,Z,T) :-
	apply_replace_goal_rule(X,Y,Z,T).

%this should never be called
apply_replace_goal_rule(error(Alert),_,Law,_) :-
	\+ is_a_law_number(Law), !,
         Alert='Law has been deleted since loaded.'.

apply_replace_goal_rule(error(Alert),CN,_,_) :- 
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.

apply_replace_goal_rule(Output,CN,LN,PosAtoms) :-
	replacement(CN,PosAtoms,LN,NewCl), !,
	Output=out(label(replace_goal(CN,LN,PosAtoms)), [CN], [NewCl]),
	assert_once_relation_prog_state(law_applied,[LN]).

apply_replace_goal_rule(error(Alert),_,_,_) :-
	Alert='Impossible replacement.'.

replacement(CN,PosAtoms,LN,NewCl) :-
	(PosAtoms=[all] ->
	    law_definition(_,LN,Left,_),
		 ( Left=[true] ->
			  PosAtoms2=[]
		 ;
			  length(Left,L),
			  some_pos_body(CN,L,PosAtoms2)
		 )
	;
		 PosAtoms=PosAtoms2
	),
	clause_hd_bd(CN,H,bd([],B1)),
	mysubstitute(B1,PosAtoms2,LN,B),
	NewCl=(H,bd([],B)).
 
mysubstitute(B,PosAtoms,LN,NewB) :- 
	law_definition(_,LN,Left,Right), 
	list_to_ord_set(PosAtoms,OrdPosAtoms),
	ord_positions_to_elts(B,OrdPosAtoms,B1,Rest),
	subsumes_lhs(Left,B1),
	(OrdPosAtoms = [Pos|_] ->
		 insert_at_position(Right,Pos,Rest,NewB)
	;
		 insert_at_position(Right,1,Rest,NewB)
	). 

subsumes_lhs(Left,B1) :- subsumes(Left,B1).
subsumes_lhs([true],_).


move_goal(X,Y,Z,T) :-
	apply_move_goal(X,Y,Z,T).

apply_move_goal(error(Alert),CN,_,_) :- 
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.

apply_move_goal(Output,CN,Atoms,Pos) :-
	move_atoms(CN,Atoms,Pos,NewCl), !,
	Output=out(label(move_goal(CN,Atoms)), [CN], [NewCl]).

move_atoms(CN,Atoms,Pos,NewCl) :- 
	clause_hd_bd(CN,H,bd(C,B)),
	(C \== [] ->
		 Msg1='{Warning:  the correctness of this transformation is not ensured by the system.',
		 Msg2=' Be sure that you are moving atoms not constraints.',
		 Msg3='\nContinue move atoms?}',
		 concat_atoms([Msg1,Msg2,Msg3],Msg),
		 tcl_warning_yesno(Msg,Answer),
		 (Answer=="yes" ->
			  true
		 ;
			  fail
		 )
	;
		 true
	),
	length(C,LC),	subtract_list(Atoms,LC,Atoms1),
	list_to_ord_set(Atoms1,OrdAtoms),
	ord_positions_to_elts(B,OrdAtoms,B1,Rest),
	Pos1 is Pos-LC,
	findall(X, (member(X,OrdAtoms), X > Pos1), L),
	length(L,L2),
	length(OrdAtoms,L1),
	NewPos is Pos1-L1+L2+1,
	insert_at_position(B1,NewPos,Rest,NewB),
	NewCl=(H,bd(C,NewB)).


subtract_list([],_N,[]).
subtract_list([X|Xs],N,[Y|Ys]) :-
	Y is X-N, Y>0,
	subtract_list(Xs,N,Ys).



add_goal(X,Y,Z) :-
	apply_add_goal(X,Y,Z).

apply_add_goal(error(Alert),CN,_) :- 
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.

apply_add_goal(Output,CN,Atoms) :-
	add_atoms(CN,Atoms,NewCl), !,
	Output=out(label(add_goal(CN,Atoms)), [CN], [NewCl]).

add_atoms(CN,Atoms,NewCl) :- 
	clause_hd_bd(CN,H,bd([],B)),
	list_to_ord_set(Atoms,OrdAtoms),
        duplicate_elts_at_positions(B,OrdAtoms,B1),
	NewCl=(H,bd([],B1)).


delete_goal(X,Y,Z) :-
	apply_delete_goal(X,Y,Z).

apply_delete_goal(error(Alert),CN,_) :- 
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.

apply_delete_goal(Output,CN,Atoms) :-
	delete_atoms(CN,Atoms,NewCl), !,
	Output=out(label(delete_goal(CN,Atoms)), [CN], [NewCl]).

apply_delete_goal(error(Alert),_,_) :- 
         Alert='Impossible deletion. Deletion applies only to duplicate goals or goals equivalent to true.'.

delete_atoms(CN,Atoms,NewCl) :- 
	clause_hd_bd(CN,H,bd([],B)),
	list_to_ord_set(Atoms,OrdAtoms),
	ord_positions_to_elts(B,OrdAtoms,B1,Rest),
	(belong(B1,Rest) ;
	    true_goals(B1)), !,
	NewCl=(H,bd([],Rest)).

true_goals([]).
true_goals([L|Ls]) :-
	is_a_negative_literal(L,Atom), !,
	is_always_false(Atom),
	true_goals(Ls).

true_goals([L|Ls]) :-
	is_always_true(L),
	true_goals(Ls).

