:- use_module(library(lists), [ append/3 ]).

tcl_ask_theory_overwrite(T,Answer) :-
	concat_atoms(['ask_theory_overwrite ',T],Cmd),
	tcl_eval(Cmd, Answer).

tcl_ask_clause_deletion_by_unfolding(F/A,CN,Answer) :-
	concat_atoms(['ask_cl_deletion ',F,' ',A,' ',CN],Cmd),
	tcl_eval(Cmd, Answer).

update_proofText(Label,OldCNs,NewCNs) :-
%tcl_eval(unmarkHotTag),
	compoundtoatom(OldCNs,Old1),
	compound_to_atom_list(OldCNs,Old),
	concat_atoms(['removeTagsClNums ',Old1],RemTags),
	tcl_eval(RemTags),
	( Label = label(unf(CN,AN)) ->
	    concat_atoms(['markUnfoldedTag ',CN,'  ',AN],CmdUTag),
	    tcl_eval(CmdUTag),
	    (NewCNs==[] ->
		( concat_atoms(['addProofText2 ',
				'{\ndeletion of ',CN,' by unfolding.\n}'],Cmd),
		    tcl_eval(Cmd))
	    ;
		(
		  concat_atoms(['addProofText2 ',
				'{\nby unfolding ',CN,':\n}'],Cmd),
		  tcl_eval(Cmd),
		  tcl_add_clauses_proofText(NewCNs)))
	;
	    ( Label = label(fold(_,_)) ->
		( concat_atoms(['addProofText2 ','{\nby folding ',Old,':\n}'],Cmd),
		    tcl_eval(Cmd),
		    tcl_add_clauses_proofText(NewCNs))
	    ;
		( Label = label(definition(_)) ->
		    ( concat_atoms(['addProofText2 ','{\nnew definition:\n}'],Cmd),
			tcl_eval(Cmd),
				%	    tcl_eval('destroy .wdefinition'),
			tcl_add_clauses_proofText(NewCNs))
		;
		    ( Label = label(restriction(_,_)) ->
			(( OldCNs=[] ->
			   concat_atoms(['addProofText2 ',
					 '{\nby restriction, program unchanged.\n}'],Cmd)
			   ;
			   (	compound_to_atom_list(OldCNs,AtomOldCNs),
			       concat_atoms(['addProofText2 ',
					     '{\nby restriction, deletion of ',
					     AtomOldCNs,'.\n}'],Cmd))),
			    tcl_eval(Cmd))
		    ;
			( Label = label(replace_goal(CN,_,_)) ->
			    (concat_atoms(['addProofText2 ',
					   '{\nby goal replacement in ',CN,':\n}'],Cmd),
				tcl_eval(Cmd),
				tcl_add_clauses_proofText(NewCNs))
			)
		    ;
			( Label = label(move_goal(CN,_)) ->
			    (concat_atoms(['addProofText2 ',
					   '{\nby goal move in ',CN,':\n}'],Cmd),
				tcl_eval(Cmd),
				tcl_add_clauses_proofText(NewCNs))
			)
		    ;
			( Label = label(add_goal(CN,_)) ->
			    (concat_atoms(['addProofText2 ',
					   '{\nby goal addition in ',CN,':\n}'],Cmd),
				tcl_eval(Cmd),
				tcl_add_clauses_proofText(NewCNs))
			)
		    ;
			( Label = label(delete_goal(CN,_)) ->
			    (concat_atoms(['addProofText2 ',
					   '{\nby goal deletion in ',CN,':\n}'],Cmd),
				tcl_eval(Cmd),
				tcl_add_clauses_proofText(NewCNs))
			)
		    ;
			( Label = label(move_clause(C1,C2)) ->
			    (number_atom(C1,N1), number_atom(C2,N2),
				(C2==0 -> concat_atoms(['addProofText2 ',
							'{\nclause ',N1,' has moved to first position.\n}'],Cmd)
				    ;
				    concat_atoms(['addProofText2 ',
						  '{\nclause ',N1,' has moved after clause ',N2,'.\n}'],Cmd)),
				tcl_eval(Cmd))
			)
		    ;
			( Label = label(case_split(CN)) ->
			    (concat_atoms(['addProofText2 ',
					   '{\nby splitting ',CN,':\n}'],Cmd),
				tcl_eval(Cmd),
				tcl_add_clauses_proofText(NewCNs))
			)
		    ;
			( Label = label(generalize(CN)) ->
			    (concat_atoms(['addProofText2 ',
					   '{\nby generalizing ',CN,':\n}'],Cmd),
				tcl_eval(Cmd),
				tcl_add_clauses_proofText(NewCNs))
			)
		    ;
			( Label = label(ccr) ->
			    (concat_atoms(['addProofText2 ',
					   '{\nby contextual constraint replacement:\n}'],Cmd),
				tcl_eval(Cmd),
				tcl_add_clauses_proofText(NewCNs))
			)		    
		    ;
			( Label = label(replace_clauses(_)) ->
			    ( (NewCNs==[] ->
				  (concat_atoms(['addProofText2 ',
						 '{\nreplace clause(s) ',Old,' by true (deletion).\n}'],Cmd),
				      tcl_eval(Cmd))
			      ;
				  (concat_atoms(['addProofText2 ',
						 '{\nreplace clause(s) ',Old,' by:\n}'],Cmd),
				      tcl_eval(Cmd),
				      tcl_add_clauses_proofText(NewCNs)))
			    )
			;
			    ( Label = label(add_clauses(_)) ->
				(concat_atoms(['addProofText2 ',
					       '{\nadd clause(s):\n}'],Cmd),
				    tcl_eval(Cmd),
				    tcl_add_clauses_proofText(NewCNs))
			    )
			;
			    ( Label = label(delete_clauses(_)) ->
				(concat_atoms(['addProofText2 ',
					       '{\ndelete clause(s) ',Old,'.\n}'],Cmd),
				    tcl_eval(Cmd)
				))
			;
			    ( Label = label(solve_constraints(_,IntVars,_)) ->
					  (concat_atoms(['addProofText2 ',
										  '{\nsolve clause ',IntVars,'\n}'],Cmd),
							tcl_eval(Cmd),
							tcl_add_clauses_proofText(NewCNs))
			    )
			;
			    ( Label = label(caf(_Solver,CafV,_,_)) ->
					  (concat_atoms(['addProofText2 ','{\nby ',CafV ,' constrained atomic folding ',Old,':\n}'],Cmd),
							tcl_eval(Cmd),
							tcl_add_clauses_proofText(NewCNs))
			    )
			;
			    ( Label = label(make_prop) ->
				(concat_atoms(['addProofText2 ','{\nPropositional program:\n}'],Cmd),
				    tcl_eval(Cmd),
				    tcl_add_clauses_proofText(NewCNs))
			    )
			;

			    ( Label = label(cs) ->
				(concat_atoms(['addProofText2 ','{\nby  contextual specialization:\n}'],Cmd),
				    tcl_eval(Cmd),
				    tcl_add_clauses_proofText(NewCNs))
			    )
			;
			    ( Label = label(ds) ->
				(concat_atoms(['addProofText2 ','{\nby the determinazation strategy:\n}'],Cmd),
				    tcl_eval(Cmd),
				    tcl_add_clauses_proofText(NewCNs))
			    )
			;
			    ( Label = label(solve_program(_,IntVars)) ->
							(concat_atoms(['addProofText2 ','{\nby  solving program clauses w.r.t. ',IntVars,' :\n}'],Cmd),
				    tcl_eval(Cmd),
				    tcl_add_clauses_proofText(NewCNs))
			    )
			;
			    ( Label = label(cs,_,_,_,_,_) ->
				(concat_atoms(['addProofText2 ','{\nby  contextual specialization:\n}' ],Cmd),
				    tcl_eval(Cmd),
				    tcl_add_clauses_proofText(NewCNs))
			    )
			;
			    ( Label = label(ds,_,_,_) ->
				(concat_atoms(['addProofText2 ','{\nby determinization:\n}' ],Cmd),
				    tcl_eval(Cmd),
				    tcl_add_clauses_proofText(NewCNs))
			    )			    
			))))).

tcl_add_clauses_proofText([]).
tcl_add_clauses_proofText([CN|NewCNs]) :-
	make_clause_readable(CN,Num,Head,Body),
	concat_atoms(['addProofTextClause2 ',Num,' {',Head,'} ',Body],Cmd2),
	tcl_eval(Cmd2),
	tcl_add_clauses_proofText(NewCNs).

/*tcl_update_current_prog(PN) :-
        program_clauses(PN,ClNs),
	tcl_add_clauses_current_prog(ClNs).
tcl_add_clauses_current_prog([]).
tcl_add_clauses_current_prog([CN|NewCNs]) :-
	make_clause_readable(CN,Num,Head,Body),
	concat_atoms(['addCurrentProg ',Num,' {',Head,'} ',Body],Cmd2),
	tcl_eval(Cmd2),
	tcl_add_clauses_current_prog(NewCNs).
*/

tcl_add_clauses_current_prog([]).
tcl_add_clauses_current_prog([CN|NewCNs]) :-
	make_clause_readable(CN,Num,Head,Body),
	concat_atoms(['addCurrentProg ',Num,' {',Head,'} ',Body],Cmd2),
	tcl_eval(Cmd2),
	tcl_add_clauses_current_prog(NewCNs).


tcl_update_current_theory :-
	findall(L, T^LHS^RHS^law_definition(T,L,LHS,RHS), Laws),
	tcl_add_laws_current_theory(Laws).
tcl_add_laws_current_theory([]).
tcl_add_laws_current_theory([LawN|LawNs]) :-
	make_law_readable(LawN,Num,LHS,RHS),
	concat_atoms(['addCurrentTheory ',Num,' {',LHS,'} ',' {',RHS,'} '],Cmd),
	tcl_eval(Cmd),
	tcl_add_laws_current_theory(LawNs).

tcl_update_solver :-
	findall(L, is_a_solver(L), Solvers),
	tcl_add_solver(Solvers).

tcl_add_solver([]).
tcl_add_solver([Solver|Solvers]) :-
	concat_atoms(['addSolver ',Solver],Cmd),
	tcl_eval(Cmd),
	tcl_add_solver(Solvers).



make_law_readable(LawN,LawN,Lhs,Rhs) :-
	law_definition(_,LawN,LHS,RHS),
        numbervars((LHS,RHS),0,_),
	make_law_text(lhs,LHS,Lhs),
	make_law_text(rhs,RHS,Rhs).


make_clause_readable_entry(CN,Entry) :-
	clause_hd_bd(CN,Hd,Bd),
	numbervars((Hd,Bd),0,_),
	Bd = bd(Constrs,Atoms), append(Constrs,Atoms,Bd1),
	make_tcl_list_entry(Hd,Bd1,Entry).

make_tcl_list_entry(Hd,[],L):-!,
	term_atom(Hd,A), concat_atoms([A,'.'],L).
make_tcl_list_entry(Hd,Bd,L) :-
	make_tcl_list_entry2(Bd,'',TclL),
	term_atom(Hd,A),
	concat_atoms([A,' :- ',TclL],L).
	
make_tcl_list_entry2([Lit],I,L):- !,
	term_atom(Lit,ALit),
	concat_atoms([I,ALit,'.'],L).

make_tcl_list_entry2([Lit|Lits],I,L):-
	term_atom(Lit,ALit),
	concat_atoms([I,ALit,', '],I2),
	make_tcl_list_entry2(Lits,I2,L).


make_clause_readable(CN,Num,Head,Body) :-
	clause_hd_bd(CN,Hd,Bd),
	number_atom(CN,Num),
	Bd = bd(Constrs,Atoms), append(Constrs,Atoms,Bd1),	
	make_clause_readable_(Hd,Bd1,Head,Body).

make_clause_readable_(Hd,Bd,Head,Body) :-
	numbervars((Hd,Bd),0,_),
	term_atom(Hd,Head),
	make_tcl_list(Bd,Body).
	
make_tcl_list([],{}):-!.
make_tcl_list(L,TclL) :-
	make_tcl_list(L,'',TclL).
make_tcl_list([],I,L):-
	concat_atoms(['{ ',I,' }'],L).
make_tcl_list([Lit|Lits],I,L):-
	term_atom(Lit,ALit),
	concat_atoms([I,' {',ALit,'}'],I2),
	make_tcl_list(Lits,I2,L).
	
	     
%%%make_clauses_readable(CNs,Text) :- make_clauses_readable('',CNs,Text).
%%%make_clauses_readable(Text,[],Text).
%%%make_clauses_readable(Text,[CN|CNs],AllText) :-
%%%	clause_hd_bd(CN,Hd,Bd),
%%%	numbervars((Hd,Bd),0,_),
%%%	term_atom(Hd,AtomHd),
%%%	(Bd==[] ->	concat_atoms([Text,CN,'. ',  AtomHd,'.\n'],NewText) ;
%%%	    (make_body_text(Bd,TextBd),
%%%	    concat_atoms([Text,CN,'. ', AtomHd,' :- ', TextBd],NewText))),
%%%	make_clauses_readable(NewText,CNs,AllText).


make_law_text(LR,HS,Text) :-
	make_law_text(LR,'',HS,Text).

make_law_text(LR,Text,[A],TextBd) :- !,
	term_atom(A,AA),
	(LR=rhs -> concat_atoms([Text,AA],TextBd) ;  concat_atoms([Text,AA],TextBd)).
make_law_text(LR,Text,[A|Bd],TextBd) :-
	term_atom(A,AA),
	concat_atoms([Text,AA, ', '],NewText),
	make_law_text(LR,NewText,Bd,TextBd).


make_body_text(Bd,TextBd) :-
	make_body_text('',Bd,TextBd).
make_body_text(Text,[A],TextBd) :- !,
	term_atom(A,AA), concat_atoms([Text,AA,'.\n'],TextBd).
make_body_text(Text,[A|Bd],TextBd) :-
	term_atom(A,AA),
	concat_atoms([Text,AA, ' , '],NewText),
	make_body_text(NewText,Bd,TextBd).

tcl_warning_ok(Msg) :-
	concat_atoms(['msg_warning_ok ',Msg],Cmd),
	tcl_eval(Cmd).
tcl_warning_yesno(Msg,Answer) :-
	concat_atoms(['msg_warning_yesno ',Msg],Cmd),
	tcl_eval(Cmd,Answer).

tcl_warning_not_in_current_program([]) :- !.
tcl_warning_not_in_current_program(L) :-
	compound_to_atom_list(L,AtomL),
	(L=[_] ->
	    concat_atoms(['{Warning: Clause ',AtomL,' is not in current program.}'],Msg) ;
	    	    concat_atoms(['{Warning: Clauses ',AtomL,' are not in current program.}'],Msg)),
	concat_atoms(['msg_warning_ok ',Msg],Cmd),
	tcl_eval(Cmd).


tcl_warning_new_definition([],[],no) :-!.
tcl_warning_new_definition(S,U,Answer) :-
	tcl_warning_new_definition(S,U,'',Msg),
	concat_atoms(['msg_warning_yesno ',Msg],Cmd),
	tcl_eval(Cmd,Answer).


tcl_warning_new_definition([A=_|Ss],U,Msg,Cmd) :-
	concat_atoms([Msg,'\n\[',A,'\] - singleton variable.'],Msg1),
	tcl_warning_new_definition(Ss,U,Msg1,Cmd).
tcl_warning_new_definition([],[P/A|Us],Msg,Cmd) :-
	concat_atoms([Msg,'\n\"',P,'/',A,'\" - undefined predicate.'],Msg1),
	tcl_warning_new_definition([],Us,Msg1,Cmd).
tcl_warning_new_definition([],[],Msg,MsgQ) :-
	concat_atoms(['{Warning:',Msg,'\nContinue define ?}'],MsgQ).




tcl_warning_new_definition2([],[],no) :-!.
tcl_warning_new_definition2(S,U,Answer) :-
	tcl_warning_new_definition2(S,U,'',Msg),
	concat_atoms(['msg_warning_yesno ',Msg],Cmd),
	tcl_eval(Cmd,Answer).

% are these five clauses really needed?
% the only difference w.r.t. tcl_warning_new_definition/4
% is the message in the 3rd clause

tcl_warning_new_definition2([A=_|Ss],U,Msg,Cmd) :-
	concat_atoms([Msg,'\n\[',A,'\] - singleton variable.'],Msg1),
	tcl_warning_new_definition2(Ss,U,Msg1,Cmd).
tcl_warning_new_definition2([],[P/A|Us],Msg,Cmd) :-
	concat_atoms([Msg,'\n\"',P,'/',A,'\" - undefined predicate.'],Msg1),
	tcl_warning_new_definition2([],Us,Msg1,Cmd).
tcl_warning_new_definition2([],[],Msg,MsgQ) :-
	concat_atoms(['{Warning:',Msg,'\nContinue replace ?}'],MsgQ).

tcl_error_pred_exists(F/A,P) :-
	defined_predicate(F/A,P),
	concat_atoms(['msg_error_pred_exists ',F,' ',A,' ',P],Cmd),
	tcl_eval(Cmd).


tcl_error_def_not_same_pred :-
	tcl_eval(msg_error_def_not_same_pred).


tcl_syntax_error_theory :-
	tcl_eval(msg_syntax_error_theory).


tcl_css_param_info(Param,D,Desc) :-
	( css_param_info(Param,L) ->
	    (memberchk(display(D),L) ->
		  true
	      ;
		  D=Param
	    ),
	    (memberchk(description(Desc),L) ->
		  true
	      ;
		  Desc=Param
	    )	    
	;
	    D=Param,Desc=Param
	).





















showRules :-
	findall(R, is_a_transformation_rule(default,R), L),
	compoundtoatom(L,L_tcl),
	concat_all(['showRS rule ', L_tcl ], Cmd),
	tcl_eval(Cmd).

showRulesHelp :-
	findall(R, X^is_a_transformation_rule(X,R), L),
	compoundtoatom(L,L_tcl),
	concat_all(['showRSHelp  ', L_tcl ], Cmd),
	tcl_eval(Cmd).


showStrategies :-
	findall(R, is_a_strategy(R), L),
	compoundtoatom(L,L_tcl),
	concat_all(['showRS strategy ', L_tcl ], Cmd),
	tcl_eval(Cmd).

showStrategiesHelp :-
	findall(R, is_a_strategy(R), L),
	compoundtoatom(L,L_tcl),
	concat_all(['showRSHelp  ', L_tcl ], Cmd),
	tcl_eval(Cmd).


show_current_program(PN) :-
        concat_atoms(['showCurrentProg ',PN],CmdProgramLine),
        tcl_eval(CmdProgramLine),
        tcl_update_current_prog(PN).

update_show_program :-
%        show_current_program(PN).
	tcl_eval(clearProofText),
        current_program(PN,ClNs),
	concat_atoms(['addProofText2 ','{Program ',PN,':\n}'],Cmd),
	tcl_eval(Cmd),
	tcl_add_clauses_proofText(ClNs),
	tcl_eval(clearProofSelText).

	    
update_show_history :-
        current_program(PN,_),
	show_current_history_buffer(PN),
	concat_atoms(['showCurrentHistory ',PN, ' map_tmp_current_history'],Cmd),
	tcl_eval(Cmd).



update_show_init_defs :-
        current_program(PN,_),
        concat_atoms(['showCurrentInit_Defs ',PN],CmdProgramLine),
        tcl_eval(CmdProgramLine),
				%	tcl_eval('clearInit_Defs '),
	show_init_defs.

show_init_defs :-
	program_clauses_defs(1,All),
	tcl_add_clauses_current_init_defs(All).



tcl_add_clauses_current_init_defs([]).
tcl_add_clauses_current_init_defs([CN|NewCNs]) :-
	make_clause_readable(CN,Num,Head,Body),
	concat_atoms(['addCurrentInit_Defs ',Num,' {',Head,'} ',Body],Cmd2),
	tcl_eval(Cmd2),
	tcl_add_clauses_current_init_defs(NewCNs).


definition_for_folding(CN) :-
	is_a_definition_clause(CN,F/A),
	is_defined_in_current_prog(F/A).

	


show_law_definitions_print :-
	(all_laws_applied -> show_law_definitions_all ;
	    tcl_eval(show_law_definitions_print)).
show_law_definitions_save :-
	(all_laws_applied -> show_law_definitions_all ;
	    tcl_eval(show_law_definitions_save)).

all_laws_applied :-
	findall(X, T^Lhs^Rhs^law_definition(T,X,Lhs,Rhs), L),
	all_laws_applied(L).

all_laws_applied([]).
all_laws_applied([L|Ls]) :-
	clause(law_applied(L),_),
	all_laws_applied(Ls).

show_law_definitions_used :-
	setof(N, X^clause(law_applied(N),X), L), !,
	show_law_definitions(L).
show_law_definitions_used.

show_law_definitions_all :-
	findall(X,  T^Lhs^Rhs^law_definition(T,X,Lhs,Rhs), L),
	show_law_definitions(L).

show_law_definitions([]).
show_law_definitions([N|L]) :-
	law_definition(_,N,LHS,RHS),
	numbervars((LHS,RHS),0,_),
	write_commas(LHS), write(' -> '), write_commas(RHS), nl,
	show_law_definitions(L).


show_current_history_buffer(PN) :-
	tell('map_tmp_current_history'),
	show_current_history(PN),
	told.

show_current_history(PN) :-
	ancestors(PN,Anc),
	show_short_history([PN|Anc]).

show_short_history([N1]) :-!,
      node_clauses(N1,CNs),
      write('Initial Program:'), nl,
      show_numbered_clauses(CNs).

show_short_history([N,N1]) :-!,
      node_label_clauses(N,LA,CNs),
      node_clauses(N1,CNsFather),
      write('Initial Program:'), nl,
      show_numbered_clauses(CNsFather),
      subtract(CNs,CNsFather,_,Diff),
      nl, write_label_history(LA), write(' -> program '), write(N), nl,
      show_numbered_clauses(Diff).
	
show_short_history([N,N1|Ns]) :- 
      show_short_history([N1|Ns]),
      node_clauses(N1,CNsFather),
      node_label_clauses(N,LA,CNs),
      subtract(CNs,CNsFather,_,Diff),
      nl, write_label_history(LA),  write(' -> program '), write(N), nl,
      show_numbered_clauses(Diff).

write_label_history(unf(CN,AN)) :-
	write('unfold '), write(CN), write(' at position '),
	write(AN).

write_label_history(fold(CNs,DNs)) :-
	write('fold '), write_l(CNs), write(' wrt '),
	write_l(DNs).

write_label_history(definition(_)) :-
	write('new definition ').

write_label_history(restriction(_,CNs)) :-
	(CNs=[] ->
	    (write('restriction - program unchanged')) ;
	    (write('restriction - deletion of clauses '),
		write_l(CNs))).

write_label_history(replace_goal(C,L,_)) :-
	write('goal replacement in '), write(C),
	write(' using law '), write(L).

write_label_history(move_goal(C,_)) :-
	write('goal move in '), write(C).

write_label_history(add_goal(C,_)) :-
	write('goal addition in '), write(C).

write_label_history(delete_goal(C,_)) :-
	write('goal deletion in '), write(C).

write_label_history(move_clause(C,_)) :-
	write('move of clause '), write(C).

write_label_history(case_split(C)) :-
	write('split clause '), write(C).

write_label_history(generalize(C)) :-
	write('generalize clause '), write(C).

write_label_history(add_clauses(_)) :-
	write('add clauses ').

write_label_history(delete_clauses(Cls)) :-
	write('deletion of '), write(Cls).

write_label_history(replace_clauses(Cls)) :-
	write('replacement of '), write(Cls).

write_label_history(solve_constraints(CN,Pos,Vars,Solver)) :-
	write('solve constraints in clause '), write(CN),
	write(' at positions '), write_l(Pos),
	write(' using the '), write(Solver), write(' solver '),
	write(' w.r.t. variables '), write(Vars ).


write_label_history(caf(CafV,ToFoldN,ToUseN)) :-
	write(CafV), write(' c.a.f. of clause '), write_l(ToFoldN),
	write(' by using '), write_l(ToUseN).

write_label_history(def_fold(Leaves,_FldCls)) :-
	write('Define-fold of clauses '), write_l(Leaves).


write_l([]).
write_l([E|L]) :- write(E), write(' '), write_l(L).

write_commas([]).
write_commas([E]) :- !, write(E).
write_commas([E|L]) :- write(E), write(', '), write_commas(L).


show_rootleaves([]).
show_rootleaves([X|Xs]) :-
	X=(R,L),
	write('Root'), nl,
	portray_clause(R),
	write('Leaves'), nl,
	portray_clauses(L),	
	show_rootleaves(Xs).


portray_clause_n(C) :-
	clause_hd_bd(C,H,B),
	B = bd(Constrs,Lits), append(Constrs,Lits,B1),	
	portray_clause((H :- B1)).

portray_clauses_n([]).
portray_clauses_n([C|Cs]) :-
	writeq(C),write('. '),
	portray_clause_n(C),
	portray_clauses_n(Cs).

portray_clauses([]).
portray_clauses([(H,B)|Cs]) :-
	B = bd(Constrs,Lits), append(Constrs,Lits,B1),
	portray_clause((H :- B1)),
	portray_clauses(Cs).

portray_clauses_lists([]).
portray_clauses_lists([Cs|Css]) :-
	portray_clauses(Cs),
	portray_clauses_lists(Css).