% name/2 is obsolescent: use number_codes or atom_codes instead


:- module(conversion,
	[ term_atom/2,
	  term_atom/3,
	  terms_atoms/2,
	  atom_term/2,
  	  atoms_terms/2,
	  list_term/2,
	  number_atom/2,
	  atom_number/2,
%%%	  tk_num_clause/2,
	  compoundtoatom/2,
	  compound_to_atom_list/2
	]).

:- use_module( library(charsio), [ format_to_chars/3,
	                           read_from_chars/2 ]).

:- use_module( mapconcat ).

:- use_module(library(lists), [append/3, is_list/1 ]).


list_term(ListOfAtoms,Term) :-
	concat_all(ListOfAtoms,Atom0,' '),
	concat(Atom0,'.',Atom),
	name(Atom,ListOfChars),
	read_from_chars(ListOfChars,Term).

atom_term(Atom,Term) :-
	concat(Atom,' .',Atom1),   % space, because Atom may end with 
                                   % funny character
	name(Atom1,Chars),
	read_from_chars(Chars,Term).

atoms_terms([],[]).
atoms_terms([A|As],[T|Ts]) :-
	atom_term(A,T),
	atoms_terms(As,Ts).


term_atom(Term,Atom) :-
	(  number(Term)
	-> number_chars(Term,Chars),
	   atom_chars(Atom,Chars)
	;  format_to_chars('~p',[Term],Chars), 
		% ~w had problems with terms of the form X=0
		% which were printed as X=rat(0,1)
	   name(Atom,Chars)
	).

term_atom(Term,Atom,Length) :-
	(  integer(Term)
	-> number_atom(Term,Atom),
           name(Atom,AtomChars),
           length(AtomChars,Length)
	;  format_to_chars('~w',[Term],Chars),
	   length(Chars,Length),
	   name(Atom,Chars)
	).

number_atom(Num,Atom) :-
	(  number(Num)
	-> number_chars(Num,String),
	   atom_chars(Atom,String)
	;  atom_chars(Atom,String),
	   number_chars(Num,String)
	).


term_atom_l([],[]).
term_atom_l([T|Ts],[A|As]) :- term_atom(T,A), term_atom_l(Ts,As).


terms_atoms([],[]).
terms_atoms([T|Ts],[A,' '|As]) :- term_atom(T,A), terms_atoms(Ts,As).

%%%tk_num_clause([Num,Hd,Bd], TkNumCl) :-
%%%	numbervars([Hd|Bd], 0, _),

%%%	terms_atoms(NumCl,TkNumCl).	
	


% compound_to_atom(+Compound, -Atom)
% Used to convert a general Prolog term to an atom suitable to pass to Tcl
% For each case there will generally be an inverse function atom_to_compound/2
% which can recover enough structure to pass back to Prolog on user input.

compound_to_atom(Compound, Atom):-
    compoundtoatom(Compound, Atom).

compoundtoatom({X}, Atom):-
    !,
    compound_to_atom(X, Atom).

compoundtoatom(Number, Number):-
    number(Number),
    !.

compoundtoatom(Atom, Atom):-
    atom(Atom),
    \+ Atom = [],
    !.

compoundtoatom(L, Atom):-
    is_list(L),
    !,
    compound_to_atom_list(L, AtomL),
    concat_atoms(['{', AtomL, '}'], Atom).


compoundtoatom(Term, Atom):-
    Term =..[Functor|Args],
    !,
    compound_to_atom(Functor, FAtom),
    compound_to_atom_list(Args, AAtom),
    concat_atoms([FAtom, '(', AAtom, ')'], Atom).

%%% Recursive list processing.
%% compound_to_atom_list(+List, -Atom)
%% Turns something of the form [x,y,z] into 'x y z'

compound_to_atom_list([], '').
compound_to_atom_list([H], Atom):-
    !,
    compound_to_atom(H, Atom).
compound_to_atom_list([H|T], AtomL):-
    compound_to_atom_list(T, RHS),
    compound_to_atom(H, Atom),
%    concat_atoms([Atom, ',', RHS], AtomL).
    concat_atoms([Atom, ' ', RHS], AtomL).



