:- module(flags, [
            initialize_flag/2,
	    flag/1,
            flag/2,
            flag/3,
	    set_flag/2,
            wr_flag/1,
            wr_flag/2,
	    wr_flag/3 ]).

% TODO: modularize flags, i.e. each key is prefixed with
% Module name...

% setting/reading global variables that have a single value
% with flag/2,3

% sets flag to Val only if Flag is currently undefined
% this is useful to have in as file directives; once such
% files are reconsulted in e.g. a debugging session, these
% variables are not touched
initialize_flag(Flag,Val) :-
	flag(Flag,Old),
	(  Old == undefined    % if undefined 
	-> flag(Flag,_,Val)    % change it
	;  true                % else don't
	).

set_flag(Flag,Val):-
	flag(Flag,_Val,Val).

flag(Flag) :-
	% flag is a currently-defined flag
	recorded(Flag,'$flag'(Flag,_),_).

flag(Flag,Val):-              % to read Val of variable Flag
    flag(Flag,Val,Val).


% note: only the first flag that matches is changed...
flag(Flag,Old,New):-          % to set Flag from Old to New value
	(  ground(Flag)
	-> true
	;  flag(Flag)
	),
	(  recorded(Flag, '$flag'(Flag,Old), Ref)
	-> true
	;  Old = undefined
	),
	(  Old == New
	-> true
	;  ((  var(Ref)
	   -> true
	   ;  erase(Ref)
	   ),
	   (  New == undefined
	   -> true
	   ;  recorda(Flag, '$flag'(Flag,New),_)
	   ))
	).

wr_flag(A):-
	wr_flag(user_error,A,_V).

wr_flag(A,V):-
	wr_flag(user_error,A,V).

wr_flag(Stream,A,V):-
	flag(A,V),
	print(Stream,A=V),
	nl(Stream).



