:- module(mapconcat, [
			concat/3,
		    concat_all/2,
		    concat_all/3,
		   concat_atoms/2
	      ]).

/*:- use_module(version).
*/

:- use_module( library(charsio), [ format_to_chars/3,
	                           read_from_chars/2 ]).
:- use_module( library(lists),   [ append/3          ]).

concat_atoms([],' ').
concat_atoms([H|T],Atom):-
    concat_atoms(T, RHS),
    concat2(H, RHS, Atom).


concat2(A,B,C) :-
    name(A,AA),name(B,BB),
    append(AA,BB,CC),
    name(C,CC).

concat(A,B,C) :-
	(  atomic(A),
	   atomic(B) -> concat_concat(A,B,C)
	;  atomic(A),
	   atomic(C) -> dif(B,'$fail'),
	              concat_delprefix(C,A,B)
	;  atomic(B),
	   atomic(C) -> dif(A,'$fail'),
	              concat_delsuffix(C,B,A)
	;  	raise_exception(instantiation_error(concat(_,_,_),1))
	).


concat_all(List,Atom) :-
	concat_all(List,Atom,'').

concat_all([],'',_):-!.
concat_all([E],E,_):-!.
concat_all([E|List],Result,Sep):-
	concat_all(List,Res2,Sep),
	concat(Sep,Res2,Res1),
	concat(E,Res1,Result),
	!.

	
%if the foreign interface doesn't work, use: 
:- use_module(library(lists)).
concat_concat(A,B,C) :-
	name(A,Astr),
	name(B,Bstr),
	append(Astr,Bstr,Cstr),
	name(C,Cstr).

concat_delprefix(C,A,B):-
	name(A,Astr),
	name(C,Cstr),
	append(Astr,Bstr,Cstr),
	name(B,Bstr).

concat_delsuffix(C,B,A):-
	name(B,Bstr),
	name(C,Cstr),
	append(Astr,Bstr,Cstr),
	name(A,Astr).

/*
foreign(concat_concat,    c, concat_concat(+string,+string,[-string])).
foreign(concat_delprefix, c, concat_delprefix(+string,+string,[-string])).
foreign(concat_delsuffix, c, concat_delsuffix(+string,+string,[-string])).
foreign(concat_all,       c, concat_all(+term,[-string],+string)).


foreign_file('concat.o',[concat_concat,
			 concat_delprefix,
			 concat_delsuffix,
			 concat_all]).

%%% 3.5
foreign_resource(concat,[concat_concat,
			 concat_delprefix,
			 concat_delsuffix,
			 concat_all]).

:- parse_version(Maj,Min),
	(  Maj==3
	-> (  Min == 3
	   -> load_foreign_files( ['concat.o'], [])
	   ;  Min >= 4
	   -> load_foreign_resource( concat )
	   )
	;  format(user_error,
             "library(concat): SICStus version ~w #~w not supported~n",
	     [Maj,Min])
	).


*/

