/*
:- module(xtk, [
	       % inherited from ProTcl/tk.pl:

	tcl_new/0,
	tcl_new/1,
	tcl_delete/0,
	tcl_delete/1,
	tcl_eval/1,
	tcl_eval/2,
	tcl_eval/3,
	tcl_event/3,
	tk_new/1,
	tk_new/2,
	tk_main_window/1,
	tk_main_window/2,
	tk_destroy_window/1,
	tk_make_window_exist/1,
	tk_num_main_windows/1,
	tk_do_one_event/0,
	tk_do_one_event/1,
	tk_next_event/2,
	tk_next_event/3,
	tk_main_loop/0,

	       % defined here:
	tk_init/0,

	tcl/1,
	tcl/2,
	tcl/3,

	tk_atom/2,

	call_from_tcl/1,
	call_from_tcl_all/1,

	tk_flag/1, tk_flag/2, tk_flag/3,
	menu_flag/2,
	help_flag0/1
              ]).

*/

:- use_module(library(tcltk)).

:- use_module(library(charsio), [format_to_chars/3]).

:- use_module(conversion, [term_atom/2, 
				atom_term/2]).

:- use_module(mapconcat, [concat_all/3]).

:- use_module(flags).

%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% utilities %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%

tcl_delete :-
	flag(map_tcl_interp,I),
	tcl_delete(I).

tcl_new :-
	tcl_new(I),
	flag(map_tcl_interp,_,I).

tcl_eval(Cmd) :-
	flag(map_tcl_interp,I),
	tcl_eval(I,Cmd,_).
tcl_eval(Cmd,Return) :-
	flag(map_tcl_interp,I),
	tcl_eval(I,Cmd,Return).

tcl(Expr) :-
	tcl(Expr,[],_).

tcl(Expr,Subs) :-
	tcl(Expr,Subs,_).

tcl(Expr,Subs,Return) :-
	flag(map_tcl_interp,Interp),
	tcl_eval(Interp,format(Expr,Subs),Return0),
	atom_chars(Return,Return0).

tk_new(Options) :-
	flag(map_tcl_interp,I),
	(  I \== undefined
	-> tcl_delete(I)
	;  true
	),
	tk_new(Options,Int),
	flag(map_tcl_interp,_,Int).

tk_main_window(W) :-
	flag(map_tcl_interp,Int),
	tk_main_window(Int,W).

%% tk_atom(Term,Atom)
%% converts between Prolog term and atom with appropriate
%% symbols escaped for Tcl/Tk 

% don't trust this!
tk_atom(A,B) :-
	integer(A),!,
	number_chars(A,Astr),
	atom_chars(B,Astr).
tk_atom(A0,A) :-
	term_atom(A0,A1),
	name(A1,Chars1),
	tk_escape_l(Chars1,Chars),!,
	name(A,Chars).

tk_escape_l([],[]).
tk_escape_l([H|T],Out):-
	tk_escape_c(H,Rest,Out),
	tk_escape_l(T,Rest).

% 92: \

% 59: ;
% 32: <SPACE>
% 34: "
% 91: [
% 93: ]
% 123 {
% 125 }

tk_escape_c(34,R,[92,34|R]).
tk_escape_c(59,R,[92,59|R]).
tk_escape_c(91,[93|R],[92,91,92,32,92,93|R]).
tk_escape_c(91,R,[92,91|R]).
tk_escape_c(92,R,[92,92|R]).
tk_escape_c(93,R,[92,93|R]).
tk_escape_c(32,R,[92,32|R]).
tk_escape_c(123,R,[92,123|R]).
tk_escape_c(125,R,[92,125|R]).

% catch all:
tk_escape_c(C,R,[C|R]).

% call_from_tcl(Fun)
% communicates the succuss of calling user:Fun/1
% to the Tcl variable SP_answer

/*
% note that the ProTcl documentation says that answers
% are put in var array, but this only works for Eclipse, it
% seems...
call_from_tcl(Atom) :-
	(  flag(tk_debug,on)
	-> writeq(Atom), nl
	;  true
	),
	atom_term(Atom,Var^Call),
	(  call(user:Call)
	-> tcl("set SP_answer ~w",Var)
	;  tcl_eval('set SP_answer 0')
	).


call_from_tcl_all(Atom) :-
	atom_term(Atom,Var^Call),
	(  flag(tk_debug,on)
	-> writeq(findall(Var,Call,Vars)),nl
	;  true
	),
	findall(Var,user:Call,Vars),
	term_atom_list(Vars,Atoms),
	concat_all(Atoms,Answer,' '),
	tcl('set SP_answer "~w"',[Answer]).


term_atom_list([],[]).
term_atom_list([H0|T0],[H|T]) :-
	term_atom(H0,H),
	term_atom_list(T0,T).
*/


%%%%%%%%%%%%%%%%%%%%%%
%% Global variables %%
%%%%%%%%%%%%%%%%%%%%%%

%%% the Tcl array variable `flag' contains a number of flags that are 
%%% important for both Prolog and Tcl. If either side changes this flag
%%% then the other side should be informed.
%%%
%%% use tk_flag/1, 2, 3 to change a flag from the Prolog side
%%% SP_set_flag         to change a flag from the Tcl/tk side
%%%
%%% note that we also check whether tcl/tk is actually initialized.

% sets the flag Flag, and in some cases passes the flag to Tcl/Tk
% flag is supposedly ground
tk_flag(Flag) :-
	flag(Flag,Val),
	tk_flag(Flag,Val,Val).

tk_flag(Flag,Val) :-
	tk_flag(Flag,Val,Val).

tk_flag(Flag,Old,Val) :-
	flag(Flag,Old,Val),                    % first do it on Prolog side
	(  tk_num_main_windows(X),
	   X > 0                               % if tcl/tk is initialized
	-> tcl("set flag(~w) ~w",[Flag,Val])   % then send it to tk
	;  true
	).

% menu_flag(Flag,Values)
% adds a radiobutton for Flag under the Options menu, with Values
% an instantiated list of possible values
%:- use_module(library(lists)).

menu_flag(Flag,Values) :-
	flag(Flag,Value),
	(  member(Value,Values)
	-> true
	;  format(user_error,"menu_flag: value ~w not in ~w~n",[Value,Values]),
	   fail
	),
	tcl("set flag(~w) ~w",[Flag,Value]),
	menu_flag('.menu.options.m',Flag,Values).

menu_flag(Menu,Flag,Values) :-
	tcl('~w add cascade -label "~w" -menu ~w.~w',
	    [Menu,Flag,Menu,Flag]),
	tcl('menu ~w.~w',[Menu,Flag],SubMenu),
	menu_flag_values(Values,Flag,SubMenu).

menu_flag_values([],Flag,Menu) :-
	tcl('~w add command -label Help -command "prolog help_flag0(~w)"',
	    [Menu,Flag]).
menu_flag_values([V0|V],Flag,Menu) :-
	tcl('~w add radiobutton -label ~w -value ~w -variable flag(~w) \
	      -command "prolog flag(~w,_,~w) "',
	      [Menu,V0,V0,Flag,Flag,V0]),
	menu_flag_values(V,Flag,Menu).

tk_init :-
	tk_new([top_level_events]).

help_flag0(Flag) :-
	(  user:help_flag1(Flag)
	-> true
	;  tcl('tk_dialog .dialog Help "No help available for ~w" "" 0 OK',
                [Flag])
	).

wrap_write([],[]).
wrap_write([H|T0],[write(H)|T]) :-
	wrap_write(T0,T).

send_all(Var,Call,TclVar):-
	findall(Var,user:Call,Vars0),
	wrap_write(Vars0,Vars),
	tcl_eval([set,TclVar,br(Vars)]).



