
:- use_module(library(lists), [
										 append/3,
										 memberchk/2,
										 member/2
										] ).


tcl_solve_clause(Err,CN,IntVarsCode,Solver,PosC) :-
	tcl_apply_solve_clause(Err,CN,IntVarsCode,Solver,PosC).

%this should never be called
tcl_apply_solve_clause(error(Alert),_,_,Solver,_) :-
	\+ is_a_solver(Solver), !,
         Alert='Solver does not exists.'.

tcl_apply_solve_clause(error(Alert),CN,_,_,_) :- 
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.

tcl_apply_solve_clause(Output,CN,IntVarsCode,Solver,PosConstrs) :-
	clause_hd_bd(CN,Hd,bd(C,Bd)),
	( var(PosConstrs) ->
		 solve_clause((Hd,bd(C,Bd)),IntVarsCode,Solver,NewCl)
	;
		 %%% remove this cumbersome piece of code?
		 %%% does anyone want to solve only part of a constraint?
		 append(C,Bd,Bd1),
		 solve_clause_pos((Hd,Bd1),PosConstrs,IntVarsCode,Solver,NewCl)
	),	!,
	Output=out(label(solve_constraints(CN,IntVarsCode,Solver)), [CN], [NewCl]),
	assert_once_relation_prog_state(solve,[Solver,IntVarsCode]).

tcl_apply_solve_clause(error(Alert),_,_,_,_) :-
	Alert='Impossible solving.'.



solve_clauses_n([],_,_,[]).
solve_clauses_n([CN|CNs],IntVarsCode,S,[NewCl|NewCls]) :-
	clause_hd_bd(CN,H,B),
	solve_clause((H,B),IntVarsCode,S,NewCl),
	solve_clauses_n(CNs,IntVarsCode,S,NewCls).

solve_clause((Hd,bd(C,Bd)),IntVarsCode,Solver,NewCl) :-
	(IntVarsCode = linking_vars ->
		 linking_variables(C,(Hd,Bd),IntVars)
	; (IntVarsCode = head_vars ->
			free_variables(Hd,[],[],IntVars)
	  ;
			fail
	  )
	),
	solve(Solver,C,IntVars,SC),
	NewCl = (Hd,bd(SC,Bd)).

solve_clause_pos((H,B1),PosConstrs,IntVarsCode,Solver,NewCl) :-
	(IntVarsCode = linking_vars ->
	    list_to_ord_set(PosConstrs,OrdPosConstrs),
	    ord_positions_to_elts(B1,OrdPosConstrs,ConstrList,Rest),
	    linking_variables(ConstrList,(H,Rest),LinkingVars),
	    substitute_constraints(B1,PosConstrs,LinkingVars,Solver,B),
	    NewCl=(H,B)
	; fail).

solve_clauses([],_IntVarsCode,_Solver,[]).
solve_clauses([Cl|Cls],IntVarsCode,Solver,[NewCl|NewCls]) :-
	solve_clause(Cl,IntVarsCode,Solver,NewCl),
	solve_clauses(Cls,IntVarsCode,Solver,NewCls).



quick_solve_clauses([],_Solver,[]).
quick_solve_clauses([Leaf|Leaves],Solver,NFLeaves) :-
	Leaf = (_,bd(C,_)),
	(satisfiable(Solver,C) ->
	    NFLeaves = [Leaf|NFLeaves1],
	    quick_solve_clauses(Leaves,Solver,NFLeaves1)
	;
	    quick_solve_clauses(Leaves,Solver,NFLeaves)
	).


	
substitute_constraints(B,PosConstrs,Vars,Solver,NewB) :- 
      list_to_ord_set(PosConstrs,OrdPosConstrs),
      ord_positions_to_elts(B,OrdPosConstrs,ConstrList,Rest),
      solve(Solver,ConstrList,Vars,SolvedConstrList),
      (OrdPosConstrs = [Pos|_] ->
	  insert_at_position(SolvedConstrList,Pos,Rest,NewB)
      ;
	  insert_at_position(SolvedConstrList,1,Rest,NewB)).


has_failing_body_n(CN,Solver) :-
	clause_hd_bd(CN,_H,B),
	failing_goal(B,Solver).

has_failing_body((_H,B),Solver) :-
	failing_goal(B,Solver).

failing_goal(bd(C,_G),Solver) :-
	(satisfiable(Solver,C) ->
%%%	    (undefined_lits_in_body(G,UndefLits),
%%%		UndefLits \== [] ->
%%%		true
%%%	    ;
		fail
%%%	    )
	;
	    true
	).


is_a_constrained_fact_n(CN,Solver) :-
	clause_hd_bd(CN,Hd,BdD),
	is_a_constrained_fact((Hd,BdD),Solver).
	
is_a_constrained_fact((_Hd,bd(_C,[])),_Solver).