%%% solve CONSTRAINTS

:- use_module(library(lists), [
										 append/3,
										 memberchk/2,
										 member/2
										] ).



%solve_constraint_list(ConstrList,Vars,Solver,SolvedConstrList) :-
%	list_to_constraint(Solver,ConstrList,Constr),
%	solve(Solver,Constr,Vars,SolvedConstrList).


%list_variables([],[]) :- !.
%list_variables([X|Xs],V) :-
%	(var(X) -> (V=[X|Vs], list_variables(Xs,Vs))
%	; list_variables(Xs,V)).


% find_positions(X,L,P):
% the i-th element of list P is the index
% of the first occurrence of the i-th element of list X in list L

find_positions([],_,[]) :- !.
find_positions(_,[],[]) :- !.
find_positions([X|Xs],L,Pos) :-
	(occurs_at_pos(X,L,P) ->
	    Pos=[P|Ps],
	    find_positions(Xs,L,Ps)
	;
	    find_positions(Xs,L,Pos)
	).


occurs_at_pos(X,[Y|Ys],Pos) :-
	occurs_at_pos(X,[Y|Ys],Pos,1).
occurs_at_pos(X,[Y|_],Pos,Start) :-
	X==Y,
	!,
	Pos is Start.
occurs_at_pos(X,[_|Ys],Pos,Start) :-
	Start1 is Start+1,
	occurs_at_pos(X,Ys,Pos,Start1).


% recover_positions(-L1,+L2,+P) iff
% L1 is a permutation of L2 such that find_positions(L1,L2,P).
recover_positions(L1,L2,P) :-
	length(L2,L),
	length(L1,L),
	recover_positions_(L1,L2,P). 

recover_positions_([],_,[]).
recover_positions_([X|Xs],L,[N|Ns]) :-
	nth(N,L,X),
	recover_positions_(Xs,L,Ns).




%% same_pred_pos(X,AtomList,Pos,A)
%% Pos is the position in AtomList of the first
%% (positive or negative) atom A with the same predicate of X
same_pred_pos(X,[Y|Ys],Pos) :-
	same_pred_pos(X,[Y|Ys],Pos,1).
same_pred_pos(X,[Y|Ys],Pos,Start) :-
	(is_a_negative_literal(Y,PY) ->
	    Y1=PY
	;
	    Y1=Y
	),
	(same_functor(X,Y1,_) ->
	    Pos is Start
	;
	    Start1 is Start+1,
	    same_pred_pos(X,Ys,Pos,Start1)
	).
	


is_constraint(C,S) :- member(AC,C), \+ is_atomic_constraint(AC,S),!,fail.
is_constraint(_C,_S).

is_atomic_constraint(C,Solver) :-
	functor(C,F,A),
	is_a_constraint_predicate(Solver,F,A).


% find_constraints_pos(Solver,List, ConstrPos).
% List is a list of atoms and constraints
% ConstrPos is the list of the positions in List where Solver constraints occur
find_constraints_pos(Solver,List,ConstrPos) :-
	find_constraints_pos(Solver,List,ConstrPos,1).
find_constraints_pos(_Solver,[],[],_).
find_constraints_pos(Solver,[El|Rest],[Pos|PosRest],Pos) :-
	is_atomic_constraint(El,Solver),
	!,
	Pos1 is Pos+1,
	find_constraints_pos(Solver,Rest,PosRest,Pos1).
find_constraints_pos(Solver,[_El|Rest],C,Pos) :-
	Pos1 is Pos+1,
	find_constraints_pos(Solver,Rest,C,Pos1).	


% find_constraints(Solver,BdD,CD,BdDwoCD)
% BdD is a sequence obtained by merging
% - a list CD of Solver constraints
% - a list BdDwoCD of atoms

%%% Optimized version: we assume that all the constraints 
%%% occur at the beginning of the body
find_constraints(_Solver,[],[],[]).
find_constraints(Solver,[El|Rest],CList,BdwoC) :-
	(is_atomic_constraint(El,Solver) ->
		 CList = [El|CRest],
		 find_constraints(Solver,Rest,CRest,BdwoC)
	;
		 CList = [], BdwoC = [El|Rest]
	).



%%% Unoptimized version: constraints may anywhere in the body
find_constraints2(_Solver,[],[],[]).
find_constraints2(Solver,[El|Rest],[El|CRest],BdwoC) :-
	is_atomic_constraint(El,Solver),
	!,
	find_constraints2(Solver,Rest,CRest,BdwoC).
find_constraints2(Solver,[El|Rest],C,[El|BdwoC]) :-
	find_constraints2(Solver,Rest,C,BdwoC).


find_constraints2_clauses([],_,[]).
find_constraints2_clauses([(H,bd(C,Lits))|Cls],S,[(H,bd(NewC,NewLits))|NewCLs]) :-
	find_constraints2(S,Lits,C1,NewLits),
	append(C,C1,NewC),
	find_constraints2_clauses(Cls,S,NewCLs).



%%%% catom_subsumed(Solver,CD,AD,C,A): (A,C) subsumes (AD,CD) 
%%%%   - A,AD are atoms
%%%%   - C,CD are constraints __represented as lists__
%%%catom_subsumed(Solver,CD,AD,C,A):-
%%%	subsumes_chk(AD,A),
%%%	unify_with_occurs_check(AD,A),
%%%	entails(Solver,C,CD).




% By default, constraints are represented as lists

%% entailment between constraints represented as list
%list_entails(Solver,C,D):-
%	list_to_constraint(Solver,C,C1),
%	list_to_constraint(Solver,D,D1),
%	entails(Solver,C1,D1).
%% entails(Solver,C,D): C entails D 
%%  - C,D are constraints of the form {c1,...,cn}


equivalent_constraints(Solver,C,D) :-
	entails(Solver,C,D), entails(Solver,D,C).


satisfiable(Solver,C) :-
	solve(Solver,C,[],C1),
	(memberchk(false,C1) -> fail
	;
	    true
	).



%%% residue_to_list([],[]).
%%% residue_to_list([[_]-{C}|R],[C1|N]):-
%%% 	fix_equality(C,C1),
%%% 	residue_to_list(R,N),!.
%%% residue_to_list([[_]-(clpq:{C})|R],[C1|N]):-
%%% 	fix_equality(C,C1),	
%%% 	residue_to_list(R,N).
%%% residue_to_list([[_]-(clpfd:{C})|R],[C|N]):-
%%% 	fix_equality(C,C1),	
%%% 	residue_to_list(R,N).


fix_equalities([C|Cs],[C1|C1s]) :-
	fix_equality(C,C1),
	fix_equalities(Cs,C1s).
fix_equalities([],[]).

%%%fix_equality(C,C1) :-
%%%	(C=(LHS=RHS) ->
%%%	    C1=(LHS=:=RHS)
%%%	;
%%%	    C1=C
%%%	).
fix_equality(LHS=RHS,LHS=:=RHS) :- !.
fix_equality(C,C).

post([]).
post([X|Xs]) :- X, post(Xs).

post_curly([]).
post_curly([X|Xs]) :- {X}, post_curly(Xs).


%%% primitives(C1,C2)
%%% C2 is the result of transforming equalities in C1 to a conjunction of inequalities
primitives([],[]).
primitives([C|Cs],PC) :-
	(functor(C,'=:=',2) ->
	    arg(1,C,LHS), arg(2,C,RHS),
	    PC1 =.. ['=<',LHS,RHS], PC2 =.. ['>=',LHS,RHS],
	    PC=[PC1,PC2|PCs]
	;
	  functor(C,'#=',2) ->
	    arg(1,C,LHS), arg(2,C,RHS),
	    PC1 =.. ['#=<',LHS,RHS], PC2 =.. ['#>=',LHS,RHS],
	    PC=[PC1,PC2|PCs]
	;
	    PC=[C|PCs]
	
	),
	primitives(Cs,PCs).

%%% keep_entailed(Solver,C1,C2,C3)
%%% C3 is the subset of the constraints in C1 entailed by C2
keep_entailed(S,C1,C2,C3) :-
	keep_entailed(S,C1,C2,C3,_).


keep_entailed(_Solver,[],_C2,[],[]).
keep_entailed(Solver,C1,C2,E,NE) :-
	C1=[PC1|PC1s],
	(entails(Solver,C2,[PC1]) ->
	    E=[PC1|E1],
	    keep_entailed(Solver,PC1s,C2,E1,NE)
	;
	    NE=[PC1|NE1],
	    keep_entailed(Solver,PC1s,C2,E,NE1)
	).



propagate_and_remove_equalities([],[]).
propagate_and_remove_equalities([C|Cs],NewC) :-
	functor(C,F,2),arg(1,C,LHS), arg(2,C,RHS),
	( (F == '=:='; F == '=')   ->
		 ( (var(LHS),ground(RHS)) ->
			  LHS=RHS,
			  propagate_and_remove_equalities(Cs,NewC)
		 ;
			  ( LHS==RHS ->
					propagate_and_remove_equalities(Cs,NewC)
			  ;
					NewC = [LHS is RHS|NewC1],
					propagate_and_remove_equalities(Cs,NewC1)
			  )
		 )
	;
		 NewC = [C|NewC1],
		 propagate_and_remove_equalities(Cs,NewC1)
	).


%%% example of usage:	generate_constraint([X,Y],['=:=','>=','>','=<','<'],[0],PC1)

generate_constraint(Vars,CPreds,Constants,Ls) :-
%%%% quick and dirty solution for arity 2 constraints	
	(setof(C,generate_constraint2(Vars,CPreds,Constants,C),Ls) ->
		 true
	;
		 Ls=[]
	).

generate_constraint2(Vars,CPreds,Constants,C) :-
	member(F,CPreds),
	member(X,Vars),
	(
	  (member(Y,Vars),X\==Y)
	;
	  member(Y,Constants)
%%%	;
%%%	  member(Op,[+,-]), RHS =.. [Op,Y,1],
%%%	  arg(2,C,RHS)
	),
	functor(C,F,2),
	arg(1,C,X),
	arg(2,C,Y).
