

:- discontiguous list_to_constraint/3.
:- discontiguous solve/4.
:- discontiguous entails/3.
:- discontiguous negate/3.
:- discontiguous is_a_solver/1.
:- discontiguous is_a_constraint_predicate/3.

:- multifile css_param_info/2.
:- discontiguous css_param_info/2.



:- use_module(library(clpq), [ {}/1, entailed/1, dump/3 ]).
:- use_module(library(lists), [ memberchk/2]).
%%%:- use_module(library(terms), [term_variables/2]).
%%%:- use_module(library(charsio), [atom_to_chars/2,write_to_chars/2,
%%%				 read_from_chars/2
%%%				]).



is_a_constraint_predicate(_,'false',0).
is_a_constraint_predicate(_,'=:=',2).

%%% IMPORTANT NOTICE
%%% Sets of constraint predicate symbols corresponding to different solvers must be disjoint. This is a consequence of the fact that the current "solver" is guessed when a new derivation is started (see parse_init_prog_clause in file environment.pl). 


is_a_solver(clpq).

is_a_constraint_predicate(clpq,'=\\=',2).
is_a_constraint_predicate(clpq,'<',2).
is_a_constraint_predicate(clpq,'>',2).
is_a_constraint_predicate(clpq,'>=',2).
is_a_constraint_predicate(clpq,'=<',2).


%%% solve(Solver,C,IntVars,PC):
%%% PC is the projection of C over the variables in the list IntVars
%%% using solver Solver
%%%  - C,PC are constraints
%%%  - T is a term
%%% IMPORTANT NOTICE: vars(IntVars) must be included in vars(C)

solve(clpq,ConsLst, IntVars, PConsLst) :-
	!,
	(ConsLst==[] ->
		 PConsLst=[]
	;
	    ( memberchk(false,ConsLst) ->
			  PConsLst=[false]
	    ;
		list_to_constraint(clpq,ConsLst,C),
		copy_term((C,IntVars),(C1,IntVars1)),

		retractall(projection_temporary(_)),
		assert((projection_temporary(IntVars1) :- C1)),
		( call_residue(projection_temporary(IntVars1), _R1) ->
		    get_singletons(IntVars,IntVars1,Eqs,DumpVars,DumpVars1),
		    C1,dump(DumpVars1,DumpVars,D),
		    fix_equalities(D,DF),
		    append(Eqs,DF,PConsLst)
		;
		    PConsLst=[false]
		)
	    )
	).
%%%%%% test whether this is a projecting solver
%%%	free_variables(PConsLst,IntVars,[],ExtraVars),
%%%	( ExtraVars==[] ->
%%%	    true % solver is projecting
%%%	;
%%%	    throw(clpq_solver_not_projecting(ConsLst, IntVars, PConsLst))
%%%	).

entails(clpq,CL,DL) :-
	!,
	(CL==[] -> DL=[]
	;
	    (DL==[] ->
		true
	    ;
		(\+ satisfiable(clpq,DL) ->
		    CL=[false]
		;

		    (satisfiable(clpq,CL) ->
			list_to_constraint(clpq,CL,C),
			list_to_constraint(clpq,DL,D),
			entails_clpq(C,D)
		    ;
			true
		    )
		)
	    )
	).


% entails_clpq(C,D): C entails D 
%  - C,D are constraints of the form {c1,...,cn}
%
% Note for CLPQ: (does this apply also to  other domains?)
% copy_term needed to not bind variables to rationals
% this would cause trouble with   numbervars/3

entails_clpq(C,D) :-
	(copy_term(C,C1),\+C1 -> true
	;
	% entailed does not like curly brackets
	    (copy_term((C,D),(C2,{D2})),
		C2,entailed(D2) -> true
	    ;
		fail)
	).


%%% the negation of every basic constraint
%%% can be partitioned

negate(clpq,C,NC) :- negate_clpq(C,NC).


negate_clpq( A<B,   A>=B) :- !.
negate_clpq( A>B,   A=<B) :- !.
negate_clpq( A=<B,  A>B) :- !.
negate_clpq( A>=B,  A<B) :- !.
negate_clpq( A=:=B, A>B).
negate_clpq( A=:=B, A<B).

get_singletons([],[],[],[],[]).
get_singletons([VC|VCs],[VC1|VC1s],P,DV,DV1) :-
	(ground(VC1) ->
	    % build equality
	    functor(PC,=:=,2),
	    arg(1,PC,VC),
	    arg(2,PC,VC1),
	    P=[PC|PCs],
	    get_singletons(VCs,VC1s,PCs,DV,DV1)
	;
	    % collect vars to dump on
	    DV=[VC|DVs],DV1=[VC1|DV1s],
	    get_singletons(VCs,VC1s,P,DVs,DV1s)
	).


list_to_constraint(clpq,L,C) :-
	( L=[] -> C={1=1}
	;
	    list_to_clpq_constraint(L,C)
	).

list_to_clpq_constraint([X],{X}).
list_to_clpq_constraint([X,Y|Xs],{X,CXs}):-
	list_to_clpq_constraint([Y|Xs],{CXs}).


%%%list_to_clpq_constraint(L,C) :-
%%%	term_variables(L,VL),
%%%	write_to_chars(L,LC),
%%%	substitute(91,LC,123,LC1), % replace '[' by '{'
%%%	substitute(93,LC1,125,LC2),  % replace ']' by '}'
%%%	atom_to_chars(.,Dot),	
%%%	flatten([LC2,Dot],LC3),
%%%	read_from_chars(LC3,C),
%%%	term_variables(C,VC),
%%%	VC=VL.


css_param_info(clpq,[display('clpq'),
		     description('A constraint solver based on Holzbaur\'s clp(Q,R) solver.')]).


%%% NONE
is_a_solver(none).

is_a_constraint_predicate(none,_,_) :- fail.

solve(none,[],_,[]) :- !.
entails(none,[],[]) :- !.

css_param_info(none,[display('None')]).


%%% CLPFD

%:- use_module(library(clpfd)).
%:- assert(clpfd:full_answer).
%:- use_module(library(lists), [ memberchk/2 ]).

%%:- set_current(solver,clpfd).


%is_a_solver(clpfd).


%is_a_constraint_predicate(clpfd,'false',0).
%is_a_constraint_predicate(clpfd,'#=',2).
%%is_a_constraint_predicate(clpfd,'=',2).
%is_a_constraint_predicate(clpfd,'#\=',2).
%is_a_constraint_predicate(clpfd,'#<',2).
%is_a_constraint_predicate(clpfd,'#>',2).
%is_a_constraint_predicate(clpfd,'#>=',2).
%is_a_constraint_predicate(clpfd,'#=<',2).
%% Reified constraints
%is_a_constraint_predicate(clpfd,'#<=>',2). 
%% Membership constraints
%is_a_constraint_predicate(clpfd,domain,3).
%is_a_constraint_predicate(clpfd,in,2).
%is_a_constraint_predicate(clpfd,in_set,2).
%% Combinatorial constraints
%is_a_constraint_predicate(clpfd,all_different,1).
%is_a_constraint_predicate(clpfd,all_distinct,1).




%% solve(Solver,C,IntVars,PC):
%% PC is the projection of C over the variables in the list IntVars
%% using solver Solver
%%  - C,PC are constraints
%%  - T is a term
%% IMPORTANT NOTICE: vars(IntVars) must be included in vars(C)

%solve(clpfd,C, IntVars, PC) :-
%	(C==[] ->
%		 PC=[]
%	;
%	    ( memberchk(false,C) ->
%			  PC=[false]
%	    ;
%		copy_term((C,IntVars),(C1,IntVars1)),
%		retractall(clpfd_temp(_)),
%		assert((clpfd_temp(IntVars1) :- post(C1))),
%		( call_residue(clpfd_temp(IntVars1), R1) ->
%		    get_clpfd_singletons(IntVars,IntVars1,Eqs,RVars,RVars1),
%		    RVars1=RVars,
%		    convert_call_residue(IntVars1,R1,R),
%%%%		    normalise_clpfd_constraint(CR,R),
%		    append(Eqs,R,PC)
%		;
%		    PC=[false]
%		)
%	    )
%	),
%	free_variables(PC,IntVars,[],ExtraVars),
%	( ExtraVars==[] ->
%	    true % solver is projecting
%	;
%	   true %%% throw(clpfd_solver_not_projecting)
%	).


%convert_call_residue(_,[],[]).
%convert_call_residue(Vars,['-'(RVars,Constr)|T],Res) :-
%    ( ( \+ intersection_vars_empty(Vars,RVars))
%       /* improve: check whether relationship necessary */
%     -> (Res = [Constr|CT]) ; (Res = CT)
%    ),
%    convert_call_residue(Vars,T,CT).

%normalise_clpfd_constraint([],[]).
%normalise_clpfd_constraint([in(Var,'..'(Low,Up))|T],Res) :- !,
%    Res = [Var#>=Low, Var#=<Up|NT],
%    normalise_clpfd_constraint(T,NT).
%normalise_clpfd_constraint([C|T],[C|NT]) :-
%    normalise_clpfd_constraint(T,NT).

%entails(clpfd,C,D) :-
%	(C==[] -> D=[]
%	;
%	    (D==[] -> true
%	    ;
%		(satisfiable(clpfd,C) ->
%		    entails_clpfd(C,D)
%	    ;
%		    true
%		)
%	    )
%	).


%% entails_clpfd(C,D): C entails D 
%%  - C,D are constraints lists of the form c1,...,cn
%%

%entails_clpfd(C,D) :-
%	(copy_term(C,C1),\+ post(C1) -> true
%	;
%	    copy_term((C,D),(C2,D2)),
%	    list_to_constraint(clpfd,C2,C2t),
%	    list_to_constraint(clpfd,D2,D2t),
%	    negate_clpfd( D2t, D2n),	    
	    
%	    ( call(( C2t, \+ (D2n))) ->
%		true
%	    ;
%		fail
%	    )
	    
%	).



%list_to_constraint(clpfd,[X|Xs],C) :-
%	(Xs == [] ->
%	    C=X
%	;
%	    C = ','(X,CXs),
%	    list_to_constraint(clpfd,Xs,CXs)
%	).



%entailed_clpfd2( CL) :-
%	list_to_constraint(clpfd,CL,C),
%	negate_clpfd( C, Cn),
%	free_variables(Cn,[],[],Vars),
%	\+ ( call(Cn)).
%	%%% \+ ( call( (Cn,  labeling([],Vars)) )).


%negate_clpfd( Rel,   _) :- var( Rel), !, throw(instantiation_error(negate_clpfd(Rel))).
%negate_clpfd( (A,B), (Na;Nb)) :- !, negate_clpfd( A, Na), negate_clpfd( B, Nb).	
%negate_clpfd( (A;B), (Na,Nb)) :- !, negate_clpfd( A, Na), negate_clpfd( B, Nb).	
%negate_clpfd( A#<B,   A#>=B) :- !.
%negate_clpfd( A#>B,   A#=<B) :- !.
%negate_clpfd( A#=<B,  A#>B) :- !.
%negate_clpfd( A#>=B,  A#<B) :- !.
%negate_clpfd( A#=B,   A#\=B) :- !.
%negate_clpfd( A#\=B, A#=B) :- !.
%negate_clpfd( Rel,   _) :- throw( type_error(negate_clpfd(Rel))).
	   
	   

%entailed_clpfd([]).
%% entailed_clpfd([X|Xs]) :- (X #<=> 1), entailed_clpfd(Xs).
%% didn't work
%entailed_clpfd([X|Xs]) :-
%	reifyall([X|Xs],R),
%	post(R).

%reifyall([],[]).
%reifyall([X|Xs],[X #<=> 1 | Rs]) :-
%	reifyall(Xs,Rs).



%get_clpfd_singletons([],[],[],[],[]).
%get_clpfd_singletons([VC|VCs],[VC1|VC1s],P,DV,DV1) :-
%	(ground(VC1) ->
%	    % build equality
%	    functor(PC,#=,2),
%	    arg(1,PC,VC),
%	    arg(2,PC,VC1),
%	    P=[PC|PCs],
%	    get_clpfd_singletons(VCs,VC1s,PCs,DV,DV1)
%	;
%	    % collect vars to unify later
%	    DV=[VC|DVs],DV1=[VC1|DV1s],
%	    get_clpfd_singletons(VCs,VC1s,P,DVs,DV1s)
%	).



