:- use_module(library(lists),[
			      append/3,
			      nth/3,
			      nth/4,
			      prefix/2
			      ]).

:- use_module(library(ugraphs),[
				add_vertices/3,
				add_edges/3,
				neighbors/3
				]).



%%% "LEGACY" CASE SPLIT

split_clause(X,Y,Z) :- apply_split_clause(X,Y,Z).

apply_split_clause(error(Alert),CN,_) :-
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.
apply_split_clause(Out,CN,A) :-
	apply_split_clause(Out,CN,A,_).


apply_split_clause(Out,CN,A,Cls) :-
	clause_hd_bd(CN,H,bd(C,Lits)),
	split_clause_wrt_instance_of_head((H,bd(C,Lits)),A,Cls),
	Out=out(label(case_split(CN)), [CN], Cls).


split_clause_wrt_instance_of_head((H,bd(C,Lits)),A,Cls) :-
	(variant(H,A) ->
	    Cls = [(H,bd(C,Lits))]
	;
	    append(C,Lits,B),
	    apply_split_clause_(H,B,A,Cls1,_),
	    apply_split_clauses_(Cls1,A,Cls2),
	    (current(solver,S) -> true ; (S = none)),
	    find_constraints2_clauses(Cls2,S,Cls)
	).


apply_split_clauses_([(H,bd([],B))|L],A,Cls) :-
	apply_split_clause_(H,B,A,Cls1,Applied),
	Applied==split, !,
	append(Cls1,L,NewL),
	apply_split_clauses_(NewL,A,Cls).
apply_split_clauses_([(H,bd([],B))|L],A,[(H,bd([],B))|Cls]) :-
	apply_split_clauses_(L,A,Cls).
apply_split_clauses_([],_,[]).

apply_split_clause_(H,B,A,Cls,Applied) :-
	anti_unify(H,A, Theta_H, Theta_H2,_),
	recover_split_clause_arg(H,B,Theta_H,Theta_H2,Cls,Applied).

recover_split_clause_arg(H,B,[V1=A_H|_],[V1=A_H2|_],Cls,split):-
    var(A_H), nonvar(A_H2),
    (belongs( \+(A_H=A_H2),B) -> fail ;
         case_split_arg((H,B),A_H,A_H2,Cls)).
recover_split_clause_arg(H,B,[_|L1],[_|L2],Cls,Applied) :-
    recover_split_clause_arg(H,B,L1,L2,Cls,Applied).
recover_split_clause_arg(H,B,[],[],[(H,bd([],B))],_).

case_split_arg((H,B),A_H,K, [(H,bd([],B3)), (H2, bd([],[\+(A_H2=K)| B2]))]):-
   copy_term((A_H,H,B),(A_H2,H2,B2)), A_H=K,
   singletons_vars_body(B,Singletons),
   simpl_equalities_body(Singletons,B,B3).

simpl_equalities_body(_,[],[]).
simpl_equalities_body(Singletons,[\+(A=B)|Bd1],Bd2):-
   ground(A),ground(B),A\==B,!,
   simpl_equalities_body(Singletons,Bd1,Bd2).
simpl_equalities_body(Singletons,[(A=B)|Bd1],Bd2):-
   A==B,!,
   simpl_equalities_body(Singletons,Bd1,Bd2).
simpl_equalities_body(Singletons,[([A|B]=[C|D])|Bd1],[B=D|Bd2]):-
   A==C,!, simpl_equalities_body(Singletons,Bd1,Bd2).
simpl_equalities_body(Singletons,[\+([A|B]=[C|D])|Bd1],[\+(A=C)|Bd2]):-
   (belongs(D,Singletons) ; belongs(B,Singletons)),
   simpl_equalities_body(Singletons,Bd1,Bd2).
simpl_equalities_body(Singletons,[P|Bd1],[P|Bd2]):-
   simpl_equalities_body(Singletons,Bd1,Bd2).

singletons_vars_body(B,Sings):-
   free_variables(B,[],[],Vars), occurrences_of_vars(Vars,B,Sings).

occurrences_of_vars([V|Vars],B,[V|Sings]):-
   occurrences_of_var(V,B,1),!, occurrences_of_vars(Vars,B,Sings).
occurrences_of_vars([_|Vars],B,Sings):- occurrences_of_vars(Vars,B,Sings).
occurrences_of_vars([],_,[]).


check_entry_case_split(N,A) :-
	clause_hd_bd(N,H,_),
	subsumes_chk(H,A),
	\+ variant(H,A).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




apply_split_determ(error(Alert),CN,_) :-
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.

apply_split_determ(Out,CN,(A,CA)) :-
	clause_hd_bd(CN,H,bd(C,B)),
	current(solver,S),
	determ_clause((H,bd(C,B)),S,(A,bd(CA,[])),Cls),
	Out=out(label(case_split(CN)), [CN], Cls).








%%% split_determ(Cls,S,Packets)
%%% Cls is a list of clauses defining the same predicate
%%% S is a solver
%%% Packets is a list of packets (lists of clauses with same heads and equivalent input constraints)



split_determ(Cls,S,Packs) :-
	determ_clauses(Cls,S,DCls),
	create_packets(DCls,S,Packs).

%%%%%% split_determ_(Clss,S,Packetss) similar to split_determ
%%%%%% but works on lists of lists
%%%split_determ_([],_,[]).
%%%split_determ_([Cls|Clss],S,[Packs|Packss]) :-
%%%	split_determ(Cls,S,Packs),
%%%	split_determ_(Clss,S,Packss).


%%% case_split_strategy_(Clss,S,Packetss) similar to case_split_strategy
%%% but works on lists of lists
case_split_strategy_([],_,[]).
case_split_strategy_([Cls|Clss],S,Packs) :-
	case_split_strategy_term_constr(Cls,S,Packs1),
	append(Packs1,Packs2,Packs),
	case_split_strategy_(Clss,S,Packs2).

case_split_strategy_term_constr(Cls,S,Packs) :-
	case_split_strategy(Cls,S,term,NewCls),
	(S == none ->
	    NewCls2 = NewCls
	;
	    case_split_strategy(NewCls,S,constr,NewCls2)
	),
	equivalence_classes_input(NewCls2,S,term,EqClassesTerm),
	equivalence_classes_input_(EqClassesTerm,S,constr,Packs).



%%% case_split_strategy(Cls,S,T,NewCls)
case_split_strategy([],_,_,[]) :- !.
case_split_strategy([Cl],_,_,[Cl]) :- !.
%%% more complex and hopefully more efficient :-) solution
case_split_strategy(Cls,S,T,NewCls) :-
 %	equivalence_classes_input(Cls,S,T,EqClasses),
 %	get_representative(EqClasses,ReprCls),
 %	fake_clauses(ReprCls,S,T,FakeCls),
	overlapping_graph(Cls,S,T,Graph,LN),
	split_all_clauses(Cls,LN,Cls,LN,Graph,S,T,SplitClss),
	flatten(SplitClss,SplitCls),
	equivalence_classes_input(SplitCls,S,T,Packs),
 %%%	minimal_hd_gen_packets(Packs,Packs2),
 %%%	flatten(Packs2,NewCls).
	flatten(Packs,NewCls).

%%% "natural" solution (does not work! :-( )
%%%case_split_strategy(Cls,S,T,NewCls) :-
%%%	nth(_,Cls,Cl1,R),
%%%	nth(_,R,Cl2),
%%%	clauses_overlap_input(Cl1,Cl2,S,T),!,
%%%	split_clause(Cl1,Cl2,S,T,SplitCl1),
%%%	append(SplitCl1,R,NewInCls),
%%%	case_split_strategy(NewInCls,S,T,NewCls).


	
minimal_hd_gen_packets([],[]).
minimal_hd_gen_packets([P|Ps],[P1|P1s]) :-
%%% minimal_hd_generalization_clauses_(P,_,_,P1),
%%% generalizes too much	
	minimal_hd_gen_packet(P,P1),
	minimal_hd_gen_packets(Ps,P1s).


minimal_hd_gen_packet(P,P1) :-
	P \== [],
	clauses_heads(P,Heads),
	anti_unify_list(Heads,GH),
	minimal_hd_gen_packet(P,GH,P1).

minimal_hd_gen_packet([],_,[]).
minimal_hd_gen_packet([Cl|Cls],GH,[(GH,bd(C,Lits1))|GCls]) :-
	Cl = (H,bd(C,Lits)),
	subsumes_chk(GH,H),	% sanity check
	anti_unify(H,GH,Theta,_,GH),
	call_eq_vars(Theta),
	simpl_equalities_body([],Theta,ThetaS),
	append([ThetaS,Lits],Lits1),
	minimal_hd_gen_packet(Cls,GH,GCls).
	



%%%reconstruct_packets([],_,_,[]).
%%%reconstruct_packets([Class|Cs],[P|Ps],S,Cls) :-
%%%	clauses_heads(Class,Heads),
%%%	anti_unify_list(Heads,GH),	
%%%	reconstruct_packet(Class,P,GH,S,Cls1),
%%%	append(Cls1,Cls2,Cls),
%%%	reconstruct_packets(Cs,Ps,S,Cls2).

%%%reconstruct_packet([],_P,_,_,[]).
%%%reconstruct_packet([Cl|Cls],P,GH,S,NewCls) :-
%%%	reconstruct_clauses(Cl,P,GH,S,NewCls1),
%%%	append(NewCls1,NewCls2,NewCls),
%%%	reconstruct_packet(Cls,P,S,NewCls2).

%%%reconstruct_clauses(_,[],_,_,[]).
%%%reconstruct_clauses(Cl1,[Cl2|Cl2s],GH,S,[Cl3|Cl3s]) :-
%%%	reconstruct_clause(Cl1,Cl2,GH,S,Cl3),
%%%	reconstruct_clauses(Cl1,Cl2s,GH,S,Cl3s).

%%%%%% reconstruct_clause(Cl1,Cl2,S,H,NewCl)
%%%%%% Cl1 is the original clause
%%%%%% Cl2 is one of the fake clauses obtained by splitting
%%%%%% the fake clause representing the equivalence class of Cl1
%%%reconstruct_clause(Cl1,Cl2,H,S,NewCl) :-
%%%	copy_term(Cl1,Cl1copy),
%%%	Cl1copy = (H1,bd(C1,Lits1)),
%%%	Cl2 = (H2,bd(C2,Lits2)),
%%%	subsumes_chk(H,H2),	% sanity check
%%%	anti_unify(H2,H,Theta2,_,GH),
%%%	call_eq_vars(Theta2),
%%%	simpl_equalities_body([],Theta2,Theta2S),
%%%	input_args(H1,InArgs1),input_args(H2,InArgs2),
%%%	subsumes_chk(InArgs1,InArgs2),
%%%	unify_with_occurs_check(InArgs1,InArgs2),
%%%	append(C2,C1,C3),
%%%	append([Theta2S,Lits2,Lits1],Lits3), % does this inserts dups ?
%%%	solve_clause((GH,bd(C3,Lits3)),linking_vars,S,NewCl).


call_eq_vars([]).
call_eq_vars([A=B|Ls]) :-
	var(A),var(B),!,A=B,
	call_eq_vars(Ls).
call_eq_vars([_|Ls]) :-
	call_eq_vars(Ls).


	
split_all_clauses([],[],_AllCls,_ToUseNs,_G,_S,_T,[]).
split_all_clauses([Cl|Cls],[CN|CNs],AllCls,ToUseNs,G,S,T,[SplitCl|SplitCls]) :-
	neighbors(CN,G,Neighbs),
	list_intersection(Neighbs,ToUseNs,ToUseN1),
	split_all_clauses_([Cl],AllCls,ToUseN1,G,S,T,SplitCl),
	split_all_clauses(Cls,CNs,AllCls,ToUseNs,G,S,T,SplitCls).
	

split_all_clauses_(Cls,_AllCls,[],_G,_S,_T,Cls).
split_all_clauses_(Cls,AllCls,[N|Ns],G,S,T,SplitCls) :-
	nth(N,AllCls,ToUse),
	split_clauses__(Cls,ToUse,S,T,SplitCls1),
	split_all_clauses_(SplitCls1,AllCls,Ns,G,S,T,SplitCls).


	
%%%	%%% we could optimize as follows
%%%	split_clauses__(Cls,ToUse,S,T,Cls_and_ToUse,Cls_and_NotToUse),

%%%	neighbors(N,G,Neighbs),
%%%	list_intersection(Neighbs,Ns,ToUseNs),
%%%	split_all_clauses_(Cls_and_ToUse,AllCls,ToUseNs,G,S,T,SplitCls1),
%%%	split_all_clauses_(Cls_and_NotToUse,AllCls,Ns,G,S,T,SplitCls2),
%%%	append(SplitCls1,SplitCls2,SplitCls).



split_clauses__([],_ToUse,_S,_T,[]).
split_clauses__([Cl|Cls],ToUse,S,T,SplitCls) :-
	split_clause(Cl,ToUse,S,T,SplitCls1),
	append(SplitCls1,SplitCls2,SplitCls),
	split_clauses__(Cls,ToUse,S,T,SplitCls2).
	
	
split_clause(Cl,ToUse,S,constr,SplitCls) :-
	determ_clause(Cl,S,ToUse,SplitCls).
split_clause(Cl,ToUse,_S,term,SplitCls) :-
	Cl = (H1,_), ToUse = (H2,_),
	copy_term(H1,H1c),copy_term(H2,H2c),
	input_args(H1c,InArgs1),input_args(H2c,InArgs2),
	(unify_with_occurs_check(InArgs1,InArgs2) ->
	    split_clause_wrt_instance_of_head(Cl,H2c,SplitCls)
	;
	    SplitCls = [Cl]
	).


	
%%% Improve: make it a _directed_ graph
overlapping_graph(Cls,S,T,Graph,LN) :-
	list_of_ints(Cls,LN),
	add_vertices([],LN,G1),
	
	(setof([N1-N2,N2-N1],
	%(setof([N1-N2], % to make overlapping graph is directed
	      Cl1^Cl2^(
		       nth(N1,Cls,Cl1),
		       nth(N2,Cls,Cl2),
		       % N1 \== N2,  % to make overlapping graph is directed
		       N1 < N2,
		       clauses_overlap_input(Cl1,Cl2,S,T)
		      ),
	      Edges1) ->
	    flatten(Edges1,Edges),
	    add_edges(G1,Edges,Graph)
	;
	    Graph = G1
	).
	
	
list_of_ints(L,LN) :- (setof(N,El^nth(N,L,El),LN) -> true ; LN = []).

	     
fake_clauses([],_,_,[]).
fake_clauses([(H,bd(C,_))|Cls],S,T,[FCl|FCls]) :-
	( T == term ->
	    %%% TODO: put vars in noninput positions (cleaner)
	    FCl = (H,bd([],[]))
	;
	    (T == constr ->
		input_vars(H,InVars),
		solve(S,C,InVars,D),
		FCl = (H,bd(D,[]))
	    )
	),
	fake_clauses(Cls,S,T,FCls).


clauses_overlap_input(Cl1,Cl2,S,constr) :-
	Cl1 = (H1,bd(C1,_G1)),
	copy_term(Cl2,Cl2copy),
	Cl2copy = (H2,bd(C2,_G2)),
	input_args(H1,InArgs1),	input_args(H2,InArgs2),
	variant(InArgs1,InArgs2),
	unify_with_occurs_check(InArgs1,InArgs2),
	(equivalent_constraints(S,C1,C2) ->
	    fail
	;
	    append(C1,C2,C),
	    satisfiable(S,C)
	    %, \+ entails(S,C1,C2) % to make overlapping graph is directed
	).

clauses_overlap_input(Cl1,Cl2,_S,term) :-
	copy_term(Cl1,Cl1copy),
	Cl1copy = (H1,bd(_C1,_G1)),
	copy_term(Cl2,Cl2copy),
	Cl2copy = (H2,bd(_C2,_G2)),
	input_args(H1,InArgs1),	input_args(H2,InArgs2),
	\+ variant(InArgs1,InArgs2),
	unify_with_occurs_check(InArgs1,InArgs2).



determ_clauses([],_S,[]).
determ_clauses([Cl|Cls],S,[Cl|DCls]) :-
	determ_clauses_(Cls,S,[Cl|Cls],DCls).

%%% determ_clauses_(ToSplit,S,OrigCls,Split)
determ_clauses_(TS,_,[],TS).
determ_clauses_(TS,S,[Cl|Cls],SplitCls) :-
	determ_clauses__(TS,S,Cl,SplitCl),
	flatten(SplitCl,SplitCl1),
	determ_clauses_(SplitCl1,S,Cls,SplitCls).


determ_clauses__([],_,_,[]).
determ_clauses__([Cl1|Cls],S,Cl2,[DCl1s|DCls]):-
	determ_clause(Cl1,S,Cl2,DCl1s),
	determ_clauses__(Cls,S,Cl2,DCls).


%%% determ_clause(Cl1,S,Cl2,DCl1s)
%%% DCl1s is a disjunction (as list) of "deterministic" clauses
%%% obtained by splitting Cl1 w.r.t. Cl2
%%% IMPORTANT: we assume that heads are variants
%%% TODO: extend to the more general case
determ_clause(Cl1,S,Cl2,DCl1s) :-
%%%	copy_term(Cl1,Cl1copy),
%%%	Cl1copy = (H1,bd(C1,_G1)),
	Cl1 = (H1,bd(C1,_G1)),	
	copy_term(Cl2,Cl2copy),
	Cl2copy = (H2,bd(C2,_G2)),
	input_vars(H1,InVars1),
	solve(S,C1,InVars1,D1),
	input_vars(H2,InVars2),
	solve(S,C2,InVars2,D2),
	unify_with_occurs_check(H1,H2),
	(equivalent_constraints(S,D1,D2) ->
	    DCl1s = [Cl1]
	;
	    append(D1,D2,D),
	    (satisfiable(S,D) ->
		partition_negation(D2,S,NotD2),
		merge_constrs(C1,NotD2,C1NotD2),
		NewCs = [D|C1NotD2],
		determ_clause_(Cl1,NewCs,S,DCl1s)
	    ;
		DCl1s = [Cl1]
	    )
	).

determ_clause_(_Cl,[],_S,[]).
determ_clause_((H,bd(C,G)),[D|Ds],S,DCls) :-
	append(C,D,C1),
	(satisfiable(S,C1) ->
	    free_variables((H,G),[],[],VC1),
	    solve(S,C1,VC1,C2),
	    DCls = [(H,bd(C2,G))|DCls1],
	    determ_clause_((H,bd(C,G)),Ds,S,DCls1)
	;
	    determ_clause_((H,bd(C,G)),Ds,S,DCls)
	).

%%% merge_constrs(C1,C2,C1C2)
%%% C1 is a constraint
%%% C2 is a disjunction (as list) of basic constraints
%%% C1C2 is a disjunction (as list) of constraints

merge_constrs(_C,[],[]).
merge_constrs(C,[D|Ds],[CD|CDs]) :-
	append(C,D,CD),
	merge_constrs(C,Ds,CDs).

%%% partition_negation(C,S,NotC)
%%% NotC is a disjunction (as list) of "deterministic" constraints
%%% equivalent to not(C)
partition_negation([],_,[false]).
partition_negation([C|Cs],S,NotC) :-
	(satisfiable(S,[C|Cs]) ->
	    partition_negation_([C|Cs],S,NotC)
	;
	    NotC = []
	).

partition_negation_(C,S,NotC) :-
	free_variables(C,[],[],VC),
	setof(NC,
	      P^NBC^L^L1^BC^(
			     prefix(P,C),
			     length(P,L),
			     L1 is L+1,
			     nth(L1,C,BC),
			     negate(S,BC,NBC),
			     solve(S,[NBC|P],VC,NC),
			     satisfiable(S,NC)
			    ),
	      NotC).
%	length(C,L),
%	partition_negation_basic(C,S,[],1,L,NotC).

%%%partition_negation_basic(C,S,CurrNotC,Pos,MaxP,NotC) :-
%%%	(Pos =< MaxP ->
%%%	    nth(Pos,C,BC,RestC),
%%%	    setof(NC,NBC^(negate(S,BC,NBC),NC = [NBC|RestC]),NCs),
%%%	    append(NCs,CurrNotC,NewNotC),
%%%	    NextPos is Pos + 1,
%%%	    partition_negation_basic(C,S,NewNotC,NextPos,MaxP,NotC)
%%%	;
%%%	    NotC = CurrNotC
%%%	).


atom_modes(A,Modes) :-
	A =.. [F|Args],
	(mode_definition(F,Modes) ->
	    true
	;
	    same_length(Args,Modes),
	    all_input(Modes)
	).


all_input([]).
all_input(['+'|Xs]) :- all_input(Xs).


input_args(H,InArgs) :- input_args(H,InArgs,_).
input_args(H,InArgs,NonInArgs) :-
	atom_modes(H,Modes),
	H =.. [_|Args],
	get_input_args(Modes,Args,InArgs,NonInArgs).
	


get_input_args(Modes,Args,InArgs) :- 	get_input_args(Modes,Args,InArgs,_).

get_input_args([],[],[],[]).
get_input_args(['+'|Ms],[Arg|Args],[Arg|InArgs],NonInArgs) :- !,
	get_input_args(Ms,Args,InArgs,NonInArgs).
get_input_args([_|Ms],[Arg|Args],InArgs,[Arg|NonInArgs2]) :- 
	get_input_args(Ms,Args,InArgs,NonInArgs2).



input_vars(H,InVars) :- input_args_vars(H,_,InVars).

input_args_vars(H,InArgs,InVars) :-
	input_args(H,InArgs),
	free_variables(InArgs,[],[],InVars).


equivalence_classes_input_([],_S,_T,[]).
equivalence_classes_input_([Class|Cs],S,T,Packs) :-
	equivalence_classes_input(Class,S,T,Packs1),
	append(Packs1,Packs2,Packs),
	equivalence_classes_input_(Cs,S,T,Packs2).

equivalence_classes_input([],_S,_T,[]).
equivalence_classes_input([Cl|Cls],S,T,[[Cl|Equiv]|Packets]) :-
	(T == constr ->
	    equivalent_input_constraints(Cls,Cl,S,Equiv,NotEquiv)
	;
	    (T == term ->
		equivalent_input_terms(Cls,Cl,S,Equiv,NotEquiv)
	    )
	),
	equivalence_classes_input(NotEquiv,S,T,Packets).


get_representative([],[]).
get_representative([[Cl|_]|EqCls],[Cl1|Rest]) :-
	copy_term(Cl,Cl1),
	get_representative(EqCls,Rest).

equivalent_input_constraints([],_Cl,_S,[],[]).
equivalent_input_constraints([Cl1|Cls],Cl2,S,Equiv,NotEquiv) :-
	copy_term(Cl1,Cl1copy),
	Cl1copy = (H1,bd(C1,_G1)),
	input_args_vars(H1,InArgs1,InVars1),
	solve(S,C1,InVars1,D1),
	copy_term(Cl2,Cl2copy),
	Cl2copy = (H2,bd(C2,_G2)),
	input_args_vars(H2,InArgs2,InVars2),
	solve(S,C2,InVars2,D2),
	( variant(InArgs1,InArgs2) -> 
	    unify_with_occurs_check(InArgs1,InArgs2),
	    (equivalent_constraints(S,D1,D2) ->
		Equiv = [Cl1|Equiv2],
		NotEquiv = NotEquiv2
	    ;
		Equiv = Equiv2,
		NotEquiv = [Cl1|NotEquiv2]
	    )
	;
	    Equiv = Equiv2,
	    NotEquiv = [Cl1|NotEquiv2]
	),
	equivalent_input_constraints(Cls,Cl2,S,Equiv2,NotEquiv2).



equivalent_input_terms([],_Cl,_S,[],[]).
equivalent_input_terms([Cl1|Cls],Cl2,S,Equiv,NotEquiv) :-
	copy_term(Cl1,Cl1copy),
	Cl1copy = (H1,_),
	input_args(H1,InArgs1),
	copy_term(Cl2,Cl2copy),
	Cl2copy = (H2,_),
	input_args(H2,InArgs2),
	(variant(InArgs1,InArgs2) -> 
	    unify_with_occurs_check(InArgs1,InArgs2),
	    Equiv = [Cl1|Equiv2],
	    NotEquiv = NotEquiv2
	;
	    Equiv = Equiv2,
	    NotEquiv = [Cl1|NotEquiv2]
	),
	equivalent_input_terms(Cls,Cl2,S,Equiv2,NotEquiv2).



check_entry_split_determ(N,(A,C)) :-
	clause_hd_bd(N,H,_),
	subsumes_chk(H,A),
	variant(H,A),
	is_constraint(C,_).
