:- use_module(library(terms), [ subsumes_chk/2 ]).
:- use_module(library(lists), [
			       last/2,
			       remove_duplicates/2,
			       memberchk/2,
			       nth/3,
			       suffix/2
			      ]).


clause_hd_bd(Num,Head,Body) :-
	clause_definition(Num,Head,Body,_).


clause_hd_bd_list([],[]).
clause_hd_bd_list([CN|CNs],[(H,B)|Cls]) :-
	clause_hd_bd(CN,H,B),
	clause_hd_bd_list(CNs,Cls).


clauses_heads([],[]).
clauses_heads([(H,_)|Cls],[H|Hs]) :-
	clauses_heads(Cls,Hs).


atom_pred_arity(Atom,F,A) :- functor(Atom,F,A).

pred_hd_of_clause(CN,F,A) :-
	clause_hd_bd(CN,Head,_),
	atom_pred_arity(Head,F,A).
pred_hd_of_atom(CN,Pos,F,A) :-
	clause_hd_bd(CN,_,bd(Constrs,Atoms)),
	append(Constrs,Atoms,Body),
	pos_atom_bd(Pos,Body,Atom),
	atom_pred_arity(Atom,F,A).


pos_atom_bd(1,[Atom|_],Atom) :- !.
pos_atom_bd(N,[_|Bd],Atom) :-
	N2 is N-1,
	pos_atom_bd(N2,Bd,Atom).


negative_literal(\+(A), A, P, N) :- 
	functor(A,P,N).
negative_literal(not(A), A, P, N) :-
	functor(A,P,N).


clause_skeleton([Head|Body],HdPred/Ar,[HdPred|BodySkeleton],LengthBd) :-
	functor(Head,HdPred,Ar),
	clause_body_skeleton(Body,BodySkeleton,LengthBd).
clause_body_skeleton([],[],0).
clause_body_skeleton([L|Ls],[Skel_L|Skel_Ls],Length) :-
	clause_body_skeleton(Ls,Skel_Ls,Length2),
	Length is Length2 + 1,
	(negative_literal(L,_,P,_) -> Skel_L= \+(P)
	      ; Skel_L=L).


% in CLP a unit clause may have a non-empty body including only constraints
% check it out!!
is_a_unit_clause(N) :-
	clause_hd_bd(N,_,bd([],[])).




is_a_fact_n(CN) :-
	clause_hd_bd(CN,Hd,BdD),
	is_a_fact((Hd,BdD)).

is_a_fact((H,bd([],[]))) :-
	H =.. [_|Args],
	free_variables(Args,[],[],Args1),
	vars_seteq(Args,Args1).




get_predicates_body(L,C,L2) :-
	clause_hd_bd(C,_,bd(Constrs,Atoms)),
	append(Constrs,Atoms,Body),
	get_pred_arity(L,Body,L2).

get_pred_arity(L,[],L).
get_pred_arity(L,[B|Body],L2) :-
   atom_pred_arity(B,P,A),
   (member(P/A,L) -> get_pred_arity(L,Body,L2) ; get_pred_arity([P/A|L],Body,L2) ).


%%%% get_clauses(CNs,CLs)
%%%% CNs is a list of clause numbers
%%%% CLs is alist of clauses
%%%get_clauses([],[]).
%%%get_clauses([CN|CNs],[(H,B)|CLs]) :-
%%%	clause_hd_bd(CN,H,B),
%%%	get_clauses(CNs,CLs).

program_clauses_defs(P,All) :-
	program_clauses(P,Init),
	findall(CN, definition_for_folding(CN),New),
	append(Init,New,All).

is_an_equality(Atom) :- is_an_equality(Atom,_,_).
is_an_equality(Atom,LHS,RHS) :- Atom = (LHS=RHS).


is_an_inequality(Atom) :- is_an_inequality(Atom,_,_).
is_an_inequality(Atom,LHS,RHS) :- Atom = \+(LHS=RHS).


opposite_literals(A,NotA) :-
	(is_a_negative_literal(A,A2), A2==NotA) ;
	(is_a_negative_literal(NotA,A2), A2==A).

is_a_negative_literal(Lit,Atom) :- nonvar(Lit), Lit = (not(Atom)).
is_a_negative_literal(Lit,Atom) :- nonvar(Lit), Lit = (\+(Atom)).


pos_neg_literals([],[],[]).
pos_neg_literals([Lit|Ls],P,N) :-
	(is_a_negative_literal(Lit,_Atom) ->
	    P=P1,N=[Lit|N1]
	;
	    P=[Lit|P1],N=N1
	),
	pos_neg_literals(Ls,P1,N1).



undefined_lits_in_body([],[]).
undefined_lits_in_body([Lit|Lits],[F/A|Rels]) :-
	(is_a_negative_literal(Lit,Atom) ->
	    (functor(Atom,F,A),
	    (\+ is_a_bip(F/A), \+ is_defined_in_current_prog(F/A))) ;
	    (functor(Lit,F,A),
	     (\+ is_a_bip(F/A), \+ is_defined_in_current_prog(F/A)))), !,
	undefined_lits_in_body(Lits,Rels).
undefined_lits_in_body([_|Lits],Rels) :-
	undefined_lits_in_body(Lits,Rels).

is_a_bip(is/2).
is_a_bip(var/1).
is_a_bip(nonvar/1).
is_a_bip('='/2).
is_a_bip('\\='/2).
is_a_bip('<'/2).
is_a_bip('>'/2).
is_a_bip('=<'/2).
is_a_bip('>='/2).
%is_a_bip('=\='/2). % commented out because of a conflict with the clpq pred.
is_a_bip('=='/2).

is_defined_in_current_prog(F/A) :-
	current_program(P,_),
	is_defined_in_prog(F/A,P).

is_defined_in_prog(F/A,P) :-
	program_clauses(P,Cls),
	member(Cl,Cls),
	clause_hd_bd(Cl,Hd,_),
	atom_pred_arity(Hd,F,A), !.

is_in_current_prog(CN) :-
	current_program(_,CNs),
	member(CN,CNs), !.

ancestors(P,L) :-
	findall(X, is_an_ancestor(X,P), L).

in_ancestor(CN) :-
	current_program(P,CNs),
	(member(CN,CNs) ;
	is_an_ancestor(A,P),
	program_clauses(A,ACNs),
 	member(CN,ACNs)).

is_an_ancestor(A,P) :-
	node(P,A,_,_,_), A =\= 0.
is_an_ancestor(A,P) :-
	node(P,A2,_,_,_),
	is_an_ancestor(A,A2).

some_pos_body(N,Length,Pos) :-
	length_body_clause(N,L),
	make_sublist_count(L,Length,Pos).

length_body_clause(N,L) :-
	get_param_clause(N,length_bd,L).

is_a_definition_clause(N) :-
	get_param_clause(N,label,L),
	functor(L,definition,_).

is_a_definition_clause(N,Pred) :-
	get_param_clause(N,label,L),
	functor(L,definition,_),
	arg(1,L,Pred).

is_a_law_number(N) :-
	law_definition(_,N,_,_).

current_program_is_a_leaf :-
	current_program(P,_),
	is_a_leaf(P).

is_a_leaf(N) :-
	node(N,_,[],_,_).

node_father(N,Father) :-
	node(N,Father,_,_,_).

node_label(N,Label) :-
	node(N,_,_,label(Label),_).

node_clauses(N,CNs) :-
	node(N,_,_,_,CNs).

node_label_clauses(N,Label,CNs) :-
	node(N,_,_,label(Label),CNs).

new_clauses_at_prog(N,L) :- findall(CN, new_clause_at_prog(N,CN), L).
new_clause_at_prog(N,CN) :-
	get_param_clause(CN,prog,N).


get_params_clause(CN,Params) :-
	clause_definition(CN,_,_,Params).
get_param_clause(CN,Param,Param_Value) :-
	get_params_clause(CN,Params),
	get_param(Param,Params,Param_Value).
get_param(Param,[P|Ps],V) :-
	(functor(P,Param,_) ->
		 arg(1,P,V)
	;
		 get_param(Param,Ps,V)
	).

current_theory_empty :-
	clause(law_definition(_,_,_,_),_), !,
	fail.
current_theory_empty.


heads_unifiable([],H,H).
heads_unifiable([C|L],H,H3) :-
   clause_hd_bd(C,Hd_C,_),
   (var(H) -> H2 = Hd_C ; anti_unify(Hd_C,H,H2)), heads_unifiable(L,H2,H3).

body_without_inequalities([\+(A=B)|IneqsBd_C],[\+(A=B)|Ineqs],Bd_C):-!,
    body_without_inequalities(IneqsBd_C,Ineqs,Bd_C).
body_without_inequalities(Bd_C,[],Bd_C).

body_without_equalities([A=B|IneqsBd_C],[A=B|Ineqs],Bd_C):-!,
    body_without_equalities(IneqsBd_C,Ineqs,Bd_C).
body_without_equalities(Bd_C,[],Bd_C).


has_a_failing_body(CN) :-
	clause_hd_bd(CN,_,bd(_,B)),
	member(Lit,B),
	( is_a_negative_literal(Lit,Atom) -> 
	    is_always_true(Atom) ;
	    is_always_false(Lit)).

is_always_false(A) :-
	is_an_equality(A,LHS,RHS), !,
	\+ unify_with_occurs_check(LHS,RHS).
is_always_false(A) :-
	atom_pred_arity(A,F,N),
	is_always_false(A,F,N).
% ToDo. Nothing is done for the bips yet...
is_always_false(_,F,N) :-
	is_a_bip(F/N), !, fail. 
is_always_false(A,F,N) :-
	is_defined_in_current_prog(F/N), !,
	is_a_failing_goal(A).
is_a_failing_goal(A) :-
	program_clauses_defs(_,Cls_P),
	member(C,Cls_P),
	clause_hd_bd(C,A2,_),
	unify_with_occurs_check(A2,A), !,
	fail.
is_a_failing_goal(_).

is_always_true(A) :-
	is_an_equality(A,LHS,RHS), !,
	LHS==RHS.
is_always_true(A) :-
	atom_pred_arity(A,F,N),
	is_always_true(A,F,N).
%ToDo
is_always_true(_,F,N) :-
	is_a_bip(F/N), !, fail. 
is_always_true(A,F,N) :-
	is_defined_in_current_prog(F/N), !,
	is_an_instance_of_fact(A).
is_an_instance_of_fact(A) :-
	current_program(P,_),
	program_clauses_defs(P,Cls_P),
	member(C,Cls_P),
	clause_hd_bd(C,A2,bd([],[])),
	subsumes_chk(A2,A).


% this is called in rules.tcl
no_solver_available :-
	setof(X,is_a_solver(X),_),
	!,
	fail.
no_solver_available.

no_solver_selected :-
	current(solver,_), !,
	fail.
no_solver_selected.

no_widening_available :-
	current(solver,Solver),	
	setof(X,is_a_widening(X,Solver),_),
	!,
	fail.
no_widening_available.

no_widening_selected :-
	current(widening,_), !,
	fail.
no_widening_selected.



no_wqounf_available :-
	current(solver,Solver),	
	setof(X,is_a_wqounf(X,Solver),_),
	!,
	fail.
no_wqounf_available.

no_wqounf_selected :-
	current(wqounf,_), !,
	fail.
no_wqounf_selected.


no_wqogen_available :-
	current(solver,Solver),	
	setof(X,is_a_wqogen(X,Solver),_),
	!,
	fail.
no_wqogen_available.

no_wqogen_selected :-
	current(wqogen,_), !,
	fail.
no_wqogen_selected.


no_uf_available :-
	setof(X,is_a_uf(X,_Solver),_),
	!,
	fail.
no_uf_available.

no_uf_selected :-
	current(uf,_), !,
	fail.
no_uf_selected.


%%
%


% a tree si represented as the list of the paths
% from the leaves to the root

tree_root([Path|_Paths], Root) :-
	last(Path,Root).
% Note: to check consistency one may add  tree_root(Paths,Root)

% tree_leaves(Tree,Leaves)
% Leaves is the list of the leaves of Tree
tree_leaves([], []).
tree_leaves([[Leaf|_Ancs] | Paths ], [Leaf|Leaves]) :-
	tree_leaves(Paths,Leaves).


% tree_nodes(Tree,Nodes)
% Nodes is the list of the nodes of Tree
tree_nodes([], []).
tree_nodes([Path|Paths],Nodes) :-
	flatten([Path|Paths],FlatTree),
	remove_duplicates(FlatTree,Nodes).



% tree_addleaves(NewLeaves,OldPath,NewPaths)
% NewLeaves and OldPath are lists of clause numbers
% and NewPaths is the list of the lists obtained by inserting
% an element of NewLeaves as first element of OldPath

tree_addleaves([],Path,[Path]).
tree_addleaves([L],Path,[NewPath]) :-
	append([L],Path,NewPath).
tree_addleaves([L|Ls],Path,[NewPath|NewPaths]) :-
	append([L],Path,NewPath),
	tree_addleaves(Ls,Path,NewPaths).


%%%tree_addleaves([],_Path,[]).
%%%tree_addleaves([L|Ls],Path,[[L|Path]|NewPaths]) :-
%%%	tree_addleaves(Ls,Path,NewPaths).



% forest_nodes(Forest,Nodes)
% Nodes is the list of the nodes of Forest
forest_nodes([], []).
forest_nodes([Tree|Trees],Nodes) :-
	flatten([Tree|Trees],Nodes1),
	remove_duplicates(Nodes1,Nodes).
	
% create_forest(Roots,Forest)
% Forest is a list of trees, one for each node R in Roots,
% which consists of the root R
create_forest([],[]).
create_forest([R|Roots],[ [[R]] | Forest]) :-
	create_forest(Roots,Forest).

% update_forest(NewDef,Parent,DefsForest,NDefsForest) 
% NDefsForest is the forest obtained by making
% NewDef son of Parent in DefsForest

update_forest(_NewDef,_Parent,[],[]).
update_forest(NewDef,Parent,[DT| DTs],[NDT|NDTs]) :-
	tree_nodes(DT,Nodes),
	(memberchk(Parent,Nodes) ->
	    update_tree(NewDef,Parent,DT,NDT)
	;
	    NDT=DT
	),
	update_forest(NewDef,Parent, DTs,NDTs).

update_tree(_NewDef,_Parent,[],[]).
update_tree(NewDef,Parent,[P|Ps],[NP|NPs]) :-
	(memberchk(Parent,P) ->
	    (nth(1, P, Parent) -> % Parent is a leaf
		replace_in_list([Parent],P,[NewDef,Parent],NP),
		NPs=Ps
	    ;
		PathtoParent=[Parent|_Rest],
		suffix(PathtoParent,P),
		NP=[NewDef|PathtoParent],
		NPs=[P|Ps]
	    )
	;
	    NP=P, update_tree(NewDef,Parent,Ps,NPs)
	).
	
newdefs_forest(NewForest,OldForest,NewDefs) :-
	forest_nodes(OldForest,OldFNs),
	forest_nodes(NewForest,NewFNs),
	delete_elts_sublist(OldFNs,NewFNs,NewDefs).

	

% node_ancestors_forest(Root,Forest,Ancs)
% Ancs is the list of the ancestors of Root 
% in a tree in Forest
node_ancestors_forest(_Root,[],_Ancs) :- fail, !.
node_ancestors_forest(Root,[Tree|_Ts],Ancs) :-
	node_ancestors_tree(Root,Tree,Ancs),!.
node_ancestors_forest(Root,[_Tree|Ts],Ancs) :-
	node_ancestors_forest(Root,Ts,Ancs).

node_ancestors_tree(_Root,[],_Ancs) :- fail, !.
node_ancestors_tree(Root,[Path|_Paths],Ancs) :-
	node_ancestors_path(Root,Path,Ancs), !.
node_ancestors_tree(Root,[_Path|Paths],Ancs) :-
	node_ancestors_tree(Root,Paths,Ancs).

node_ancestors_path(_Root,[],_Ancs) :- fail, !.
node_ancestors_path(Root,[Root|Ancs],Ancs) :- !.
node_ancestors_path(Root,[_L|As],Ancs) :-
	node_ancestors_path(Root,As,Ancs).


reorder_clauses([],[]).
reorder_clauses(InCls,OutCls) :-
	group_clauses(InCls,GInCls),
	flatten(GInCls,OutCls).

%%% group_clauses(CNs,GCNs): CNs is a list of clause numbers
%%% GCNs is a list of lists of clause numbers defining the same predicate
group_clauses([],[]).
group_clauses([CN|U],[ [CN|S1]|Rest]) :-
	clause_hd_bd(CN,H,_B),
	functor(H,F,A),
	pred_def(F,A,U,S1,URest),
	group_clauses(URest,Rest).


pred_def(_F,_A,[],[],[]).
pred_def(F,A,[CN|U],S,Rest) :-
	clause_hd_bd(CN,H,_B),
	( functor(H,F,A) ->
	    S=[CN|S1],
	    Rest1=Rest
	;
	    S=S1,
	    Rest=[CN|Rest1]
	),
	pred_def(F,A,U,S1,Rest1).
