:- use_module(library(lists), [ nth/3, same_length/2 ]).

:- discontiguous is_a_uf/2.
:- discontiguous uf/5.


:- multifile css_param_info/2.
:- discontiguous css_param_info/2.


%%% if the  unfolding tree contains the root definition only
%%% unfold the leftmost atom in the body
uf(_,Cl,[[Cl]],_Solver,AN) :-
	Cl = (_,bd(_,Bd)),
	leftmost_atom(Bd,_Solver,1,AN).


leftmost_atom(Lits,S,Pos,AN) :-
	leftmost_atom(Lits,S,Pos,AN,_).  


leftmost_atom([],_Solver,_Pos,_AN,_) :- fail.
leftmost_atom([Lit|Lits],_Solver,Pos,AN,Atom) :-
	((is_a_negative_literal(Lit,_);
	  is_an_equality(Lit);
	  is_an_inequality(Lit)) ->
	    NextPos is Pos+1,
	    leftmost_atom(Lits,_Solver,NextPos,AN,Atom)
	;
	    AN=Pos,
	    Atom = Lit
	).


is_a_uf(uf_ctl,_).
uf(uf_ctl,Cl,_UTree,_Solver,AN) :-
	Cl = (_H,bd(_C,Bd)),
	leftmost_safe_ctl(Bd,_Solver,1,AN).


leftmost_safe_ctl([],_Solver,_Pos,_AN) :- fail.
leftmost_safe_ctl([Lit|Lits],_Solver,Pos,AN) :-
	(safe_ctl(Lit) ->
	    AN=Pos
	;
	    NextPos is Pos+1,
	    leftmost_safe_ctl(Lits,_Solver,NextPos,AN)
	).

css_param_info(uf_ctl,[display('CTL')]).




is_a_uf(uf_leftmost,_).
uf(uf_leftmost,(_,bd(_,Bd)),_UTree,_Solver,AN) :-
	leftmost_atom(Bd,_Solver,1,AN).
css_param_info(uf_leftmost,[display('Leftmost')]).


is_a_uf(uf_leftmost_determinate,_).
uf(uf_leftmost_determinate,Cl,_UTree,Solver,AN) :-
	Cl = (_H,bd(_C,Bd)),
	leftmost_determinate(Bd,Cl,Solver,1,AN).

leftmost_determinate([],_Cl,_Solver,_Pos,_AN) :- fail.
leftmost_determinate([Lit|Lits],Cl,Solver,Pos,AN) :-
	(is_a_negative_literal(Lit,_) ->
	    NextPos is Pos+1,
	    leftmost_determinate(Lits,Cl,Solver,NextPos,AN)
	;
	    unfold_init(Cl,Pos,UCls),
	    solve_clauses(UCls,linking_vars,Solver,SCls),
	    nonfailed_clauses(SCls,Solver,OutCls),
	    length(OutCls,L),
	    ( (L==0 ; L==1) -> AN=Pos
	    ;
		NextPos is Pos+1,
		leftmost_determinate(Lits,Cl,Solver,NextPos,AN)
	    )
	).
css_param_info(uf_leftmost_determinate,[display('Leftmost determinate')]).




is_a_uf(uf_left_allnonvar_input,_).
uf(uf_left_allnonvar_input,Cl,_UTree,Solver,AN) :-
	Cl = (_H,bd(_C,Bd)),
	left_input_args_cond(Bd,Cl,Solver,forall,nonvar,1,AN).
css_param_info(uf_left_allnonvar_input,[display('Leftmost all nonvar input'),description('Selects the leftmost atom containing non-variable arguments at all input positions.')]).




is_a_uf(uf_left_exnonvar_input,_).
uf(uf_left_exnonvar_input,Cl,_UTree,Solver,AN) :-
	Cl = (_H,bd(_C,Bd)),
	left_input_args_cond(Bd,Cl,Solver,exists,nonvar,1,AN).
css_param_info(uf_left_exnonvar_input,[display('Leftmost exists nonvar input'),description('Selects the leftmost atom containing a non-variable argument at an input position.')]).









left_input_args_cond([],_Cl,_Solver,_Quant,_Cond,_Pos,_AN) :- fail.
left_input_args_cond([Lit|Lits],Cl,Solver,Quant,Cond,Pos,AN) :-
	(is_a_negative_literal(Lit,_) ->
	    NextPos is Pos+1,
	    left_input_args_cond(Lits,Cl,_Solver,Quant,Cond,NextPos,AN)
	;
	    unfold_init(Cl,Pos,UCls),
	    solve_clauses(UCls,linking_vars,Solver,SCls),
	    nonfailed_clauses(SCls,Solver,OutCls),
	    (length(OutCls,0) ->   %%% Cl can be removed by unfolding Lit
		AN=Pos
	    ;
		Lit =.. [_F|Args],
		atom_modes(Lit,Modes),
		
		(input_args_cond(Modes,Args,Quant,Cond)  ->
		    AN=Pos
		;
		    NextPos is Pos+1,
		    left_input_args_cond(Lits,Cl,_Solver,Quant,Cond,NextPos,AN)
		)
	    )
	).
	
	
	
input_args_cond(Modes,Args,forall,Cond) :-
	nth(N,Modes,'+'),
	nth(N,Args,Arg),
	G =.. [Cond,Arg],
	\+ G, !,
	fail.
input_args_cond(_Modes,_Args,forall,_Cond).
input_args_cond(Modes,Args,exists,Cond) :-
	nth(N,Modes,'+'),
	nth(N,Args,Arg),
	G =.. [Cond,Arg],
	G.




%%%is_a_uf(uf_leftmost_determinate_dontsolve,_).
%%%uf(uf_leftmost_determinate_dontsolve,Cl,_UTree,_Solver,AN) :-
%%%	Cl = (H,bd(C,Bd)),
%%%	leftmost_determinate_dontsolve(Bd,Cl,1,AN).


%%%leftmost_determinate_dontsolve([],_Cl,_Pos,_AN) :- fail.
%%%leftmost_determinate_dontsolve([Lit|Lits],Cl,Pos,AN) :-
%%%	unfold_init(Cl,Pos,UCls),
%%%	length(UCls,L),
%%%%	      length(NonFacts,L),
%%%	( (L==0 ; L==1) -> AN=Pos
%%%	      ;
%%%		 NextPos is Pos+1,
%%%		 leftmost_determinate_dontsolve(Lits,Cl,NextPos,AN)
%%%	).
	


is_a_uf(uf_left_consumer,_).
uf(uf_left_consumer,Cl,_UTree,Solver,AN) :-
	Cl = (_H,bd(_C,Bd)),
	left_consumer(Bd,Cl,Solver,1,AN).
css_param_info(uf_left_consumer,[display('Leftmost consumer atom'),description('Selects the leftmost consumer atom.')]).


left_consumer([],_Cl,_Solver,_Pos,_AN) :- fail.
left_consumer([Lit|Lits],Cl,Solver,Pos,AN) :-
	(is_a_negative_literal(Lit,_) ->
	    NextPos is Pos+1,
	    left_consumer(Lits,Cl,Solver,NextPos,AN)
	;
	    get_initial_clauses(InitCls),
	    atom_modes(Lit,Modes),
	    (consumer_atom(InitCls,Lit,Modes,Solver,Cl) ->
		AN=Pos
	    ;
		NextPos is Pos+1,
		left_consumer(Lits,Cl,Solver,NextPos,AN)
	    )
	).

consumer_atom([],_Atom,_Modes,_S,_OrigCl).
consumer_atom([Cl|Cls],Atom,Modes,S,OrigCl) :-
	is_a_constrained_fact(Cl,S),!,
	consumer_atom(Cls,Atom,Modes,S,OrigCl).
consumer_atom([Cl|Cls],Atom,Modes,S,OrigCl) :-
	copy_term(tuple(Cl,Atom,OrigCl),tuple(ClC,AtomC,OrigClC)),
	OrigClC = (_,bd(OC,_)),
	ClC = (H,bd(C,_)),
	( \+ unify_with_occurs_check(AtomC,H) ->
	    true
	;
	    append(OC,C,NewC),
	    \+ satisfiable(S,NewC)
	),
	consumer_atom(Cls,Atom,Modes,S,OrigCl).
consumer_atom([Cl|Cls],Atom,Modes,S,OrigCl) :-
	copy_term(tuple(Cl,Atom,OrigCl),tuple(ClC,AtomC,OrigClC)),
	OrigClC = (_,bd(OC,_)),
	list_to_constraint(S,OC,OC1),
	call(OC1),
	ClC = (H,bd(_C,_)),
%%%	list_to_constraint(S,C,C1),
%%%	call(C1),
	
	copy_term(AtomC,AtomCC),
	unify_with_occurs_check(AtomC,H),

	input_args(AtomC,InArgsC),
	input_args(AtomCC,InArgsCC),
	variant(InArgsC,InArgsCC),
	consumer_atom(Cls,Atom,Modes,S,OrigCl).

