:- use_module(library(lists), [ append/3, member/2 ]).
%:- use_module(library(terms), [ variant/2, unify_with_occurs_check/2 ]).

% unify_with_occurs_check/2 is now built-in
:- use_module(library(terms), [ variant/2]).

unfold(X,Y,Z) :- apply_unfold(X,Y,Z).

%% Give an alert message if clause CN is not in the current program
apply_unfold(error(Alert),CN,_) :- 
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.
	
	
apply_unfold(Out,CN,AN) :- 
   apply_unfold(1,CN,AN,NewCls), !, %default is initial prog
   Out=out(label(unf(CN,AN)), [CN], NewCls). 

apply_unfold(error(Alert),_,_) :-	Alert='Program unchanged.'.

apply_unfold(P,CN,AN,Cls) :-
    clause_hd_bd(CN,Head,bd(Constrs,Body1)),
    split(Body1,AN,Before,Lit,After),
%%%    writeq('Unfolding '), writeq(Lit), nl,
    (is_an_equality(Lit,LHS,RHS) ->
        (unify_with_occurs_check(LHS,RHS),
         append(Before,After,Body),
         Cls = [(Head,bd(Constrs,Body))])
	 ;
		  (is_a_negative_literal(Lit,Atom) ->
				%%% todo: manage constraints
				apply_unfold_neg(P,CN,Head,Before,Atom,After,AN,Cls)
		  ;
				functor(Lit,Fun,Ar),
				apply_unfold_pos(P,CN,AN,Fun/Ar,Cls)
		  )
	 ).

apply_unfold_pos(P,CN,AN,Fun/Ar,Cls) :-
	(P==1 ->
%%% we cache only for the clauses in the initial program for now,
%%% since it's the only being used for unfolding in the strategy.
%%% If you want to extend it, you'll have to manage program undos.
		 (call(predicate_program_clause(Fun/Ar,P,_)) ->
			  true
		 ;
			  program_clauses(P,PCls),
			  cache_predicate_program_clause(P,PCls)
		 ),
		 setof(FunCN,(predicate_program_clause(Fun/Ar,P,FunCN)),UnfCNs)
	  ;
		 program_clauses_defs(P,UnfCNs)
	),
	bagof((Hd,Bd), one_unf(UnfCNs,CN,AN,(Hd,Bd)), Cls), !.

apply_unfold_pos(P,CN, AN,F/A, Cls) :-
	( is_defined_in_prog(F/A,P) ->  
		 Cls=[]
	;
	    (is_a_bip(F/A) ->
			  apply_unfold_bip(CN, AN, Cls)
	    ;	
			  (tcl_ask_clause_deletion_by_unfolding(F/A,CN,Answer),
					(Answer=="yes" -> Cls=[] ; fail )
			  )
	    )
	).
	    
one_unf(Cls_P,CN,AN,(Head,bd(Constrs,Bd))) :-
    clause_hd_bd(CN,Head,bd(C1,Body1)),
    split(Body1,AN,Before,Atom,After),
    member(C,Cls_P),
    clause_hd_bd(C,Atom2,bd(C2,Bd2)),
    unify_with_occurs_check(Atom2,Atom),
	 append(C1,C2,Constrs),
    append(Before,Bd2,After,Bd).


get_initial_clauses(InitCls) :-
	(initial_clauses(InitCls) ->
	    true
	;
	    program_clauses(1,InitCNs),
	    clause_hd_bd_list(InitCNs,InitCls),
	    assert(initial_clauses(InitCls))
	).


unfold_init((Head,bd(C1,Body1)),AN,Cls) :-
	get_initial_clauses(InitCls),
	
%%%	(bagof(Cl1,one_unf_(InitCls,Cl,AN,Cl1),Cls) ->
%%%		 true
%%%	;
%%%		 Cls = []
%%%	).
	split(Body1,AN,Before,Atom,After),
	unfold_init_(InitCls,cl(Head,C1,Before,Atom,After),Cls).


unfold_init_([],_,[]).
unfold_init_([(Atom2,bd(C2,Bd2))|PCls],Cl,UCls) :-
	copy_term(Cl,cl(Head,C1,Before,Atom,After)),
	(unify_with_occurs_check(Atom2,Atom) ->
	    append(C1,C2,Constrs),
	    append(Before,Bd2,After,Bd),
	    UCls=[(Head,bd(Constrs,Bd))|RUCls]
	    ;
	    UCls=RUCls
	),
	unfold_init_(PCls,Cl,RUCls).

one_unf_(Cls,(Head,bd(C1,Body1)),AN,(Head,bd(Constrs,Bd))) :-
    split(Body1,AN,Before,Atom,After),
    member(Cl2,Cls),
    copy_term(Cl2,(Atom2,bd(C2,Bd2))),
    unify_with_occurs_check(Atom2,Atom),
    append(C1,C2,Constrs),
    append(Before,Bd2,After,Bd).


%TO DO



apply_unfold_bip(_,_,_) :-
	tcl_warning_ok('{Unfolding rule for a bip is not yet implemented.}'), fail.
apply_unfold_neg(_,_,_,_,_,_,_,_) :-
	tcl_warning_ok('{Unfolding rule for a negative lit is not yet implemented.}'), fail.
