# display.tcl


proc positionWindow w {
    wm geometry $w +300+300
}

proc setDefaultFonts {} {
    global font  clause_font history_font
    set w .defaultfonts
    if [winfo exists  $w] { wm deiconify $w ; raise $w ; return }    
    catch {destroy $w}
    toplevel $w
    wm title $w "Fonts"
    positionWindow $w
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 5 -padx 5
    button $w.buttons.dismiss -text Close -foreground black \
		  -command "wm withdraw $w"
    button $w.buttons.save -text Save -foreground black \
		  -command {
				saveNewSizeFonts $size
				wm withdraw .defaultfonts
				return}
    pack  $w.buttons.save $w.buttons.dismiss -side left -expand 1
    frame $w.left
    pack $w.left -pady 5 -padx 5 -side left -expand 1
    foreach i {10 12 14 16 18 20} {
		  radiobutton $w.left.c$i -text "$i"  -foreground black \
				-font "Courier $i" \
				-variable size -relief flat -value $i
		  pack $w.left.c$i  -side top -pady 2 -anchor w
	 }   
    tkButtonInvoke $w.left.c12
}

proc print {} {
    set w .print
    if [winfo exists  $w] { wm deiconify $w ; raise $w ; return }    
    catch {destroy $w}
    toplevel $w
    wm title $w "MAP: Print"
    positionWindow $w
    frame $w.cmd 
    label $w.cmd.msg -wraplength 2i -foreground black  -justify left \
	    -anchor w -text "Print command:"
    entry $w.cmd.e1 -relief sunken -foreground black \
	    -textvariable printcommand
    $w.cmd.e1 insert end lpr
    pack $w.cmd -pady 15 -padx 5
    pack $w.cmd.msg $w.cmd.e1 -side left -pady 5 -padx 5
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 5m
    button $w.buttons.dismiss -foreground black \
	    -text Close -command "wm withdraw $w"
    button $w.buttons.save -foreground black -text Print -command {
	prolog apply_print('$printcommand',$what)}
    pack $w.buttons.dismiss $w.buttons.save -side left -expand 1
    frame $w.left
    pack $w.left -side left -expand 1 -pady 5 -padx 5
    radiobutton $w.left.prog -foreground black -text Program \
		-variable what -relief flat -value program
    radiobutton $w.left.init_defs -foreground black -text "Init + Defs" \
		-variable what -relief flat -value init_defs
    radiobutton $w.left.theory -foreground black -text Theory \
		-variable what -relief flat -value theory
    radiobutton $w.left.all -foreground black \
	    -text "Theory + Init + Defs + Program" \
		-variable what -relief flat -value all
    radiobutton $w.left.history -foreground black -text History \
		-variable what -relief flat -value history
    pack $w.left.prog $w.left.init_defs $w.left.theory $w.left.all \
	    $w.left.history -side top -pady 2 -anchor w   
    tkButtonInvoke $w.left.prog
}

proc saveNewSizeFonts {size} {
    global fontcl fonthistory
    if [null $size]  {return}
    set fontcl courier$size 
    set fonthistory bold$size 
}

proc saveNewFonts {clause_font history_font} {
    global fontcl fonthistory 
    puts stdout "$clause_font $history_font"
    if {![null $clause_font]}  {set fontcl $clause_font}
    if {![null $history_font]}  {set fonthistory $history_font}
}

proc enableActions {} {
    enablePrinting
    enableTopButtons
}

proc enablePrinting {} {
    .menuBar.file entryconfigure 2 -state normal
}

proc enableTopButtons {} {
    set m .top.session.m
# 6 is a separator
    foreach i {3 4 5 7 8} {
	$m entryconfigure $i -state normal }
    }


proc clearProofText {} {
    global ProofText
    $ProofText configure -state normal  
    $ProofText delete 1.0 end
    $ProofText configure -state disabled
    
}

proc clearAllSel  {} {
    clearProofSelText
    clearDefSelText
    clearLawSelText
#    clearSolverSelText
}

proc clearProofSelText  {} {
    global  ProofText proofSelection Position tagsSelection white
    foreach sel $tagsSelection {
	$ProofText tag  configure $sel  -background $white -foreground black
    }
    set proofSelection ""
    set Position ""
    set tagsSelection ""
}

proc clearDefSelText {} {
    global defSelection deftagsSelection white
    set w .currentInit_Defs.current.text
    foreach sel $deftagsSelection {
	$w tag configure $sel  -background $white -foreground black
    }
    set defSelection ""
    set deftagsSelection ""
}

proc clearLawSelText {} {
    global lawSelection white
    set w .currentTheory.current.text
    foreach sel $lawSelection {
	$w tag  configure law.$sel  -background $white -foreground black }
    set lawSelection ""
}

# proc clearSolverSelText {} {
#     global solverSelection white
#     set w .wSolver.current.text
#     foreach sel $solverSelection {
# 	$w tag  configure solver.$sel  -background $white -foreground black }
#     set solverSelection ""
# }

proc removeTagsCls {nums numsfather} {
    global ProofText fontclnum fontcl
    set w $ProofText
    foreach num $numsfather {
		  $w tag configure clausenum.$num -font $fontclnum
    }
    foreach num $nums {
		  $w tag configure clausenum.$num -font $fontcl
		  prolog "length_body_clause($num,L)"
		  set length $prolog_variables(L)
		  set i 0
        while { $i <= $length} {
				$w tag remove atom.$num.$i 1.0 end
				incr i 
		  } 
	 }
}


proc removeTagsClNums {nums} {
    global ProofText fontcl
    set w $ProofText
    foreach num $nums { 
		  $w tag configure clausenum.$num -font $fontcl
	 }
}

proc markUnfoldedTag {cn an} {
    global ProofText tagsSelection
    set w $ProofText
    $w tag configure $tagsSelection -underline on
}

proc markHotHd {numcl} {
    markHotAtom $numcl 0
}

proc markHotClause {numcl} {
    global ProofText proofSelection white
    prolog "length_body_clause($numcl,L)"
    set length $prolog_variables(L)
    set i 0
    lapp_proof_sel "clause $numcl"
    while { $i <= $length} {
		  lapp_tags_sel atom.$numcl.$i
		  $ProofText tag configure atom.$numcl.$i \
				-background black -foreground $white
		  incr i
	 }
    puts stdout "proofSelection=$proofSelection" 
}

proc markHotDef {numcl} {
    global defSelection white
    set w .currentInit_Defs.current.text
    lapp_defs_sel "clause $numcl"
    lapp_tags_defs_sel def.$numcl
    $w tag configure def.$numcl \
		-background black -foreground $white
    puts stdout "defSelection=$defSelection" 
}

proc markHotLaw {numlaw} {
    global lawSelection white 
    lappend lawSelection $numlaw
    .currentTheory.current.text tag configure law.$numlaw \
		  -background black -foreground $white
    puts stdout "lawSelection=$lawSelection" 
}

proc markHotAtom {numcl posatom} {
    global proofSelection  ProofText white
    lapp_tags_sel atom.$numcl.$posatom
    lapp_proof_sel "atom $numcl $posatom"
    puts stdout "proofSelection=$proofSelection" 
    $ProofText tag configure atom.$numcl.$posatom \
		  -background black -foreground $white
}

proc markHotBody {numcl} {
    global proofSelection  ProofText white
    prolog "length_body_clause($numcl,L)"
    set length $prolog_variables(L)
    set i 1
    while {$i <= $length} {
		  lapp_tags_sel atom.$numcl.$i
		  lapp_proof_sel "atom $numcl $i"
		  $ProofText tag configure atom.$numcl.$i \
				-background black -foreground $white
		  incr i
	 }
    puts stdout "proofSelection=$proofSelection" 
}


# proc markHotSolver {solver} {
#     global solverSelection white 
#     lappend solverSelection $solver
#     .wSolver.current.text tag configure solver.$solver \
# 	    -background black -foreground $white
#     puts stdout "solverSelection=$solverSelection" 
# }





proc emptyproofSelection {} {
    global proofSelection Position
    set proofSelection ""
    set Position ""
}
proc emptyDefSelections {} {
    global  defSelection
    set defSelection ""
}

proc clear_alertnote {} {
    infonote "  "
}

proc clear_css_alertnote {} {
    css_infonote "  "
}

proc clear_ds_alertnote {} {
    ds_infonote "  "
}


proc alertnote {msg} {
    global font
    bell
    catch {destroy .msg}
    message .msg  -font $font -width 20c \
	    -background #ffffcc -foreground black -relief raised -text $msg
    pack .msg -side bottom -pady 10
    } 

proc infonote {msg} {
 infonote2 $msg .msg
}


proc css_infonote {msg} {
 infonote2 $msg .css_param.body.msg
} 

proc ds_infonote {msg} {
 infonote2 $msg .ds_param.body.msg
} 

proc infonote2 {msg w} {
    catch {destroy $w}
    message $w -width 20c -background #ffffcc -foreground black\
	    -relief raised -text $msg
    pack  $w -side bottom -pady 10
} 


proc describe_button_theories {} {
    infonote "Theories menu for modifying the current program."
}
proc describe_button_strategies {} {
    infonote "Strategies menu for applying a strategy on the current program."
}
proc describe_button_rules {} {
    infonote "Transformation rules menu for applying a rule on the current program."
}
proc describe_button_help {} {
    infonote "Open and close the Help window."
}
proc describe_button_quit {} {
    infonote "Exit the system."
}

proc describe_button_close {} {
    infonote "Close window."
}

proc describe_button_edit {} {
    infonote "Edit the selected file."
}
proc describe_button_load_prog {} {
    infonote "Start a new derivation with the selected file as initial program."
}
proc describe_button_load_theory {} {
    infonote "Load the selected theory, i.e. add its laws to the current theory."
}
proc describe_button_load_session {} {
    infonote "Restore a derivation constructed in a previous (saved) session."
}
proc describe_button_load_solver {} {
    infonote "Load the selected solver."
}
proc describe_button_load_unfoldfunc {} {
    infonote "Load the selected unfolding function."
}
proc describe_button_load_wqounf {} {
    infonote "Load the selected well-quasi-order for unfolding."
}
proc describe_button_load_wqogen {} {
    infonote "Load the selected well-quasi-order for generalization."
}
proc describe_button_load_widening {} {
    infonote "Load the selected widening."
}

#proc msg_warning_yesno {msg} {
#    tk_messageBox -icon warning -type yesno -message $msg 
#}

proc msg_warning_yesno {msg} {
    return [tk_messageBox -icon question \
	    -type yesno -message $msg]
}

proc msg_warning_okcancel {msg} {
    return [tk_messageBox -icon warning \
	    -type okcancel -message $msg]
}

proc msg_warning_ok {msg} {
    tk_messageBox -icon warning -type ok  -message $msg
}

proc todo {} {
    tk_messageBox -icon warning  -type ok  -message "Not yet available"
}

proc prologfailed {} {
    alertnote "Prolog goal failed."
}

proc msg_error_def_not_same_pred {} {
    tk_messageBox -icon error  \
	    -type ok -message "Definition clauses must use the same predicate in their heads."
}


proc msg_error_pred_exists {pred arity program} {
    tk_messageBox -icon error  \
	    -type ok -message "\"$pred/ $arity\" is already used.\
Choose another predicate name."
}

proc loadMenuRules {} {
    global R_or_S font
    if {$R_or_S=="Rules"} {return}
    set R_or_S "Rules"
    .middle.control.buttons delete 0 end
    prolog showRules
}

proc loadMenuStrategies {} {
    global R_or_S font
    if {$R_or_S=="Strategies"} {return}
    set R_or_S "Strategies"
    prolog showStrategies
}

# proc loadMenuSolvers {} {
#     global R_or_S font
#     if {$R_or_S=="Solvers"} {return}
#     set R_or_S "Solvers"
#     prolog showSolvers  # in interface.pl
# }


proc showRS {RorS L_tcl } {
    global font
    set w .middle.control.buttons
    $w configure -state normal
    $w delete 1.0 end
    foreach i $L_tcl {
	prolog "button_name_rule(Name,$i) "
	set text $prolog_variables(Name)
	button $w.$i -text $text -width 10\
	    -height 1 -foreground black -font $font -cursor top_left_arrow 
	bind $w.$i <1> "apply_transformation_rs $RorS $i"
	$w window create end -window $w.$i -padx 3 -pady 2
    }
    $w configure -state disabled
}

proc showRSHelp {L_tcl } {
 global font
 set w .help
 if {[winfo depth $w] > 1} {
  set bold "-background pink -foreground black -relief raised \
	    -borderwidth 1"
  set normal "-background {} -foreground black -relief flat" } else {
   set bold "-foreground $white -background black"
   set normal "-foreground {} -background {}" }
 $w.mess configure -state normal
 foreach i $L_tcl {
  prolog "button_name_rule(Name,$i) "
  set text $prolog_variables(Name)
  $w.mess insert end $text d$i
  $w.mess insert end {  }
  $w.mess tag configure d$i  -underline on 
  $w.mess tag bind d$i <Any-Enter> "$w.mess tag configure d$i $bold"
  $w.mess tag bind d$i <Any-Leave> "$w.mess tag configure d$i $normal"
  $w.mess tag bind d$i <1> "helpRule $i"
 }
 $w.mess insert end \n\n\n
 $w.mess configure -state disabled

}





# How to put a window from background to foreground ?
# instead of doing withdraw then deiconify...
# answer: raise! Does not work if window is closed

proc show {what} {
    global dirNameProg dirNameTheories dirNameSessions 
#    global dirNameSolvers dirNameUnfoldFuncs dirNameWqounfs 
#    global dirNameWqogens dirNameWidenings

	 set dir "/"

	 if {$what== "program"} {set dir $dirNameProg} 
	 if {$what== "theory"} {set dir $dirNameTheories} 
	 if {$what== "session"} {set dir $dirNameSessions} 
# 	 if {$what== "solver"} {set dir  dirNameSolvers} 
# 	 if {$what== "unfoldfunc"} {set dir dirNameUnfoldFuncs} 
# 	 if {$what== "wqounf"} {set dir dirNameWqounfs} 
# 	 if {$what== "wqogen"} {set dir dirNameWqogens} 
# 	 if {$what== "widening"} {set dir dirNameWidenings} 


 set types {
		  {"Prolog files"		{.pl}	}
		  {"Text files"		{}		TEXT}
		  {"All files"		*}
	 }


	 set file [tk_getOpenFile -filetypes $types -initialdir $dir]
	 if {[string compare $file ""] == 0} { return -code return }
	 
	 if { $what== "program" || $what== "session" } {
		  if { [prolog clause(current_program(_,_),_)] == 1} {
				if {[ msg_warning_yesno "The current derivation will be lost if not saved.\n\nContinue start a new derivation ?"] == "no"} { return }}
		  clearProofText 
		  if { $what== "program" } {
				if { [prolog "apply_start('$file')"] == 1} {
					 destroy4
					 enableActions } 
		  }
		  if { $what== "session" } {
				if { [prolog "apply_load_session('$file')"] == 1} {
					 destroy4
					 enableActions } 
		  }
	 }
	 if { $what== "theory" } then {
		  prolog "apply_add_theory('$file')"
	 }

}

proc showPrograms {} {
	 show program
}


# proc showPrograms {} {
#     global dirNameProg font
#     set w .fprograms
#     if [winfo exists  $w] {  
# 	wm deiconify $w
# 	raise $w 
# 	$w.dirNameProg delete 0 end 
# 	$w.dirNameProg insert end $dirNameProg 
# 	tkEntrySetCursor $w.dirNameProg end
# 	pack $w.dirNameProg   -pady 0 -anchor w 
# 	puts stdout $dirNameProg
# 	$w.f.list delete 0 end
# 	foreach i [exec ls -a $dirNameProg] { $w.f.list insert end $i } 
# 	bind $w.dirNameProg <Return> { 
# 	    set dirNameProg $dirEntryLabel  
# 	    puts stdout $dirEntryLabel
# 	    showPrograms}
# 	    bind_double_1_box $w program
#     return }
#     catch {destroy .fprograms}
#     toplevel $w 
#     wm title $w "Programs"
#     positionWindow $w
#     label $w.fileLabel -text "Files:"  -font $font -foreground black 
#     label $w.dirLabel -text "Directory:"  -font $font -foreground black 
#     entry $w.dirNameProg -width 30 -textvariable dirEntryLabel \
# 	     -foreground black -font $font -foreground black 
#     $w.dirNameProg insert end $dirNameProg
#     tkEntrySetCursor $w.dirNameProg end
#     frame $w.f
#     frame $w.bot
#     button $w.bot.cancel -text Close -font $font -foreground black \
# 	    -command "wm withdraw $w"
#     bind $w.bot.cancel <Enter> describe_button_close
#     bind $w.bot.cancel <Leave> clear_alertnote
#     listbox $w.f.list -font $font -foreground black -width 25 -height 15 \
# 	    -yscrollcommand "$w.f.scroll set"
#     scrollbar $w.f.scroll -command "$w.f.list yview"
#     foreach i [exec ls -a $dirNameProg] { $w.f.list insert end $i }
#     pack $w.fileLabel -side top -pady 20 -anchor w
#     pack $w.f -padx 10 -pady 5
#     pack $w.f.list $w.f.scroll -side left -fill y -padx 1 -expand 1
#     pack $w.dirLabel $w.dirNameProg -pady 10 -anchor w
#     pack $w.dirNameProg  -padx 20 -pady 5 -anchor w 
#     pack $w.bot -side bottom -pady 20
#     pack $w.bot.cancel -side right -padx 10 -anchor w
#     showPrograms

# }


proc showTheories {} {
    global dirNameTheories font
    set w .ftheories
    if [winfo exists  $w] { 
	wm deiconify $w
	raise $w 
	$w.dirNameTheories delete 0 end 
	$w.dirNameTheories insert end $dirNameTheories
	tkEntrySetCursor $w.dirNameTheories end
	pack $w.dirNameTheories   -pady 0 -anchor w 
	puts stdout $dirNameTheories
	$w.f.list delete 0 end
	foreach i [exec ls -a $dirNameTheories] { $w.f.list insert end $i } 
	bind $w.dirNameTheories <Return> { 
	    set dirNameTheories $dirEntryLabel  
	    puts stdout $dirEntryLabel
	    showTheories}
	    bind_double_1_box $w theory
	return }
    catch {destroy .ftheories}
    toplevel $w 
    wm title $w "Theories"
    positionWindow $w
    label $w.fileLabel -text "Files:"  -font $font -foreground black 
    label $w.dirLabel -text "Directory:"  -font $font -foreground black 
    entry $w.dirNameTheories -width 30 -textvariable dirEntryLabel \
	    -font $font -foreground black 
    $w.dirNameTheories insert end $dirNameTheories
    tkEntrySetCursor $w.dirNameTheories end
    frame $w.f
    frame $w.bot
    button $w.bot.cancel -text Close -font $font -foreground black \
	    -command "wm withdraw $w"
    bind $w.bot.cancel <Enter> describe_button_close
    bind $w.bot.cancel <Leave> clear_alertnote
    listbox $w.f.list -font $font -foreground black -width 25 -height 15 \
	    -yscrollcommand "$w.f.scroll set"
    scrollbar $w.f.scroll -command "$w.f.list yview"
    foreach i [exec ls -a $dirNameTheories] { $w.f.list insert end $i }
    pack $w.fileLabel -side top -pady 20 -anchor w
    pack $w.f -padx 10 -pady 5
    pack $w.f.list $w.f.scroll -side left -fill y -padx 1 -expand 1
    pack $w.dirLabel $w.dirNameTheories -pady 10 -anchor w
    pack $w.dirNameTheories  -padx 20 -pady 5 -anchor w 
    pack $w.bot -side bottom -pady 20
    pack $w.bot.cancel -side right -padx 10 -anchor w
    showTheories

}

proc showSessions {} {
 show session
}

# proc showSessions {} {
#     global dirNameSessions font
#     set w .fsessions
#     if [winfo exists  $w] { 
# 	wm deiconify $w
# 	raise $w 
# 	$w.dirNameSessions delete 0 end 
# 	$w.dirNameSessions insert end $dirNameSessions
# 	tkEntrySetCursor $w.dirNameSessions end
# 	pack $w.dirNameSessions   -pady 0 -anchor w 
# 	puts stdout $dirNameSessions
# 	$w.f.list delete 0 end
# 	foreach i [exec ls -a $dirNameSessions] { $w.f.list insert end $i } 
# 	bind $w.dirNameSessions <Return> { 
# 	    set dirNameSessions $dirEntryLabel  
# 	    puts stdout $dirEntryLabel
# 	    showSessions}
# 	    bind_double_1_box $w session
# 	return }
#     catch {destroy .fsessions}
#     toplevel $w 
#     wm title $w "Sessions"
#     positionWindow $w
#     label $w.fileLabel -text "Files:"  -font $font -foreground black 
#     label $w.dirLabel -text "Directory:"  -font $font -foreground black 
#     entry $w.dirNameSessions -width 30 -textvariable dirEntryLabel \
# 	    -font $font -foreground black 
#     $w.dirNameSessions insert end $dirNameSessions
#     tkEntrySetCursor $w.dirNameSessions end
#     frame $w.f
#     frame $w.bot
#     button $w.bot.cancel -text Close -font $font \
# 	    -foreground black -command "wm withdraw $w"
#     bind $w.bot.cancel <Enter> describe_button_close
#     bind $w.bot.cancel <Leave> clear_alertnote
#     listbox $w.f.list -font $font -foreground black -width 25 -height 15 \
# 	    -yscrollcommand "$w.f.scroll set"
#     scrollbar $w.f.scroll -command "$w.f.list yview"
#     foreach i [exec ls -a $dirNameSessions] { $w.f.list insert end $i }
#     pack $w.fileLabel -side top -pady 20 -anchor w
#     pack $w.f -padx 10 -pady 5
#     pack $w.f.list $w.f.scroll -side left -fill y -padx 1 -expand 1
#     pack $w.dirLabel $w.dirNameSessions -pady 10 -anchor w
#     pack $w.dirNameSessions  -padx 20 -pady 5 -anchor w 
#     pack $w.bot -side bottom -pady 20
#     pack $w.bot.cancel -side right -padx 10 -anchor w
#     showSessions

# }



proc bind_double_1_box {w what} {
    global dirNameProg dirNameTheories dirNameSessions 
    global dirNameSolvers dirNameUnfoldFuncs dirNameWqounfs 
    global dirNameWqogens dirNameWidenings
    if {$what=="program"} { 
		  bind $w.f.list <Double-1> {
				foreach i [selection get] {browse program $dirNameProg $i} } }
    if {$what=="theory"} { 
		  bind $w.f.list <Double-1> {
				foreach i [selection get] {browse theory $dirNameTheories $i} } }
    if {$what=="session"} { 
		  bind $w.f.list <Double-1> {
				foreach i [selection get] {browse session $dirNameSessions $i} } }
    if {$what=="solver"} { 
		  bind $w.f.list <Double-1> {
				foreach i [selection get] {browse solver $dirNameSolvers $i} } }
    if {$what=="unfoldfunc"} { 
		  bind $w.f.list <Double-1> {
				foreach i [selection get] {browse unfoldfunc $dirNameUnfoldFuncs $i} } }
    if {$what=="wqounf"} { 
		  bind $w.f.list <Double-1> {
				foreach i [selection get] {browse wqounf $dirNameWqounfs $i} } }
    if {$what=="wqogen"} { 
		  bind $w.f.list <Double-1> {
				foreach i [selection get] {browse wqogen $dirNameWqogens $i} } }
    if {$what=="widening"} { 
		  bind $w.f.list <Double-1> {
				foreach i [selection get] {browse widening $dirNameWidenings $i} } }
}



proc browse {what dir file} {
    global dirNameProg  dirNameTheories dirNameSessions 
    global dirNameSolvers dirNameUnfoldFuncs dirNameWqounfs 
    global dirNameWqogens dirNameWidenings

    if {[string compare $dir "."] != 0} {set file $dir/$file}
    if [file isdirectory $file] {
	if {$what== "program"} {set dirNameProg $file ; showPrograms} 
	if {$what== "theory"} {set dirNameTheories $file ; showTheories} 
	if {$what== "session"} {set dirNameSessions $file ; showSessions} 
	if {$what== "solver"} {set dirNameSolvers $file ; showSolvers} 
	if {$what== "unfoldfunc"} {set dirNameUnfoldFuncs $file ; showUnfoldFuncs} 
	if {$what== "wqounf"} {set dirNameWqounfs $file ; showWqounfs} 
	if {$what== "wqogen"} {set dirNameWqogens $file ; showWqogens} 
	if {$what== "widening"} {set dirNameWidenings $file ; showWidenings} 
    } else {
        if [file isfile $file]  { 
	    if {$what== "program"} { edit_or_load program }
	    if {$what== "theory"} { edit_or_load theory }
	    if {$what== "session"} { edit_or_load session }
	    if {$what== "solver"} { edit_or_load solver }
	    if {$what== "unfoldfunc"} { edit_or_load unfoldfunc }
	    if {$what== "wqounf"} { edit_or_load wqounf }
	    if {$what== "wqogen"} { edit_or_load wqogen }
	    if {$what== "widening"} { edit_or_load widening }
        }   else {
            msg_warning_ok "\"$file\" isn't a directory or regular file"
        }
    }
}

proc edit_or_load { what }	{
    global font
    if {$what=="program"} { set w .fprograms }
    if {$what=="theory"} { set w .ftheories }
    if {$what=="session"} { set w .fsessions }
    if {$what=="solver"} { set w .fsolvers }
    if {$what=="unfoldfunc"} { set w .funfoldfuncs }
    if {$what=="wqounf"} { set w .fwqounfs }
    if {$what=="wqogen"} { set w .fwqogens }
    if {$what=="widening"} { set w .fwidenings }
    if [winfo exists $w.bot.load] { return }
    if {$what=="program"} {
 	button $w.bot.edit -text Edit -font $font \
		-foreground black -command { edit program }
	button $w.bot.load -text Load -font $font \
		-foreground black -command { load program }
	bind $w.bot.load <Enter> "describe_button_load_prog"}
    if {$what=="session"} { 
	    button $w.bot.load -text Load -font $font \
		    -foreground black -command { load session }
	    bind $w.bot.load  <Enter> "describe_button_load_session"}
    if {$what=="solver"} { 
	    button $w.bot.edit -text Edit -font $font \
		    -foreground black -command { edit solver }
	    button $w.bot.load -text Load -font $font \
		    -foreground black -command { load solver }
	    bind $w.bot.load  <Enter> "describe_button_load_solver"}
    if {$what=="unfoldfunc"} { 
	    button $w.bot.edit -text Edit -font $font \
		    -foreground black -command { edit unfoldfunc }
	    button $w.bot.load -text Load -font $font \
		    -foreground black -command { load unfoldfunc }
	    bind $w.bot.load  <Enter> "describe_button_load_unfoldfunc"}
    if {$what=="wqounf"} { 
	    button $w.bot.edit -text Edit -font $font \
		    -foreground black -command { edit wqounf }
	    button $w.bot.load -text Load -font $font \
		    -foreground black -command { load wqounf }
	    bind $w.bot.load  <Enter> "describe_button_load_wqounf"}
    if {$what=="wqogen"} { 
	    button $w.bot.edit -text Edit -font $font \
		    -foreground black -command { edit wqogen }
	    button $w.bot.load -text Load -font $font \
		    -foreground black -command { load wqogen }
	    bind $w.bot.load  <Enter> "describe_button_load_wqogen"}
    if {$what=="widening"} { 
	    button $w.bot.edit -text Edit -font $font \
		    -foreground black -command { edit widening }
	    button $w.bot.load -text Load -font $font \
		    -foreground black -command { load widening }
	    bind $w.bot.load  <Enter> "describe_button_load_widening"}

    if {$what!="session"} {
	bind $w.bot.edit <Enter> "describe_button_edit"
	pack $w.bot.edit -side right -padx 10 -pady 0}
    bind $w.bot.load  <Leave> clear_alertnote
    pack $w.bot.load -side left -padx 10 -pady 0
}


proc edit { what } {
    global env dirNameProg dirNameTheories dirNameSessions
    global dirNameSolvers dirNameUnfoldFuncs dirNameWqounfs 
    global dirNameWqogens dirNameWidenings

    puts stdout "here4="
    puts stdout "$what"
    if {$what=="program"} {set dir $dirNameProg ; set w .fprograms.f.list}
    if {$what=="theory"} {set dir $dirNameTheories ; set w .ftheories.f.list}
    if {$what=="session"} {set dir $dirNameSessions ; set w .fsessions.f.list}
    if {$what=="solver"} {set dir $dirNameSolvers ; set w .fsolvers.f.list}
    if {$what== "unfoldfunc"} {set dir $dirNameUnfoldFuncs ; set  w .funfoldfuncs.f.list} 
    if {$what== "wqounf"} {set dir $dirNameWqounfs  ; set  w .fwqounfs.f.list} 
    if {$what== "wqogen"} {set dir $dirNameWqogens  ; set  w .fwqogens.f.list} 
    if {$what== "widening"} {set dir $dirNameWidenings  ; set  w .fwidenings.f.list} 

#    clear_alertnote
# in case the selection in the box has disappeared (in editing)
    tkListboxBeginSelect $w [$w index active]
    set i [selection get] 
    if [file isfile $dir/$i]  { } {msg_warning_ok "\"$dir/$i\" isn't a directory or regular file" ; return}
    if [info exists env(EDITOR)] {
	eval exec $env(EDITOR)  $dir/$i &
    } else { 
	exec xedit $i &
    }
}

proc load { what } {
    global dirNameProg dirNameTheories dirNameSessions
    global dirNameSolvers dirNameUnfoldFuncs dirNameWqounfs 
    global dirNameWqogens dirNameWidenings

    if {$what=="program"} { set w .fprograms.f.list}
    if {$what=="theory"} {set w .ftheories.f.list}
    if {$what=="session"} {set w .fsessions.f.list}
    if {$what=="solver"} {set w .fsolvers.f.list}
    if {$what=="unfoldfunc"} {set w .funfoldfuncs.f.list}
    if {$what=="wqounf"} {set w .fwqounfs.f.list}
    if {$what=="wqogen"} {set w .fwqogens.f.list}
    if {$what=="widening"} {set w .fwidenings.f.list}
# in case the selection in the box has disappeared (in editing)
    tkListboxBeginSelect $w [$w index active]
    set i [selection get] 
    if {$what=="program"} { 
	if [file isfile $dirNameProg/$i]  { } {
	    msg_warning_ok "\"$dirNameProg/$i\" isn't a directory or regular file" ; return}
	if { [prolog clause(current_program(_,_),_)] == 1} {
	    if {[ msg_warning_yesno "The current derivation will be lost if not saved.\n\nContinue start a new derivation ?"] == "no"} { return }}
  	clearProofText 
	if { [prolog apply_start('$dirNameProg/$i')] == 1} {
	    destroy4
	    enableActions } }
    if {$what=="theory"} {
	if [file isfile $dirNameTheories/$i]  { } {
	    msg_warning_ok "\"$dirNameTheories/$i\" isn't a directory or regular file" 
	    return
	}
	prolog "apply_add_theory('$dirNameTheories/$i')"
    }
    if {$what=="session"} {
	if [file isfile $dirNameSessions/$i]  { } {
	    msg_warning_ok "\"$dirNameSessions/$i\" isn't a directory or regular file" 
	    return
	}
	if { [prolog clause(current_program(_,_),_)] == 1} {
	    if {[ msg_warning_yesno "Be sure you have saved your work before starting a new derivation.\n\nContinue start a new derivation ?"] == "no"} { return }}
	clearProofText 
	if { [prolog apply_load_session('$dirNameSessions/$i')]==1 } {
	    destroy4
	    enableActions } 
    }
    if {$what=="solver"} {
	if [file isfile $dirNameSolvers/$i]  { } {
	    msg_warning_ok "\"$dirNameSolvers/$i\" isn't a directory or regular file" 
	    return
	}
	prolog "consult(\'$dirNameSolvers/$i\')"
    }
    if {$what== "unfoldfunc"} {
	if [file isfile $dirNameUnfoldFuncs/$i]  { } {
	    msg_warning_ok "\"$dirNameUnfoldFuncs/$i\" isn't a directory or regular file" 
	    return
	}
	prolog "consult(\'$dirNameUnfoldFuncs/$i\')"
    } 
    if {$what== "wqounf"} {
	if [file isfile $dirNameWqounfs/$i]  { } {
	    msg_warning_ok "\"$dirNameWqounfs/$i\" isn't a directory or regular file" 
	    return
	}
	prolog "consult(\'$dirNameWqounfs/$i\')"
    }
    if {$what== "wqogen"} {
	if [file isfile $dirNameWqogens/$i]  { } {
	    msg_warning_ok "\"$dirNameWqogens/$i\" isn't a directory or regular file" 
	    return
	}
	prolog "consult(\'$dirNameWqogens/$i\')"
    }
    if {$what== "widening"} { 
	if [file isfile $dirNameWidenings/$i]  { } {
	    msg_warning_ok "\"$dirNameWidenings/$i\" isn't a directory or regular file" 
	    return
	}
	prolog "consult(\'$dirNameWidenings/$i\')"
    }

}


proc addProofTextClause2 {num hd bd} {
    global ProofText Position fontclnum
    global fontcl
	 
    set real_fontcl {Courier 14}
#    set real_fontcl $fontcl
#	 puts stdout $fontcl

    set MaxPos [max_pos]
    set w $ProofText
    clear_alertnote
    $w configure -state normal
    $w insert end "$num.\ " 
    set char [$w index insert]
    set begincl $char
    set pos0 [$w index "insert linestart"]
    $w tag add clausenum.$num $pos0 $char
    $w tag configure clausenum.$num -font $fontclnum
    $w insert end "$hd" $real_fontcl
    $w tag add atom.$num.0 $char [$w index insert]
    $w tag configure atom.$num.0 -font $real_fontcl 
    $w tag bind atom.$num.0 <1> \
	    "clearProofSelText ; markHotClause $num"
    $w tag bind atom.$num.0 <Control-1> \
	    "markHotClause $num"
    $w tag bind atom.$num.0 <Double-1> \
	    "clearProofSelText ; markHotClause $num"
    $w tag bind atom.$num.0 <Control-Double-1> \
	    "markHotClause $num"
    $w tag bind atom.$num.0 <2> \
		"set Position \"atom $num 0\" "
    set i 1
    foreach lit $bd { 
		  set before [$w index insert]
		  if {$i==1} {  
				$w insert end "\ :-" $real_fontcl } {  $w insert end "," $real_fontcl }
		  set curpos2 \
				[expr [pos_x [$w index insert]] + [string length $lit] + 1]
		  if {$curpos2 > $MaxPos} {$w insert end "\n\ \ \ \ \ " $real_fontcl} {
				$w insert end " " $real_fontcl}
		  set beginbd [$w index insert]
		  $w tag add atom.$num.0 $before $beginbd
		  $w insert end "$lit" $real_fontcl
		  $w tag add atom.$num.$i $beginbd [$w index insert]
		  $w tag configure atom.$num.$i -font $real_fontcl -underline off
		  $w tag bind atom.$num.$i <1> \
				"clearProofSelText ; markHotAtom $num $i"
		  $w tag bind atom.$num.$i <Control-1> \
				"markHotAtom $num $i"
		  $w tag bind atom.$num.$i <Double-1> \
				"clearProofSelText ; markHotBody $num"
		  $w tag bind atom.$num.$i <Control-Double-1> "markHotBody $num"
        $w tag bind atom.$num.$i <2> \
				"set Position \"atom $num $i\" "
		  incr i}
    set endcl  [$w index insert]
    $w insert end ".\n"
    $w tag add atom.$num.0 $endcl [$w index insert]
    $w configure -state disabled
    tkTextSetCursor $ProofText end
}

proc addCurrentInit_Defs {num hd bd} {
    global fontcl fontclnum
    set w .currentInit_Defs.current.text
    set MaxPos [max_pos]  
    $w configure -state normal
    $w insert end "$num.\ " $fontcl 
    set char [$w index insert]
    set begincl $char
    set pos0 [$w index "insert linestart"]
    $w insert end "$hd" $fontcl
    $w tag add atom.$num.0 $char [$w index insert]
    $w tag configure atom.$num.0 -font $fontcl  -underline off
    set i 1
    foreach lit $bd { 
	if {$i==1} {  $w insert end "\ :-" $fontcl} {
	    $w insert end "," $fontcl
	}
	set curpos2 \
	    [expr [pos_x [$w index insert]] + [string length $lit] + 1]
	if {$curpos2 > $MaxPos} {$w insert end "\n\ \ \ \ \ \ \ "  $fontcl} {
	    $w insert end " " $fontcl
	}
	set beginbd [$w index insert]
	$w insert end "$lit" $fontcl
        
	incr i}
    $w tag add def.$num $begincl [$w index insert]
    $w tag configure def.$num -font $fontcl -underline off
    $w tag bind def.$num <1> \
	    "clearDefSelText ; markHotDef $num"
    $w tag bind  def.$num <Control-1> \
	    "markHotDef $num"
    $w tag bind  def.$num <Double-1> \
	    "clearDefSelText ; markHotDef $num"
    $w tag bind  def.$num <Control-Double-1> \
	    "markHotDef $num"
    set endcl  [$w index insert]
    $w insert end ".\n"
    $w configure -state disabled
    tkTextSetCursor $w end
}

proc addCurrentProg {num hd bd} {
    set w .currentProg.current.text
    set MaxPos 79
    $w configure -state normal
    $w insert end "$num.\ " 
    set char [$w index insert]
    set begincl $char
    set pos0 [$w index "insert linestart"]
    $w insert end "$hd" 
    set i 1
    foreach lit $bd { 
	if {$i==1} {  $w insert end "\ :-" } {
	    $w insert end "," 
	}
	set curpos2 \
	    [expr [pos_x [$w index insert]] + [string length $lit] + 1]
	if {$curpos2 > $MaxPos} {$w insert end "\n\ \ \ \ \ \ \ "  } {
	    $w insert end " " 
	}
	set beginbd [$w index insert]
	$w insert end "$lit" 
	incr i
    }
    set endcl  [$w index insert]
    $w insert end ".\n"
    $w configure -state disabled
    tkTextSetCursor $w end

}

proc addProofText2 {line} {
    global ProofText fonthistory
    clear_alertnote
    $ProofText configure -state normal
    $ProofText insert end $line $fonthistory
    $ProofText configure -state disabled
    tkTextSetCursor $ProofText end
}



proc ask_cl_deletion {pred arity clause} {
    tk_messageBox -icon question \
	    -type yesno -message "\"$pred/ $arity\" is undefined.\
Do you want to delete clause $clause ?"
}

proc ask_theory_overwrite {theory} {
    tk_messageBox -icon question \
	    -type yesno -message "This theory has already been loaded.\
Do you want to overwrite it ?"
}

proc msg_syntax_error_theory {} {
    tk_messageBox -icon error \
	    -type ok -message "Syntax error in theory. A theory must contain clauses of the form:
\"LHS -> RHS.\"
 where LHS and RHS are conjunctions of litterals."
}

proc showCurrentProg { i} {
    global white
    set w .currentProg 
    if [winfo exists  $w] {  
	wm deiconify $w
	raise $w 
	refreshCurrentProg ; return}    
#    catch {destroy $w}
    toplevel $w
    positionWindow $w
    wm title $w "Program"
    wm iconname $w "prog"
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.dismiss -text Close -foreground black \
	    -command "wm withdraw $w"
    button $w.buttons.refresh -text Update -foreground black \
	    -command "	refreshCurrentProg "
    pack $w.buttons.dismiss $w.buttons.refresh -side left -expand 1
    frame $w.current
    text $w.current.text -relief sunken -bd 2 -yscrollcommand "$w.current.vscrool set" \
	    -setgrid 1 	-background $white -foreground black -height 15
    scrollbar $w.current.vscrool -orient vertical -command "$w.current.text yview"
    set progNumBar $w.progNumBar
    text $progNumBar -height 2 -borderwidth 0
    $progNumBar tag configure boldcenter -justify center -font {Times 14 bold}
    $progNumBar configure -state normal
    $progNumBar insert end "Program $i" boldcenter
    $progNumBar configure -state disabled
    pack $w.current -padx 2 -pady 2
    pack $w.current.vscrool -padx 1 -pady 3 -side right -fill y
    pack $w.current.text -padx 1 -pady 3 -fill x -fill y 
    pack $progNumBar -side bottom -fill x -fill y 
}

proc showCurrentHistory {i f} {
    global white
    set w .currentHistory
#   catch {destroy $w}    
    if [winfo exists  $w] { 
	wm deiconify $w
	raise $w 
        refreshCurrentHistory2 $i $f
        return}    
    toplevel $w
    positionWindow $w
    wm title $w "History"
    wm iconname $w "hist"
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.dismiss -text Close -command "wm withdraw $w"
    button $w.buttons.refresh -text Update \
	    -command "	refreshCurrentHistory $f"
    pack $w.buttons.dismiss $w.buttons.refresh -side left -expand 1
    frame $w.current
    text $w.current.text -relief sunken -bd 2 -yscrollcommand "$w.current.vscrool set" \
	    -setgrid 1 	-background $white -foreground black -height 15
    scrollbar $w.current.vscrool -orient vertical -command "$w.current.text yview"
    set progNumBar $w.progNumBar
    text $progNumBar -height 2 -borderwidth 0
    $progNumBar tag configure boldcenter -justify center -font {Times 14 bold}
    pack $w.current -padx 2 -pady 2
    pack $w.current.vscrool -padx 1 -pady 3 -side right -fill y
    pack $w.current.text -padx 1 -pady 3 -fill x -fill y 
    pack $progNumBar -side bottom -fill x -fill y 
    refreshCurrentHistory2 $i $f
}

proc showCurrentInit_Defs { i} {
    global white
    set w .currentInit_Defs
    if [winfo exists  $w] {  
	wm deiconify $w
	raise $w
	return}    
    toplevel $w
    positionWindow $w
    wm title $w "Init+Defs"
    wm iconname $w "defs"
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.dismiss -text Close -command "wm withdraw $w"
    button $w.buttons.refresh -text Clear \
	    -command "clearDefSelText"
    pack $w.buttons.dismiss $w.buttons.refresh -side left -expand 1
    frame $w.current
    text $w.current.text -relief sunken -bd 2 -yscrollcommand \
	    "$w.current.vscrool set" \
	    -setgrid 1 	-background $white -foreground black -height 15
    scrollbar $w.current.vscrool -orient vertical -command \
	    "$w.current.text yview"
#    set progNumBar $w.progNumBar
#    text $progNumBar -height 2 -borderwidth 0
#    $progNumBar tag configure boldcenter -justify center -font {Times 14 bold}
#    $progNumBar configure -state normal
#    $progNumBar insert end "Program $i" boldcenter
#    $progNumBar configure -state disabled
    pack $w.current -padx 2 -pady 2
    pack $w.current.vscrool -padx 1 -pady 3 -side right -fill y
    pack $w.current.text -padx 1 -pady 3 -fill x -fill y 
#    pack $progNumBar -side bottom -fill x -fill y 
}

proc refreshCurrentProg { } {
    prolog "current_program(N,_)"
    set new $prolog_variables(N)
    set progNumBar .currentProg.progNumBar 
    set line1 [$progNumBar get 1.0 "insert lineend"]
    puts stdout "line1=$line1"
    puts stdout "new=$new"
    if {$line1=="Program $new"} { } {  
	set w .currentProg.current.text
	$w  configure -state normal  
	$w delete 1.0 end
	prolog "tcl_update_current_prog($new)"
	tkTextSetCursor $w end
	$w configure -state disabled
	$progNumBar configure -state normal
	$progNumBar delete 1.0 end
	$progNumBar insert end "Program $new" boldcenter
	$progNumBar configure -state disabled
}
}

proc refreshCurrentInit_Defs { } {
    prolog "current_program(N,_)"
    set new $prolog_variables(N)
    set progNumBar .currentInit_Defs.progNumBar 
    set line1 [$progNumBar get 1.0 "insert lineend"]
    if {$line1=="Program $new"} { puts stdout bla} {  
	set w .currentInit_Defs.current.text
	$w  configure -state normal  
	$w delete 1.0 end
	prolog "show_init_defs"
	tkTextSetCursor $w end
	$w configure -state disabled
	$progNumBar configure -state normal
	$progNumBar delete 1.0 end
	$progNumBar insert end "Program $new" boldcenter
	$progNumBar configure -state disabled
}
}

proc refreshCurrentHistory { f} {
    prolog "current_program(N,_)"
    refreshCurrentHistory2 $prolog_variables(N) $f
}
proc refreshCurrentHistory2 { new f} {
    global env
    set progNumBar .currentHistory.progNumBar 
    set line1 [$progNumBar get 1.0 "insert lineend"]
    puts stdout "line1=$line1"
    if {$line1=="History of program $new"} { return } {  
	set w .currentHistory.current.text
	prolog "show_current_history_buffer($new)"
	$w  configure -state normal  
	$w delete 1.0 end
	puts stdout $env(NEWMAP_DIR)/$f
	set file [open $f]
	while {![eof $file]} {
	    $w insert end [read $file 1024]
	}
	tkTextSetCursor $w end
	$w configure -state disabled
	close $file
	$progNumBar configure -state normal
	$progNumBar delete 1.0 end
	$progNumBar insert end "History of program $new" boldcenter
	$progNumBar configure -state disabled
    }
}

proc showCurrentTheoryIf { } {
    if [winfo exists .currentTheory] {
	   if [winfo viewable .currentTheory] {
	       showCurrentTheory } }
}

proc showCurrentTheory { } {
    global white
    set w .currentTheory 
    if {[prolog  current_theory_empty] == 1} {
	msg_warning_ok "Theory is empty!" ; return }
    if [winfo exists  $w] {  
	wm deiconify $w
	raise $w ; refreshCurrentTheory ; return}    
    toplevel $w
    positionWindow $w
    wm title $w "Theory"
    wm iconname $w "theory"
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.dismiss -text Close -command "wm withdraw $w"
    button $w.buttons.refresh -text Clear \
	    -command "clearLawSelText"
    pack $w.buttons.dismiss $w.buttons.refresh -side left -expand 1
    frame $w.current
    text $w.current.text -relief sunken -bd 2 -yscrollcommand "$w.current.vscrool set" \
	    -setgrid 1 	-background $white -foreground black -height 15
	$w.current.text tag configure courier10 -font {Courier 10}
	$w.current.text tag configure courier12 -font {Courier 12}
	$w.current.text tag configure courier14 -font {Courier 14}
	$w.current.text tag configure courier16 -font {Courier 16}
	$w.current.text tag configure courier18 -font {Courier 18}
	$w.current.text tag configure courier20 -font {Courier 20}
    scrollbar $w.current.vscrool -orient vertical -command "$w.current.text yview"
    pack $w.current -padx 2 -pady 2
    pack $w.current.vscrool -padx 1 -pady 3 -side right -fill y
    pack $w.current.text -padx 1 -pady 3 -fill x -fill y 
    refreshCurrentTheory 
}

proc refreshCurrentTheory { } {
	set w .currentTheory.current.text
        clearLawSelText
	$w  configure -state normal  
	$w delete 1.0 end
	prolog "tcl_update_current_theory"
	tkTextSetCursor $w end
	$w configure -state disabled
}




proc get_head_case_split {} {
    puts stdout blabla
}


proc get_head_case_split_old {c} {
    global newhd white
    set w .entrycasesplit
    catch {destroy $w}
    toplevel $w
    positionWindow $w
    wm title $w "case split"
    label $w.msg -wraplength 4i  -justify left -foreground black \
	    -text "enter an instance of the head:" 
    pack $w.msg -side top -pady 5 -padx 3 -anchor w 
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    entry $w.e1 -relief sunken -background $white -foreground black\
	    -width 40  -textvariable newhd
    pack $w.e1 -side top -pady 5 -padx 5 -expand yes
    $w.e1 delete 0 end
    if {[prolog make_entry_case_split($c,NewHd)] == 1}  {
	$w.e1 insert end $prolog_variables(NewHd)}
    button $w.buttons.cancel -text Close -foreground black \
	    -command {destroy .entrycasesplit ; return "error"}
    button $w.buttons.ok -foreground black \
	    -text Split -command {prolog write(bla)}
    pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
}
#$prolog_variables(NewHd)} 

proc  get_n_and_goto_n {} {
    global white
    set w .entrygoto
    catch {destroy $w}
    toplevel $w
    positionWindow $w
    wm title $w "Goto"
    label $w.msg -wraplength 2i  -foreground black \
	    -justify left -text "enter program number:"
    pack $w.msg -side top
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.cancel -text Close -foreground black \
	    -command "destroy $w"
    button $w.buttons.ok -text Goto -foreground black -command { 
	if [null $n] {get_n_and_goto_n}
	if {[prolog node_label($n,_)] == 0}  {
	      msg_warning_ok "Unknown program number" 
	    return }
	if {[prolog current_program($n,_)] == 1}  {
	      msg_warning_ok "Program $n is already the current program." 
	    return }
	    destroy .entrygoto
	    prolog set_current_program($n) }
    entry $w.e1 -relief sunken -background $white -foreground black\
	    -textvariable n
    pack $w.e1 -side top -pady 5
    pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
}

proc get_namefile_initial_prog {} {
    global dirNameProg
    set w .entryinitialprog
    catch {destroy $w}
    toplevel $w
    wm title $w "Save"
    positionWindow $w
    label $w.msg -wraplength 2i  -justify left -foreground black \
	    -text "give a name to the new initial program:"
    pack $w.msg -side top
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.cancel -text Close -foreground black \
	    -command "destroy $w"
    button $w.buttons.ok -text Save -foreground black -command { 
	 if [file isfile $dirNameProg/$filename]  {
	     if {[ msg_warning_yesno "\"$filename\" already exists in directory $dirNameProg
Do you want to overwrite it ?"] == "no"} { return } }
	    destroy .entryinitialprog
 	    destroy .entryinitialprog.e1
            if {[null $filename]} {alertnote "Not a file name." ; return}
	    prolog "apply_save_program('$dirNameProg/$filename')"
            clearProofText
	    if { [prolog apply_restart('$dirNameProg/$filename')] == 1} {
	    destroy4
	    enableActions }
	}
    entry $w.e1 -relief sunken -textvariable filename
    pack $w.e1 -side top -pady 5
    pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
}

proc get_namefile_and_save {what} {
 global dirNameProg dirNameTheories dirNameSessions dirNameHistories

 set types {
  {"Prolog files"		{.pl}	}
  {"Text files"		{}		TEXT}
  {"All files"		*}
 }

 if {$what== "program"} {set dir $dirNameProg} 
 if {$what== "program_as_term"} {set dir $dirNameProg} 
 if {$what== "program_for_sicstus"} {set dir $dirNameProg}  
 if {$what== "program_for_xsb"} {set dir $dirNameProg} 
 if {$what== "theory"} {set dir $dirNameTheories} 
 if {$what== "session"} {set dir $dirNameSessions} 
 if {$what== "history"} {set dir $dirNameHistories} 


 set file [tk_getSaveFile -filetypes $types -initialdir $dir]
 if {[string compare $file ""] == 0} { return -code return }
 
 if {$what== "program"} {prolog "apply_save_program('$file')"} 
 if {$what== "program_as_term"} {prolog "apply_save_program_as_term('$file')"}
 if {$what== "program_for_sicstus"} {prolog "apply_save_program_for_sicstus('$file')"}
 if {$what== "program_for_xsb"} {prolog "apply_save_program_for_xsb('$file')"}
 if {$what== "theory"} {prolog "apply_save_theory('$file')"}
 if {$what== "session"} {prolog "apply_save_session('$file')"}
 if {$what== "history"} {prolog "apply_save_history('$file')"}

}



# proc get_namefile_and_save_program {} {
#     global dirNameProg
#     set w .entryprog
#     catch {destroy $w}
#     toplevel $w
#     wm title $w "Save"
#     positionWindow $w
#     label $w.msg -wraplength 2i  -justify left -foreground black \
# 	    -text "enter program name:"
#     pack $w.msg -side top
#     frame $w.buttons
#     pack $w.buttons -side bottom -fill x -pady 2m
#     button $w.buttons.cancel -text Close -foreground black \
# 	    -command "destroy $w"
#     button $w.buttons.ok -text Save -foreground black -command { 
# 	 if [file isfile $dirNameProg/$filename]  {
# 	     if {[ msg_warning_yesno "\"$filename\" already exists in directory $dirNameProg
# Do you want to overwrite it ?"] == "no"} { return } }
# 	    destroy .entryprog 
#  	    destroy .entryprog.e1
#             if {[null $filename]} {alertnote "Not a file name." ; return}
# 	    prolog "apply_save_program('$dirNameProg/$filename')"
#      }
#     entry $w.e1 -relief sunken  -foreground black -textvariable filename
#     pack $w.e1 -side top -pady 5
#     pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
#  }


# proc get_namefile_and_save_program_as_term {} {
#     global dirNameProg
#     set w .entryprog
#     catch {destroy $w}
#     toplevel $w
#     wm title $w "Save"
#     positionWindow $w
#     label $w.msg -wraplength 2i  -justify left -foreground black \
# 	    -text "enter program name:"
#     pack $w.msg -side top
#     frame $w.buttons
#     pack $w.buttons -side bottom -fill x -pady 2m
#     button $w.buttons.cancel -text Close -foreground black \
# 	    -command "destroy $w"
#     button $w.buttons.ok -text Save -foreground black -command { 
# 	 if [file isfile $dirNameProg/$filename]  {
# 	     if {[ msg_warning_yesno "\"$filename\" already exists in directory $dirNameProg
# Do you want to overwrite it ?"] == "no"} { return } }
# 	    destroy .entryprog 
#  	    destroy .entryprog.e1
#             if {[null $filename]} {alertnote "Not a file name." ; return}
# 	    prolog "apply_save_program_as_term('$dirNameProg/$filename')"
#      }
#     entry $w.e1 -relief sunken  -foreground black -textvariable filename
#     pack $w.e1 -side top -pady 5
#     pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
#  }

# proc get_namefile_and_save_program_for_sicstus {} {
#     global dirNameProg
#     set w .entryprog
#     catch {destroy $w}
#     toplevel $w
#     wm title $w "Save"
#     positionWindow $w
#     label $w.msg -wraplength 2i  -justify left -foreground black \
# 	    -text "enter program name:"
#     pack $w.msg -side top
#     frame $w.buttons
#     pack $w.buttons -side bottom -fill x -pady 2m
#     button $w.buttons.cancel -text Close -foreground black \
# 	    -command "destroy $w"
#     button $w.buttons.ok -text Save -foreground black -command { 
# 	 if [file isfile $dirNameProg/$filename]  {
# 	     if {[ msg_warning_yesno "\"$filename\" already exists in directory $dirNameProg
# Do you want to overwrite it ?"] == "no"} { return } }
# 	    destroy .entryprog 
#  	    destroy .entryprog.e1
#             if {[null $filename]} {alertnote "Not a file name." ; return}
# 	    prolog "apply_save_program_for_sicstus('$dirNameProg/$filename')"
#      }
#     entry $w.e1 -relief sunken  -foreground black -textvariable filename
#     pack $w.e1 -side top -pady 5
#     pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
#  }


# proc get_namefile_and_save_program_for_xsb {} {
#     global dirNameProg
#     set w .entryprog
#     catch {destroy $w}
#     toplevel $w
#     wm title $w "Save"
#     positionWindow $w
#     label $w.msg -wraplength 2i  -justify left -foreground black \
# 	    -text "enter program name:"
#     pack $w.msg -side top
#     frame $w.buttons
#     pack $w.buttons -side bottom -fill x -pady 2m
#     button $w.buttons.cancel -text Close -foreground black \
# 	    -command "destroy $w"
#     button $w.buttons.ok -text Save -foreground black -command { 
# 	 if [file isfile $dirNameProg/$filename]  {
# 	     if {[ msg_warning_yesno "\"$filename\" already exists in directory $dirNameProg
# Do you want to overwrite it ?"] == "no"} { return } }
# 	    destroy .entryprog 
#  	    destroy .entryprog.e1
#             if {[null $filename]} {alertnote "Not a file name." ; return}
# 	    prolog "apply_save_program_for_xsb('$dirNameProg/$filename')"
#      }
#     entry $w.e1 -relief sunken  -foreground black -textvariable filename
#     pack $w.e1 -side top -pady 5
#     pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
#  }



# proc get_namefile_and_save_theory {} {
#     global dirNameTheories
#     set w .entrytheory
#     catch {destroy $w}
#     toplevel $w
#     wm title $w "Save"
#     positionWindow $w
#     label $w.msg -wraplength 2i  -justify left -text "enter theory name:"
#     pack $w.msg -side top
#     frame $w.buttons
#     pack $w.buttons -side bottom -fill x -pady 2m
#     button $w.buttons.cancel -text Close -foreground black \
# 	    -command "destroy $w"
#     button $w.buttons.ok -text Save -foreground black -command { 
# 	 if [file isfile $dirNameTheories/$filename]  {
# 	     if {[ msg_warning_yesno "\"$filename\" already exists in directory $dirNameTheories
# Do you want to overwrite it ?"] == "no"} { return } } 
# 	    destroy .entrytheory 
#  	    destroy .entrytheory.e1
#             if {[null $filename]} {alertnote "Not a file name." ; return}
# 	    prolog "apply_save_theory('$dirNameTheories/$filename')"
#      }
#     entry $w.e1 -relief sunken  -foreground black -textvariable filename
#     pack $w.e1 -side top -pady 5
#     pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
#  }

# proc get_namefile_and_save_session {} {
#     global dirNameSessions
#     set w .entrysession
#     catch {destroy $w}
#     toplevel $w
#     wm title $w "Save"
#     positionWindow $w
#     label $w.msg -wraplength 2i -foreground black \
# 	    -justify left -text "enter session name:"
#     pack $w.msg -side top
#     frame $w.buttons
#     pack $w.buttons -side bottom -fill x -pady 2m
#     button $w.buttons.cancel -text Close -foreground black \
# 	    -command "destroy $w"
#     button $w.buttons.ok -text Save -foreground black -command { 
# 	 if [file isfile $dirNameSessions/$filename]  {
# 	     if {[ msg_warning_yesno "\"$filename\" already exists in directory $dirNameSessions
# Do you want to overwrite it ?"] == "no"} { return } } 
# 	    destroy .entrysession 
#  	    destroy .entrysession.e1
#             if {[null $filename]} {alertnote "Not a file name." ; return}
# 	    prolog "apply_save_session('$dirNameSessions/$filename')"
#      }
#     entry $w.e1 -relief sunken  -foreground black -textvariable filename
#     pack $w.e1 -side top -pady 5
#     pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
#  }


# proc get_namefile_and_save_history {} {
#     global dirNameHistories
#     set w .entryhistory
#     catch {destroy $w}
#     toplevel $w
#     wm title $w "Save"
#     positionWindow $w
#     label $w.msg -wraplength 2i  -foreground black \
# 	    -justify left -text "enter history name:"
#     pack $w.msg -side top
#     frame $w.buttons
#     pack $w.buttons -side bottom -fill x -pady 2m
#     button $w.buttons.cancel -text Close -foreground black \
# 	    -command "destroy $w"
#     button $w.buttons.ok -text Save -foreground black -command { 
# 	 if [file isfile $dirNameHistories/$filename]  {
# 	     if {[ msg_warning_yesno "\"$filename\" already exists in directory $dirNameHistories
# Do you want to overwrite it ?"] == "no"} { return } } 
# 	    destroy .entryhistory 
#  	    destroy .entryhistory.e1
# if {[null $filename]} {alertnote "Not a file name." ; return}
# 	    prolog "apply_save_history('$dirNameHistories/$filename')"
#      }
#     entry $w.e1 -relief sunken  -foreground black -textvariable filename
#     pack $w.e1 -side top -pady 5
#     pack $w.buttons.cancel $w.buttons.ok -side left -expand 1
#  }



proc addCurrentTheory {num lhs rhs} {
    global fontcl fontclnum
    set MaxPos [max_pos]
    set w .currentTheory.current.text
    $w configure -state normal
     $w insert end "$num. "
    set beginlaw [$w index insert]
    $w insert end "$lhs" $fontcl
    set endlhs [$w index insert]
    set curpos2 \
	    [expr [pos_x $endlhs] + [string length $rhs] + 9]
    if {$curpos2 > $MaxPos} {
	$w insert end "\n\ \ \ \ \ \ \ -> " $fontcl} { 
	    $w insert end " -> " $fontcl} 
    set beginrhs [$w index insert]
    $w insert end "$rhs" $fontcl
    $w insert end ".\n"
    set endlaw [$w index insert]
    $w tag add law.$num $beginlaw $endlaw
    $w tag configure law.$num -font $fontcl
    $w tag bind law.$num <1> \
	    "clearLawSelText ; markHotLaw $num"
    $w tag bind law.$num <Double-1> \
	    "clearLawSelText ; markHotLaw $num"
    $w tag bind law.$num <Control-1> \
		"markHotLaw $num"
    $w configure -state disabled
    tkTextSetCursor $w end
    puts stdout "add tag law.$num"
	}


proc prova {} {
    global ProofText
    set w $ProofText
    $w configure -state normal
    set msg  "abcdefghijabcdefghijabcdefghijabcdefghijabcdefghijabcdefghijabcdefghijabcdefghij\n"
    $w delete 1.0 end
    $w insert end 10\n
    $w insert end $msg courier10
    $w insert end 12\n
    $w insert end $msg   courier12 
    $w insert end 14\n
    $w insert end $msg   courier14 
    $w insert end 16\n
    $w insert end $msg   courier16 
    $w insert end 18\n
    $w insert end $msg   courier18
    $w insert end 20\n
    $w insert end $msg courier20
}

proc show_law_definitions_print {} {
    if {[ msg_warning_yesno "Do you want to print all laws or only those which have been used ?\n\nSelect \"Yes\" for all, \"No\" otherwise." ] == "yes"} { 
	prolog show_law_definitions_all } { prolog show_law_definitions_used }
}

proc show_law_definitions_save {} {
    if {[ msg_warning_yesno "Do you want to save all laws or only those which have been used ?\n\nSelect \"Yes\" for all, \"No\" otherwise." ] == "yes"} { 
	prolog show_law_definitions_all } { prolog show_law_definitions_used }
}


proc addClause {w hd bd} {
#    set w .wdefinition.text
    set MaxPos 79
    $w insert end "$hd" 
    set i 1
    foreach lit $bd { 
	if {$i==1} {  $w insert end "\ :-" } {
	    $w insert end "," 
	}
	set curpos2 \
	    [expr [pos_x [$w index insert]] + [string length $lit] + 1]
	if {$curpos2 > $MaxPos} {$w insert end "\n\ \ \ \ \ \ \ "  } {
	    $w insert end " " 
	}
	set beginbd [$w index insert]
	$w insert end "$lit" 
	incr i
    }
    $w insert end ".\n"
    tkTextSetCursor $w end

}


puts stdout "display.tcl compiled"
