# rules.tcl

proc restartTransformation {} {
    if { [prolog clause(current_program(_,_),_)] == 1} {
		if {[ msg_warning_yesno "This will start a new derivation, with the current program as initial program. The current derivation will be lost if not saved.\n\nContinue restart ?"] == "no"} { return }} 
    get_namefile_initial_prog
}


proc apply_cs_strategy {} {
    global enable_ccr
    global enable_bup
    global enable_quick_solve
    global enable_mc

    if {$enable_ccr == 1} { prolog set_current(ccr,yes) }
    if {$enable_ccr == 0} { prolog set_current(ccr,no) }
    if {$enable_bup == 1} { prolog set_current(bup,yes) }
    if {$enable_bup == 0} { prolog set_current(bup,no) }
    if {$enable_quick_solve == 1} { prolog set_current(quick_solve,yes) }
    if {$enable_quick_solve == 0} { prolog set_current(quick_solve,no) }
    if {$enable_mc == 1} { prolog set_current(mc,yes) }
    if {$enable_mc == 0} { prolog set_current(mc,no) }

    apply_transformation_rs "strategy" cs
}

proc apply_ds_strategy {} {
 apply_transformation_rs "strategy" ds
}


proc apply_transformation_rs { RorS rule_strat } {
    global help ProofText
    clear_alertnote
    if {[prolog "clause(initial_prog(P),_)"]==0} {
		  alertnote "Load a program first."
		  return
    }
    set Goal [check_selection $RorS $rule_strat]
    if {$Goal=="error"} {
		  if $help { helpRule $rule_strat} ; return}
    if {$Goal=="not yet implemented"} {
		  puts stdout "not yet available" ; return}
	 #	clearProofSelText
}
proc check_selection { rule_or_strat rsname } {
    global proofSelection defSelection lawSelection Position\
	solverSelection ufSelection wqounfSelection wqogenSelection wideningSelection


    if {$rsname=="define"} { 
	 enter_new_definition 
	 return ""
	 }
    if {$rsname=="restrict"} { 
		  set check [check_selection_for_restrict\
							  $proofSelection]
    }
    if {$rsname=="unfold"} {
		  set check [check_selection_for_unfolding\
							  $proofSelection]
    }
    if {$rsname=="fold"} {
		  set check [check_selection_for_folding\
							  $proofSelection $defSelection]
    }
    if {$rsname=="replace_goal"} { 
		  set check [check_selection_for_replace_goal \
							  $lawSelection $proofSelection]
    }
    if {$rsname=="move_goal"} { 
		  set check [check_selection_for_move_goal \
							  $proofSelection $Position]
    }
    if {$rsname=="add_goal"} { 
		  set check [check_selection_for_add_goal \
							  $proofSelection]
    }
    if {$rsname=="delete_goal"} { 
		  set check [check_selection_for_delete_goal \
							  $proofSelection]
    }
    if {$rsname=="replace_clauses"} { 
		  set check [check_selection_for_replace_clauses \
							  $proofSelection]
		  if {$check== "error"} {return $check}
		  puts stdout "prosel=$proofSelection"
		  #	puts stdout "check replace clauses"
		  enter_replace_clauses
		  return ""
    }
    if {$rsname=="move_clause"} { 
		  set check [check_selection_for_move_clause \
							  $proofSelection $Position]
    }
    if {$rsname=="add_clauses"} { 
		  enter_new_clauses
		  return ""
    }
    if {$rsname=="delete_clauses"} { 
		  set check [check_selection_for_delete_clauses\
							  $proofSelection]
    }
    if {$rsname=="generalize"} { 
		  set check [check_selection_for_generalize\
							  $proofSelection]
		  if {$check== "error"} {return $check}
    }
    if {$rsname=="case_split"} { 
		  set check [check_selection_for_case_split\
							  $proofSelection]
		  if {$check== "error"} {return $check}
    }

	 
	 # note: $rsname should be changed --> see reg. exp.
    
    if { $rsname=="solve_constraints(linking_vars)"} {
		  
		  # try to match a regular expression and extract the "context"
		  # e.g. solve_constraints(*) and extract * 
		  # 	set startpos [string first "(" $rsname]
		  # 	set endpos [string last ")" $rsname]
		  # 	if { $startpos != -1 && $endpos != -1} {
		  # 	    set intvarscode [string range $rsname $startpos+1 $endpos-1]
		  # 	} 
		  
		  # 	puts $startpos
		  # 	puts $endpos
		  
		  # 	else {set check "error"; return $check}
		  # set lawSelection "clpq"
		  # $solverSelection?
		  
		  set check [check_selection_for_solve_constraints \
				 $proofSelection]
		  # was: 		 $lawSelection $proofSelection] 
		  
		  if {$check== "error"} {
				puts stdout "error in check solve constraints"
				return $check
		  }
		  
		  prolog "current(solver,S)"
		  set solver $prolog_variables(S)
		  regsub {#solver#} { solve_constraints(#solver#,linking_vars)} \
$solver rsname
		  }


		  
	if {$rsname=="caf(soft)"} {
	    set check [check_selection_for_caf\
				 $proofSelection $defSelection]
	    prolog "current(solver,S)"
	    set solver $prolog_variables(S)
	    set rsname "caf($solver,soft)"
	}
	if {$rsname=="split_determ"} { 
		  set check [check_selection_for_split_determ\
							  $proofSelection]
		  if {$check== "error"} {return $check}
	}	
		  
	# STRATEGIES
	if {$rsname== "uve"} {todo ; return }
	if {$rsname== "lap"} {todo ; return }
	if {$rsname== "determ"} {todo ; return }

	if {$rsname== "cs"} {
	 css_param

	 set check [check_selection_for_cs $defSelection ]
	 if {$check== "error"} {
	  puts stdout "error in check contextual specialization"
	  return $check
	 }
	 
	 prolog "cs_env(S,UF,WU,WG,W)"
	 set solver $prolog_variables(S)
	 set uf $prolog_variables(UF)
	 set wqounf $prolog_variables(WU)
	 set wqogen $prolog_variables(WG)
	 set widening $prolog_variables(W)
	    
	 puts stdout $widening
	 set rsname cs($solver,$uf,$wqounf,$wqogen,$widening)
	}

	
	if {$rsname== "ds"} {
	 ds_param
	 set check [check_selection_for_ds $defSelection ]
	 if {$check== "error"} {
	  puts stdout "error in check determinization"
	  return $check
	 }
	 
	 prolog "ds_env(S,UF,WU)"
	 set solver $prolog_variables(S)
	 set uf $prolog_variables(UF)
	 set wqounf $prolog_variables(WU)
	 
	 set rsname ds($solver,$uf,$wqounf)
	}
		  if { $rsname=="solve_program_head_vars"} {
		  	    set check [check_selection_for_solve_program ]
				if {$check== "error"} {
					 puts stdout "error in check solve program"
					 return $check
				}
		  prolog "current(solver,S)"
		  set solver $prolog_variables(S)
		  regsub {#solver#} { solve_program(#solver#,head_vars)} \
$solver rsname
		  }

	if { $rsname=="make_prop"} {set check ""}
	if { $rsname=="ccr"} {set check ""}


	if {$check== "error"} { return $check }
	
	set sel [makeprologselection $proofSelection]
	set defsel [makeprologselection $defSelection]
	set lawsel [makeprologselection $lawSelection]

	# rule or strategy?
	if {$rule_or_strat== "rule"} {
		 puts stdout "apply_transformation_rule($rsname,$sel,$defsel,$lawsel)"
		 prolog "apply_transformation_rule($rsname,$sel,$defsel,$lawsel)"
	}
	if {$rule_or_strat== "strategy"} {
		 puts stdout "apply_strategy($rsname,$sel,$defsel,$lawsel)"
		 prolog "apply_strategy($rsname,$sel,$defsel,$lawsel)"
	}

}


proc check_selection_for_restrict {selection} {
    if [empty $selection] { 
	alertnote "Click on some atom(s) having the predicate(s) wrt which\
		you want to restrict the program." ; return "error" }
	return ""
}

proc check_selection_for_unfolding {selection} {
	if [empty $selection] {
	    alertnote "Select an atom first." ; return "error" 
	}
	if [more_than_one_selection $selection] {
	    alertnote "Select one atom only." ; return "error" 
	}
	if [not_a_body_atom $selection] {
	    alertnote "Select an atom in a clause body." ; return "error" 
	}
    return ""
}

proc check_selection_for_folding {to_fold to_use} { 
    puts stdout $to_fold,$to_use"
    set w .currentInit_Defs
    if [winfo exists  $w] {  raise $w } {
	prolog update_show_init_defs }
    set msg1 "-the clause(s) to be folded in the main window."
    set msg2 "-the clause(s) to be used for folding in the \"Init+Defs\" window."
    if {[expr [empty $to_fold] && [empty $to_use]]} { 
	alertnote "Select:\n$msg1\n$msg2" ; return "error"}
    if [empty $to_fold] { 
	alertnote "Select:\n$msg1" ; return "error"}
    if [empty $to_use] { 
	alertnote "Select:\n$msg2" ; return "error"}
    if {![expr [llength $to_fold]==[llength $to_use]]} {
	alertnote "Bad selection. Select as many clauses in the \"Init+Defs\" window as in the main window." ; return "error"}
    if {![all_clauses $to_fold]} {
	alertnote "Bad selection. Only clauses can be selected in the current implementation." 
	return "error"}
    return ""
}

proc check_selection_for_replace_goal {lawSel ProofSel} {    
    if {[prolog  current_theory_empty] == 1} {
	msg_warning_ok "Theory is empty!" 
	return "error"
    }
    if {[expr [empty $lawSel] && [empty $ProofSel]]==1} { 
	alertnote "Select first a goal and a law."
	return "error"
    }
    if {[expr [llength $lawSel] > 1] } {
	alertnote "Select one law only."
	return "error"
    } 
    if {[empty $ProofSel]} { 
	alertnote "Select a goal."
	return "error"
    }
    if {[empty $lawSel]} { 
	if [winfo exists .currentTheory] {wm deiconify .currentTheory ; raise .currentTheory}
	alertnote "Select a law."
	return "error"
    } 
   return ""
}

proc check_selection_for_move_goal {ProofSel Pos} {    
    if {[empty $ProofSel]} { 
	alertnote "Select a goal."
	return "error"
    }
    if {[empty $Pos]} { 
	alertnote "Click with the second button of the mouse on the atom after which to do the move.\n(Head of clause for the first position of its body).Then, press again the rule button."
	return "error"
    }
    lapp_proof_sel $Pos
   return ""
}

proc check_selection_for_add_goal {ProofSel} {    
    if {[empty $ProofSel]} { 
	alertnote "Select a goal."
	return "error"
    }
    if {![all_atoms $ProofSel]} { 
	alertnote "Select some goal(s) only."
	return "error"
    }
   return ""
}

proc check_selection_for_delete_goal {ProofSel} {    
    if {[empty $ProofSel]} { 
	alertnote "Select a goal."
	return "error"
    }
    if {![all_atoms $ProofSel]} { 
	alertnote "Select some goal(s) only."
	return "error"
    }
   return ""
}

proc check_selection_for_replace_clauses {ProofSel} {
    if {[empty $ProofSel]} { 
	alertnote "Select some clause(s)."
	return "error"}
     if {![all_clauses $ProofSel]} {
     alertnote "Bad selection. Select clause(s) only."
	return "error"
     }
     return " "
}

proc check_selection_for_move_clause {ProofSel Pos} {
    set Posdefault "clause 0"
    if {[empty $ProofSel]} { 
	alertnote "Select a clause."
	return "error"
    }
    if {![is_a_clause $ProofSel]} {
	alertnote "Select one clause."
	return "error"
    }
    if {[null $Pos]} { 
	if {[ msg_warning_okcancel "The selected clause will become the first clause of the program. Press \"Ok\" to perform this move. Otherwise, press \"Cancel\", click with the second button of the mouse on an atom of the clause after which to do the move. Then, press again the rule button."] == "cancel"} { return "error"} }
    if {[null $Pos]} {set Pos $Posdefault } 
    puts stdout "$ProofSel, $Pos"
    lapp_proof_sel $Pos
    return ""
}

proc check_selection_for_delete_clauses {ProofSel} {
    if {[empty $ProofSel]} { 
	alertnote "Select a clause first."
	return "error"
    }
    if {![all_clauses $ProofSel]} {
	alertnote "Bad selection. Select some clause(s) only."
	return "error"
    }
    return ""
}

proc check_selection_for_case_split {ProofSel} {
    if {[empty $ProofSel]} { 
	alertnote "Select a clause."
	return "error"
    }
    if {![is_a_clause $ProofSel]} {
	alertnote "Select one clause."
	return "error"
    }
    return ""
}

proc check_selection_for_split_determ {ProofSel} {
    if {[empty $ProofSel]} { 
	alertnote "Select a clause."
	return "error"
    }
    if {![is_a_clause $ProofSel]} {
	alertnote "Select one clause."
	return "error"
    }
    if {[prolog  no_solver_selected] == 1} {
	msg_warning_ok "No solver selected." 
	return "error"
    }
    return ""
}

proc check_selection_for_generalize {ProofSel} {
    if {[empty $ProofSel]} { 
	alertnote "Select a clause."
	return "error"
    }
    if {![is_a_clause $ProofSel]} {
	alertnote "Select one clause."
	return "error"
    }
    return ""
}


proc check_selection_for_solve_constraints { ProofSel} {    

    if {[prolog  no_solver_selected] == 1} {
	msg_warning_ok "No solver selected." 
	return "error"
    }
    if {[empty $ProofSel]} { 
	alertnote "Select the constraints or the clause to solve."
	return "error"
    }

   return ""
}


proc check_selection_for_caf {to_fold to_use} { 
    puts stdout $to_fold,$to_use"
    set w .currentInit_Defs
    if [winfo exists  $w] {  raise $w } {
	prolog update_show_init_defs }
    set msg1 "-the clause to be folded in the main window."
    set msg2 "-the clause to be used for folding in the \"Init+Defs\" window."
    if {[expr [empty $to_fold] && [empty $to_use]]} { 
	alertnote "Select:\n$msg1\n$msg2" ; return "error"}
    if [empty $to_fold] { 
	alertnote "Select:\n$msg1" ; return "error"}
    if [empty $to_use] { 
	alertnote "Select:\n$msg2" ; return "error"}
    if {![expr [llength $to_fold]==1 && [llength $to_use]==1]} {
	alertnote "Bad selection. Select one clause the \"Init+Defs\" window and one clause in the main window." ; return "error"}
    if {![all_clauses $to_fold]} {
	alertnote "Bad selection. Only clauses can be selected in the current implementation." 
	return "error"}
    return ""
}



proc check_selection_for_cs {defselection} {
    global enable_ccr
    global enable_bup
    global enable_quick_solve
    global enable_mc

    if {$enable_ccr == 1} { prolog set_current(ccr,yes) }
    if {$enable_ccr == 0} { prolog set_current(ccr,no) }
    if {$enable_bup == 1} { prolog set_current(bup,yes) }
    if {$enable_bup == 0} { prolog set_current(bup,no) }
    if {$enable_quick_solve == 1} { prolog set_current(quick_solve,yes) }
    if {$enable_quick_solve == 0} { prolog set_current(quick_solve,no) }
    if {$enable_mc == 1} { prolog set_current(mc,yes) }
    if {$enable_mc == 0} { prolog set_current(mc,no) }

	 
    if {[prolog  no_solver_selected] == 1} {
	msg_warning_ok "No solver selected." 
	#showSolvers
	return "error"
    }
    if {[prolog  no_uf_selected] == 1} {
	msg_warning_ok "No unfolding function selected." 
	return "error"
    }
    if {[prolog  no_wqounf_selected] == 1} {
	msg_warning_ok "No wqo selected for unfolding." 
	return "error"
    }
    if {[prolog  no_wqogen_selected] == 1} {
	msg_warning_ok "No wqo selected for generalization." 
	return "error"
    }
    if {[prolog  no_widening_selected] == 1} {
	msg_warning_ok "No widening operator selected." 
	return "error"
    }
    # check the compatibility of the selected components
    prolog "cs_env(S,UF,WU,WG,W)"
    set solver $prolog_variables(S)
    set uf $prolog_variables(UF)
    set wqounf $prolog_variables(WU)
    set wqogen $prolog_variables(WG)
    set widening $prolog_variables(W)

    if {[prolog  is_a_uf($uf,$solver)] == 0} {
	msg_warning_ok "The unfolding function $uf was not designed to work with solver $solver." 
	return "error"
    }
    if {[prolog  is_a_wqounf($wqounf,$solver)] == 0} {
	msg_warning_ok "The unfolding wqo $wqounf was not designed to work with solver $solver." 
	return "error"
    }
    if {[prolog  is_a_wqogen($wqogen,$solver)] == 0} {
	msg_warning_ok "The generalization wqo $wqogen was not designed to work with solver $solver." 
	return "error"
    }
    if {[prolog  is_a_widening($widening,$solver)] == 0} {
	msg_warning_ok "The widening operator $widening  was not designed to work with solver $solver." 
	return "error"
    }


    if {[llength $defselection]==0} {
	alertnote "Bad selection. Select the initial definitions in the \"Init+Defs\" window."
	return "error"}
    
    return ""
}


proc check_selection_for_ds {defselection} {

  if {[prolog  no_solver_selected] == 1} {
   msg_warning_ok "No solver selected." 
   #showSolvers
   return "error"
  }
  if {[prolog  no_uf_selected] == 1} {
   msg_warning_ok "No unfolding function selected." 
   return "error"
  }
  if {[prolog  no_wqounf_selected] == 1} {
     msg_warning_ok "No wqo selected for unfolding." 
     return "error"
  }
    # check the compatibility of the selected components
  prolog "ds_env(S,UF,WU)"
  set solver $prolog_variables(S)
  set uf $prolog_variables(UF)
  set wqounf $prolog_variables(WU)

  if {[prolog  is_a_uf($uf,$solver)] == 0} {
   msg_warning_ok "The unfolding function $uf was not designed to work with solver $solver." 
   return "error"
  }
  if {[prolog  is_a_wqounf($wqounf,$solver)] == 0} {
   msg_warning_ok "The unfolding wqo $wqounf was not designed to work with solver $solver." 
   return "error"
  }
  
  
  if {[llength $defselection]==0} {
     alertnote "Bad selection. Select the initial definitions in the \"Init+Defs\" window."
   return "error"
  }
    
  return ""
}


proc check_selection_for_solve_program {} {
    if {[prolog  no_solver_selected] == 1} {
		  msg_warning_ok "No solver selected." 
		  #showSolvers
		  return "error"
    }
}


proc enter_new_definition {} {
    global font white proofSelection
    set w .wdefinition
    catch {destroy $w}
    toplevel $w
    wm title $w "define"
    label $w.msg  -wraplength 7i -foreground black -justify left \
	    -text "Enter new definition clause(s)." -font $font
    pack $w.msg -side top
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.cancel -text Close -command "destroy $w"
    button $w.buttons.ok -text Define \
	    -command {
	    parseNewDefinition .wdefinition.text 
            prolog "apply_transformation_rule(definition,\[\],\[\],\[\])"
    if [winfo exists .currentInit_Defs] {
	   if [winfo viewable .currentInit_Defs] {
	       prolog update_show_init_defs } } }
    pack $w.buttons.ok $w.buttons.cancel -side left -expand 1
    text $w.text -relief sunken -bd 2 -yscrollcommand "$w.vscroll set" \
	    -foreground black -xscrollcommand "$w.hscroll set" \
	    -setgrid 1 	-background $white -height 5
    scrollbar $w.vscroll -command "$w.text yview"
    scrollbar $w.hscroll -orient horiz -command "$w.text xview"
    pack $w.vscroll -side right -fill both
    pack $w.hscroll  -side bottom -fill x
    pack $w.text -expand yes -pady 2m -padx 2m -fill both
     if {[all_clauses $proofSelection]} {
	 set sel [makeprologselection $proofSelection]
	 prolog tcl_prepare_new_definition($sel)}
    $w.text mark set insert 0.0
}

proc enter_new_clauses {} {
    global font white proofSelection
    set w .wnewclauses
    catch {destroy $w}
    toplevel $w
    wm title $w "New clauses"
    label $w.msg  -wraplength 7i -justify left -foreground black \
	    -text "Enter new clause(s)." -font $font
    pack $w.msg -side top
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.cancel -text Close -command "destroy $w"
    button $w.buttons.ok -text Add \
	    -command {
	    parseNewDefinition .wnewclauses.text 
	    prolog "apply_transformation_rule(add_clauses,\[\],\[\],\[\])"}
    pack $w.buttons.ok $w.buttons.cancel -side left -expand 1
    text $w.text -relief sunken -bd 2 -yscrollcommand "$w.vscroll set" \
	    -foreground black -xscrollcommand "$w.hscroll set" \
	    -setgrid 1 	-background $white -height 5
    scrollbar $w.vscroll -command "$w.text yview"
    scrollbar $w.hscroll -orient horiz -command "$w.text xview"
    pack $w.vscroll -side right -fill both
    pack $w.hscroll  -side bottom -fill x
    pack $w.text -expand yes -pady 2m -padx 2m -fill both
    if {[all_clauses $proofSelection]} {
	 set sel [makeprologselection $proofSelection]
	 prolog tcl_prepare_new_clauses('.wnewclauses.text',$sel)}
    $w.text mark set insert 0.0
}

proc enter_replace_clauses {} {
    global font white proofSelection
    set w .wrepclauses
    catch {destroy $w}
    toplevel $w
    wm title $w "Replace clauses"
    label $w.msg  -wraplength 7i -justify left -foreground black \
	    -text "Enter the new clauses and press \"Replace\" to\
	             apply the replacement." -font $font
    pack $w.msg -side top
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.cancel -text Close -command "destroy $w"
    set sel [makeprologselection $proofSelection]
	    puts stdout"proofSelection=$proofSelection
    button $w.buttons.ok -text Replace \
	    -command {
	    parseNewDefinition .wrepclauses.text 
	    set sel [makeprologselection $proofSelection]
	    puts stdout"proofSelection=$proofSelection
	    prolog "apply_transformation_rule(replace_clauses,$sel,\[\],\[\])"}
    pack $w.buttons.ok $w.buttons.cancel -side left -expand 1
    text $w.text -relief sunken -bd 2 -yscrollcommand "$w.vscroll set" \
	    -foreground black -xscrollcommand "$w.hscroll set" \
	    -setgrid 1 	-background $white -height 5
    scrollbar $w.vscroll -command "$w.text yview"
    scrollbar $w.hscroll -orient horiz -command "$w.text xview"
    pack $w.vscroll -side right -fill both
    pack $w.hscroll  -side bottom -fill x
    pack $w.text -expand yes -pady 2m -padx 2m -fill both
    set sel [makeprologselection $proofSelection]
    prolog tcl_prepare_new_clauses('.wrepclauses.text',$sel)
    $w.text mark set insert 0.0
}


proc enter_new_laws {} {
    global font dirNameTheories white
    set w .wlaws
    catch {destroy $w}
    toplevel $w
    wm title $w "Law addition"
    label $w.msg  -wraplength 7i -foreground black -justify left \
	    -text "Enter new law(s). A law must be a (prolog) term of the form" -font $font
    label $w.msg2  -wraplength 7i -foreground black -justify left \
	    -text "L1, ..., Ln -> R1,..., Rm. " -font {Courier 12 bold}
    pack $w.msg $w.msg2 -side top
    frame $w.buttons
    pack $w.buttons -side bottom -fill x -pady 2m
    button $w.buttons.cancel -text Close -command "destroy $w"
    button $w.buttons.ok -text Add \
	    -command {
	set f map_tmp_user_laws
	parseNewLaws .wlaws.text $f
	if {[ prolog "apply_add_theory('$f')"]==1} {} {
		    prologfailed
    }   }
    pack $w.buttons.ok $w.buttons.cancel -side left -expand 1
    text $w.text -relief sunken -bd 2 -yscrollcommand "$w.vscroll set" \
	    -foreground black -xscrollcommand "$w.hscroll set" \
    -setgrid 1 	-background $white -height 5
    scrollbar $w.vscroll -command "$w.text yview"
    scrollbar $w.hscroll -orient horiz -command "$w.text xview"
    pack $w.vscroll -side right -fill both
    pack $w.hscroll  -side bottom -fill x
    pack $w.text -expand yes -pady 2m -padx 2m -fill both
    $w.text mark set insert 0.0
}

proc parseNewDefinition {w} {
    set str [$w get 1.0 end]
    set out [open map_tmp_newdef w]
    puts  $out $str
    close $out 
}

proc parseNewLaws { w f} {
    set str [$w get 1.0 end]
    set out [open $f w]
    puts  $out $str
    close $out 
}

proc save {what} {
 if {$what=="theory" && [prolog  current_theory_empty] == 1} {
  msg_warning_ok "Theory is empty!" ; return }
# was:	get_namefile_and_save_$what
 get_namefile_and_save $what
}

proc delete_laws {} {
    global lawSelection
    set w .currentTheory
    if { [prolog current_theory_empty]==1 } {
	msg_warning_ok "Theory is empty!" ; return 
    }
    if [winfo exists  $w] { wm deiconify $w }
    if {[empty $lawSelection]} {
	alertnote "Select some law(s) first"     } {
	    if { [prolog "apply_delete_laws([makeprologlist $lawSelection])"]==1 
	} { if {  [prolog current_theory_empty]==1 } { wm withdraw $w } {
	    showCurrentTheory } } {
	    prologfailed
	    }
	}
    }



puts stdout "rules.tcl compiled"










